/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2012, 2013.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
(function(){
    dojo.provide("rmps.ui.ErrorReportController");
    
    dojo.require("dijit._Widget");
    dojo.require("dijit._Templated");
    dojo.require("dojo.i18n");
    dojo.require("rmps.core.error");
    
    dojo.require("jazz.ui.MessageBox");
    dojo.require("dojo.date.stamp");
    dojo.require("rmps.ui.PagingControl");
    
    dojo.requireLocalization("rmps.ui", "ErrorReportControllerMessages");
	var messages = dojo.i18n.getLocalization("rmps.ui", "ErrorReportControllerMessages");
    
    /**
     * <p>
     * Error controller.  Listens for error events on a channel (publish/subscribe)
     * and manages an error report viewer.
     * </p>
     * <p>To use a custom viewer, extend this class and override the
     * <code>#newViewer(sev)</code> method.
     * </p>
     */
    dojo.declare("rmps.ui.ErrorReportController", [dijit._WidgetBase, dijit._Templated], {
    
        // CSS styling class
        "class": "errorReportController",
        
        // Layout template
        templateString: "<div dojoAttachPoint=\"errorReportContainer\"></div>",
        
        // Order to display the severities in.  To add a new severity or to 
		// re-arrange the levels 
        displayOrder: [ rmps.core.error.Severity.CRITICAL,  // +---Top----+
                        rmps.core.error.Severity.ERROR,     // | ======== |
                        rmps.core.error.Severity.WARNING,   // | ======== |
                        rmps.core.error.Severity.INFO],     // +--Bottom--+

        // Default (static) channel
        DEFAULT_CHANNEL: rmps.core.error.Channel,
        
        // Initial list of channels to listen on.  Not kept current!
        channels: null,
        handles: [],
        
        /*  
         * Maps (2D array) of errors, depending if they've been acknowledged or
         * remain unacknowledged.  Maps the errors to their severity, ex:
         * "acknowledgedErrors[severity][errors]".  The errors array is a FILO queue,
         * so index 0 is always the last added error.
         */
        acknowledgedErrors: null,
        unacknowledgedErrors: null,
        
        // Array of viewer objects.
        viewers: null,
        
        /**
         * Arguments:
         * <ul>
         *  <li>channels (optional): Channels to listen on, or use default if 
         *                           none provided.</li>
         * </ul>
         *
         * @param {Object} args Arguments
         */
        constructor: function(args){           
            // Init the arrays as empty
            this.acknowledgedErrors = [];
            this.unacknowledgedErrors = [];
            this.viewers = [];

            // Use provided channel(s), or default if none given
            if (!this.channels) {
                this.channels = [this.DEFAULT_CHANNEL];
            }
        },
        
        // After Dojo initialisation, subscribe to the channel(s)
        postCreate: function(){
            this.inherited(arguments);
			this.addChannels(this.channels);
        },
        
		/**
		 * <p>
		 * Subscribes the controller to an error channel or list of channels.
		 * </p>
		 * <p>
		 * The error report controller listens to a dojo channel, which errors are
		 * published to.  When an error is published on a channel, the controller
		 * picks up this error and displays it. 
		 * </p>
		 * 
		 * @param {Object} channelName New channel(s) to listen on
		 * @see #removeChannel(channelName)
		 */
		addChannels: function(channelNames){
			if (dojo.isArray(channelNames)) {
				dojo.forEach(channelNames, dojo.hitch(this, function(channel){
					this.addChannels(channel);
				}));
			} else {
				var handle = this.subscribe(channelNames, dojo.hitch(this, this._handleNewError));
				this.handles.push(handle);
			}
		},
		
		/**
		 * <p>
		 * Removes a channel (or list of channels) from the list of channels 
		 * the controller is listening on.
		 * </p>
		 * <p>
		 * The error report controller listens to a dojo channel, which errors are
		 * published to.  When an error is published on a channel, the controller
		 * picks up this error and displays it. 
		 * </p>
		 * @param {Object} channelName Channel to no longer listen on
		 * @see #addChannel(channelName)
		 */
		removeChannels: function(handles){
			if (dojo.isArray(handles)) {
				dojo.forEach(handles, this.removeChannels, this);
			} else {
				if (handles){
					handles.remove();
					var index= dojo.indexOf(this.handles, handles);
					if(index>-1){
						this.handles.splice(index, 1);
					}
				}
			}
		},
		
        /**
         * <p>
         * Destructor, need to clean up the views and unsubscribe from all out subscribed channels
         * </p>
         */
		destroy: function(){
			// Make sure we call the default destructor
			this.inherited(arguments);
			
			// Unsubscribe from the channels
			this.removeChannels(this.handles);		
			
			// Explicitly destroy the viewers
			dojo.forEach(this.viewers, function(view){
				view.destroy();
			});
			
			// Make sure we call the default destructor
			this.inherited(arguments);
		},
        
        
        /**
         * <p>
         * Handler for new error events.  When an error comes in, it will be sorted
         * into the proper map (using severity as the key)  Maps and viewers are
         * lazily initialized.
         * </p>
         *
         * @param {Object} error Error to add
         */
        _handleNewError: function(error){
            if (error.severity) {
                var sev = error.severity;
                
                // Lazy initialisation 
                if (!this.acknowledgedErrors[sev]) {
                    this.acknowledgedErrors[sev] = [];
                    this.unacknowledgedErrors[sev] = [];
					this.viewers[sev] = [];
                    
                    // Create the new view, add it to the views array, then add to DOM
                    var view = this._newViewer(sev);
					view.startup();
                    this.viewers[sev.toString()].push(view);
                    
                    // Insert the new viewer in the proper index location 
                    // (see this.displayOrder).  Default is end of the line.	
                    var orderIndex = dojo.indexOf(this.displayOrder, sev);
					// Might cause an error here if adding to 3, but only 0 has been initialised?
                    this.errorReportContainer.appendChild(view.domNode, orderIndex > 0 ? orderIndex : null);
                }
                
                // Add the error to the beginning of the array and notifiy it's view
                this.unacknowledgedErrors[sev].unshift(error);
                this._updateViewForError(error);
                
            } else {
                /*
                 * Could not recognize what 'error' is - wrap it in an error and re-dispatch it.  Not the best
                 * idea but arguably better then silently burying the error
                 */
				var e = rmps.core.error.createError(messages.unrecognized_error, error);
				e.dispatch();
            }
        },
        
        /**
         * <p>
         * Creates a new viewer for the given error severity.
         * </p>
         * <p>
         * Can be overridden for different viewers.
         * </p>
         *
         * @param {Object} sev Severity to be displayed
         * @return The newly created view
         */
        _newViewer: function(sev){
            var view = new rmps.ui.ErrorReportController._errorReportMessageBox({
                //note: below is a workaround because jazz.ui.MessageBox does not have a "CRITICAL" message box
                //      even though the server can respond with a "CRITICAL" message 
                severity: sev === rmps.core.error.Severity.CRITICAL? rmps.core.error.Severity.ERROR: sev,
                controller: this
            });
            view.severity = sev;
            return view;
        },
        
        /**
         * <p>
         * Marks an error as having been acknowledged by the user
         * </p>
         *
         * @param {Object} err The error
         */
        acknowledgeError: function(err){
            var index = dojo.indexof(this.unacknowledgedErrors, err);
            
            // Error doesn't exist in the unacknowledged list
            if (index < 0) {
                return;
            }
            
            this.unacknowledgedErrors.splice(index, 1);
            this.acknowledgedErrors.shift(err);
            
            // Errors can be acknowledged anywhere so we need to make sure we update
            // all the views. 
            this._updateViewForError(err);
        },
        
        /**
         * <p>
         * Updates the view for an error
         * </p>
         * <p>
         * Note: The index for the error viewer is hard-coded to 0 for now as we are only 
         * expecting there to be one viewer per error severity.  In the future there may be
         * more then one viewer per error severity, this is part of the groundwork towards
         * this.
         * </p>
         *
         * @param {Object} err The error
         */
        _updateViewForError: function(err){
			this.viewers[err.severity][0].setError(err);
        },
        
        /**
         * <p>
         * Gets the Nth error in the unacknowledgedErrors array.
         * </p>
         *
         * @param {String} sev Severity of the error
         * @param {int} index Index of the error in the map
         * @return {rmps.error} The Nth error, or <code>NULL</code> if the index is out of bounds
         */
        getError: function(sev, index){
            if( 0 <= index && index < this.unacknowledgedErrors[sev].length){
				return this.unacknowledgedErrors[sev][index];
            }
            return null;
        },
        
        /**
         * <p>
         * Gets the index for an error in the apporopriate unacknowledged errors array
         * </p>
         * 
         * @param {rmps.error.Severity} sev Severity of the error
         * @param {rmps.error.Error} error Error to look up
         * @return {int} The index of the error
         */
        getIndexForError: function(sev, error){
        	return dojo.indexOf(this.unacknowledgedErrors[sev], error);
        },
        
        /**
         * <p>
         * Counts how many unacknowledged errors exist
         * </p>
         */
        countErrors: function(sev){
            return this.unacknowledgedErrors[sev].length;
        },
    
    });
		
    /** 
     * <p>
     * Displays error reports of a given severity to the user as a message box.
     * </p>
     */
    dojo.declare("rmps.ui.ErrorReportController._errorReportMessageBox", jazz.ui.MessageBox, {
		
		// Controller for the message box
		controller: null,
		
		// Severity being displayed
		severity: null,
		
		// Cells for the visible stuff
		contentCell: null,		// Overall DOM node
		detailsCell: null,		// Details of the error
		paginationCell: null,	// Paginator
		showMoreLessCell: null,	// Show more/less toggle
		acknowledgeCell: null,  // Acknowledge button
		
		// Paginator, show more/less toggle
		paginator: null,
		showMoreLess: null,
		
		// Error variables
		currError: null,	// Currently visible error
		
		// Current state of the viewer
		detailState: null,
		
		// Current details obj
		details: null,

		
		// Valid states for the viewer
		States: {
			MORE: messages.more,
			LESS: messages.less,
			ALL: messages.all,	// Not used yet (M9)
		},
		
		popup: null,

		postCreate: function(){
			// Create the main content node
			this.contentCell = dojo.create("span");
			dojo.addClass(this.contentCell, "contentCell");
	
			// Create the main content cells within the main content node
			this.detailsCell = this._makeContentCell("detailsCell");
			this.showMoreLessCell = this._makeContentCell("cell showMoreLessCell");
			this.acknowledgeCell = this._makeContentCell("cell acknowledgeCell");

			// Add the show more/less, which also sets the default state to LESS
			this.showMoreLess = new rmps.ui.ErrorReportController._errorReportShowMoreLess({
				viewer: this,
				States: this.States
			});
			this.showMoreLessCell.appendChild(this.showMoreLess.domNode);
			
			this.showMessage({
				type: this.severity,
				summary: this.contentCell,
				closable: true
			});
			
			this.refresh();
		},
		
		/**
		 * <p>
		 * Helper for creating new content cells
		 * </p>
		 */
		_makeContentCell: function(style){
			var cell = dojo.create("span");
			dojo.addClass(cell, style);
			this.contentCell.appendChild(cell);
			return cell;
		},
		
		
		/**
		 * <p>
		 * Updates the visible content of the message box
		 * </p>
		 */
		refresh: function(){						
			// No error, no problem
			if(!this.currError){
				//TODO: Hide instead of do nothing
				return;
			}
		
			// Create the details depending on view state.  State is determined by the text
			// shown to the user, i.e. when you're in the "show more" mode, the text will
			// say "Show Less".  
			if(this.state === this.States.LESS){
				this.details = new rmps.ui.ErrorReportController._errorReportMoreDetails({
					viewer: this,
					error: this.currError
				});
			} else {
				this.details = new rmps.ui.ErrorReportController._errorReportLessDetails({
					viewer: this,
					error: this.currError
				});
			}
			
			// Refresh the details content
			dojo.empty(this.detailsCell);
			this.detailsCell.appendChild(this.details.domNode);
			
			// Refresh the paginator, removing it if there are 0 or 1 errors, and 
			// creating/updating accordingly it if there are more then 1 errors
			var errCount = this.controller.countErrors(this.severity);
			if(errCount > 1) {
				if(this.paginator){
					// Update the paginator
					this.paginator.update(null, null, null, errCount, this.paginator.item);
				} else { 
					// Create the paginator
					this.paginationCell = dojo.create("span");
					errCount = this.controller.countErrors(this.severity);
					this.paginator = new rmps.ui.ErrorReportController._errorReportPagingControls({
						viewer: this,
						size: errCount,
						item: 1,
					});
					this.paginationCell.appendChild(this.paginator.domNode);
					dojo.addClass(this.paginationCell, "paginationCell cell");
					dojo.place(this.paginationCell, this.detailsCell, "after");
				}
			} else {
				// Remove the paginator
				if (this.paginator) {
					this.paginator.destroy();
				}
				if (this.paginationCell) {
					dojo.destroy(this.paginationCell);
				}
			}
		},
		
		
		/**
		 * <p>
		 * Sets the error to be displayed.
		 * </p>
		 * 
		 * @param {rmps.core.Error} error Error to be displayed
		 */
		setError: function(error){
			if (this.currError !== error) {
				this.currError = error;
				var errIndex = this.controller.getIndexForError(this.severity, error);
				if(errIndex >= 0 && this.paginator && this.paginator.item - 1 !== errIndex){
					this.paginator.item = errIndex + 1;
				} 
				this.refresh();
			}
		},
		
		/**
		 * <p>
		 * Sets the error to be displayed by looking up the error in the controller.
		 * </p>
		 * 
		 * @param {Object} index Index of the error to show
		 */
		setErrorByIndex: function(index){
			this.setError(this.controller.getError(this.severity, index));
		},
		
		/**
		 * <p>
		 * Sets the show more/less state
		 * </p>
		 * 
		 * @param {Object} state New state
		 */
		setState: function(state){
			if (this.state !== state) {
				this.state = state;
				this.refresh();
			}
		}
		
	});
	
	/**
	 * <p>
	 * Shows the bare minimum of error details
	 * </p>
	 */
	//TODO: NLS
	dojo.declare("rmps.ui.ErrorReportController._errorReportLessDetails", [dijit._WidgetBase, dijit._Templated], {
		
		"class": "lessDetails",
		
		// Parent viewer
		viewer: null,
		
		// Error being displayed
		error: null,
		
		// Template
		templateString: '<div dojoAttachPoint="lessDetailsContainer">' + 
							'<span dojoAttachPoint="messageContent" class="message"></span>' +
							'<span dojoAttachPoint="dateContent" class="date"></span>' +
						'</div>',
		
		// Create the details content
		postCreate: function(){
			var dateText = dojo.date.locale.format(this.error.date, {formatLength: "short", selector: "time"});
			
			// Create the content
			var message = dojo.doc.createTextNode(this.error.message);
			var date = dojo.doc.createTextNode(dateText);
			
			// Add it to the DOM
			this.messageContent.appendChild(message);
			this.dateContent.appendChild(date);
		}
	});

	/**
	 * <p>
	 * Shows the full error details
	 * </p>
	 */
	//TODO: NLS
	dojo.declare("rmps.ui.ErrorReportController._errorReportMoreDetails", dijit._WidgetBase, {
		
		"class": "moreDetails",
		
		// Parent viewer
		viewer: null,
		
		// Error being displayed
		error: null,
		table: null,
		
		popupContent: null,
		
		// Create the details content
		postCreate: function(){
			var dateText = dojo.date.locale.format(this.error.date, {formatLength: "medium"});
			
			// Create the base table
			this.table = dojo.create("table");
			this.domNode.appendChild(this.table);
			dojo.addClass(this.table, "showMoreTable");
			
			// Create the rows
			this._makeTitleRow(this.error.message, "message");	
	
			if(this.error.status) {
				this._makeRow(messages.status, this.error.status, "status");
			}
            if (this.error.subStatus) {
                this._makeRow(messages.subStatus, this.error.subStatus, "substatus");
            }
			if(this.error.severity)	{
				this._makeRow(messages.severity, dojo.hitch(this, function(){
					var container = dojo.create("span");
					var sev = this.error.severity;
					
					var img = dojo.create("img");
					img.src = dojo.moduleUrl("rmps.ui", "graphics/icons/" + sev.toLowerCase() + ".gif");
					img.alt = sev;
					
					var text = dojo.doc.createTextNode(sev);

					dojo.addClass(img, "icon");
					
					container.appendChild(img);
					container.appendChild(text);
					
					return container;
				}), "severity");
			}
			
			if(this.error.action !== null){
				this._makeRow(messages.action, this.error.action, "action");
			} else if(this.error.cause && this.error.cause.action){
				this._makeRow(messages.action, this.error.cause.action, "action");
			}
			
			if(this.error.detailedMessage){
				this._makeRow(messages.detailedMessage, this.error.detailedMessage, "detailed");
			}
			this._makeRow(messages.date, dateText, "date");
			
			if (this.error.seeAlso) {
				this._makeRow(messages.seeAlso, dojo.hitch(this, function(){
					var link = dojo.doc.createElement("a");
					link.appendChild(dojo.doc.createTextNode(this.error.seeAlso));
					link.href = this.error.seeAlso;
					link.target = "_blank";
					return link;
				}), "seeAlso");
			}
			if(this.error.cause) {
				this._makeRow(messages.cause, dojo.hitch(this, function(){
					var link = dojo.create("a");			
					link.appendChild(dojo.doc.createTextNode(this.error.cause.message ? this.error.cause.message : "Error"));
					dojo.addClass(link, "fauxLink");
					dojo.connect(link, 'onclick', dojo.hitch(this, function(){
						this.error.cause.caused = this.error;
						this.viewer.setError(this.error.cause);
					}));
					return link;
				}), "causedBy");
			}
			if(this.error.caused) {
				this._makeRow(messages.caused, dojo.hitch(this, function(){
					var link = dojo.create("a");			
					link.appendChild(dojo.doc.createTextNode(this.error.caused.message ? this.error.caused.message : "Go Back"));
					dojo.addClass(link, "fauxLink");
					dojo.connect(link, 'onclick', dojo.hitch(this, function(){
						this.viewer.setError(this.error.caused);
					}));
					return link;
				}), "causedBy");
			}
		},
		
		/**
		 * <p>
		 * Helper for creating a row in the table body
		 * </p>
		 * 
		 * @param {Object} label Label for the row
		 * @param {Object} style CSS style to apply to the row
		 * @param {Object} content Content for the row
		 * @param {Object} hoverLabel Help message to display if the user hovers over the label
		 * @param {Object} hoverMessage Help message to display if the user hovers over the message
		 */
		_makeRow: function(label, content, style, hoverLabel, hoverMessage){		
			// Create the row
			var row = dojo.create("tr");
			dojo.addClass(row, "content");
			
			// Label
			var labelCell = dojo.create("td");
			dojo.addClass(labelCell, "label " + style );
			var labelNode = dojo.doc.createTextNode(label);
			labelCell.appendChild(labelNode);
			
			// Add a help message popup - not working at the moment, disabled
//			if(hoverLabel && hoverLabel != ""){
//				this.popup = new jazz.ui.HoverPopup({
//					around: labelNode,
//					createContents: dojo.hitch(this, function(helpMessage){
//						if(this.popupContent){
//							this.popupContent.destroy();
//						}
//						
//						// Create the nodes
//						var contentCell = dojo.create("div");
//						var labelCell = dojo.create("span");
//						var labelContent = dojo.doc.createTextNode("Description:");
//						var descriptionCell = dojo.create("span");
//						var descriptionContent = dojo.doc.createTextNode(helpMessage);
//						
//						// Build up the DOM
//						contentCell.appendChild(labelCell);
//						contentCell.appendChild(descriptionCell);
//						labelCell.appendChild(labelContent);
//						descriptionCell.appendChild(descriptionContent);
//					}),
//				});
//			}
			
			// Content
			// TODO: if performance is an issue, then maybe we can suppress one of the else if's below
			var contentCell = dojo.create("td");
			dojo.addClass(contentCell, "data " + style );
			if(dojo.isString(content)){
				contentCell.appendChild(dojo.doc.createTextNode(content));
			} else if (dojo.isFunction(content)) {
				contentCell.appendChild(content());
			} else if(!(content === null || content === undefined)) {				
				//if it's not a function or a string, then it's either a number, a boolean, or an Object(Array is object too)
				//all of these types respond to toString(), allowing their contents to be displayed without error, NaN included.
				
				if(content.innerHTML){ //if it's a DOM object, display its contents
				    contentCell.appendChild(dojo.doc.createTextNode(content.innerHTML));
				}else{
				    contentCell.appendChild(dojo.doc.createTextNode(content.toString()));
				}			
			} else if (content === null){ 
				contentCell.appendChild(dojo.doc.createTextNode("null"));
			} else { //if it made it here, it's undefined
				contentCell.appendChild(dojo.doc.createTextNode("undefined"));
			}
			
			// Add the label and content to the row
			row.appendChild(labelCell);
			row.appendChild(contentCell);
			
			// Add the row to the table
			this.table.appendChild(row);
		},
		
		/**
		 * <p>
		 * Creates a title row, spanning 2 columns 
		 * </p>
		 * 
		 * @param {Object} content Text or DOM for the title row
		 * @param {Object} style Additional styles to apply
		 */
		_makeTitleRow: function(content, style){
			var row = dojo.create("tr");
			var td = dojo.create("td");
			td.colSpan = 2;
			td.appendChild(dojo.doc.createTextNode(content));
			row.appendChild(td);
			
			dojo.addClass(td, "title " + style);
			dojo.addClass(row, "title " + style);
			
			this.table.appendChild(row);
		}
				
	});
	
	/**
	 * <p>
	 * Paging controls for the error viewer.
	 * </p>
	 */
	dojo.declare("rmps.ui.ErrorReportController._errorReportPagingControls", rmps.ui.PagingControl, {
		
		// Parent viewer
		viewer: null,
		
		// Called when the paging control is updated, update the viewer with the new error index
		update: function(){
			this.inherited(arguments);
			this.viewer.setErrorByIndex(this.item - 1);
		},
		
		// Next item hover label
		getNextItemLabel: function(){
			var message = this.viewer.controller.getError(this.viewer.severity, this.item).message;
			var label = dojo.string.substitute(messages.gotoCausedBy, [message]);
			return label;
		}, 
		
		// Previous item hover label
		getPreviousItemLabel: function(){
			var message = this.viewer.controller.getError(this.viewer.severity, this.item - 2).message;
			var label = dojo.string.substitute(messages.gotoCausedBy, [message]);
			return label;
		}
	});
	
	/**
	 * <p>
	 * Toggles between showing more and showing less details
	 * </p>
	 */
	dojo.declare("rmps.ui.ErrorReportController._errorReportShowMoreLess", [dijit._WidgetBase/*, dijit._Templated*/], {
		
		"class": "showMoreLess",
		
		// View for this showMoreLess
		viewer: null,
				
		// Current state of the show moreLess button.
		state: null,
		
		// List of valid states, read off of the viewer
		States: null,
		
		toggleLabelButton: null,
				
		postCreate: function(){
			// Create the DOM elements		
			this.toggleLabelButton = dojo.create("a");
			dojo.addClass(this.toggleLabelButton, "fauxLink");
			this.domNode.appendChild(this.toggleLabelButton);
			
			// Initialise self via the toggle method
			this.toggle();
			
			// Hook up the onClick event
			dojo.connect(this.toggleLabelButton, 'onclick', dojo.hitch(this, this.toggle));
		},
		
		// Toggles between showMore and showLess
		toggle: function(){
			this.state = this.state === this.States.MORE ? this.States.LESS : this.States.MORE;
			
			// Correct the label
			dojo.empty(this.toggleLabelButton);
			var label = dojo.doc.createTextNode(this.state);	// Note: the State contains its' own message
			this.toggleLabelButton.appendChild(label);
			
			// Set the viewers state
			this.viewer.setState(this.state);
		},
	});
})();
