/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010, 2014.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
(function() {
dojo.provide("rmps.ui.SidebarContainer");

dojo.require("dijit._Widget");
dojo.require("dijit._Templated");

dojo.require("dojo.i18n");

dojo.requireLocalization("rmps.ui", "SidebarContainerMessages");

var messages = dojo.i18n.getLocalization("rmps.ui", "SidebarContainerMessages");

var CONTENT_AREA_MIN_WIDTH = 200;
var SIDEBAR_SNAP_WIDTH = 80;
var SIDEBAR_MIN_WIDTH = 180;
var SIDEBAR_MAX_WIDTH = 600;
var SIDEBAR_DEFAULT_WIDTH = 240;
var SIDEBAR_SASH_DELAY = 50;

/**
 * @author Christian Vogt (cvogt)
 */
dojo.declare("rmps.ui.SidebarContainer", dijit._Widget, {

	"class": "rmps-ui-sidebarContainer",

	dragResizeDelay: 300,

	_sidebars: null,
	_isDisplayed: true,

	_cookieNameSuffix: '.sidebarState',
	
	constructor: function() {
		this._sidebars = [];
	},

	postCreate: function() {
		this.containerNode = dojo.create("div", {style: "overflow: hidden;"}, this.domNode);
	},

	startup: function() {
		this.inherited(arguments);

		for (var i = 0, sidebar; sidebar = this._sidebars[i]; i++) {
			if (sidebar && !sidebar._started) {
				sidebar.sidebarContainer = this;
				sidebar.startup();
			}
		}

		if (this.contentWidget && !this.contentWidget._started) {
			this.contentWidget.startup();
		}
		
		// Trigger a resize to make sure all the editors are sized properly.
		if (window.onresize) window.onresize({});
	},

	destroy: function() {
		dojo.forEach(this._sidebars, function(sidebar) {
			sidebar.destroyRecursive();
		});
		this._sidebars = [];
		this.inherited(arguments);
	},
	
	/**
	 * <p>
	 * Adds a collapseable side bar to the canvas.  If <code>preserveState</code> is true, then any value found in the state
	 * cookie will override the <code>isOpen</code> parameter. 
	 * </p>
	 * 
	 * @param widget Widget to add to the sidebar container.
	 * @param isLeft <code>TRUE</code> if the sidebar is on left side, <code>FALSE</code> for the right. 
	 * @param isOpen <code>TRUE</code> if the initial position should be open, <code>FALSE</code> for closed
	 * @param preserveState <code>TRUE</code> if the open/closed state of the widget should be remembered and preserved across sessions,
	 * 						<code>FALSE</code> if it should not be.  
	 * 
	 */
	addSidebar: function(widget, isLeft, isOpen, preserveState) {
		
		// Recalls the previous state for the widget.
		if(preserveState){
			var previousState = this.getStateFromCookie(widget);
			if(previousState != null){	
				// Override the isOpen state.  If previousState is null (cookie was not set), the call to open the widget will
				// take care of setting the cookie.
				isOpen = previousState;
			}
		}
		
		var sidebar = new rmps.ui.SidebarContainerSash({
			container: this,
			sidebarWidget: widget,
			isLeft: !!isLeft,	// Ensure this is a boolean through type coersion 
			isOpen: isOpen === undefined ? true : isOpen
		});
		
		// Connect the on show/hide handles
		if(preserveState){
			// Need to use a hitch to preserve the meaning of 'widget' at call-time.  
			dojo.connect(sidebar, "onHide", dojo.hitch(this, function(){
				this.setWidgetClosed(widget);
			}));
			dojo.connect(sidebar, "onShow", dojo.hitch(this, function(){
				this.setWidgetOpen(widget);
			}));
		}
		
		this._sidebars.push(sidebar);		
		this.domNode.insertBefore(sidebar.domNode, this.containerNode);
		if (this._started) {
			sidebar.startup();
		}
	},
	
	/**
	 * <p>
	 * Returns the common name of the state cookie for a widget.
	 * </p>
	 * 
	 * @return Name for the state cookie of a widget.
	 */
	getStateCookieName: function(widget){
		return widget.declaredClass + this._cookieNameSuffix
	},
	
	/**
	 * <p>
	 * Checks the state cookie for the last used setting for the widget, returning the <code>isOpen</code> state for that widget.
	 * </p>
	 * <p>
	 * State per widget is rememberd by using the class name of the widget, retrieved from the widgets <code>declaredClass</code> attribute.  
	 * If this attribute is missing, then state will <em>not</em> be preserved.
	 * </p>
	 * 
	 * @see #setStateInCookie(widget, stateToSet)
	 * @see #clearStatecookie(widget)
	 * @param widget The widget to retrieve the state for.
	 * @return <code>TRUE</code> if the widget should be open, <code>FALSE</code> if it should be closed.  
	 * 							 <code>NULL</code> is returned if the cookie was not set.
	 * 
	 */
	getStateFromCookie: function(widget){
		var cookieVal = dojo.cookie(this.getStateCookieName(widget));
		if(cookieVal == null){
			// Return NULL if the cookie was not set
			return null;
		} 
		return cookieVal === "true"; // Convert to boolean
	},
	
	/**
	 * <p>
	 * Sets the state for the provided widget in the state cookie.
	 * </p>
	 * <p>
	 * State per widget is rememberd by using the class name of the widget, retrieved from the widgets <code>declaredClass</code> attribute.  If
	 * this attribute is missing, then state will <em>not</em> be preserved.
	 * </p>
	 * 
	 * @see #getStateFromCookie(widget)
	 * @see #clearStatecookie(widget)
	 * @param widget The widget who's state is to be preserved.
	 * @param stateToSet The <code>BOOLEAN</code> state of the widget to preserve.
	 * @return The state which was set, allowing this call to be easily chained
	 */
	setStateInCookie: function(widget, stateToSet){
		var cookieName = this.getStateCookieName(widget);
		var cookieValue = dojo.cookie(cookieName, stateToSet);
		return stateToSet;
	},
	
	/**
	 * <p>
	 * Sets the state for a widget to open and saves it
	 * </p>
	 * 
	 * @see #setStateCookie(widget, stateToSet)
	 * @param widget The widget to set
	 * @return The state it was set
	 */
	setWidgetOpen: function(widget){
		return this.setStateInCookie(widget, true);
	},
	
	/**
	 * <p>
	 * Sets the state for a widget to closed and saves it
	 * </p>
	 * 
	 * @see #setStateCookie(widget, stateToSet)
	 * @param widget The widget to set
	 * @return The state it was set
	 */
	setWidgetClosed: function(widget){
		return this.setStateInCookie(widget, false);
	},
	
	/**
	 * <p>
	 * Clears the state cookie for a widget
	 * </p>
	 * 
	 * @see #getStateFromCookie(widget)
	 * @see #setStateInCookie(widget, stateToSet)
	 * @param widget The widget who's state is to be cleared.
	 * @return The last known value of the cookie, as retrieved by #getStateFromCookie(widget).
	 */
	clearStateCookie: function(widget){
		var cookieName = this._cookieNamePrefix + widget.declaredClass;
		var cookieValue = this.getStateFromCookie(widget);
		dojo.cookie(cookieName, null, {expires: -1}); // Set expires to -1 to delete the cookie
		return cookieValue;
	},
	
	setContent: function(widget, considerHeight) {
		if (this.contentWidget) {
			this.contentWidget.destroyRecursive();
		}
		this.contentWidget = widget;
		this.containerNode.appendChild(this.contentWidget.domNode);
		if (!this.contentWidget._started) {
			this.contentWidget.startup();
		}
		this._considerHeight = considerHeight;
		// Trigger a resize to make sure all the editors are sized properly.
		if(window.onresize){
			window.onresize({});
		} else {
			// TODO: It would be better to wait for a proper event for this, but how do we know it's done?
			setTimeout(function() {
				// double-check that it's still not null. Who sets this?
				window.onresize && window.onresize({});
			},250);
		}
	},

	removeSidebar: function(widget) {
		for (var i = 0, sidebar; sidebar = this._sidebars[i]; i++) {
			if (sidebar.sidebarWidget === widget) {
				// only destroy the sidebar and not its content widget
				sidebar.destroy();
				this._sidebars.splice(i, 1);
				return;
			}
		}
	},

	resize: function(size) {
		// Intentionally blank.  See WI33654
	},

	resizeContentWidth: function(height) {
		// Intentionally blank.  See WI33654
	},
	
	display: function(display) {
		if (display !== this._isDisplayed) {
			// with IE8, scrollbar gets increasingly smaller
			if (dojo.isIE!=8) {
				var displaystyle = display ? "" : "none";
				dojo.forEach(this._sidebars, function(sidebar) {
					sidebar.domNode.style.display = displaystyle;
				});
			}
			this._isDisplayed = display;
			this.resizeContentWidth();
		}
		window.onresize({});
	},
	
	openSidebars: function(bOpen) {
		var res = false;
		dojo.forEach(this._sidebars, function(sidebar) {
			if (sidebar.isOpen !== bOpen) {
				sidebar._toggleBtnDown = true;
				res = sidebar._toggleSidebar();
			}
		});
		return res;
	} 
});

dojo.declare("rmps.ui.SidebarContainerSash", [dijit._Widget, dijit._Templated], {

	templateString: dojo.cache("rmps.ui", "templates/SidebarContainer.html"),

	isLeft: false,

	container: null,
	sidebarWidget: null,
	isOpen: true,
	_isFirstResize: true,

	_sidebarWidth: SIDEBAR_DEFAULT_WIDTH,
	
	_isResizing : false,

	postCreate: function() {
		this.sidebarWidget.__sidebar = this;

		this.sash.title = messages.resize;
		this.toggleBtn.title = this.isOpen ? messages.collapse : messages.expand;
		this.containerNode.style.width = this.isOpen ? this._sidebarWidth + "px" : 0;

		dojo.addClass(this.domNode, this.isLeft ? "left" : "right");

		if (!this.isOpen) {
			dojo.addClass(this.domNode, "closed");
		}
		
		if (this.sidebarWidget) {
			this.containerNode.appendChild(this.sidebarWidget.domNode);
		}

		this.connect(this.toggleBtn, "onclick", this._toggleSidebar);
		this.connect(this.toggleBtn, "onmousedown", this._onToggleBtnDown);
		this.connect(this.toggleBtn, "onkeydown", this._onToggleBtnDown);
		this.connect(this.toggleBtn, "onkeypress", this._toggleSidebar);
		this.connect(this.toggleBtn, "onfocus", function() {
			this._toggleBtnFocused = true;
			this._showSash();
			dojo.addClass(this.domNode, "sashHover");
		});
		
		this.connect(this.toggleBtn, "onblur", function() {
			this._toggleBtnFocused = false;
			this._hideSash();
			dojo.removeClass(this.domNode, "sashHover");
		});
	

		this.connect(this.sashContainer, "onmouseover", this._delayedShowSash);
		this.connect(this.sashContainer, "onmouseout", this._hideSash);
		
		this.connect(this.sash, "ondblclick", this._onSashDblClick);
		this.connect(this.sash, "onmousedown", this._startDragging);
				
		// cancel text selection and text dragging
		this.connect(this.sash, "ondragstart", dojo.stopEvent);
		this.connect(this.sash, "onselectstart", dojo.stopEvent);

		this.connect(this.sash, "onmouseenter", function() {
				dojo.addClass(this.domNode, "sashHover");
			});
	
		this.connect(this.sash, "onmouseleave", function() {
				dojo.removeClass(this.domNode, "sashHover");
			});
	},

	startup: function() {
		this.inherited(arguments);
		// Hold startup of sidebar widget if initially collapsed
		if (this.sidebarWidget && !this.sidebarWidget._started && this.isOpen) {
			this.sidebarWidget.sidebarContainer = this.sidebarContainer;
			this.sidebarWidget.startup();
		}
		
		dojo.connect("onresize", dojo.hitch(this, this.resizeDelayed));
	},
	
	getWidgetHeight: function(newSize) {
		var osh = this.domNode.offsetHeight;
		var height = newSize && newSize.h ? newSize.h : this._height ? this._height : osh;
		this._height = height;
		return height - osh + dojo.contentBox(this.border).h;
	},

	resizeDelayed: function() {
		// let things settle down before doing resize
		if (this._isResizing) {
			return;
		}

		this._settleTimer1 = setTimeout(dojo.hitch(this, function() {
			delete this._settleTimer1;
			this.resize();		
		}), (dojo.isChrome ? 1000 : 0));
		
	},
	
	resize: function() {
		// Make sure we have a border to set, otherwise this is all a waste.
		if(!this.border){
			return;
		}
		
		// First resize we only use the footer position
		if(this._isFirstResize){
			this._isFirstResize = false;
			return;
		}
		
		// size of sidebar is the total size of window - the footer height - the y position of this sidebar - 7 (misc pixels)
		var myDim = dojo.position(this.border, true);
		var footerHeight = dojo.byId("jazz_app_internal_Footer_0")?dojo.position(dojo.byId("jazz_app_internal_Footer_0"), true).h:0;
		var windowDim = dojo.position(document.getElementsByTagName("body")[0], false);
		var minHeight = windowDim.h - myDim.y - footerHeight - 7;
		
		// If content is provided, also consider it when setting height
		if(this.container.contentWidget && this.container._considerHeight){
			var contentDim = dojo.position(this.container.contentWidget.domNode, true);
			minHeight = Math.max(minHeight, (contentDim.y - myDim.y + contentDim.h));
		}
		this.border.style.minHeight = minHeight + "px";
		
		// also resize sidebar widget
		if (this.isOpen) {
			if (this.sidebarWidget.resize) {
				//prevent recursive resizing in IE8 
				//See WI 41721
				if (dojo.isIE) {
					this._isResizing = true;
				}

				this.sidebarWidget.resize({
					h: minHeight, 
					w: this._sidebarWidth});
				if (dojo.isIE) {
					this._isResizing = false;
				}
			}
		}
	},

	getWidth: function() {
		return this.domNode.offsetWidth;
	},
	
	_delayedShowSash: function() {
		if (!this._showSashTimer) {
			this._showSashTimer = setTimeout(dojo.hitch(this, function() {
				clearTimeout(this._showSashTimer);
				delete this._showSashTimer;
				this._showSash();
			}), SIDEBAR_SASH_DELAY);
		}
	},

	_showSash: function() {
		dojo.removeClass(this.sashContainer, "rmps-ui-sidebarContainer-sash-unfocused");
		dojo.addClass(this.sash, "rmps-ui-sidebarContainer-sash-focused");
		if (this.isOpen && !this._isDragging && !this._toggling && (!this._sashAnimation || !this._sashAnimation.__open)) {
			if (this._sashAnimation) {
				this._sashAnimation.stop();
				delete this._sashAnimation;
			}
			if (parseInt(this.sash.style.width, 10) !== 8) {
				this._animateSash(true);
			}
		}
	},

	_hideSash: function() {
		if (this._showSashTimer) {
			clearTimeout(this._showSashTimer);
			delete this._showSashTimer;
			return;
		}
		if (!this._toggleBtnFocused) {
			dojo.removeClass(this.sash, "rmps-ui-sidebarContainer-sash-focused");
			dojo.addClass(this.sashContainer, "rmps-ui-sidebarContainer-sash-unfocused");
		}
		if (this.isOpen && !this._toggleBtnFocused && !this._isDragging && !this._toggling && (!this._sashAnimation || this._sashAnimation.__open)) {
			this._animateSash(false);
		}
	},

	_animateSash: function(opening) {
		var margin = this.isLeft ? "marginRight" : "marginLeft";
		var leftright = this.isLeft ? "right" : "left";

		var currentSize = parseInt(this.sash.style.width, 10);

		var args = {
			curve: opening ? [currentSize, 8] : [8, 0],
			duration: opening ? 100 : 350,
			delay: opening ? null : 800,
			onEnd: dojo.hitch(this, function() {
				if (!opening) {
					// necessary to solve certain scenarios which depend on the offsetWidth of the content
					this.container.resizeContentWidth();
				}
				if (animation === this._sashAnimation) {
					delete this._sashAnimation;
				}
			}),
			onAnimate: dojo.hitch(this, function(value) {
				this.content.style[margin] = this.sash.style.width = parseInt(value, 10) + "px";
				this.container.resizeContentWidth();
			})
		};

		var animation = this._sashAnimation = new dojo.Animation(args);
		this._sashAnimation.__open = opening;
		this._sashAnimation.play();
	},

	_handleOpened: function() {
		if (!this.isOpen) {
			dojo.removeClass(this.domNode, "closed");
			this.isOpen = true;
			this.toggleBtn.title = messages.collapse;

			if (this.sidebarWidget.onShow) {
				this.sidebarWidget.onShow();
			}
		}
		this._settleTimer2 = setTimeout(dojo.hitch(this, function() {
			delete this._settleTimer2;
			window.onresize({});
		}), 500);
	},

	_handleClosed: function() {
		if (this.isOpen) {
			dojo.addClass(this.domNode, "closed");
			this.isOpen = false;
			this.toggleBtn.title = messages.expand;
			if (this.sidebarWidget.onHide) {
				this.sidebarWidget.onHide();
			}
		}
		window.onresize({});
	},

	_onSashDblClick: function() {
		this._toggleBtnDown = true;
		this._toggleSidebar();
	},

	_onToggleBtnDown: function(e) {
		if(e.keyCode && e.keyCode == dojo.keys.TAB)
			return;	
		this._toggleBtnDown = true;
	},
	
	_ensureWidgetStartup: function() {
		// If the sidebar widget has not been started yet due to initially collapsed sash
		// then do this now.
		if (this.sidebarWidget && !this.sidebarWidget._started) {
			this.sidebarWidget.sidebarContainer = this.sidebarContainer;
			this.sidebarWidget.startup();
			return true;
		}
		return false;
	},

	_toggleSidebar: function() {
		if (!this._toggleBtnDown) {
			return;
		}
		delete this._toggleBtnDown;

		var res = this._ensureWidgetStartup();
		
		if (this._sashAnimation) {
			this._sashAnimation.stop();
			delete this._sashAnimation;
		}

		// animate by simultaneously changing margin-right/left
		var args = {
				node: this.containerNode,
				properties: {
					left: {unit: "px"},
					marginRight: {unit: "px"}
				}
			};

		var anim;
		if (this.isOpen) {
			// hide
			this.onHide();
			args.properties.left.end = -this._sidebarWidth;
			args.properties.marginRight.end = -this._sidebarWidth;

			args.onEnd = dojo.hitch(this, this._onHideAnimationEnd);
			anim = dojo.animateProperty(args);

			this._prevWidth = 0;

		} else {
			// show
			this.onShow();
			args.properties.left.end = 0;
			args.properties.marginRight.end = 0;
			args.properties.width = {end: this._sidebarWidth, unit: "px" };

			args.onEnd = dojo.hitch(this, this._onShowAnimationEnd);
			args.beforeBegin = dojo.hitch(this, this._onShowAnimationBeforeBegin);
			anim = dojo.animateProperty(args);

			this._prevWidth = this._sidebarWidth;
		}

		this._toggling = true;
		anim.play();
		return res;
	},
	
	// Hook for when the sidebar is shown
	onShow: function(){},

	// Hook for when the sidebar is hidden
	onHide: function(){},

	_onShowAnimationBeforeBegin: function() {
		this._handleOpened();
		this.resize();
	},

	_onHideAnimationEnd: function() {
		this._handleClosed();
		this._toggling = false;
		if(window.onressize){
			window.onresize({});
		}
		this.container.resizeContentWidth();
	},

	_onShowAnimationEnd: function() {
		this._toggling = false;
		this._hideSash();
		if(window.onressize){
			window.onresize({});
		}
		this.container.resizeContentWidth();
	},

	_startDragging: function(e) {
		dojo.stopEvent(e);

		this._isOpenBeforeDrag = this.isOpen;
		var d = this.domNode.ownerDocument;
		this._dragListeners = [
			this.connect(d, "onmouseup", this._stopDragging),
			this.connect(d, "onmousemove", this._drag),
			this.connect(d, "ondragstart", dojo.stopEvent),
			this.connect(dojo.body(), "onselectstart", dojo.stopEvent)
		];

		this._startOffset = e.pageX - dojo.coords(this.sash).x;

		if (e.originalTarget === e.currentTarget) {
			this._onFirstDrag();
		}
		
		this._ensureWidgetStartup();
	},

	_stopDragging: function(e) {
		if (this._dragListeners) {
			dojo.forEach(this._dragListeners, this.disconnect, this);
			delete this._dragListeners;
		}

		if (this._isDragging) {
			dojo.stopEvent(e);

			if (this.isOpen) {
				this.toggleBtn.title = messages.collapse;
			} else {
				this.toggleBtn.title = messages.expand;
			}

			delete this._px;
			this._isDragging = false;
			dojo.removeClass(this.domNode, "resizing");
			dojo.removeClass(dojo.body(), "rmps-ui-sidebarContainerSash-dragCursor");

			var sidebarWidth = this.containerNode.clientWidth;
			if (sidebarWidth >= SIDEBAR_MIN_WIDTH) {
				this._sidebarWidth = sidebarWidth;
			}

			this._isDragging = false;
			if (this._dragResizer && this._dragResizer.timeout) {
				this._dragResize();
			}

			// close sash
			if (!dojo.isDescendant(e.target, this.sashContainer)) {
				this._hideSash();
			}
			
			if (this._isOpenBeforeDrag !== this.isOpen) {
				this.isOpen ? this.onShow() : this.onHide();
			}
			

		}
	},

	_onFirstDrag: function() {
		if (!this._isDragging) {
			delete this._toggleBtnDown;
			this._isDragging = true;

			this.toggleBtn.title = "";

			dojo.addClass(this.domNode, "resizing");
			dojo.addClass(dojo.body(), "rmps-ui-sidebarContainerSash-dragCursor");

			if (this.container.dragResizeDelay != null && this.container.dragResizeDelay > 0) {
				this._dragResizer = {};
			}
		}
	},

	_drag: function(e) {
		this._onFirstDrag();
		if (this._px === e.pageX) {
			return;
		}

		this._px = e.pageX;

		var diff = this.isLeft
				? e.pageX - dojo.coords(this.sash).x - this._startOffset
				: dojo.coords(this.sash).x - e.pageX + this._startOffset;

		var x = this.containerNode.offsetWidth + diff;

		if (this._prevWidth == null) {
			if (x < SIDEBAR_SNAP_WIDTH) {
				this._prevWidth = 0;
			} else {
				this._prevWidth = this.containerNode.offsetWidth;
			}
		}

		var width = this._prevWidth;
		if (x < SIDEBAR_SNAP_WIDTH) {
			if (this.isOpen) {
				// snap closed
				width = 0;

				this.containerNode.style.left = -this._sidebarWidth + "px";
				this.containerNode.style.marginRight = -this._sidebarWidth + "px";

				this._handleClosed();
			}
		} else {
			this._handleOpened();

			this.containerNode.style.left = "0px";
			this.containerNode.style.marginRight = "0px";

			if (x < SIDEBAR_MIN_WIDTH) {
				// prevent further shrinking
				width = SIDEBAR_MIN_WIDTH;

			} else if (x > SIDEBAR_MAX_WIDTH) {
				// prevent further expansion
				width = SIDEBAR_MAX_WIDTH;

			} else {
				var sidebarContainerWidth = dojo.isIE < 8
						? dojo.contentBox(this.domNode.parentNode).w : this.domNode.parentNode.clientWidth;

				if (x > (sidebarContainerWidth - CONTENT_AREA_MIN_WIDTH)) {
					// prevent further expansion
					width = (sidebarContainerWidth - CONTENT_AREA_MIN_WIDTH);

				} else {
					// slide with mouse
					width = x;
				}
			}
		}

		if (this._prevWidth !== width) {
			if (this._dragResizer) {
				this._dragResizer.sideBarWidgetWidth = null;
				if (this._dragResizer.timeout) {
					clearTimeout(this._dragResizer.timeout);
					delete this._dragResizer.timeout;
				}
			}

			if (width !== 0) {
				this.containerNode.style.width = width + "px";
				if (!this._dragResizer) {
					this.container.resizeContentWidth();
				}
				if (this.sidebarWidget && this.sidebarWidget.resize) {
					if (this._dragResizer) {
						this._dragResizer.sideBarWidgetWidth = width;
					} else {
						this.sidebarWidget.resize({w: width});
					}
				}
			}

			this._prevWidth = width;

			if (this._dragResizer) {
				this._dragResizer.timeout = setTimeout(dojo.hitch(this, this._dragResize), this.container.dragResizeDelay);
			}
		}
	},

	_dragResize: function() {
		if (this._dragResizer.timeout) {
			clearTimeout(this._dragResizer.timeout);
			delete this._dragResizer.timeout;
		}
		if (this.isOpen && this._dragResizer.sideBarWidgetWidth != null && this.sidebarWidget.resize) {
			this.sidebarWidget.resize({
					w: this._dragResizer.sideBarWidgetWidth,
					h: this.getWidgetHeight()
				});
		}
		this.container.resizeContentWidth();
	}
});
}());
