/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010, 2014.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.diagrams.render.Viewport");

dojo.require("dojox.gfx.matrix");
dojo.require("dojox.gfx.fx");
dojo.require("dojo.fx.easing");
dojo.require("rmpw.diagrams.render.decorations.Grid");

/**
 * A figure that remains fixed on the screen to implement a "hand tool" to drag the diagram,
 * and zooming.  Both of which are accomplished by applying transforms to the group that contains
 * the entire diagram.
 * 
 * @author Marc Gobeil / Steve Shaw
 */
dojo.declare("rmpw.diagrams.render.Viewport", null, {
	
	viewer: null,
	
	constructor: function(surface, container) {
		this.surface = surface;
		this.contents = surface.createGroup();
		this.container = container;
		
		// wire grid toggle event to hiding/showing its layer
		dojo.subscribe('gef/diagram/grid', dojo.hitch(this, function(visible) {
			if (!visible) {
				if( this._gridLayer ){
					this._gridLayer.destroy();
				}
			}
			else {
				this._gridLayer = new rmpw.diagrams.render.decorations.Grid(this.contents);
				this._gridLayer.moveToBack();
			}
		}));
		
		this._connectWindowResize = dojo.connect(window, "onresize", this, "onViewportChanged");
	},
	
	destroy: function() {
		dojo.disconnect(this._connectWindowResize);
	},
	
	pan: function(dx, dy, doNotTransform) {
		if(dx || dy) {						
			if (!doNotTransform) {
				var m = dojox.gfx.matrix;
				var t = this._getContentsTransform();
				t = m.multiply(t, m.translate(dx / t.xx, dy / t.yy));
				this.contents.setTransform(t);
			}

			this.onViewportChanged();
			
			if(this._isPanning) {
				// hold back on firing event, just track changes
				this._panX += dx;
				this._panY += dy;
			} else {
				// if not in a pan transaction, fire the events immediately
				this.onBeginPan();
				this.onEndPan(dx, dy);
			}
		}
	},
	
	/**
	 * Begins a pan transaction, during which the sum of all pan() operations is tracked,
	 * and one onPanEnd() event is fired at the end.  Otherwise, onPanEnd() will be called
	 * on every call to pan().
	 */
	beginPan: function() {
		this._isPanning = true;
		this._panX = 0;
		this._panY = 0;
		this.onBeginPan();
	},
	
	/**
	 * Ends a pan transaction.
	 */
	endPan: function() {
		this._isPanning = false;
		this.onEndPan(this._panX, this._panY);
	},

	/**
	 * for listeners
	 */
	onBeginPan: function() {
		dojo.publish("/gfx/move/start");
	},
	
	isPanning: function() {
		return this._isPanning;
	},
	
	/**
	 * for listeners
	 */
	onEndPan: function(totalX, totalY) {
		dojo.publish("/gfx/move/stop");
	},
	
	getZoom: function() {
		return this._getContentsTransform().xx;
	},
	
	setZoom: function(scale) {
		// given the current zoom of the diagram, 
		// how much do we have to zoom to get to the new scale
		var zoom = scale/ this.getZoom();
		var bounds = dojo.position(this.container);

		var cx = bounds.w / 2;
		var cy = bounds.h / 2;
		if (this.viewer && this.viewer.diagramLayer.outline.isUsingFixedDiagram()) {
			cx = cy = 0;
		}

		this.zoomAt(zoom, cx, cy, 200, dojo.fx.easing.cubicOut);
	},
	
	simpleZoom: function(zoom) {
		zoom = scale / this.getZoom();
		this.zoomAt(zoom, 0, 0, 200, dojo.fx.easing.cubicOut);
	},
	
	/**
	 * Zoom by an incremental amount at a point in the viewport's coordinates. 
	 * 
	 * @param zoom multiplier (zoom in 10% = 1.10)
	 * @param x center x
	 * @param y center y
	 */
	zoomAt: function(zoom, cx, cy, duration, easing) {
		this.zoomAndPan(zoom, cx, cy, 0, 0, duration, easing);
	},
	
	zoomToBounds: function(x, y, width, height, duration, easing, padding) {
		var bounds = dojo.position(this.container);
		var zoom = Math.min((bounds.w - padding) / width, (bounds.h - padding) / height);
		var midTargetX = bounds.w / 2;
		var midTargetY = bounds.h / 2;
		var midSourceX = x + width / 2;
		var midSourceY = y + height / 2;
		this.zoomAndPan(zoom, midSourceX, midSourceY, midTargetX - midSourceX, midTargetY - midSourceY, duration, easing);
	},

	zoomAndPan: function(zoom, cx, cy, dx, dy, duration, easing) {
		var m = dojox.gfx.matrix,
		  stack;
		if(duration) {
			stack = [];
			if(dx || dy) {
				stack.push({ name: "translate", start: [0, 0], end: [dx, dy] });
			}
			if(zoom !== 1) {
				stack.push({ name: "scaleAt", start:[1, 1, cx, cy], end: [zoom, zoom, cx, cy] });
			}
			stack.push({name: "original"});
			var anim = dojox.gfx.fx.animateTransform({
				shape: this.contents,
				duration: duration,
				easing: easing,
				transform: stack
			});
			dojo.connect(anim, "onAnimate", this, "onViewportChanged");
			dojo.connect(anim, "onEnd", this, "onViewportFinished");
			anim.play();
		} else {
			var t = this._getContentsTransform();
			stack = [];
			if(dx || dy) {
				stack.push(m.translate(dx, dy));
			}
			if(zoom !== 1) {
				stack.push(m.scaleAt(zoom, zoom, cx, cy));
			}
			stack.push(t);
			t = m.multiply.apply(null, stack);
			this.contents.setTransform(t);
			this.onViewportChanged();
			this.onViewportFinished();
		}
	},
	
	onViewportChanged: function() {
		// for listeners
	},
	
	onViewportFinished: function() {
		// for listeners
	},
	
	/*
	 * Return the absolute offset of the canvas div
	 */
	getAbsoluteCanvasOffset: function() {
		var offset = dojo.position(this.container.parentNode);
		if (this.viewer && this.viewer.diagramLayer.outline.isUsingFixedDiagram()) {
			offset.x = offset.x - this.viewer.diagramLayer.domNode.scrollLeft;
			offset.y = offset.y - this.viewer.diagramLayer.domNode.scrollTop;
		}
		
		return offset;
	},
	
	/*
	 * Given a mouse coordinate - find the corresponding absolute coordinate on the
	 * diagram surface accounting for toolbar offset etc.
	 */
	getSurfaceAbsoluteMousePosition: function(mouseX, mouseY) {
		var offset = this.getAbsoluteCanvasOffset();
		return new rmpw.diagrams.render.util.Point(mouseX - offset.x, mouseY - offset.y);
	},
	
	/*
	 * Given a mouse coordinate - find the corresponding coordinate relative to the
	 * diagram surface accounting for toolbar offset etc.
	 * 
	 * TODO (MG): convert to matrix math
	 */
	getSurfaceRelativeMousePosition: function(mouseX, mouseY) {
		var m = dojox.gfx.matrix;
		var offset = this.getAbsoluteCanvasOffset();
		
		var surfaceMatrix = m.invert(this.contents.matrix);
		var tp = m.multiplyPoint(surfaceMatrix, mouseX - offset.x, mouseY - offset.y);
		
		return new rmpw.diagrams.render.util.Point(tp.x, tp.y);
	},
	
	getOffset: function() {
		return dojox.gfx.matrix.multiplyPoint(this.contents.matrix, 0, 0);
	},
	
	reset: function() {
		this.contents.setTransform({ xx: this.zoom, yy: this.zoom,  dx: 0, dy: 0 });
	},
	
	invert: function(){										
		var newMatrix = new dojox.gfx.matrix.Matrix2D();
		this.contents.setTransform( newMatrix );		
	},
	
	_getContentsTransform: function() {
		var t = this.contents.getTransform();
		if(!t) {
			t = new dojox.gfx.matrix.Matrix2D();
		}
		return t;
	}
});

