/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2009, 2011.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
(function() {
dojo.provide("rmps.core.client");

dojo.require("rmps.core.util.xhrUtil");
dojo.require("jazz.client.xhr");
dojo.require("jazz.app.context");
dojo.require("rmps.core.util.IRmpsConstants");
dojo.require("dojox.uuid.generateRandomUuid");

function addContext(data, sendProjectId, context) {
	/*
	 * private
	 * 
	 * Takes the XHR request data and adds baseline/stream context if there is one
	 * If the sendProjectId parameter is truthy, the projectID will be also appended to the request URL
	 * Unless opted out, the IBM internal header is added
	 */
	if (!context || context === undefined) {
		// Baseline
		context = jazz.app.context.get().baseline;
	}

	var params = {};

	if (context) {
		params[rmps.core.util.IRmpsConstants.BASELINE] = context;
	}

	// Add changeset if it exists
	// TODO: revisit if we want to only send the changeset for the services that will use it
	var changeset = jazz.app.context.get().changeset;
	if (changeset) {
		params[rmps.core.util.IRmpsConstants.CHANGESET] = changeset;
	}
	
	
	if (sendProjectId || (sendProjectId === undefined)) {
		// send by default or if specified
		params[rmps.core.util.IRmpsConstants.PROJECT_ID] = jazz.app.context.get().itemId;
	}

	rmps.core.util.xhrUtil.addParamater(data, params);

	data.headers = data.headers || {};

	if (!("X-ibm-rmps-internal" in data.headers)) {
		data.headers["X-ibm-rmps-internal"] = "true";
	}
}

/**
 * <p>
 * Adds a correlation ID to a request, allowing it to be tracked as it moves through the server.  
 * </p>
 * 
 * @param headers The headers for a request.
 */
function addCorrelationID(headers){
	// Random time based UUID
	var uuid = dojox.uuid.generateRandomUuid();
	
	// Set the correlation ID in the headers.
	headers["X-ibm-internal-dm-request-correlation-ID"] = uuid;
}


rmps.core.client = {
	
		xhrHead: function(data, sendProjectId /* bool? */, context /* String ? */) {
			/*
			 * This API is to be called just like the jazz.client.xhr call, but contains some additional optional parameters.
			 *   
			 * Performs a XHR HEAD adding the stream/baseline context implicitly.
			 * Adds the IBM Internal header unless the data already contains it (i.e you must opt out)
			 *  @param sendProjectId - optional: if false, the current projectID will not be sent (default is to send)
			 *  @param context - optional: a stream/baseline override, if not undefined, this context will be used
			 */
			addContext(data, sendProjectId, context);
			addCorrelationID(data.headers);
			return jazz.client.xhrHead(data);
		},
		
		xhrGet: function(data, sendProjectId /* bool? */, context /* String ? */) {
			/*
			 * This API is to be called just like the jazz.client.xhr call, but contains some additional optional parameters.
			 *   
			 * Performs a XHR GET adding the stream/baseline context implicitly.
			 * Adds the IBM Internal header unless the data already contains it (i.e you must opt out)
			 * 	@param sendProjectId - optional: if false, the current projectID will not be sent (default is to send)
			 *  @param context  a stream/baseline override, if not undefined, this context will be used
			 *  
			 *  example:
			 *  var def = rmps.app.core.util.xhrGet({
			 *			url: someURL,
			 *			handleAs: 'json',
			 *			headers: {
			 *				"Accept" : "application/json",
			 *			},
			 *			load: dojo.hitch(this, function(response, ioArgs) {
			 *				// work
			 *				return response;
			 *			}),
			 *		}, 
			 *		true); // send project ID
			 */
			addContext(data, sendProjectId, context);
			addCorrelationID(data.headers);
			return jazz.client.xhrGet(data);
		},
		
		xhrPut: function(data, sendProjectId /* bool? */, context /* String ? */) {
			/*
			 * This API is to be called just like the jazz.client.xhr call, but contains some additional optional parameters.
			 *   
			 * Performs a XHR PUT adding the stream/baseline context implicitly.
			 * Adds the IBM Internal header unless the data already contains it (i.e you must opt out)
			 * @param sendProjectId - optional: if false, the current projectID will not be sent (default is to send)
			 *  @param context  a stream/baseline override, if not undefined, this context will be used
			 */
			addContext(data, sendProjectId, context);
			addCorrelationID(data.headers);
			return jazz.client.xhrPut(data);
		},
		
		xhrPost: function(data, sendProjectId /* bool? */, context /* String ? */) {
			/*
			 * This API is to be called just like the jazz.client.xhr call, but contains some additional optional parameters.
			 *   
			 * Performs a XHR POST adding the stream/baseline context implicitly.
			 * Adds the IBM Internal header unless the data already contains it (i.e you must opt out)
			 * 	@param sendProjectId - optional: if false, the current projectID will not be sent (default is to send)
			 *  @param context  a stream/baseline override, if not undefined, this context will be used
			 */
			addContext(data, sendProjectId, context);
			addCorrelationID(data.headers);
			return jazz.client.xhrPost(data);
		},
		
		xhrDelete: function(data, sendProjectId /* bool? */, context /* String ? */) {
			/*
			 * This API is to be called just like the jazz.client.xhr call, but contains some additional optional parameters.
			 *   
			 * Performs a XHR DELETE adding the stream/baseline context implicitly.
			 * Adds the IBM Internal header unless the data already contains it (i.e you must opt out)
			 * 	@param sendProjectId - optional: if false, the current projectID will not be sent (default is to send)
			 *  @param context  a stream/baseline override, if not undefined, this context will be used
			 */
			addContext(data, sendProjectId, context);
			addCorrelationID(data.headers);
			return jazz.client.xhrDelete(data);
		}
			
};

}());

