/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.common.command.CommandStack");

dojo.declare("rmpw.common.command.CommandStack", null, {
	constructor: function(repository) {
		this.repository = repository;
		
		this._connectResetStack = dojo.connect(this.repository, "resetStack", this, this._resetCmdStack);
		this._resetCmdStack();
	},
	
	_resetCmdStack: function() {
		this.undoStack = [this.repository.getCurrentRevision()];
		this.undoStackPosition = 0;
	},
	
	destroy: function() {
		dojo.disconnect(this._connectResetStack);
		delete this._connectResetStack;
	},
	
	/*
	 * accessor determining if a command is currently being executed by the command stack
	 */
	isExecuting: function() {
		return this._executingSuccess != undefined;
	},
	
	execute: function(command, args) {
		var success;
		
		// check for command recursion
		if (this.isExecuting()) {
			success = command.execute(this, args);
			this._executingSuccess &= success;
		} else {
			this._executingSuccess = true;
			
			try {
				this._executingSuccess &= command.execute(this, args);
			} catch (e) {
				this._executingSuccess = false;
				console.error("command failed:", e, ", command:", command.declaredClass, ", args:", args);
			} finally {
				success = this._executingSuccess;
				this.postExecuteHook();
				delete this._executingSuccess;
			}
			
			if(success) {
				this._updateStack(command.description);
			} else {
				// revert to state before command executed
				this.repository.rollbackTo(this.undoStack[this.undoStack.length - 1]);
			}
		}
		return success;
	},
	
	/*
	 * accessor determining if undo or redo is currently being run by the command stack.
	 */
	isUndoRedoExecuting: function() {
		return this._undoredoActive != undefined;
	},
	
	commit: function(changeSet) {
		this.repository.commit(changeSet);	
		this._updateStack("Properties Popup Change");
	},
	
	_updateStack: function(description) {
		// if repository commit was delayed, don't update undo stack until repository has committed
		if (this.repository.wasDelayedCommit || (this._queuedDescriptions && this._queuedDescriptions.length>0)) {
			if (!this._connectRepositoryCommit) {
				this._queuedDescriptions = [];
				this._connectRepositoryCommit = dojo.connect(this.repository, "afterCommit", this, dojo.hitch(this, function(changeSet) {
					// add saved description to changeset
					changeSet.description = this._queuedDescriptions.shift();
					// if no more saved descriptions, kill this connection
					if (this._queuedDescriptions.length==0) {
						dojo.disconnect(this._connectRepositoryCommit);
						delete this._connectRepositoryCommit;
						delete this._queuedDescriptions;
					}
					this.undoStack.length = ++this.undoStackPosition;
					this.undoStack.push(changeSet);
				}));
			}
			this._queuedDescriptions.push(description);
		} else {
			// truncate redo segment of stack, and increment stack position for following push
			this.undoStack.length = ++this.undoStackPosition;
			
			// store revision marker to roll back to in case of undo
			var endingRevision = this.repository.getCurrentRevision();
			endingRevision.description = description;
			this.undoStack.push(endingRevision);
		}
	},
	
	undo: function() {
		var wasUndone = false;
		
		try {
			this._undoredoActive = true;
			if(this.undoStackPosition > 0) {
				var previousRevision = this.undoStack[this.undoStackPosition - 1];
				wasUndone = this.repository.rollbackTo(previousRevision);
				if(wasUndone) {
					this.undoStackPosition--;
				}
			}
		}
		catch (e) {
			this._executingSuccess = false;
			console.debug("Undo failed:", e);
		}
		finally {
			this.postExecuteHook();
			
			delete this._undoredoActive;
		}
		
		return wasUndone;
	},
	
	redo: function() {
		var wasRedone = false;
		
		try {
			this._undoredoActive = true;
			if((this.undoStackPosition + 1) < this.undoStack.length) {
				var nextRevision = this.undoStack[this.undoStackPosition + 1];
				wasRedone = this.repository.recommitTo(nextRevision);
				if(wasRedone) {
					this.undoStackPosition++;
				}
			}
		}
		catch (e) {
			this._executingSuccess = false;
			console.debug("Undo failed:", e);
		}
		finally {
			this.postExecuteHook();
			
			delete this._undoredoActive;
		}
		return wasRedone;
	},
	
	/*
	 * Hook that is called after execute, undo, redo to allow clients to do post-processing (through dojo.connect)
	 */
	postExecuteHook: function() {
		// clients can attach to this api to do post processing
	}
});