/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.common.command.CompoundCommand");

/**
 * Simple compound command which will contain an array of commands to execute
 * 
 * @author Steve Shaw
 */
dojo.declare("rmpw.common.command.CompoundCommand", null, {
	
	constructor: function() {
		this.commands = [];
	},

	/*
	 * Adds a command to the compound command
	 * input: a string representing an expected input from a previous command execution.  
	 * It should match an input string previously specified.
	 * input :is a set of properties which map to property name with-in the command.
	 * chainedInput : same format as input except translated from a previous command's output
	 * output: a object representing the output of the command.  Can be a set of string 
	 * properties on the output object.  If any of the output key matches any of the 
	 * input of a future executing command, then it the result from that command will be
	 * fed as input to the future command.  
	 */
	add: function(command, input, chainedInput, output) {
		this.commands.push(command);
		command.__input = input;
		command.__chainedInput = chainedInput;
		command.__output = output;
	},
	
	execute: function(commandStack, args) {
		var success = true;
		var prevResult = null;
		var blackboard = {};
		for (var i=0, lg=this.commands.length; i<lg; i++) {
			var command = this.commands[i];
			
			// transfer .__input to args
			var cmdArgs = {};
			for (var prop in args) {
				cmdArgs[prop] = args[prop];
			}
			
			if (command.__input) {
				for (var prop in command.__input) {
					cmdArgs[prop] = command.__input[prop];
				}
			}
			
			// copy blackboard arguments to command
			if (command.__chainedInput) {
				for (var name in command.__chainedInput) {
					for (var prop in blackboard) {
						if (command.__chainedInput[name] == prop) {
							cmdArgs[name] = blackboard[prop];
						}
					}
					
					// if the args wasn't set from the blackboard then check the passed in args
					if (!cmdArgs[name]) {
						for (var prop in args) {
							if (command.__chainedInput[name] == prop) {
								cmdArgs[name] = args[prop];
							}
						}
					}
				}
			}
			
			success &= command.execute(commandStack, cmdArgs);
			if (!success) {
				return false;
			}
			
			// copy output to blackboard
			if (command.__output && command.__output != null) {
				var output = command.getOutput();
				for (var prop in output) {
					for (var name in command.__output) {
						if (command.__output[name] == prop) {
							// if property already exists then convert / add new property to array
							if (blackboard[name]) {
								if (blackboard[name].length > 0) {
									blackboard[name].push(output[prop]);
								}
								else {
									blackboard[name] = [blackboard[name], output[prop]];
								}
							}
							else {
								blackboard[name] = output[prop];
							}
						}
					}
				}
			}
		};
		
		this.output = blackboard;
		return success;
	},
	
	getOutput: function() {
		return this.output;
	}
	
});