/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.common.command.ReparentElement");

dojo.require("rmpw.common.command.Command");

/**
 * Command that will reparent a graph node
 * 
 * @author John Swanke
 */
dojo.declare("rmpw.common.command.ReparentElement", rmpw.common.command.Command, {
	id: "rmpw.common.command.ReparentElement",
	
	updateChangeSet: function(changeSet, args) {
		// calculate node locations in new container
		var top = 1000000, left = 1000000;
		var points = {};
		for(var i = 0, len = args.nodes.length; i < len; i++) {
			var mappingToReparent = args.nodes[i].renderMapping;
			var pt = this._getElementPoint(mappingToReparent, args);
			pt.x+=args.dx;
			pt.y+=args.dy;
			var oldParentMapping = mappingToReparent.policies.graphNode.getGraphContainer().host;
			this._translateToContainer(oldParentMapping, pt, args, true);
			this._translateToContainer(args.newParentMapping, pt, args, false);
			points[mappingToReparent.id] = pt;
			if (left > pt.x) left = pt.x;
			if (top > pt.y) top = pt.y;
		}
		
		// make sure all locations are in container 
		var dX=0, dY=0;
		if (!args.newParentMapping.policies.graphContainer.isTopLevel()) {
			if (left < 10) dX = 10 - left;
			if (top < 20) dY = 20 - top;
		}
		
		// do the reparent and position
		var newParentElement = args.newParentMapping.getDefaultSource();
		var addChange = changeSet.createChange(newParentElement);
		var changeMap = {};
		var delChangeMap = {};
		var movedConnectorMap = {};
		var destroy = [];
		for(var i = 0, len = args.nodes.length; i < len; i++) {		
			var mappingToReparent = args.nodes[i].renderMapping;
			if (mappingToReparent.policies.graphEdge) continue;
			var elementToReparent = mappingToReparent.getDefaultSource();
			
			// create a change for deleting this node from its old container
			// may come from multiple containers so one change per source container
			var delChange;
			var oldParentMapping = mappingToReparent.policies.graphNode.getGraphContainer().host;
			if (oldParentMapping.id in delChangeMap) {
				delChange = delChangeMap[oldParentMapping.id];
			} else {
				var oldParentElement = oldParentMapping.getDefaultSource();
				delChange = delChangeMap[oldParentMapping.id] = changeMap[oldParentMapping.id] = changeSet.createChange(oldParentElement);
			}
			
			// move shapes between containers
			delChange.removeArray(args.ownerChildrenProperty, elementToReparent);
			addChange.push(args.ownerChildrenProperty, elementToReparent);
			
			// if the moved shape also owns edges, move those edges to new container
			var graphAnchorable = mappingToReparent.policies.anchorable;
			if (graphAnchorable) {
				if (graphAnchorable.ownedConnections) {
					for (var j=0, edgeMappingToReparent; edgeMappingToReparent=graphAnchorable.ownedConnections[j]; j++) {
						var edgeToReparent = edgeMappingToReparent.getDefaultSource();
						delChange.removeArray(args.connectionOwnerChildrenProperty, edgeToReparent);
						addChange.push(args.connectionOwnerChildrenProperty, edgeToReparent);
						movedConnectorMap[edgeMappingToReparent.id] = true;
						
						// remove breakpoints from connections
						// TODO
						
					}
				}
			}
			
			// position shapes in new container
			// inject the x and y positions
			var newMappingState = new rmpw.common.transform.MappingState(mappingToReparent, true);
			var pt = points[mappingToReparent.id];
			newMappingState.state.x = pt.x + dX;
			newMappingState.state.y = pt.y + dY;
			elementToReparent.__inactiveMappings = [newMappingState];
		}

		// refresh the edges which target the moved nodes
		changeMap[args.newParentMapping.id] = addChange;
		this._refreshTargets(args.nodes, movedConnectorMap, changeSet, changeMap, args.connectionOwnerChildrenProperty);

		// destroy now so it doesn't affect layout policies
		while (destroy.length>0) {
			destroy.pop().destroy();
		}
		
		return true;
	},
	
	// refresh target connectors on reparented node as well as target connectors on all child nodes
	_refreshTargets: function(nodes, movedConnectorMap, changeSet, changeMap, property) {
		if (nodes) {
			for(var i = 0, len = nodes.length; i < len; i++) {
				var node = nodes[i];
				var mapping = node.renderMapping ? node.renderMapping : node;
				if (!mapping.policies.graphEdge && mapping.policies.anchorable) {
					var graphAnchorable = mapping.policies.anchorable;
					if (graphAnchorable.targetConnections) {
						for (var j=0, edgeMappingToRefresh; edgeMappingToRefresh=graphAnchorable.targetConnections[j]; j++) {
							if (!(edgeMappingToRefresh.id in movedConnectorMap) && edgeMappingToRefresh.policies.graphElement) {
								// we need to refresh the container that this edge lives in
								var edgeContainer = edgeMappingToRefresh.policies.graphElement.getGraphContainer().host;
								var change = null;
								if (edgeContainer.id in changeMap) {
									change = changeMap[edgeContainer.id];
								} else {
									var edgeContainerElement = edgeContainer.getDefaultSource();
									change = changeSet.createChange(edgeContainerElement);
									changeMap[edgeContainer.id] = change;
								}
								var edgeToRefresh = edgeMappingToRefresh.getDefaultSource();
								change.pushRefresh(property, edgeToRefresh);
							}
						}
					}
				}
				this._refreshTargets(mapping.children, movedConnectorMap, changeSet, changeMap, property);	
			}
		}
	},
	
	_translateToContainer: function (containerMapping, pt, args, absolute) {
		var walker = containerMapping;
		while (walker.policies.graphContainer && !walker.policies.graphContainer.isTopLevel()) {
			var delta = this._getElementPoint(walker, args);
			if (delta) {
				if (absolute) {
					pt.x+=delta.x;
					pt.y+=delta.y;
				} else {
					pt.x-=delta.x;
					pt.y-=delta.y;
				}
			}
			walker = walker.policies.graphNode.getGraphContainer().host;			
		}
	},
	
	_getElementPoint: function(mapping, args) {
		var x = mapping.traces[args.xTraceName].get();					
		var y = mapping.traces[args.yTraceName].get();						
		return { x :x, y : y};
	}
	
	
});
