/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010, 2013.  All Rights Reserved.
 * 
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
(function() {
dojo.provide("rmps.ui.editor.EditorMRU");

dojo.require("dojo.i18n");
dojo.require("rmps.ui.editor.EditorTracker");

dojo.requireLocalization("rmps.ui.common", "CommonMessages");

var commonMessages = dojo.i18n.getLocalization("rmps.ui.common", "CommonMessages");

dojo.declare("rmps.ui.editor.EditorMRU", rmps.ui.editor.EditorTracker, {
	
	// mixin
	maxRecent: 5,
	closeOldestRecentlyViewed: true,
	
	// private
	_recentlyViewedCount: 0,
	_unsavedWorkCount: 0,
	_recentlyViewed: null, //{}
	_unsavedWork: null, //{}
	_editorHandles: null, //{}
	
	constructor: function(mixin) {
		if(mixin){
			dojo.mixin(this, mixin);			
		}
		
		this._editorHandles = {};
		this._recentlyViewed = {};
		this._unsavedWork = {};
	},
	
	/*****************************************************(Start) PageHelper Listeners **********************************************************/
	onEditorLoading: function(editorInput) {
	},
	
	onEditorLoadFailed: function(editorInput, error) {
		this._handleEditorClosed(editorInput);
	},
	
	onEditorClosed: function(editor) {
		this._handleEditorClosed(editor.getEditorInput());
	},
	
	onEditorOpened: function(editor) {
		var editorInput = editor.getEditorInput();
		
		if (editorInput.isRecentlyViewedSupported()) {
			this._addRecentlyViewed(editorInput);
		}
		
		this._editorHandles[editorInput.getId()] = [
			dojo.connect(editor, "dirtyStateChanged", this, function(isDirty) {
				this._handleEditorDirtied(editor, isDirty);
			})
		];
		
		this._updateEditor(editor);
	},
	
	onEditorActivated: function(editor) {
		this._updateEditor(editor);
	},
	/*****************************************************(End) PageHelper Listeners ************************************************************/
	
	
	/*****************************************************(Start Defining) EditorTracker Methods ************************************************/
	getUnsavedWorkCount: function() {
		return this._unsavedWorkCount;
	},
	
	getRecentlyViewedCount: function() {
		return this._recentlyViewedCount;
	},
	
	getRecentlyViewed: function() {
		var result = [];
		for (var id in this._recentlyViewed) {
			result.push(this._recentlyViewed[id].editorInput);
		}
		return result;
	},
	
	getUnsavedWork: function() {
		var result = [];
		for (var id in this._unsavedWork) {
			result.push(this._unsavedWork[id].editorInput);
		}
		return result;
	},
	/*****************************************************(End Defining) EditorTracker Listeners ************************************************/
	
	_handleEditorDirtied: function(editor, isDirty) {
		var editorInput = editor.getEditorInput();
		
		if (isDirty) {
			this._removeRecentlyViewed(editorInput);
		} else {
			this._removeUnsavedWork(editorInput);
		}
		
		this._updateEditor(editor);
	},
	
	_updateEditor: function(editor) {
		var editorInput = editor.getEditorInput();
		
		if (editor.isDirty()) {
			this._addUnsavedWork(editorInput);
		} else if (editorInput.isRecentlyViewedSupported()){
			this._addRecentlyViewed(editorInput, true);
		}
	},
	
	_handleEditorClosed: function(editorInput) {
		var id = editorInput.getId();
		
		var handles = this._editorHandles[id];
		if (handles) {
			dojo.forEach(handles, dojo.disconnect);
			delete this._editorHandles[id];
		}
		
		this._removeRecentlyViewed(editorInput);
		this._removeUnsavedWork(editorInput);
	},
	
	_addRecentlyViewed: function(editorInput) {
		var item = this._recentlyViewed[editorInput.getId()];
		if (!item) {
			item = {};
			this._recentlyViewed[editorInput.getId()] = item;
			this._recentlyViewedCount++;
			this.onAddRecentlyViewed(editorInput);
		} else {
			this.onUpdate(editorInput);
		}
		
		item.editorInput = editorInput;
		item.lastVisit = new Date().getTime();
		
		if (this._recentlyViewedCount > this.maxRecent) {
			this._removeOldestRecentlyViewed();
		}
	},
	
	_removeOldestRecentlyViewed: function() {
		var oldestVisit = new Date().getTime();
		var oldest;
		for (var key in this._recentlyViewed) {
			if (this._recentlyViewed[key].lastVisit < oldestVisit) {
				oldest = key;
				oldestVisit = this._recentlyViewed[key].lastVisit;
			}
		}
		if (this.closeOldestRecentlyViewed) {
			this.getPageHelper().closeEditorInput(this._recentlyViewed[oldest].editorInput, true);
		}  else {
			this._removeRecentlyViewed(this._recentlyViewed[oldest].editorInput);
		}
	},
	
	_addUnsavedWork: function(editorInput) {
		var item = this._unsavedWork[editorInput.getId()];
		if (!item) {
			item = {};
			this._unsavedWork[editorInput.getId()] = item;
			this._unsavedWorkCount++;
			this.onAddUnsavedWork(editorInput);
		} else {
			this.onUpdate(editorInput);
		}
		
		item.editorInput = editorInput;
	},
	
	_removeRecentlyViewed: function(editorInput) {
		var id = editorInput.getId();
		if (this._recentlyViewed[id]) {
			this._recentlyViewedCount--;
			delete this._recentlyViewed[id];
			this.onRemoveRecentlyViewed(editorInput);
		}
	},
	
	_removeUnsavedWork: function(editorInput) {
		var id = editorInput.getId();
		if (this._unsavedWork[id]) {
			this._unsavedWorkCount--;
			delete this._unsavedWork[id];
			this.onRemoveUnsavedWork(editorInput);
		}
	}
	
	
});

})();
