/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010, 2014.  All Rights Reserved.
 * 
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
(function() {
dojo.provide("rmps.ui.editor.PageHelper");

dojo.require("dojox.collections.Dictionary");
dojo.require("rmps.ui.editor.EditorRegistry");
dojo.require("rmps.ui.editor.EditorMRU");
dojo.require("dijit.layout._LayoutWidget");
dojo.require("jazz.util.Navigation");
dojo.require("net.jazz.ajax.ui.PlatformUI");

dojo.requireLocalization("rmps.ui.editor", "PageHelperMessages");
var messages = dojo.i18n.getLocalization("rmps.ui.editor", "PageHelperMessages");

var PlatformUI= net.jazz.ajax.ui.PlatformUI;

var pageHelperInstances = {};
/*
 * PageHelper is used to provide a reusable way between different pages and menus to track added, unsaved, etc items.
 * It's expected that an application page can initialize one PageHelper based on its given page key. In addition, the associated component menu for that
 * page must use the same page helper by passing the same page key. 
 * 
 * The way that a page or menu makes a call to get instance of PageHelper requires passing the tracker class (@see rmps.ui.editor.EditorTracker), and the page key
 * value. Thus, a PageHelper instance will be constructed with the required rmps.ui.editor.EditorUtil. And then, will create an instance of the given tracker
 * After that, the required listeners on that tracker will be associated to the EditorUtil instance that was created. After that a page or menu will use that
 * PageHelper instance to track the required changes or to get the instance of the editor util when required.
 * Important to mention that the same PageHelper MUST be used for page and its component menu. If a component menu is using a different instance for page helper
 * other than what its page is using, this means that will end up with untracked events on each side. 
 * 
 * @see: rmps.ui.editor.EditorMRU
 *      It provides an implementation for rmps.ui.editor.EditorTracker, an has the required methods implemntation to add/remove items to a given page or menu.
 *      Example to decide getting an instance of EditorMRU when getting a PageHelper instance:
 *          rmps.ui.editor.PageHelper.getInstance("MyPageKey", rmps.ui.editor.EditorMRU);
 * 
 * A client can provide similar implementation for rmps.ui.editor.EditorMRU when required.
 *
 * We can add listeners on the PageHelper instance at the place that's calling at as follows:
 *          this.connect(this.pageHelper.editorTracker, "onAddRecentlyViewed", this._onAddRecentlyViewed);
 *			this.connect(this.pageHelper.editorTracker, "onAddUnsavedWork", this._onAddUnsavedWork);
 *			this.connect(this.pageHelper.editorTracker, "onRemoveRecentlyViewed", this._onRemoveRecentlyViewed);
 *			this.connect(this.pageHelper.editorTracker, "onRemoveUnsavedWork", this._onRemoveUnsavedWork);
 *			this.connect(this.pageHelper.editorTracker, "onUpdate", this._onUpdate);
 *			this.connect(this.pageHelper, "onLastViewedChanged", this._onLastViewedChanged);
 * 
 * Any place (not necessarily a page or menu) can use PageHelper, and use the pre-mentioned listeners. 
 * 
 * OR
 * 
 * We if a widget needs to listen on PageHelper tracker, it can extend the class rmps.ui.editor.EditorTrackingWidget. Then, make a call for the method startListeningOnEditorTracker
 *  @see: rmps.app.ui.internal.pages.SideNavigation for example
 * 
 */
dojo.declare("rmps.app.ui._PageHelper", null, {

	// private properties
	_currentView: null, //{}
	_editorView: null,
	_editorCache: null, // Dictionary
	_editorTracker: null,

	started: false,
	lastSearch: null,
	lastViewed: null,
	editorTracker: null,
	pageId: null,
	defaultUri: null,

	constructor: function(mixin){
		/*No need to check if mixin is null or not as we pass the required params at the getInstance method */
		dojo.mixin(this, mixin);

		this._currentView = {};
		this._editorCache = new dojox.collections.Dictionary();
		if (registerNavigateHandler) {
			registerNavigateHandler();
			registerNavigateHandler = null;
		}
	},

	// eventing

	onEditorLoading: function(editorInput) {},

	onEditorLoadFailed: function(editorInput, error) {
		window.location.href = this.defaultUri;
	},

	onEditorClosed: function(editor) {},

	onEditorOpened: function(editor) {},

	onActiveEditorChanged: function(editor) {},

	onEditorResize: function(size){},

	onEditorDeactivated: function(editor) {},

	/*Could be connected by other clients. The implemntation here is to set the browser title when the editor gets activated */
	onEditorActivated: function(editor) {
		editor.setTitle();
		this._editorView.resizeView();
	},

	// public API

	setEditorView: function(page){
		var view= page.getView();
		this._editorView = new rmps.ui.editor.PageHelper._EditorView({
			_helper: this
		});
		if(view.addChild){
			view.addChild(this._editorView);						
		}else{
			page.getContainerNode().appendChild(this._editorView.domNode);
		}
		this._editorView.startup();
	},

	setDefaultUri: function(defaultUri){
		this.defaultUri= defaultUri;
	},

	getActiveEditorInput: function() {
		return this._currentView.editorInput;
	},

	getActiveEditor: function() {
		return this._currentView.editor;
	},

	getEditorView: function(){
		return this._editorView;
	},

	hideActiveEditor: function(){
		var editor= this.getActiveEditor();
		if(editor){
			if (editor.deactivate) {
				editor.deactivate();
			}
			editor.domNode.style.display = "none";
			this.onEditorDeactivated(editor);
		}
		this._currentView= {};
	},

	openEditor: function(editorInput) {
		if (this.getActiveEditorInput() && editorInput.getId() === this.getActiveEditorInput().getId()) {
			this.getActiveEditorInput().setParams(editorInput.getParams());
			if (this._currentView.editor) {
				// the editor for the given resource URI is initialized
				if (this.getActiveEditor().activate) {
					this.getActiveEditor().activate();
				}
				this._onEditorActivated(this._currentView.editor);
				return;
			}
		}

		this.hideActiveEditor();

		var item = this._editorCache.item(editorInput.getId());

		if (item) {
			this._editorCacheUpdate(editorInput.getId(), item);
			item.editorInput.setParams(editorInput.getParams());

			this._currentView.editorInput = item.editorInput;

			if (item.editor) {
				this._currentView.editor = item.editor;

				this.getActiveEditor().domNode.style.display = "";

				var started = this.getActiveEditor()._started;
				if (!started) {
					this.getActiveEditor().startup();
				}
				if (this.getActiveEditor().activate) {
					this.getActiveEditor().activate();
				}
				this._onEditorActivated(this._currentView.editor);
				
				this.onActiveEditorChanged(this._currentView.editor);
				
				if (started && this.getActiveEditor().onReopen) {
					/*
					 * If editor is just being re-opened call onReopen on an editor 
					 * if it's defined for it to give editor a chance to refresh 
					 * UI components if needed
					 */
					this.getActiveEditor().onReopen();
				}

			} else {
				this._onEditorLoading(item.editorInput);
			}

			return;
		}

		this._currentView.editorInput = editorInput;

		item = this._cacheEditor(editorInput);

		this._onEditorLoading(editorInput);

		if (editorInput.isLoaded()) {
			setTimeout(dojo.hitch(this, function() {
				this._openEditor(editorInput);
			}), 0);
		} else {
			var deferred = editorInput.load();
			item.loadDeferred = deferred;

			deferred.addCallback(this, function() {
				delete item.loadDeferred;
				this._openEditor(editorInput);
			});

			deferred.addErrback(this, function(error) {
				delete item.loadDeferred;
				this._editorCache.remove(editorInput.getId());
				this.onEditorLoadFailed(editorInput, error);
				// Error was dispatched in "editorInput.load" function
			});
		}
	},

	closeEditor: function(editor, force, doNotOpenPreviousEditor) {
		return this.closeEditorInput(editor.getEditorInput(), force, doNotOpenPreviousEditor);
	},

	closeEditorInput: function(editorInput, force, doNotOpenPreviousEditor) {
		return this._closeEditorInput(editorInput, force, doNotOpenPreviousEditor);
	},

	/*23225: Investigate why closeAllEditors is commented out */
	closeAllEditors: function() {
		// TODO
//		var values = this._editorCache.getValueList();
//		for (var i = 0; i < values.length; ++i) {
//			// TODO
//		}
//		this._editorCache.clear();
//		
//		if (this._navigateHandle) {
//			this._unregisterNavigateHandler();
//		}
	},

	getEditor: function(editorInput) {
		if (editorInput) {
			var item = this._editorCache.item(editorInput.getId());
			if (item) {
				return item.editor;
			}
		} else if (this._currentView) {
			return this._currentView.editor;
		}
		return null;
	},

	// private

	_closeEditorInput: function(editorInput, force, doNotOpenPreviousEditor) {
		var item = this._editorCache.item(editorInput.getId());

		if (item && (!item.editor || !item.editor.isDirty() || force ||
				confirm(dojo.string.substitute(messages.confirmCloseIfDirty, [editorInput.getName()])))) {

			this._editorCache.remove(editorInput.getId());

			var editor = item.editor;

			var currentInputId = this.getActiveEditorInput() && this.getActiveEditorInput().getId();
			var areInputsEqual = currentInputId === editorInput.getId();
			var openNewEditor = !doNotOpenPreviousEditor && areInputsEqual;

			if (editor) {
				if (areInputsEqual) {
					if (editor.deactivate) {
						editor.deactivate();
					}
					this.onEditorDeactivated(editor);
					this._currentView = {};
				}
				// Can't remove from DOM before destroy. CKEditor for example does
				// clean-up work in the DOM when being destroyed.
				// this.getEditorView().domNode.removeChild(domNode);
				editor.destroyRecursive();
				this.onEditorClosed(editor);
			} else if (item.loadDeferred) {
				item.loadDeferred.cancel();
				delete item.loadDeferred;
			}

			editorInput.unload();

			if (openNewEditor) {
				var items = this._editorCache.getValueList();
				if (items.length > 0) {
					window.location.href = items[items.length - 1].editorInput.getPresentationUri();
				} else if (this.defaultUri){
					window.location.href = this.defaultUri;
				}
			}

			return true;
		}
		return false;
	},

	_openEditor: function(editorInput) {
		var editorDeferred = rmps.ui.editor.EditorRegistry.createEditor(editorInput);
		
		editorDeferred.then(dojo.hitch(this, function(editor) {
			if (!editor) {
				var error = new Error(dojo.string.substitute(messages.error_noEditor, [editorInput.getName()]));
				this.onEditorLoadFailed(editorInput, error);
				rmps.core.error.createError(error,
					dojo.string.substitute(messages.error_noEditor_action, [editorInput.getName()])).dispatch();
				return;
			}

			var editorView= this.getEditorView();

			if(editor.resize){
				/*Editor should be resize as long as its view is resized */
				editor.connect(this, "onEditorResize", dojo.hitch(editor, editor.resize));	
			}

			var item = this._editorCache.item(editorInput.getId());
			item.editor = editor;

			if (this.getActiveEditorInput() === editorInput) {
				this._currentView.editor = editor;
				editorView.domNode.appendChild(editor.domNode);
				editor.startup();
				this.onEditorOpened(editor);
				if (editor.activate) {
					editor.activate();
				}
				this._onEditorActivated(editor);
				this.onActiveEditorChanged(editor);		
			} else {
				editor.domNode.style.display = "none";
				editorView.domNode.appendChild(editor.domNode);
				this.onEditorOpened(editor);
			}	
		}));
	},

	_onEditorActivated: function(editor){
		this.getEditorView().domNode.style.display= "";
		this._updateEditorLastViewed(editor.getEditorInput());
		this.onEditorActivated(editor);
		if (window.onresize) {
			window.onresize({});
		}
	},

	_onEditorLoading: function(editorInput){
		this._updateEditorLastViewed(editorInput);
		this.onEditorLoading(editorInput);
	},

	_cacheEditor: function(editorInput) {
		var item = {
			editorInput: editorInput
		};
		this._editorCache.add(editorInput.getId(), item);
		return item;
	},

	_editorCacheUpdate: function(key, value) {
		this._editorCache.remove(key);
		this._editorCache.add(key, value);
	},

	init: function(/*rmps.ui.editor.PageInitializer */ page) {
		if(this.started){
			return;
		}
		var editorTrackerClass= page.getEditorTrackerClass();
		if(!editorTrackerClass){
			/*Use the default tracker class if editorTrackerClass is undefined*/
			editorTrackerClass= rmps.ui.editor.EditorMRU;
		}

		try {
			this.editorTracker = new editorTrackerClass({pageHelper: this});
		} catch (e) {
			console.error("must provide an instance of EditorTracker to init a PageHelper");
		}

		var prefix= page.getActionsPrefix()+ ".";
		var actionNames= page.getActionsList();

		var actionsRegistry= PlatformUI.getWorkbench().getActionRegistry();

		var actionsList= {};
		var publishFunc= function(){
			dojo.publish(arguments[0], [arguments[1]]);
		};

		for (var i = 0, action; action = actionNames[i]; ++i) {
			var actionName= prefix+ action;
			actionsList[action]= dojo.hitch(page, publishFunc, actionName);
			actionsRegistry.registerAction(actionName, actionsList, action);
		}

		this.setDefaultUri(page.getDefaultUri());

		/*Do not set this at startup in order to make sure that the only places that can get this flag set to rue is an instance of PageInitializer */
		this.started = true;

		this.startup();
	},

	startup: function() {/*To be connected by clients */},

	isStarted: function(){
		return this.started;
	},

	getLastSearch: function(){
		return this.lastSearch;
	},

	getLastViewed: function(){
		return this.lastViewed;
	},

	setSearch: function(queryResult, iconUri) {
		this.lastSearch = {
			queryResult: queryResult,
			iconUri: iconUri
		};

		this.lastViewed = {
			id: queryResult.id,
			object: this.lastSearch
		};
		this.onSearch(this.lastSearch);
		this.onLastViewedChanged(this.lastViewed);
	},

	onSearch: function(searchObj) {},

	_updateEditorLastViewed: function(editorInput) {
		this.lastViewed = {
			id: editorInput.getId(),
			object: editorInput
		};
		this.onLastViewedChanged(this.lastViewed);
	},

	clearLastViewed: function() {
		delete this.lastViewed;
		this.onLastViewedChanged();
	},

	getEditorTracker: function(){
		return this.editorTracker;
	},

	getSelected: function(items){
		var sel = [];

		if(!items|| items.length===0){
			return sel;
		}
		if (PlatformUI.getWorkbench().currentPageId === this.pageId) {
			var href = location.href;
			for (var i = 0; i < items.length; i++) {
				var item = items[i];

				if (item && ((item._id && this.lastViewed && item._id === this.lastViewed.id) || (item.href && href.indexOf(item.href) > 0))) {
					sel.push(item);
				}
			}
		}
		return sel;
	},

	onLastViewedChanged: function(object) {}
});

rmps.ui.editor.PageHelper = {

	/*
	 * Get an instance for PageHelper. 
	 */	
	getInstance: function(pageId) {
		if(!pageId){
			pageId= PlatformUI.getWorkbench().currentPageId;
		}
		if(!dojo.isString(pageId)){
			console.error("must provide a pageId in order to return an instance for page helper");
			return;
		}

		var instance= pageHelperInstances[pageId];
		if(instance){
			return instance;
		}
		return pageHelperInstances[pageId]= new rmps.app.ui._PageHelper({pageId: pageId});
	},
	
	/* 
	 * Returns <code>false</code> if no editor instance was dirty; otherwise, it returns <code>true</code> 
	 */
	isDirty: function(pageId){
		return getDirtyResourcesFor(rmps.ui.editor.PageHelper.getInstance(pageId)).length>0;
	},

	/*
	 * Used to update the dirty status for a given page id. It goes through all opened instances for that page id helper, and make sure to set the page status
	 * to dirty if a dirty instance was found. Otherwise, it sets it to false. 
	 */
	updateDirtyStatus: function(pageId){
		PlatformUI.getWorkbench().setPageDirty(pageId, rmps.ui.editor.PageHelper.isDirty(pageId));
	}
};


dojo.declare("rmps.ui.editor.PageHelper._EditorView", [dijit._Widget, dijit.layout._LayoutWidget], {
	// mixing properties
	_helper: null,

	// private properties
	_height: null,
	_resizeHandleId: null,

	postCreate: function() {
		this._setStyleAttr("overflow: hidden");
		if (dojo.isIE) {
			this.domNode.style.zoom = 1;
		}
		this._resizeHandleId = jazz.util.ResizeNotifier.register({context: this, callback: this.resizeView, force: true});
		this.inherited(arguments);
	},

	uninitialize: function(){
		jazz.util.ResizeNotifier.unregister(this._resizeHandleId);
		return this.inherited(arguments);
	},

	/*23226: PageHelper#EditorView: Do we need to have resize handling there? */
	resizeView: function(){
		// Intentionally disabled, see WI33654
	}
});

var registerNavigateHandler = function() {
	jazz.util.Navigation.registerNavigateHandler(dojo.hitch(this, function() {
		var dirtyResources = getDirtyResources(pageHelperInstances);
		if (dirtyResources.length > 0) {
			return dojo.string.substitute(messages.navigateMessage, [dirtyResources]);
		}
		return dirtyResources;
	}),
	function() {
		return false;
	});
};

var getDirtyResources= function(instances){
	var dirtyResources = "";

	var func= function(item){
		if (item.value.editor.isDirty()) {
			dirtyResources += "\n  " + item.value.editorInput.getName();
		}
	}

	for (var key in instances) {
		dirtyResources= getDirtyResourcesFor(instances[key], dirtyResources);
	}

	return dirtyResources;
};

var getDirtyResourcesFor= function(instance, dirtyResources){
	if(!dirtyResources){
		dirtyResources= "";
	}

	var func= function(item){
		if (item.value.editor && item.value.editor.isDirty()) {
			dirtyResources += "\n  " + item.value.editorInput.getName();
		}
	};

	instance._editorCache.forEach(func);

	return dirtyResources;
};
})();