/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010, 2013.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
dojo.provide("rmpw.diagrams.render.gfx.ScalingImage");

dojo.require('dojox.gfx.shape');
dojo.require("dojox.image");
dojo.require("rmps.core.error");
dojo.require("rmpw.diagrams.render.util.Image");

/**
 * Image Shape that scales.
 * 
 * :ScalingImage addresses the line thickness/fuzziness caused by just changing the width/height of the same bitmap image by:
 *    1) Using a new larger image when the shape gets above a certain size
 *    2) OR using an individual image per side of the shape when the shape gets even bigger
 *    
 *    takes one json string of this format:
 *    [
 *		{src:"/dm/web/rmpw.dsv.editor.sketcher/icons/dgm/oval.gif"},
 *		{w:155,h:90, src:"/dm/web/rmpw.dsv.editor.sketcher/icons/dgm/oval5.gif"}
 *		{w:200,h:250,src:
 *			[
 *				{h:50,t:"/dm/web/rmpw.dsv.editor.sketcher/icons/dgm/cylindertop.gif"},
 *				{h:18,b:"/dm/web/rmpw.dsv.editor.sketcher/icons/dgm/cylinderbot.gif"},
 *				{w:3,t:25,b:13,lr:"/dm/web/rmpw.dsv.editor.sketcher/icons/dgm/layvoutline.gif"}
 *			]
 *		}
 *	  ]
 *    where:
 *       the first source path is the base path, 
 *       the second source path is used if the image width OR height are greater then 155/90
 *       the third source path is an array of images used if width OR height are greater then 200/250
 *           within this array t/b/l/r/tb/lr specify the image to be used for the top/bottom/left/right/top AND bottom/left AND right sides
 *           if top or bottom, h is the fixed height of the image, width is set by the widht of the total shape
 *           if left or right, w is the fixed width of the image, height is set by the height og the total shape
 *           if lr, t and b are the fixed top and bottom of the image
 *
 * 
 * @author John Swanke
 */
dojo.declare("rmpw.diagrams.render.gfx.ScalingImage", dojox.gfx.Group, {

	// keep track of preloaded images
	loadedImages: [],

	constructor: function(){
		this.shape = dojo.clone(rmpw.diagrams.render.gfx.ScalingImage.defaultScalingImage);
	},
	
	setShape: function(shape) {
		if (shape.src && shape.src != this._originalSource) {
			delete this._originalSource;
		}
		this._setShape(shape);
	},
	
	resetImageSources: function(maintainAspectRatio, fillColor, lineColor) {
		var imgShape = this.getShape();
		if (!this._originalSource) {
			this._originalSource = imgShape.src;
		}
		
		var srcs = dojo.fromJson(this._originalSource); 
		for(var i = 0; i < srcs.length; i++){
			var src = srcs[i].src;
			if (dojo.isString(src)) {
				srcs[i].src = rmpw.diagrams.render.util.Image.getNewImageSource(src, maintainAspectRatio, fillColor, lineColor);
			}
			else {
				 for(i = 0; i < src.length; i++) {
                     var s = src[i];
                     if(dojo.isString(s.t)) {
                         s.t = rmpw.diagrams.render.util.Image.getNewImageSource(s.t, maintainAspectRatio, fillColor, lineColor);
                     } else if(dojo.isString(s.b)) {
                    	 s.b = rmpw.diagrams.render.util.Image.getNewImageSource(s.b, maintainAspectRatio, fillColor, lineColor);
                     } else if(dojo.isString(s.tb)) {
                    	 s.tb = rmpw.diagrams.render.util.Image.getNewImageSource(s.tb, maintainAspectRatio, fillColor, lineColor);
                     } else if(dojo.isString(s.lr)) {
                    	 s.lr = rmpw.diagrams.render.util.Image.getNewImageSource(s.lr, maintainAspectRatio, fillColor, lineColor);
                     }
                 }
			}
		}
		
		var newSrc = dojo.toJson(srcs);
		if (newSrc != imgShape.src) {
			imgShape.src = newSrc;
			this._setShape( imgShape );
		}
	},
	
	_setShape: function(shape){
		if (!shape.src) {
			shape.src = this.shape.src;
		}
		this.shape = dojo.mixin(dojo.clone(rmpw.diagrams.render.gfx.ScalingImage.defaultScalingImage), shape);
		var i;
		if (shape.src) {
			var newSrc = this._rawSrc !== shape.src;
			this._rawSrc = shape.src;
			this.srcs = dojo.fromJson(this._rawSrc); 
			// preload all images--even the ones not being used-- to avoid blank moments during resize when a new image is loaded
			var srcz = [];
			delete this._defaultImage;
			delete this._proportionalShapeMinSize;
			for(i = 0; i < this.srcs.length; i++){
				var src=this.srcs[i].src;
				if (dojo.isString(src)) {
					this._push(src, srcz);
					// if a proportional min size is declared, get default image size
					if (this.srcs[i].proportionalShapeMinSize) {
						this._proportionalShapeMinSize = this.srcs[i].proportionalShapeMinSize; 
					}
					if (this.srcs[i].customImage) {
						this._customImage = true; 
					} else if (this._customImage) {
						this._customImage = false;
					}
					
					// get default image size
					if (!this._defaultImage) {
						if (!dojo.isOffscreen) {
							this._defaultImage = dojo.create("img");
							dojo.attr(this._defaultImage,"src",src);
						}
						else {
							var callback = dojo.hitch(this, function(imageInfo) {
								this._defaultImage = {width: imageInfo.width, height: imageInfo.height, src: src};
							});
							rmpw.diagrams.render.util.Image.getImageInfo(src, callback);
						}
					}
					
				} else {
                    for(i = 0; i < src.length; i++) {
                        var s = src[i];
                        if(dojo.isString(s.t)) {
                            this._push(s.t, srcz);
                        } else if(dojo.isString(s.b)) {
                            this._push(s.b, srcz);
                        } else if(dojo.isString(s.tb)) {
                            this._push(s.tb, srcz);
                        } else if(dojo.isString(s.lr)) {
                            this._push(s.lr, srcz);
                        }
                    }
                }
			}
			
			// no need to preload images if rendering offscreen
			if (srcz.length>0 && !dojo.isOffscreen) {
				dojox.image.preload(srcz);
			}
			if (newSrc && this._lastWidth && this._lastHeight) {
				this._doScaling(this._lastWidth, this._lastHeight, this._lastContainer, true);
			}
		}
		if (shape.width || shape.height) {
			this._doScaling(this.shape.width, this.shape.height, this.shape.container);			
		}
		return this;
	},
	
	getSrc: function() {
		return this._rawSrc;
	},
	
	getProportionalShapeMinSize: function() {
		return this._proportionalShapeMinSize;
	},
	
	getDefaultImage: function() {
		return this._defaultImage;
	},
	
	isCustomImage: function() {
		return this._customImage;
	},
	
	_push: function(src, srcz) {
		if (dojo.indexOf(this.loadedImages, src) === -1) {
			srcz.push(src);
			this.loadedImages.push(src);
		}
	},

	_doScaling: function(width, height, container, force) {
        var th, bh, t, b;
        
        if(!force && width === this._lastWidth && height === this._lastHeight && container === this._lastContainer) {
            return;
        }
        this._lastWidth = width;
        this._lastHeight = height;
        this._lastContainer = container;

        if(this.srcs) {
            // start with base shape
            var src = this.srcs[0].src;

            // if a border...
            var x = 0, y = 0, i, s;
            if(this.srcs[0].w) {
                if(!this.srcs[1].src || width < height) {
                    x = width / 2;
                    this._borderWidth = width = this.srcs[0].w;
                } else {
                    src = this.srcs[1].src;
                    y = height / 2;
                    this._borderHeight = height = this.srcs[1].h ? this.srcs[1].h : 3;
                }
            } else {
                // else if there are image for bigger dimensions
                for(i = 1; i < this.srcs.length; i++) {
                    var image = this.srcs[i];
                    if(image.w && width > image.w || image.h && height > image.h) {
                        src = image.src;
                    } else if(image.container && container) {
                        src = image.src;
                    }
                }
            }

            if(!this._isSrcEqual(src, this._oldSrc)) {
                // remove all old image shapes
                if(this.children && this.children.length > 0) {
                    this.clear();
                }
                delete this._top;
                delete this._bottom;
                delete this._left;
                delete this._right;
                delete this._proportionalTop;
                delete this._proportionalTopMax;
                delete this._proportionalBot;
                delete this._proportionalBotMax;
                delete this._borderWidth;
                delete this._borderHeight;
                this._oldSrc = src;
                if(!dojo.isArray(src)) {
                    this._createImage(src, x, y, width, height);
                } else {
                    // do proportional tops and bottoms first to get sizes for sides
                   // var th, bh;
                    for(i = 0; i < src.length; i++) {
                        s = src[i];
                        if(s.t && s.pc) {
                            this._proportionalTop = s.pc;
                            this._proportionalTopMax = s.mx;
                            th = height * this._proportionalTop / 100;
                            if(this._proportionalTopMax && th > this._proportionalTopMax){
                                th = this._proportionalTopMax;
                            }
                            this._top = this._createImage(s.t, 0, 0, width, th);
                        } else if(s.b && s.pc) {
                            this._proportionalBot = s.pc;
                            this._proportionalBotMax = s.mx;
                            bh = height * this._proportionalBot / 100;
                            if(this._proportionalBotMax && bh > this._proportionalBotMax){
                                bh = this._proportionalBotMax;
                            }
                            this._bottom = this._createImage(s.b, 0, height - bh, width, bh);
                        }
                    }

                    // then do non-proportional sides
                    for(i = 0; i < src.length; i++) {
                        s = src[i];
                        if(s.tb && s.h) {
                            this._top = this._createImage(s.tb, 0, 0, width, s.h);
                            this._bottom = this._createImage(s.tb, 0, height - s.h, width, s.h);
                        } else if(s.lr && s.w) {
                            t = th ? th / 2 : 0;
                            b = bh ? bh : 0;
                            this._left = this._createImage(s.lr, 0, t, s.w, height - t - b);
                            this._right = this._createImage(s.lr, width - s.w, t, s.w, height - t - b);
                        }
                    }

                    // then do any icon
                    for(i = 0; i < src.length; i++) {
                        s = src[i];
                        if(s.icon) {
                            var w = 32, h = 32;
                            if(this._defaultImage && this._defaultImage.width !== 0 && this._defaultImage.height !== 0) {
                                if(this._defaultImage.width > this._defaultImage.height) {
                                    h = w * this._defaultImage.height / this._defaultImage.width;
                                } else {
                                    w = h * this._defaultImage.width / this._defaultImage.height;
                                }
                            }
                            this._icon = this._createImage(s.icon, 5, 5, w, h);
                        }
                    }

                }
            } else {
                if(!dojo.isArray(src)) {
                    if(this._borderWidth) {
                        this.children[0].setShape({
                            x : this._lastWidth / 2,
                            height : height
                        });
                    } else if(this._borderHeight) {
                        this.children[0].setShape({
                            y : this._lastHeight / 2,
                            width : width
                        });
                    } else {
                        this.children[0].setShape({
                            width : width,
                            height : height
                        });
                    }
                } else {
                   // var th, bh;
                    this._top.setShape({
                        width : width
                    });
                    if(this._proportionalTop) {
                        th = height * this._proportionalTop / 100;
                        if(this._proportionalTopMax && th > this._proportionalTopMax){
                            th = this._proportionalTopMax;
                        }
                        this._top.setShape({
                            height : th
                        });
                    }
                    this._bottom.setShape({
                        y : height - this._bottom.shape.height,
                        width : width
                    });
                    if(this._proportionalBot) {
                        bh = height * this._proportionalBot / 100;
                        if(this._proportionalBotMax && bh > this._proportionalBotMax){
                            bh = this._proportionalBotMax;
                        }
                        this._bottom.setShape({
                            height : bh
                        });
                    }
                    t = th ? th / 2 : 0;
                    b = bh ? bh : 0;
                    this._left.setShape({
                        y : t,
                        height : height - t - b
                    });
                    this._right.setShape({
                        x : width - this._right.shape.width,
                        y : t,
                        height : height - t - b
                    });
                }
            }
        }
    },
	
	_createImage: function(src, x, y, width, height) {
		var sh = this.shape;
		var imageShape = dojo.clone(dojox.gfx.defaultImage);
		for(var p in imageShape) {
			if (p !== "type" && sh[p]) {
				imageShape[p] = sh[p];
			}
		}
		imageShape.src = src;
		imageShape.x = x;
		imageShape.y = y;
		imageShape.width = width < 0 ? 1 : width;
		imageShape.height = height < 0 ? 1 : height;
		var shape =  this.createImage(imageShape);
		// control-click on image in FireFox opens new tab
		if (dojo.isFF) {
			shape.connect("onclick", this, function(e){
				dojo.stopEvent(e);
			});
		}
		return shape;
	},
	
	_isSrcEqual: function( src1, src2 ) {
		if (!src1 || !src2) {
			return false;
		} else if (!dojo.isArray(src1) && !dojo.isArray(src2)) {
			return src1 === src2;
		} else if (!dojo.isArray(src1) && dojo.isArray(src2) || dojo.isArray(src1) && !dojo.isArray(src2) ) {
			return false;
		} else if (src1.length !== src2.length) {
			return false;
		} else {
            for(var i = 0; i < src1.length; i++) {
                if(src1[i] !== src2[i]) {
                    return false;
                }
            }	
		
		}
		return true;
	}
	
});


/////////////////////////////// ADD createScalingImage() TO Surface/Group //////////////////////////////////////
rmpw.diagrams.render.gfx.ScalingImage.defaultScalingImage = dojo.mixin(dojo.clone(dojox.gfx.defaultImage),	{ type: 'scalingImage', width: -1, height: -1});
rmpw.diagrams.render.gfx.ScalingImage.nodeType = dojox.gfx.Group.nodeType;

function createScalingImage() {	
	var source = arguments[0].src;
	
	if (source && dojo.isString(source) && source[0] == '[' && source[source.length-1] == ']') {
		var shape = this.createObject(rmpw.diagrams.render.gfx.ScalingImage, arguments[0]);
		// Need to do vml-specific createGroup tasks
		// see dojox.gfx.vml.js/createGroup()
		if (!dojox.gfx.svg && !dojo.isOffscreen) {
			// Call Creator._overrideSize
			var node = shape.rawNode;
			this._overrideSize(node);
			
			// Need to do vml-specific createGroup tasks
			// see dojox.gfx.vml.js/createGroup()
			var r = shape.rawNode.ownerDocument.createElement("v:rect");
			r.style.left = r.style.top = 0;
			r.style.width = shape.rawNode.style.width;
			r.style.height = shape.rawNode.style.height;
			r.filled = r.stroked = "f";
			shape.rawNode.appendChild(r);
			shape.bgNode = r;
		}
		return shape;
	}
	
	// default to creating a regular image
	return this.createImage(arguments[0]);
}

dojo.extend(dojox.gfx.Surface, {createScalingImage: createScalingImage});
dojo.extend(dojox.gfx.Group,   {createScalingImage: createScalingImage});
