/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.diagrams.render.graph.AbstractEdgeAnchor");

/*
 * Default implementation for connection anchor that can be subclasses to ensure methods exist
 */
dojo.declare("rmpw.diagrams.render.graph.AbstractEdgeAnchor", null, {
	
	TERMINAL_START_CHAR: '(',
	TERMINAL_DELIMITER_CHAR: ',',
	TERMINAL_END_CHAR: ')',
	
	constructor: function(figure) {
		this.owner = figure;
	},
	
	getRelativeReferencePoint: function() {
		// subclasses should implement
	},
	
	setRelativeReferencePoint: function(relativeReference) {
		// subclasses should implement
	},
	
	getCopy: function() {
		// subclasses should implement
	},
	
	getOwner: function() {
		return this.owner;
	},
	
	getBox: function() {
		var rBox = this.getOwner().mapping.policies.graphElement.getPolygonPoints().getBounds();
		this.getOwner().translateToAbsolute(rBox);
        return rBox;
	},
	
	/**
	 * Creates a terminal string for any reference point passed in the format understandable by
	 * slidable anchors
	 */
	composeTerminalString: function() {
		if (this.isDefaultAnchor()) {
			return null;
		}
		
		var relativeReference = this.getRelativeReferencePoint();
		return this.TERMINAL_START_CHAR + 
			relativeReference.x + this.TERMINAL_DELIMITER_CHAR + 
			relativeReference.y + this.TERMINAL_END_CHAR;
	},
	
	/**
	 * Parses anchors terminal string and returns the relative reference incorporated
	 * in the terminal string
	 */
	parseTerminalString: function(terminal) {
		return new rmpw.diagrams.render.util.Point(
				parseFloat(terminal.substring(
						terminal.indexOf(this.TERMINAL_START_CHAR) + 1,
						terminal.indexOf(this.TERMINAL_DELIMITER_CHAR))),
				parseFloat(terminal.substring(
						terminal.indexOf(this.TERMINAL_DELIMITER_CHAR) + 1,
						terminal.indexOf(this.TERMINAL_END_CHAR))));
	},
	
	equals$: function (obj){
		if (obj.getRelativeReferencePoint && obj.getOwner && this.getOwner() === obj.getOwner()) {
			var objRelativeReference = obj.getRelativeReferencePoint();
			var relativeReference = this.getRelativeReferencePoint();
			if (relativeReference && objRelativeReference) {
				return relativeReference.x === objRelativeReference.x && 
				relativeReference.y === objRelativeReference.y;
			}
			
			return relativeReference === objRelativeReference;
		}
		
		return false;
	},
	
	hashCode$: function (){
		var owner = this.getOwner();
		var hashCode = owner ? owner.mapping.id : 0;
		var relativeReference = this.getRelativeReferencePoint();
		if (relativeReference) {
			hashCode = hashCode ^ relativeReference.hashCode$();
		}

		return hashCode;
	},
	
	/**
	 * Calculates the location of the anchor depending on the anchors own reference
	 * and foreign reference points
	 */
	_getLocation: function(ownReference, foreignReference) {
		var start = new rmpw.diagrams.render.util.Point(ownReference.x, ownReference.y);
		this.getOwner().translateToRelative(start);
		var end = new rmpw.diagrams.render.util.Point(foreignReference.x, foreignReference.y);
		this.getOwner().translateToRelative(end);
		var loc = this.getOwner().mapping.policies.graphElement.getClosestIntersectionPointToLine(start, end);
		if (loc) {
			this.getOwner().translateToAbsolute(loc);
		}
		return loc;
	},

	/**
	 * Calculates the relative location of the reference point with respect to the bounds
	 * of the figure. If point p is not inside of the figure's bounds then the point
	 * is mapped on the bounds and the point relative location is calculated 
	 */
	getAnchorRelativeLocation: function(p, bounds) {
		if (bounds.width === 0 || bounds.height === 0) {
			/*
			 * If figure hasn't been laid out yet, we don't want to fail the slidable anchor creation.
			 * Hence, we'll just return the (0.5, 0.5) meaning that the anchor reference point is the center of the figure.
			 */
			return new rmpw.diagrams.render.util.Point(0.5, 0.5);
		}
		var relLocation;
		var temp = {x:p.x, y:p.y};
		if (p.x < bounds.x || p.x > bounds.x + bounds.width
			|| p.y < bounds.y || p.y > bounds.y + bounds.height) {
			if (p.x < bounds.x || p.x > bounds.x + bounds.width) {
				temp.x = p.x < bounds.x ? bounds.x
					: bounds.x + bounds.width;
			}
			if (p.y < bounds.y || p.y > bounds.y + bounds.height) {
				temp.y = p.y < bounds.y ? bounds.y
					: bounds.y + bounds.height;
			}
			relLocation = new rmpw.diagrams.render.util.Point((temp.x - bounds.x)
				/ bounds.width, (temp.y - bounds.y)
				/ bounds.height);
		} else {

		relLocation = new rmpw.diagrams.render.util.Point((temp.x - bounds.x)
				/ bounds.width, (temp.y - bounds.y)
				/ bounds.height);
		}
		return relLocation;
	},
	
	/**
	 * Checks whether the point p which is a candidate for a relative reference
	 * belongs to the area where the default anchor must be created
	 */
	isDefaultAnchorArea: function(p) {
		return p.x >= 0.25 && p.x <= 1 - 0.25 && p.y >= 0.25 && p.y <= 1 - 0.25;
	},
	
	calculateRelativeReferencePoint: function(reference) {
		// check if the reference is in the tolerance area and if so - set a new reference point or reset it
		var relativeReference = this.getReferencePoint();
		if (this.isDefaultAnchor() || 
			reference.x !== relativeReference.x || 
			reference.y !== relativeReference.y) {
			
			var anchorRelPt = this.getAnchorRelativeLocation(reference, this.getBox());
			if (!anchorRelPt || this.isDefaultAnchorArea(anchorRelPt)) {
				// reset the relative reference to default
				this.setRelativeReferencePoint({x: 0.5, y: 0.5});
			}
			else {
				this.setRelativeReferencePoint(anchorRelPt);
			}
		}
	},
	
	getLocation: function(reference) {
		var tolerance = 5;
		var ownReference = this.normalizeToStraightlineTolerance(reference, this.getReferencePoint(), tolerance);
		
		var location = this._getLocation(ownReference, reference);
		if (!location) {
			location = this._getLocation(new rmpw.diagrams.render.util.Point(this.getBox().getCenter()), reference);
			if (!location) {
				// do chopbox
				var centerX, centerY, dx, dy, r, scale;
				r = new rmpw.diagrams.render.util.Rectangle(0, 0, 0, 0);
				r.setBounds(this.getOwner().getBounds());
				r.x += -1;
				r.y += -1;
				r.width += 1;
				r.height += 1;
				this.getOwner().translateToAbsolute(r);
				centerX = r.x + 0.5 * r.width;
				centerY = r.y + 0.5 * r.height;
				if (r.width <= 0 || r.height <= 0 || reference.x === ~~Math.max(Math.min(centerX, 2147483647), -2147483648) && reference.y === ~~Math.max(Math.min(centerY, 2147483647), -2147483648)) {
					return new rmpw.diagrams.render.util.Point( ~~Math.max(Math.min(centerX, 2147483647), -2147483648), ~~Math.max(Math.min(centerY, 2147483647), -2147483648));
				}
				dx = reference.x - centerX;
				dy = reference.y - centerY;
				scale = 0.5 / Math.max((dx < 0?-dx:dx) / r.width, (dy < 0?-dy:dy) / r.height);
				dx *= scale;
				dy *= scale;
				centerX += dx;
				centerY += dy;
				
				// straighten points
				location = new rmpw.diagrams.render.util.Point(centerX, centerY);
				if (Math.abs(reference.x - location.x) < tolerance) {
					location = new rmpw.diagrams.render.util.Point(reference.x, location.y);
                } else if (Math.abs(reference.y - location.y) < tolerance) {
                    location = new rmpw.diagrams.render.util.Point(location.x, reference.y);
                }
			}
		}
		
		return location;
	},

	normalizeToStraightlineTolerance: function(foreignReference, ownReference, tolerance) {
		var preciseOwnReference = new rmpw.diagrams.render.util.Point(ownReference.x, ownReference.y);
		var normalizedReference = preciseOwnReference.getCopy();
		var preciseForeignReference = new rmpw.diagrams.render.util.Point(foreignReference.x, foreignReference.y);
		if (Math.abs(preciseForeignReference.x - preciseOwnReference.x) < tolerance) {
			normalizedReference.x = preciseForeignReference.x;
			return normalizedReference;
		}
		if (Math.abs(preciseForeignReference.y - preciseOwnReference.y) < tolerance) {
			normalizedReference.y = preciseForeignReference.y;
		}
		
		return normalizedReference;
	},
	
	isDefaultAnchor: function() {
		var relativeReference = this.getRelativeReferencePoint();
		if (!relativeReference) {
			return true;
		}
		
		return relativeReference.x === 0.5 &&  relativeReference.y === 0.5;
	},
	
	/**
	 * From relative reference returns the relative coordinates of the anchor
	 * Method's visibility can be changed as needed
	 */
	getAnchorPosition: function() {
		var rBox = this.getBox();
		if (this.isDefaultAnchor()){
			return rBox.getCenter();
		}
		var relativeReference = this.getRelativeReferencePoint();
		return new rmpw.diagrams.render.util.Point(
				relativeReference.x * rBox.width + rBox.x,
				relativeReference.y * rBox.height + rBox.y);
	},
	
	getReferencePoint: function() {
		return this.getAnchorPosition();
	}
	
});