/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
dojo.provide("rmpw.diagrams.render.graph.AnimatedEdge");		

dojo.require("rmpw.diagrams.render.util.Point");
dojo.require("rmpw.diagrams.render.util.PointList");
dojo.require("rmpw.diagrams.render.graph.AnimatedElement");

/**
 * Stores state required for animating a connection
 * 
 * @author Steve Shaw
 */
dojo.declare("rmpw.diagrams.render.graph.AnimatedEdge", rmpw.diagrams.render.graph.AnimatedElement, {
	
	constructor: function(edgePolicy) {
		this.initialPoints = edgePolicy.getDojoPoints();
		this.edgePolicy = edgePolicy;
		this.normalizeRequired = false;
	},
	
	/* 
	 * method returns the initial state information of the object before the animation occurs
	 */
	getInitialState: function() {
		return this.initialPoints;
	},
	
	/* 
	 * method returns the final state information of the object before the animation occurs
	 */
	getFinalState: function() {
		return this.finalPoints;
	},
	
	startingAnimation: function() {
		this.finalPoints = this.edgePolicy.getPointList().getCopy();
		this.makePointsCompatible();
	},
	
	endAnimation: function() {
		if (this.normalizeRequired) {
			// get rid of redundant points created when making the lists compatible.
			var points = this.edgePolicy.getPointList();
			points.normalize(1);
			this.edgePolicy.setPointList(points);
		}
	},
	
	makePointsCompatible: function() {
		var points1 = this.getInitialState();
		var points2 = this.getFinalState();

		if (points1 && points1.size !== points2.size) {
			var p = new rmpw.diagrams.render.util.Point(), q = new rmpw.diagrams.render.util.Point();

			var size1 = points1.size - 1;
			var size2 = points2.size - 1;

			var i1 = size1;
			var i2 = size2;

			var current1 = 1.0;
			var current2 = 1.0;

			var prev1 = 1.0;
			var prev2 = 1.0;

			while (i1 > 0 || i2 > 0) {
				if (Math.abs(current1 - current2) < 0.1
						&& i1 > 0 && i2 > 0) {
					//Both points are the same, use them and go on;
					prev1 = current1;
					prev2 = current2;
					i1--;
					i2--;
					current1 = i1 / size1;
					current2 = i2 / size2;
				} else if (current1 < current2) {
					//2 needs to catch up
					// current1 < current2 < prev1
					points1.getPoint_0(p, i1);
					points1.getPoint_0(q, i1 + 1);

					p.x = (((q.x * (current2 - current1) + p.x * (prev1 - current2))
							/ (prev1 - current1)));
					p.y = (((q.y * (current2 - current1) + p.y * (prev1 - current2))
							/ (prev1 - current1)));

					points1.insertPoint(p, i1 + 1);
					this.normalizeRequired = true;
					
					prev1 = prev2 = current2;
					i2--;
					current2 = i2 / size2;

				} else {
					//1 needs to catch up
					// current2< current1 < prev2

					points2.getPoint_0(p, i2);
					points2.getPoint_0(q, i2 + 1);

					p.x = (((q.x * (current1 - current2) + p.x * (prev2 - current1))
							/ (prev2 - current2)));
					p.y = (((q.y * (current1 - current2) + p.y * (prev2 - current1))
							/ (prev2 - current2)));

					points2.insertPoint(p, i2 + 1);
					this.normalizeRequired = true;
					
					prev2 = prev1 = current1;
					i1--;
					current1 = i1 / size1;
				}
			}
		}
	},
	
	/**
	 * runs a step in the animation playback
	 */
	runAnimationStep: function(progress) {
		var list1 = this.getInitialState();
		var list2 = this.getFinalState();
		
		if (list1 && list2 && (list1.size === list2.size)) {
			var pt1 = new rmpw.diagrams.render.util.Point(), pt2 = new rmpw.diagrams.render.util.Point();
			
			var points = this.edgePolicy.getPointList();
			points.removeAllPoints();
			for (var i = 0; i < list1.size; i++) {
				list1.getPoint_0(pt2, i);
				list2.getPoint_0(pt1, i);
				pt1.x = Math.round(pt1.x * progress + (1 - progress) * pt2.x);
				pt1.y = Math.round(pt1.y * progress + (1 - progress) * pt2.y);
				points.addPoint_1(pt1);
			}
			
			this.edgePolicy.setPointList(points);
		}
		return true;
	}

});