/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010, 2013.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
dojo.provide("rmpw.diagrams.render.graph.BaseGraphConfigurator");
	
/**
 * A utility class for configuring a graph
 * 
 * @author Steve Shaw
 */
dojo.declare("rmpw.diagrams.render.graph.BaseGraphConfigurator", null, {
	
	IDENTITY_ROUTING: "identity",
	RECTILINEAR_ROUTING: "orthogonal",
	OBLIQUE_ROUTING: "oblique",
	TREE_ROUTING: "tree",
	
	FIXED_ANCHOR: "FixedAnchor",
	MOVEABLE_ANCHOR: "MoveableAnchor",
	
	/*
	 * Creates the graph data structure that contains the nodes and edges
	 */
	createGraph: function() {
		// do nothing, implemented by concrete graph configurators
	},
	
	/*
	 * Creates a graph node in the given graph
	 */
	createGraphNode: function(mapping, graph, rect) {
		// do nothing, implemented by concrete graph configurators
	},
	
	/*
	 * Create a graph edge between source and target nodes in the given graph
	 */
	createGraphEdge: function(mapping, graph, sourceNode, targetNode) {
		// do nothing, implemented by concrete graph configurators
	},
	
	/*
	 * Create an anchor owned by graphElement of a particular type - one of FIXED_ANCHOR,
	 * MOVEABLE_ANCHOR
	 */
	createAnchor: function(anchorType, graphElement, args) {
		// do nothing, implemented by concrete graph configurators
	},
	
	runGraphLayoutAlgorithm: function(graph, args) {
		// do nothing, implemented by concrete graph configurators
	},
	
	/*
	 * Assume self-relations are problematic for node layout by default
	 */
	_nodeLayoutSupportsSelfRelations: function() {
		return false;
	},
	
	/*
	 * Assume floating labels are problematic for node layout by default
	 */
	_nodeLayoutSupportsFloatingLabels: function() {
		return false;
	},
	
	/*
	 * Check whether layout algorithms support an incremental mode
	 */
	_nodeLayoutSupportsIncremental: function() {
		return false;
	},
	
	/*
	 * Gets the graph that the layout algorithm will run against
	 */
	_getGraph: function(args, nodes, edges) {
		var bbox = null,
		i,
		edge, node;
		if (!this._nodeLayoutSupportsIncremental() ||
			!this._nodeLayoutSupportsFloatingLabels() || 
			!this._nodeLayoutSupportsSelfRelations() || 
			args.resetGraphPosition) {
			var layoutGraph = this.createGraph();
			var nodeElements = { size: 0 };
			var numEdge = 0;
			var allNodeElements = {};
			
			for (i = 0, node; node = nodes[i]; i++) {	
				if (!this._nodeLayoutSupportsFloatingLabels()) {
					if (node.mapping.policies.graphFloatingLabel) {
						continue;
					}
				}
				
				if ((this._nodeLayoutSupportsIncremental() && this.incremental) ||
					!args.resetGraphPosition || (args.resetGraphPosition && node.nativeNode.isLayoutInvalid())) {
					var layoutNode = this.createGraphNode(node.mapping, layoutGraph, node.b);
					nodeElements[node.mapping.id] = layoutNode;
					nodeElements.size++; 
					var state = node.nativeNode.getLayoutState();
					if (!state) {
						bbox = this._updateBoundBox(node, bbox);
					}
					node.nativeNode.setLayoutState(null);
				}
				
				allNodeElements[node.mapping.id] = node;
			}
			
			// add all edges that are attached to the nodes to layout
			for (i=0, edge; edge = edges[i]; i++) {
				var source = nodeElements[edge.sId];
				var target = nodeElements[edge.tId];
				if (!source && !target) {
					continue;
				}
				
				if (!source) {
					source = allNodeElements[edge.sId];
					bbox = this._updateBoundBox(source, bbox);
					source = this.createGraphNode(source.mapping, layoutGraph, source.b);
					nodeElements[source.mapping.id] = source;
					nodeElements.size++; 
				}
				
				if (!target) {
					target = allNodeElements[edge.tId];
					if (target) {
						bbox = this._updateBoundBox(target, bbox);
						target = this.createGraphNode(target.mapping, layoutGraph, target.b);
						nodeElements[target.mapping.id] = target;
						nodeElements.size++;
					}
				}
				
				// don't allow self-transition in layout graph
				if (this._nodeLayoutSupportsSelfRelations() || (edge.source !== edge.target && edge.source && edge.target)) {
					this.createGraphEdge(edge.mapping, layoutGraph, source, target );
				}
			}
			
			return {graph: nodeElements.size > 1 ? layoutGraph : null, bbox: !bbox ? new rmpw.diagrams.render.util.Rectangle(0, 0, 0, 0) : bbox};	
		}
		else {
			return {graph: args.targetController.renderMapping ? args.targetController.renderMapping.policies.graphElement.getGraphContainer().getInternalGraphElement() :
					args.targetController.policies.graphElement.getGraphContainer().getInternalGraphElement(),
					bbox: this._calcBoundBox(nodes)};
		}
	},
	
	_updateBoundBox : function(node, bbox) {
		if (!bbox) {
			bbox = new rmpw.diagrams.render.util.Rectangle(node.b.x, node.b.y, node.b.w, node.b.h);
		} else {
			bbox.union_2(node.b.x, node.b.y, node.b.w, node.b.h);
		}
		
		return bbox;
	},
	
	_calcBoundBox : function(nodes) {
		var li, node, bbox, i;
		bbox = null;
		for (i = 0, node; node = nodes[i]; i++) {	
			bbox = this._updateBoundBox(node, bbox);
		}
		return !bbox ? new rmpw.diagrams.render.util.Rectangle(0, 0, 0, 0) : bbox;
	},

	_offsetParts : function(nodes, edges, origRect) {
		var li, newPostLayoutRect, node, postLayoutRect, translate, i, j, edge, point;
		postLayoutRect = this._calcBoundBox(nodes);
		
		if (origRect.isEmpty()) {
			// center the graph in the canvas
			node = nodes[0].mapping === "Head" ? nodes[1] : nodes[0];
			
			// check for top level diagram and if so adjust position to middle of canvas
			var offset = {x:10, y:10};
			if (node.mapping.parent.parent === node.mapping.root && !dojo.isOffscreen) {
				var d = node.mapping.viewer.getCanvasSize();
				offset = {x: d.w / 2 - postLayoutRect.width / 2, y: d.h / 2 - postLayoutRect.height / 2};
			}
			origRect = new rmpw.diagrams.render.util.Rectangle(offset.x, offset.y, postLayoutRect.width, postLayoutRect.height);
		}
		
		newPostLayoutRect = new rmpw.diagrams.render.util.Rectangle(
						origRect.x - ((postLayoutRect.width - origRect.width) / 2),
						origRect.y - ((postLayoutRect.height - origRect.height) / 2), 
						postLayoutRect.width, postLayoutRect.height);
		
		translate = {x: newPostLayoutRect.x - postLayoutRect.x, y: newPostLayoutRect.y - postLayoutRect.y};
		for (i = 0, node; node = nodes[i]; i++) {	
			node.b.x += translate.x;
			node.b.y += translate.y;
		}
		
		for (j=0, edge; edge = edges[j]; j++) {
			for (i=0, point; point=edge.pts[i]; i++) {
				point.x += translate.x;
				point.y += translate.y;
			}
		}
	},
	
	onCommit: function(graph, args, modifyNode, modifyEdge) {
		var proxyGraph = graph.getProxyGraph();
		var nodes = proxyGraph.nodes; // need to retrieve proxy nodes again after the layout
		var edges = proxyGraph.edges;
		
		// offset the graph if it's not an incremental layout
		if (!args.incremental || dojo.isOffscreen) {
			this._offsetParts(nodes, edges, args.__origRect);
		}
		
		if (this.commitResultFromGraph(modifyNode, nodes, modifyEdge, edges)) {
			return true;
		}
		
		return false;
	},
	
	/*
	 * Method that will take the current state of the graph and run passed in functions
	 * for modifyNode and modifyEdge to make corresponding changes potentially in the data model
	 * or a local cache.
	 */
	commitResultFromGraph: function(modifyNode, nodes, modifyEdge, edges) {
	    var node, edge;
		for (var i = 0; node = nodes[i]; i++) {
			if (node.mapping !== null && node.mapping !== 'Head' && node.mapping.policies) { // protect against deletion
				
				var graphNode = node.mapping.policies.graphNode;
				if (!graphNode) {
					continue;
				}
				
				modifyNode(graphNode, node.b.x, node.b.y);
			}
		}
		
		// now update the edges
		return this.commitEdges(modifyEdge, modifyNode, edges);
	},

	/*
	 * Method that will take the current state of the graph and run passed in functions
	 * for modifyEdge to make corresponding changes potentially in the data model
	 * or a local cache.
	 */
	commitEdges: function(modifyEdge, modifyNode, edges) {
		// now update the edges
		var edge,
		  insets,
		  j,
		  child, 
		  lg;
		for (var i=0; edge = edges[i]; i++) {
			var points = edge.pts;
			var sourceRef, targetRef;
			
			// protect against deletion
			if (!edge.mapping.policies) {
				continue;
			}
			
			var sourceBBox = new rmpw.diagrams.render.util.Rectangle(edge.source.b.x, edge.source.b.y, edge.source.b.w, edge.source.b.h);
			if (edge.source.mapping.policies.anchorable) {
				insets = edge.source.mapping.policies.anchorable.getAnchorInsets();
				sourceBBox = new rmpw.diagrams.render.util.Rectangle(sourceBBox.x + insets.left, 
						sourceBBox.y + insets.top, 
						sourceBBox.width - insets.left - insets.right, 
						sourceBBox.height - insets.top - insets.bottom);
			}
			else {
				sourceBBox =  new rmpw.diagrams.render.util.Rectangle(sourceBBox.x, sourceBBox.y, sourceBBox.width, sourceBBox.height);
			}
			
			var targetBBox = new rmpw.diagrams.render.util.Rectangle(edge.target.b.x, edge.target.b.y, edge.target.b.w, edge.target.b.h);
			if (edge.target.mapping.policies.anchorable) {
				insets = edge.target.mapping.policies.anchorable.getAnchorInsets();
				targetBBox = new rmpw.diagrams.render.util.Rectangle(targetBBox.x + insets.left, 
						targetBBox.y + insets.top, 
						targetBBox.width - insets.left - insets.right, 
						targetBBox.height - insets.top - insets.bottom);
			}
			else {
				targetBBox =  new rmpw.diagrams.render.util.Rectangle(targetBBox.x, targetBBox.y, targetBBox.width, targetBBox.height);
			}
			
			sourceRef = points[0];
			targetRef = points[points.length-1];
			
			var graphEdge = edge.mapping.policies.graphEdge;
			if (!graphEdge) {
				continue;
			}
			
			var bendpts = [];
			for (var j = 0, lg = points.length; j < lg; j++) {
				// create svg styled bendpoint string
				bendpts.push(points[j].x - sourceRef.x); //sourceX
				bendpts.push(points[j].y - sourceRef.y); //sourceY
				bendpts.push(points[j].x - targetRef.x); //targetX
				bendpts.push(points[j].y - targetRef.y); //targetY
			}
			var bendpoints = bendpts.join(" ");
			
			var sourceAnchor = sourceBBox.width > 0 && sourceBBox.width > 0 ? { x: (sourceRef.x - sourceBBox.x) / sourceBBox.width,
					y: (sourceRef.y - sourceBBox.y) / sourceBBox.height} : null;
			var targetAnchor = targetBBox.width > 0 && targetBBox.width > 0 ? { x:(targetRef.x - targetBBox.x) / targetBBox.width,
					y: (targetRef.y - targetBBox.y) / targetBBox.height} : null;
			modifyEdge(graphEdge, bendpoints, 
					sourceAnchor ? this._getAnchorString(sourceAnchor) : null, 
					targetAnchor ? this._getAnchorString(targetAnchor) : null);
			
			// handle edge labels if algorithm doesn't support them
			if (!this._nodeLayoutSupportsFloatingLabels()) {
				for (j = 0, child; child=edge.mapping.children[j]; j++) {
					if (child.policies.graphFloatingLabel) {
						var gfl = child.policies.graphFloatingLabel;
						modifyNode(gfl, 0, 0); // reset labels to default position
					}
				}
			}
		}
		
		return true;
	},
	
	_getAnchorString: function(pt) {
		var anchorPrototype = rmpw.diagrams.render.graph.AbstractEdgeAnchor.prototype;
		
		return anchorPrototype.TERMINAL_START_CHAR + 
			pt.x + anchorPrototype.TERMINAL_DELIMITER_CHAR + 
			pt.y + anchorPrototype.TERMINAL_END_CHAR;
	},
	
	runGraphLayout: function(layoutCallBack, args) {
		var renderPolicies = args.targetController.renderMapping ? args.targetController.renderMapping.policies : args.targetController.policies;
		var graphContainer = renderPolicies.graphContainer ? renderPolicies.graphContainer : renderPolicies.graphElement.getGraphContainer();
		var currentGraph = graphContainer.getInternalGraphElement();
		var proxyGraph = currentGraph.getProxyGraph();
		var nodes = proxyGraph.nodes;
		var edges = proxyGraph.edges;
		
		var graphStruct = this._getGraph(args, nodes, edges);
		var graph = graphStruct.graph;
		var origRect = graphStruct.bbox;
		
		if(graph) {
			args.__origRect = origRect;
			return this.runGraphLayoutAlgorithm(graph, layoutCallBack, args);
		}
		
		return false;
	},
	
	/*
	 * Utility method which determines if the layout should route the edges in an orthogonal manner
	 */
	isOrthogonalLinks: function(args) {
		return args.routerType === "Rectilinear" || 
			   args.routerType === "tree" || args.routerType === "Tree" ||
			   args.routerType === this.RECTILINEAR_ROUTING;
	},
	
	fromRawData: function(data, idLookupFn) {
		var graph = this.createGraph();
		var nodeMap = {},
		  i,
		  node,
		  edge,
		  mapping;
		for (i=0, node; node=data.nodes[i]; i++) {
			mapping = idLookupFn ? idLookupFn(node.id) : node.id;
			var realNode = this.createGraphNode(mapping, graph, node.b);
			nodeMap[node.id] = realNode;
		}
		for (i=0, edge; edge=data.edges[i]; i++) {
			mapping = idLookupFn ? idLookupFn(edge.id) : edge.id;
			var realEdge = this.createGraphEdge(mapping, graph, nodeMap[edge.sId], nodeMap[edge.tId]);
			
			var pts = new rmpw.diagrams.render.util.PointList();
			for (var j=0, pt; pt=edge.pts[j]; j++) {
				pts.addPoint_0(pt.x, pt.y);
			}
			realEdge.setPointList(pts);
		}
		
		return graph;
	}
});