/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010, 2013.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
dojo.provide("rmpw.diagrams.edit.handles.MoveableEndHandle");

dojo.require("rmpw.diagrams.render.util.Point");
dojo.require("rmpw.diagrams.edit.handles.MoveableHandle");
dojo.require("rmpw.diagrams.render.graph.AbstractEdgeAnchor");


/**
 * Class to handle movement of end bendpoints
 * 
 * @author Steve Shaw
 */
dojo.declare("rmpw.diagrams.edit.handles.MoveableEndHandle", [rmpw.diagrams.edit.handles.MoveableHandle, rmpw.diagrams.render.graph.AbstractEdgeAnchor], {
	
	constructor: function(args) {
		this.source = args.source;
		this.defaultAnchor = null;
		if (args.anchor) {
			this.defaultAnchor = args.anchor;
		}
		this.anchor = this.defaultAnchor;
		this.targetMapping = null;
		this.editController = args.editController;
		this._cursorChanged = {};
			
		// callback method for validating whether an anchor to a mapping is valid
		this._isAnchorToMappingValid = args.isAnchorToMappingValid ? args.isAnchorToMappingValid : function() { return true; };
	},
	
	/*
	 * Sets the default anchor that this handle will attach to
	 */
	setDefaultAnchor: function(anchor) {
		this.defaultAnchor = anchor;
		this.anchor = anchor;
	},
	
	/*
	 * Gets the associated anchorable policy if it is installed or null otherwise...
	 */
	getAnchorable: function(mapping) {
		if (mapping.policies.anchorable) {
			return mapping.policies.anchorable;
		}
		
		return null;
	},
	
	setCursor: function(editController, cursorType) {
		if (!this._passiveCursor) {
			var shape = editController.getDefaultTargetShape ? editController.getDefaultTargetShape() : null;
			if (shape) {
				this._cursorChanged[editController.renderMapping.id] = editController;
				shape.rawNode.style.cursor=cursorType;	
			}
		}
	},
	
	resetCursors: function() {
		// change cursors back to default
		for (var id in this._cursorChanged) {
			var shape = this._cursorChanged[id].getDefaultTargetShape();
			if (shape) {
				shape.rawNode.style.cursor="";
			}
		}
		this._cursorChanged = {};
	},
	
	setExternalAnchor: function(moveableEndHandle, nodeMapping) {
	    var anchorable = null;
		if (moveableEndHandle.source) {
			anchorable = this.getAnchorable(nodeMapping.renderMapping);
			if (anchorable) {
				moveableEndHandle.anchor = anchorable.getSourceConnectionAnchor();
			}
		}
		else {
			anchorable = this.getAnchorable(nodeMapping.renderMapping);
			if (anchorable) {
				moveableEndHandle.anchor = anchorable.getTargetConnectionAnchor();
			}
		}
		this.targetMapping = nodeMapping;

		var isValid = this._isAnchorToMappingValid(this.targetMapping);
		this.setCursor(this.targetMapping, isValid ? "pointer" : "not-allowed");
		
		// update target highlight
		if (this.targetMapping !== this.lastTargetMapping) {
			this.lastTargetMapping = this.targetMapping;
			if (nodeMapping.policies.focusable) {
				if (!this.restyled) {
				    this.restyled = {};
				}
				if (isValid && !this.restyled[nodeMapping.renderMapping.id]) {
					this.restyled[nodeMapping.renderMapping.id] = nodeMapping.policies.focusable;
					nodeMapping.policies.focusable.setFeedbackStyle("selection", 4);
				}
			}
		}
		
		
	},
	
	onFirstMove: function(/* dojox.gfx.Mover */ mover, passiveCursor) {
		this.inherited(arguments);
		
		this._passiveCursor = passiveCursor;
		// start off move assuming disconnection
		this.targetMapping = this.lastTargetMapping = null;
		
		// if touch device we need to do our own focusing 
		if (mover.isTouch) {
			this.graphManager = this.editController.renderMapping.policies.graphElement.getGraphManager();
			this.focusScope = this.editController.root.policies.focusScope;
			this.focusMapping = null;
		}
		
		// check if source is connectable 
		this.anchor = null;
		var moveableEndHandle = this;
		this._connectFocus = dojo.connect(this.editController.root.policies.focusScope, "setFocus", function(editController){
			
			// if this shape isn't anchorable, see if its parent is
			if (editController && !editController.renderMapping.policies.anchorable) {
				if (!editController.renderMapping.policies.graphContainer &&
						editController.parent && editController.parent.renderMapping.policies.anchorable) {
					editController = editController.parent;
				}
			}
			
			// Check if we are on a handle on a shape within a shape container, if so, 
			// do not jump the anchor to the edge of container until the pointer is close to it.
			moveableEndHandle.containerEditController = null;	
			if (editController
				&& moveableEndHandle.editController.parent == editController
				&& editController.renderMapping.policies.graphContainer
				&& editController.renderMapping.policies.anchorable) {
				moveableEndHandle.containerEditController = editController;
			} 

			if (!editController || !editController.renderMapping.policies.anchorable || moveableEndHandle.containerEditController) {
				moveableEndHandle.anchor = null;
				if (moveableEndHandle.targetMapping !== null) {
					moveableEndHandle.targetMapping = null;
				}
				
				if (editController) {
					moveableEndHandle.setCursor(editController, "not-allowed");
				}
			}
			else {
				moveableEndHandle.setExternalAnchor(moveableEndHandle, editController);
			}
		 });
	},
	
	_updateRelativeReferencePoint: function(pos) {
		var owner = this.defaultAnchor.getOwner();
		var rBox = owner.mapping.policies.graphElement.getPolygonPoints().getBounds();
		owner.translateToAbsolute(rBox);
		this.setRelativeReferencePoint(this.getAnchorRelativeLocation(this.pos, rBox));
	},
	
	onMove: function(/* dojox.gfx.Mover */ mover, /* Object */ shift)  {
		this.inherited(arguments);
		
		if (this.containerEditController) {
			// We've started a connection on a shape inside a container, we need to change 
			// the anchor when we get close to the edge of the container shape.
		    if (this.containerEditController.renderMapping.policies.graphElement) {
				var bounds = this.containerEditController.renderMapping.policies.graphElement.getBounds();
				this.containerEditController.renderMapping.policies.graphElement.translateToAbsolute(bounds);
				
				if (bounds.shrink(30, 30).contains_0(this.pos)) {
					this.anchor = null;
					if (this.targetMapping !== null) {
						this.targetMapping = null;
					}
					this.setCursor(this.containerEditController, "not-allowed");				
				} else {
					this.setExternalAnchor(this, this.containerEditController);
				}
		    }
		}	
		
		if (this.anchor && this.anchor.calculateRelativeReferencePoint) {
			this.anchor.calculateRelativeReferencePoint(this.pos);
		}

		if (this.defaultAnchor) {
			this._updateRelativeReferencePoint(this.pos);
		}
		
		// if touch device we need to do our own focusing 
		if (mover.isTouch) {
			var pt = this.editController.renderMapping.root.viewport.getSurfaceAbsoluteMousePosition(mover.lastX, mover.lastY);
			var mapping = this.graphManager.getFocusableFromPoint(pt);
			if (mapping !== this.focusMapping) {
				this.focusMapping = mapping;
				if (mapping &&  mapping.__editController) {
					this.focusScope.setFocus(mapping.__editController);
				} else {
					this.focusScope.setFocus(null);
				}
			}
		}
	},
	
	onMoveStop: function(/* dojox.gfx.Mover */ mover) {
		this.inherited(arguments);
		
		this._passiveCursor = false;
		
		// restore default focus feedback to any that have changed
		if (this.restyled) {
			for (var id in this.restyled) {
				this.restyled[id].setFeedbackStyle();
			}
			delete this.restyled;
		}
		
		// change cursors back to default
		this.resetCursors();
		
        // if the handle has been moved - then set it as fixed so it won't readjust when shape is moved.
		if (this.editController.renderMapping.policies.graphEdge) {
			var geObj = this.editController.renderMapping.policies.graphEdge.getInternalGraphElement();
            this.source ? geObj.setFixSourceAnchor(!this.isDefaultAnchor()) : geObj.setFixTargetAnchor(!this.isDefaultAnchor());
		}
	},
	
	onCancel: function() {
		this.inherited(arguments);
		
		dojo.disconnect(this._connectFocus);
		delete this._connectFocus;
		
		// change cursors back to default
		this.resetCursors();
		
		/*
		 * If this anchor is null, this means we didn't connect to anything.  Revert it back to default...
		 */
		if (!this.anchor) {
			this.anchor = this.defaultAnchor;
		}
	},
	
	/*
	 * Allows this handle to simulate an anchor.  If connected, it will return the node anchor location,
	 * otherwise, it will just return the XY location where the handle has been moved to.
	 */
	getLocation: function (reference){
		if (this.anchor) {
			return this.inherited(arguments);
		}
		else {
			return new rmpw.diagrams.render.util.Point(this.pos.x, this.pos.y);
		}
	},
	
	isDefaultAnchor: function() {
		if (this._relRefPt) {
			return this.isDefaultAnchorArea(this._relRefPt);
		}
		else if (this.anchor) {
			return this.anchor.isDefaultAnchor();
		}
		
		return false;
	},
	
	getRelativeReferencePoint: function() {
		if (this._relRefPt) {
			return this._relRefPt;
		}
		else if (this.anchor) {
			return this.anchor.getRelativeReferencePoint();
		}
		
		return null;
	},
	
	setRelativeReferencePoint: function(pt) {
		if (this.anchor) {
			this.anchor.setRelativeReferencePoint(pt);
		}
		
		this._relRefPt = pt;
	},

	/*
	 * Allows this handle to simulate an anchor.  If connected, it will return the node anchor reference point,
	 * otherwise, it will just return the XY location where the handle has been moved to.
	 */
	getReferencePoint: function (){
		if (this._relRefPt || this.anchor) {
			return this.inherited(arguments);
		}
		
		return new rmpw.diagrams.render.util.Point(this.pos.x, this.pos.y);
	},
	
	getOwner: function() {
		if (this.anchor) {
			return this.anchor.getOwner();
		}
		
		if (this.defaultAnchor) {
			return this.defaultAnchor.getOwner();
		}
		
		return null;
	}
});