/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2011, 2013. All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 ******************************************************************************/
dojo.provide("rmpw.text.internal.util.Converter");

dojo.require("dojo.string");
dojo.require("dojox.string.Builder");
dojo.provide("rmpw.common.util.IdentifierUtil");

/**
 * Prototype implementation of an XHTML <--> JSON Conversion algorithm.
 * Currently used client-side as a part of Rich Text doc prototype.
 * Should be moved to server-side for performance and scalability reasons.
 */

rmpw.text.internal.util.Converter = {
	STRING_ATTRIBUTE_TEMPLATE: ' ${dmxml_key} = "${dmxml_stringValue}"',
	REF_ATTRIBUTE_TEMPLATE: ' ${dmxml_key} = "${dmxml_refValue.rdf_about}"',
	EXTERNAL_REF_ATTRIBUTE_TEMPLATE: ' ${dmxml_key} = "${dmxml_refValue.$ref}"', // for external links
	EXTERNAL_REF_ATTRIBUTE_TEMPLATE1: ' ${dmxml_key} = "${dmxml_refValue}"', // for pasted from clipboard
	
	tagsWithId: ["p", "h1", "h2", "h3", "h4", "li", "td", "body"],
	idException: "imageArea",
	
    parse: function(content, use_rdf_about){
        var xml = new dojox.string.Builder();
        if (this.isXhtmlRoot(content)) {
			this.serializeObject(content.dmxhtml_body, xml, use_rdf_about);
        } else {
			this.serializeObject(content, xml, use_rdf_about);
		}
        return xml.toString();
    },
    
    isXhtmlRoot: function(object){
        if (!object || !object.rdf_type) {
            return false;
        }
        return dojo.some(object.rdf_type, function(type) { return type.rdf_about === "http://jazz.net/ns/dm/document#Document"; });
    },
    
    isCDATA: function(object){
        if (!object || !object.rdf_type) {
            return false;
        }
        return dojo.some(object.rdf_type, function(type) { return type.rdf_about === "http://jazz.net/ns/dm/xml#CDATA"; });
    },
    
    serializeObject: function(object, xml, use_rdf_about){
        if (this.isCDATA(object)) {
            if (object.dmxml_text) {
                xml.append(object.dmxml_text);
            }
        }
        else {
            if (!object.dmxml_name) {
                return xml;
            }
            xml.append("<", object.dmxml_name);
            this.serializeAttributes(object, xml);

			if (use_rdf_about && object.rdf_about && (dojo.indexOf(this.tagsWithId, object.dmxml_name) > -1)) {
				// add id only if it has been serialized, i.e. if doc is in old format with anonymous classes leave it blank;
				// no ids for pasted images
				var id = object.rdf_about.substring(object.rdf_about.lastIndexOf("#") + 1);
				xml.append(" id=", id);
			}

            if (object.dmxml_content) {
                xml.append(">");
                var i, len = object.dmxml_content.length;
                for (i = 0; i < len; i++) {
                    this.serializeObject(object.dmxml_content[i], xml, use_rdf_about);
                }
                xml.append("</", object.dmxml_name, ">");
            } else if (object.dmxhtml_body) {
				this.serializeObject(object.dmxhtml_body, xml, use_rdf_about);
                xml.append("</", object.dmxml_name, ">");				
			}
            else {
                xml.append("/>");
            }
        }
		return xml;
    },
	
    serializeAttributes: function(object, xml){
        if (object.dmxml_attributes) {
            var i, attribute;
            for (i = 0; i < object.dmxml_attributes.length; i++) {
				attribute = object.dmxml_attributes[i];
				// Handle string attributes (attributes whose key/value pair are both strings)
				if (attribute.dmxml_key && attribute.dmxml_stringValue) {
					xml.append(dojo.string.substitute(this.STRING_ATTRIBUTE_TEMPLATE, attribute));
				}
				// Handle ref attributes (attributes whose value is an object with an rdf:about pointing to the referenced resource).
				// See the dmxml.owl ontology definition and the function comments on isRefAttribute() below.
				else if (attribute.dmxml_key && attribute.dmxml_refValue) {
					if (attribute.dmxml_refValue.rdf_about) {
						xml.append(dojo.string.substitute(this.REF_ATTRIBUTE_TEMPLATE, attribute));					
					} else if (attribute.dmxml_refValue.$ref) {
						xml.append(dojo.string.substitute(this.EXTERNAL_REF_ATTRIBUTE_TEMPLATE, attribute));
					} else {
						xml.append(dojo.string.substitute(this.EXTERNAL_REF_ATTRIBUTE_TEMPLATE1, attribute));
					}
				}
            }
        }
		return xml;
    },


// XHTML to JSON


    parseToJson: function(xhtml, resourceUri, use_rdf_about, attributeToRemove){
        var stack = [];
        var root = this.createXhtmlRoot(null, resourceUri);
        var parser = new CKEDITOR.htmlParser();
		var context = this.createXhtmlBody(root, null, use_rdf_about);
		root.dmxhtml_body = context;

        parser.onTagOpen = dojo.hitch(this, function(tagName, attributes, selfClosing){
            var tag = null;
			if (attributeToRemove) {
				for (var att in attributes) {
					if (att === attributeToRemove) {
						delete attributes[att];
						break;
					}
				}
			}
            if (tagName === "html") {
                tag = this.createXhtmlRoot(attributes, resourceUri, use_rdf_about);
				tag = dojo.mixin(root, tag);
            } else if (tagName === "body") {
				tag = this.createXhtmlBody(root, attributes, use_rdf_about);
				tag = dojo.mixin(context, tag);
				root.dmxhtml_body = tag;
			} else {
				if (!context) {
					// shouldn't happen when input comes from ckeditor
					throw "Invalid input";					
				}
				context.dmxml_content = context.dmxml_content || [];
                tag = this.createTag(context, tagName, attributes, use_rdf_about);
                context.dmxml_content.push(tag);
				if (tagName !== "br" && tagName !== "img") { // see note below
					stack.push(context);
				}
            }
			if (tagName !== "br"&& tagName !== "img") {	// there won't be closing tag for br and it can't be context for anything
				context = tag;
			}
        });

        parser.onTagClose = dojo.hitch(this, function(tagName) {
            context = stack.pop();
        });

        parser.onText = dojo.hitch(this, function(text){
            context.dmxml_content = context.dmxml_content || [];
            var cdata = this.createCDATA(context, text);
            context.dmxml_content.push(cdata);
        });
		
		parser.parse(xhtml);
        return root.dmxhtml_body;
    },

    createXhtmlRoot: function(attributes, resourceUri, use_rdf_about){
        var root = dojo.mixin(this.createTag(null, "html", attributes, use_rdf_about), {
	        rdf_type: [
				{ rdf_about: "http://jazz.net/ns/dm/document#Document"},
				{ rdf_about: "http://jazz.net/ns/dm/core#Document"}
			],
			rdf_about: ""
        });
        return root;
    },

    createXhtmlBody: function(root, attributes, use_rdf_about){
        var body = dojo.mixin(this.createTag(root, "body", attributes, use_rdf_about, "T0"), {
            rdf_type: [{ rdf_about: "http://jazz.net/ns/dm/document#Body" }]
        });
        return body;
    },

    createTag: function(context, tagName, attributes, use_rdf_about, tagId) {
        var tag = this.createRdfObject("http://jazz.net/ns/dm/xml#BodyTag");
		if (tagName) {
			tag.dmxml_name = tagName;
		}
		var recordIdAsAtt = false;
		var addId = dojo.indexOf(this.tagsWithId, tagName) > -1;
		if (attributes) {
			if (use_rdf_about && attributes.id && addId) {
				if (attributes.id.indexOf(this.idException) === -1) {
					// tag already existed
					tag.rdf_about = attributes.id;
				} else {
					recordIdAsAtt = true;
				}
			}
		}
		
		if (use_rdf_about && !tag.rdf_about && context && !recordIdAsAtt && addId) {
			// tag did not exist before, generate a unique tag id
			tag.rdf_about = tagId ? tagId : rmpw.common.util.IdentifierUtil.guid();
		}

        for (var att in attributes) {
			if (!use_rdf_about || att !== "id" || recordIdAsAtt) {
				var attrib = this.createAttribute(tag, att, attributes[att]);
	            tag.dmxml_attributes = tag.dmxml_attributes || [];			
				tag.dmxml_attributes.push(attrib);
			}
        }
        return tag;
    },

    createAttribute: function(context, key, value) {
		var isRefAttribute = this.isRefAttribute(context, key);
        var attrib = this.createRdfObject(
			isRefAttribute ?
			"http://jazz.net/ns/dm/xml#RefAttribute" :
			"http://jazz.net/ns/dm/xml#StringAttribute");
		attrib.dmxml_key = key;
        if (isRefAttribute) {
			if (context.dmxml_name === "img" && key === "src" && value && (value.indexOf("file") === 0)) {
				// image src value is local path to the image (happens e.g. when copy/paste an image from Word)
				attrib.dmxml_refValue = "";
			} else {
				attrib.dmxml_refValue = {
					rdf_about: value
				};
			}
		} else {
			attrib.dmxml_stringValue = value;
		}
		return attrib;
    },

	/**
	 * Determines whether or not a given attribute is a reference to another resource.
	 * Some references in XHTML (eg: the href attribute of an <a> tag) are references
	 * to other resources.  These need to be handled differently on serialization
	 * and deserialization since the XML ontology represents these differently.
	 * 
	 * Currently we support only <img> and <a> tags, although there are many other XHTML
	 * tags that can have reference attributes.  @TODO We should visit these on a case-by-case
	 * basis to determine if they should be exposed as resource references to the system.
	 * 
	 * @param {Object} context
	 * @param {Object} key
	 */
	isRefAttribute: function(context, key) {
		var name = context.dmxml_name;
		if (name === 'img') {
			return key === 'src';
		} else if (name === 'a') {
			return key === 'href';
//		} else if (name === 'object') {
			//TODO need to define exactly how our embed objects are going to specify the URI for the embedded resource.
		}
		return false;
	},
	
    createCDATA: function(context, text){
        return dojo.mixin(this.createTag(context), {
			rdf_type: [{ rdf_about: "http://jazz.net/ns/dm/xml#CDATA" }],
            dmxml_text: text
        });
    },
    
    createRdfObject: function(rdf_types){
        if (dojo.isArray(rdf_types)) {
            rdf_types = dojo.map(rdf_types, function(type) {
				return {
					rdf_about : type
				};
			});
        } else {
            rdf_types = [{ rdf_about: rdf_types }];
        }
        return {
            rdf_type: rdf_types
        };
    },

    htmlifyText: function(source){
        // space out rdf:Description blocks
        source = source.replace(/<rdf:Description/g, "\n  <rdf:Description");
        
        // escape angle brackets
        source = source.replace(/</g, "&lt;");
        source = source.replace(/>/g, "&gt;");
        source = "<html xmlns='http://www.w3.org/1999/xhtml/'><pre><code>" + source + "</code></pre></html>";
        return source;
    },
	
	
	/**
	 * Removes attributeToRemove from all tags in given xhtml.
	 * @param {Object} xhtml
	 */
	removeAttribute: function(xhtml, attributeToRemove){
		var json;
		try {
			json = this.parseToJson(xhtml, null, false, attributeToRemove);
		} catch (err) {
			// parsing failed, must be that xhtml is input from another tool 
			// before ckeditor had a chance to clean it up; return the original 
			return xhtml;			
		}
		var xml = new dojox.string.Builder();
		var i, len = json.dmxml_content.length;
		for (i = 0; i < len; i++) {
			this.serializeObject(json.dmxml_content[i], xml, false);
		} 		
        return xml.toString();
    }
};
