/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010, 2013.  All Rights Reserved.
 * 
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
(function() {
dojo.provide("rmps.app.pickers.internal.PickerController");

/*######################################### */
/*These imports are required to invoke auth dialogs, and
 * users' credentials checking utils. Otherwise, the users,
 * will not be autheroized when trying to search for an element
 * even if they are already logged in */
dojo.require("jazz.app.proxy");
dojo.require("jazz.app.auth");
dojo.require("jazz.app.events");
/*######################################### */


dojo.require("rmps.core.query.QueryClient");
dojo.require("rmps.core.error");

dojo.require("rmps.app.core.internal.query.TextQuery");
dojo.require("rmps.app.core.internal.query.ClientPagedQueryResult");
dojo.require("rmps.core.util.TypeConstants");
dojo.require("rmps.core.resource.ResourceInfoClient");

dojo.require("dojo.i18n");
dojo.require("dojo.string");

dojo.requireLocalization("rmps.app.pickers", "RMPSPickerMessages");
dojo.requireLocalization("rmps.app.ui.internal.pages", "DesignPageMessages");

var QueryClient = rmps.core.query.QueryClient;
var ResourceInfoClient = rmps.core.resource.ResourceInfoClient;

/**
 * PickerController works with rmps.app.pickers.internal.RMPSPicker to control some aspects of search, 
 * including the actual search query execution.
 * Following properties can be used to control the customization of RMPSPicker:
 * enableMultiSelect, allowSearchSystemDomains, allowEmptySearches, enableShowAll, 
 * orderBy, pageSize, limit, limitResultTableHeight (see their documentation for more details). 
 */
dojo.declare("rmps.app.pickers.internal.PickerController", null, {

	messages: dojo.i18n.getLocalization("rmps.app.pickers", "RMPSPickerMessages"),
	designPageMessages: dojo.i18n.getLocalization("rmps.app.ui.internal.pages", "DesignPageMessages"),
	
	// private properties
	_pickerWidget: null,
	
	_prevQueryResult: null,
	_searchDeferred: null,
	
	/**
	 * if true, checkboxes will be rendered and multiselection supported
	 */
	enableMultiSelect: false,
	
	/**
	 * if true, user will have the option to include search results from dependent projects 
	 */
	allowSearchSystemDomains: false,

	/**
	 * if true, enables search without any criteria entered into the search box 
	 * (has to be true if enableShowAll is true)
	 */
	allowEmptySearches: false,

	/**
	 * when true, Show All link will be present above the search box
	 */
	enableShowAll: false,

	/**
	 * by default, controller does not chnage orderBy defined elsewhere
	 */
	orderBy: undefined,

	/**
	 * number of table rows created on one page
	 */
	pageSize: 5,

	/**
	 * maximum number of search results to be returned from server;
	 * by default, controller does not change orderBy defined elsewhere
	 */	
	limit: undefined,
	
	/**
	 * when true, table showing results will be restricted hight meaning there will be scrollbars if results don't fit; 
	 */
	limitResultTableHeight: false,
	
	constructor: function(widget) {
		this._pickerWidget = widget;
	},
	
	search: function(args) {
		if (!args || !args.query) {
			return;
		}
		
		var queryParams = dojo.queryToObject(args.query);
		var query = new rmps.app.core.internal.query.TextQuery(queryParams);
		
		var timestamp = args.ts;
		
		query.timestamp = timestamp;
		//query.name = dojo.string.substitute(this.messages.statusSearchMessage, [query.name]);
		
		/*Inform the user that the search process is in progress */
		this._pickerWidget.showStatusNode(this.messages.statusSearchMessage);
		 
		if (!timestamp) {
			timestamp = new Date().getTime();
		}
		
		if (query.text.search(/^\d+$/) !== -1) {
			var resourceUri = rmps.core.RootServices.getUri(rmps.core.Namespaces.RMPS, "models")+"/"+query.text;
			var deferred = ResourceInfoClient.getResourceInfo(resourceUri, null, this._pickerWidget.context);
			var _resourceInfos;
			deferred.addCallback(dojo.hitch(this, function(resourceInfos) {
				// if there is a result returned, check that it matches the type requirement from query, if given
				// (CLMResourcePicker sets document as default type whenever the explicit type was not provided)
				if (!query.onlyType || query.onlyType === rmps.core.util.TypeConstants.DOCUMENT || (resourceInfos[0].type === query.onlyType)) {
					_resourceInfos= resourceInfos;
					this._pickerWidget.showResultsForResourceInfo(resourceInfos,query, false);
				}
			}));
			deferred.addBoth(this, function() {
				if(!_resourceInfos){
					this._pickerWidget.hideSearchResults();
				}
				delete this._searchDeferred;
				
				/*the search is done, then, remove the status message */
				this._pickerWidget.hideStatusNode(query.name);
			});
			return;
		}
		
		if (this._prevQueryResult && query.equals(this._prevQueryResult.query) &&
				(this._prevQueryResult.query.timestamp === query.timestamp || query.timestamp === null)) {
			this._pickerWidget.hideStatusNode(query.name);
			this._pickerWidget.showSearchResults(this._prevQueryResult, true);
		} else {
			var preserveFilter = true;
			if (!this._prevQueryResult || !query.token) {
				preserveFilter = false;
			}
			
			if (this._prevQueryResult && this._prevQueryResult.id === query.token) {
				this._prevQueryResult = new rmps.app.core.internal.query.ClientPagedQueryResult({
					query: query,
					id: this._prevQueryResult.id,
					rawEntries: this._prevQueryResult.rawEntries,
					totalResults: this._prevQueryResult.totalResults,
					uri: this._prevQueryResult.uri
				});
				this._pickerWidget.hideStatusNode(query.name);
				this._pickerWidget.showSearchResults(this._prevQueryResult, preserveFilter);
			} else {
				var queryToExecute = dojo.clone(query);
				delete queryToExecute.page;
				delete queryToExecute.pageSize;
				delete queryToExecute.orderBy;
				delete queryToExecute.token;
				
				// if there is an ongoing search request, cancel it as the user requested new search
				this.cancelSearchRequest();
				
				if (!this.validateQuery(queryToExecute)) {
					return;
				}
				
				// doing a full search (not click on Next or Previous), clear up any previous selections
				this._pickerWidget.clearSelections();
				
				this._searchDeferred = this.doSearch(queryToExecute, query);
				
				this._searchDeferred.addCallback(this, function(queryResult) {
					this._prevQueryResult = new rmps.app.core.internal.query.ClientPagedQueryResult({
						query: query,
						id: queryResult.id,
						rawEntries: queryResult.entries,
						totalResults: queryResult.totalResults,
						uri: queryResult.uri	
					});
					
					this._pickerWidget.showSearchResults(this._prevQueryResult, preserveFilter);
				});
				
				this._searchDeferred.addErrback(this, function(error) {
					if (error.dojoType !== "cancel") {
						if (!dojo._isDocumentOk(this._searchDeferred.ioArgs.xhr)) {
							rmps.core.error.createXhrError(this._searchDeferred.ioArgs).dispatch();
						} else {
							rmps.core.error.createError(error).dispatch();
						}
					}
				});

				this._searchDeferred.addBoth(this, function() {
					delete this._searchDeferred;
					
					/*the search is done, then, remove the status message */
					this._pickerWidget.hideStatusNode(query.name);
				});
			}
		}
	},
	
	cancelSearchRequest: function() {
		if (this._searchDeferred) {
			this._searchDeferred.cancel();
			delete this._searchDeferred;
		}
	},
	
	validateQuery: function(queryToExecute) {
		var validationResults = queryToExecute.validate();
				
		if (validationResults && validationResults.length > 0) {
			var validationDetailsMessage = "";
			
			for (var i = 0; i<validationResults.length; i++) {
				validationDetailsMessage +=validationResults[i]+"\n";
			}
				
			rmps.core.error.create(this.designPageMessages.error_invalidQuery, validationDetailsMessage, null).dispatch();
			this._pickerWidget.hideStatusNode(query.name);
			return false;
		}
		return true;
	},
	
	doSearch: function(queryToExecute, query) {
		return QueryClient.execute(queryToExecute);
	}
});

})();
