/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010, 2013.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
dojo.provide("rmpw.diagrams.render.policies.GraphContainer");

dojo.require("rmpw.diagrams.render.policies.GraphElement");
dojo.require("rmpw.diagrams.render.policies.GraphManager");
dojo.require("rmpw.diagrams.render.graph.BaseGraphConfigurator");
dojo.require("rmpw.diagrams.render.policies.MatchSize");
dojo.require("rmpw.diagrams.render.util.ColorConstants");

/**
 * Data needed to layout the graph node
 * 
 * @author Steve Shaw
 */
dojo.declare("rmpw.diagrams.render.policies.GraphContainer", rmpw.diagrams.render.policies.GraphElement, {
	singleton: ["graphElement", "graphContainer"],
	overridable: ["graphElement"],
	
	constructor: function(mapping) {
		this.graphManager = new rmpw.diagrams.render.policies.GraphManager(mapping);
		this.invalidationController = this.host.viewer.getInvalidationController();
		
		this._connectOnInvalidate = dojo.connect(this.invalidationController, "onInvalidate", this, "enforceParentPolicies");	
		if (this.isTopLevel()) {
			this._connectOnInvalidationComplete = dojo.connect(this.invalidationController, "onInvalidationComplete", this, "refresh");
		}
		this._backgroundInset = 0;
	},
	
	destroy: function() {
		this.graphManager.destroy();
		
		dojo.disconnect(this._connectOnInvalidate);
		if (this.isTopLevel()) {
			dojo.disconnect(this._connectOnInvalidationComplete);
		}
		if (this._connectViewportSize) {
			dojo.disconnect(this._connectViewportSize);
			delete this._connectViewportSize;
		}
		if (this._connectRootResize) {
			dojo.disconnect(this._connectRootResize);
			delete this._connectRootResize;
		}
		
		this.inherited(arguments);
	},
	
	_createInternalGraphElement: function() {
		return this.host.viewer.graphConfigurator.createGraph(this.host);
	},
	
	getRearFeedbackLayer: function() {
		return this.rearFeedbackLayer;
	},
	
	isTopLevel: function() {
		if (this.host.parent === this.host.root) {
			return true;
		}
		
		return false;
	},
	
	updateBackgroundToViewPortSize: function() {
		if (!this.isTopLevel() || !this._background) {
			return;
		}
		
		// synchronize size with viewport
		var d = this.host.root.getCanvasSize();
		var viewportRect = new rmpw.diagrams.render.util.Rectangle(0, 0, d.w, d.h);
		this.translateToRelative(viewportRect);
		
		this._background.setTransform({ dx: viewportRect.x, dy: viewportRect.y });
		this._background.setShape({width: viewportRect.width, height: viewportRect.height});
		
		this.resetPolygonShapes();
	},
	
	getBounds: function() {
		if (this._background) {
			this.updateBackgroundToViewPortSize();
			var box = rmpw.diagrams.render.util.Workarounds.getTransformedBoundingBox(this._background);
			var relBox = new rmpw.diagrams.render.util.Rectangle(box.x, box.y, box.width, box.height);
			this.translateToRelative(relBox);
			return relBox;
		}
		return new rmpw.diagrams.render.util.Rectangle(0,0,0,0);
	},
	
	getBackground: function() {
		return this._background;
	},
	
	setBackground: function(background) {
		if (this._background) {
			this._background.removeShape();
		}
		
		this._background = background;
		this._background.moveToBack();
		
		if (this.isTopLevel()) {
			this.updateBackgroundToViewPortSize();
			this._connectViewportSize = dojo.connect(this.host.root.viewport, "onViewportChanged", dojo.hitch(this, function() {
				this.updateBackgroundToViewPortSize();
			}));
			// update background size every time canvas is resized
			this._connectRootResize = dojo.connect(this.host.root, "resize", dojo.hitch(this, function() {
				this.updateBackgroundToViewPortSize();
			}));
		}
		else if (this._shape) {
			this._matchSizePolicy = this.host.createPolicy(rmpw.diagrams.render.policies.MatchSize);
			this._matchSizePolicy.setMaster(this._shape);
			this._matchSizePolicy.setSlave(this._background);
		}
	},
	
	/*
	 * @deprecated - use setShape instead
	 */
	setFigure: function(figure) {
		this.setShape(figure);
	},
	
	setShape: function(figure) {
		this.inherited(arguments);
		
		this.rearFeedbackLayer = figure.createGroup();
		if (this._background) {
			this._background.moveToBack();
		}
		else {
			if (!dojo.isOffscreen) {
				var background = figure.createRect();
				background.setFill(this.host.viewer.diagram.isImageDiagram ? "white" : rmpw.diagrams.render.util.ColorConstants.CANVAS);
				this.setBackground(background);
			}
			
		}
	},
	
	setVisible: function(visible) {
		this._visible = visible;
		if (!visible) {
			if (this._matchSizePolicy) {
				this._matchSizePolicy.deactivate();
			}
			if (this._background) {
				this._background.removeShape();
				delete this._background;
			}
		}
	},
	
	setBackgroundInset: function(backgroundInset) {
		this._backgroundInset = backgroundInset;
	},
	
	/*
	 * Returns the graph container for this element
	 */
	getGraphContainer: function() {
		// if this host also has a GraphNode policy, use that the node's graph container
		if (this.host.policies.graphNode) {
			return this.inherited(arguments);
		}
		return this;
	},
	
	/*
	 * Updates the graphContainer size - 
	 * TODO more intelligent calculation based on node movement delta
	 */
	updateContainerSize : function() {
		var childrenBox = this.getChildrenBox();
		
		if (this._background) {
			if (this.host.viewer === this.host.parent) {
				this.host.viewer.updateViewport(childrenBox);
				var dim = this.host.viewer.surface.getDimensions();
				childrenBox.x = Math.min(childrenBox.x, 0);
				childrenBox.y = Math.min(childrenBox.y, 0);
				childrenBox.width = dim.width;
				childrenBox.height = dim.height;
			}
			
			// calculate the bounding box
			var backgroundBox = rmpw.diagrams.render.util.Workarounds.getTransformedBoundingBox(this._background);
			backgroundBox = new rmpw.diagrams.render.util.Rectangle(backgroundBox.x, backgroundBox.y, backgroundBox.width, backgroundBox.height);
			
			this.translateToRelative(backgroundBox);
			
			childrenBox.x -= this._backgroundInset;
			childrenBox.y -= this._backgroundInset;
			childrenBox.width += this._backgroundInset * 2;
			childrenBox.height += this._backgroundInset * 2;
				
			this._background.applyTransform({
				dx: childrenBox.x - backgroundBox.x, 
				dy: childrenBox.y - backgroundBox.y
			});
			this._background.setShape({width: childrenBox.width, height: childrenBox.height});
		}
	},
	
	getChildrenBox : function() {
		this._topLeftPt = {x:1000000, y:1000000};
		this._childrenBox = this.isTopLevel()? undefined : new rmpw.diagrams.render.util.Rectangle(0, 0, 0, 0);
		for (var i=0, child; child = this.host.children[i]; i++) {
			if (child.policies.graphNode) {
				var box = child.policies.graphNode.getTightBounds();
				if (box.x<this._topLeftPt.x){ 
				    this._topLeftPt.x = box.x;
				}
				if (box.y<this._topLeftPt.y){ 
				    this._topLeftPt.y = box.y;
				}
				if (!this._childrenBox) {
					this._childrenBox = box;
				} else {
					this._childrenBox.union_0(box);
				}
			}
		}
		if (!this._childrenBox) this._childrenBox = new rmpw.diagrams.render.util.Rectangle(0, 0, 0, 0);
		return this._childrenBox;
	},
	
	/*
	 * Returns the bounds that are flush with the viewable extent of the shape (anchorable area)
	 */
	getTightBounds: function() {
		var childrenBox = new rmpw.diagrams.render.util.Rectangle(0, 0, 0, 0);
		for (var i=0, child; child = this.host.children[i]; i++) {
			if (child.policies.graphNode) {
				var box = child.policies.graphNode.getTightBounds();
				if (childrenBox.isEmpty()) {
					childrenBox =new rmpw.diagrams.render.util.Rectangle(box.x, box.y, box.width, box.height);
				}
				else {
					childrenBox.union_0(box);
				}
			}
		}
		
		return childrenBox;
	},
	
	getTopLeftPoint: function() {
		return this._topLeftPt;
	},
	
	translateToRelative: function (t, dojoShape){
		var shape = dojoShape ? dojoShape : this.getDojoShape();
		return rmpw.diagrams.render.util.Workarounds.translateToRelative(t, shape, true);
	},
	
	enforce: function() {	
		this.inherited(arguments);
		
		if (!this.isTopLevel()) {
			this.updateContainerSize();
		}
	},
	
	// called when "mapping" has had model values applied to it and its policies enforced
	// see if we should also enforce this container's policies and its parents too
	enforceParentPolicies: function(mapping) {
		// if this node belongs to this container and this is not the top container,
		// we need to enforce this container's policies (ex: layouts) 
		// and the policies of any parent element to the top
		if ( mapping.policies && mapping.policies.graphNode && !this.isTopLevel() && this._isChild(mapping)) {  //

			// first we need to update dojo shape of invalidated element before we can run 
			// this container's layout policies since the layout policies get the size from the dojo shapes
			mapping.policies.graphElement.refresh();
			
			var walker = this.host;
			while (walker.policies.graphElement) {
				// break on top container
				if (walker.policies.graphContainer && walker.policies.graphContainer.isTopLevel()) {
					break;
				}
				
				// run this container's layout policies
				walker.enforcePolicies();
				
				// update this container's dojo shape bounds for next parent's layout policies
				walker.policies.graphElement.refresh();
				
				walker = walker.parent;
			}
		}
	},
	
	_isChild: function(mapping) {
		if (this.host.children) {
			for (var i=0, child; child = this.host.children[i]; i++) {
				if (mapping.id === child.id) {
					return true;
				}
			}
		}
		return false;
	},
	
	// called only on TOP level when ANY invalidations have occurred and all appropriate rendering policies have been applied
	refresh: function() {
		if (this.isValidNodes()) {
			this._refreshDiagrams(this.host);
		}
	},
	
	isValidNodes: function() {
		var list = this.invalidationController.getInvalidList();
		for (var i = 0, mapping; mapping = list[i]; i++) {
			if (mapping.policies && mapping.policies.graphNode) {
				return false;
			}
		}
		return !this.invalidationController.isProcessingList();
	},
	
	
	
	_refreshDiagrams: function(mapping) {
		// routes lines to shapes/pushes neighbors
		// sets model bounds into dojo shapes
		// resets highlights
		if (mapping.policies && mapping.policies.graphContainer) {
			mapping.policies.graphContainer.graphManager.refreshDiagram();
			
		}
		if (mapping.children && mapping.children.length>0) {
			for (var i=0, child; child = mapping.children[i]; i++) {
				this._refreshDiagrams(child);
			}
		}
	},
	
	setEdgeLayer: function(edgeLayer) {
		// do nothing - exists for backward compatibility only
	},
	
	setNodeLayer: function(edgeLayer) {
		// do nothing - exists for backward compatibility only
	}
});