/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010, 2013.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
dojo.provide("rmpw.diagrams.render.policies.GraphEdge");

dojo.require("rmpw.diagrams.render.policies.GraphContainer");
dojo.require("rmpw.diagrams.render.policies.GraphElement");
dojo.require("rmpw.diagrams.render.graph.AnimatedEdge");
dojo.require("rmpw.diagrams.render.util.PointListUtilities");
dojo.require("rmpw.diagrams.render.util.Workarounds");
dojo.require("rmpw.diagrams.render.util.PointList");
dojo.require("rmpw.diagrams.render.graph.BaseGraphConfigurator");
dojo.require("rmps.core.error");

/**
 * Data needed to layout the graph edge
 * 
 * @author Steve Shaw
 */
dojo.declare("rmpw.diagrams.render.policies.GraphEdge", rmpw.diagrams.render.policies.GraphElement, {
	singleton: ["graphElement", "graphEdge"],
	
	sourceMapping: null,
	targetMapping: null,
	
	constructor: function(mapping) {
		// TODO (MG): find another solution to this, level is then no longer a true tree level, can't be counted
        //       on for algorithms that make use of level equality (like finding lowest common ancestor)
        this.host.level += 0.5;
        // push edge level to be greater then the default so they are processed after

		
		this._unpositioned = true;
	},
	
	_invalidate: function() {
		this.inherited(arguments);
		
		this.getInternalGraphElement().invalidate();
		
		// special code to ensure real owning graph manager invalidates the edge
		if (this._parentMapping) {
			this.getInternalGraphElement().getGraph().mapping.policies.graphContainer.graphManager._addToInvalidMappings(this.host);
		}
	},
	
	_completeInternalTraces: function() {
		this.inherited(arguments);
		
		// determine which aspects of notation aren't traced yet and add internal traces to complete
		var internalTraces = [];
		
		this._createInternalTrace("sourceAnchor", null, internalTraces);
		this._createInternalTrace("targetAnchor", null, internalTraces);
		
		// default value for router trace will mach that defined in QVT, or Manual if default not defined
		var graphSource = this.host.sources.__metadata;		
		if (!graphSource || !graphSource.router || graphSource.router === "oblique") {
			this._createInternalTrace("router", rmpw.diagrams.render.graph.BaseGraphConfigurator.prototype.OBLIQUE_ROUTING, internalTraces);
		} else {
			this._createInternalTrace("router", rmpw.diagrams.render.graph.BaseGraphConfigurator.prototype.RECTILINEAR_ROUTING, internalTraces);
		}
		
		this._createInternalTrace("bendpoints", [], internalTraces);
		
		this._createInternalTrace("smoothness", "none", internalTraces);
		this._createInternalTrace("avoidObstructions", true, internalTraces);
		
		if (internalTraces.length > 0) {
			this.host.connectTraces(this.host.sources.__metadata, internalTraces);
		}
	},
	
	destroy: function() {
	    var anchorable;
		if(this.sourceMapping) {
			anchorable = this.getAnchorable(this.sourceMapping);
			if (anchorable) {
				anchorable.removeSourceConnection(this.host);
			}
		}
		
		if(this.targetMapping) {
			anchorable = this.getAnchorable(this.targetMapping);
			if (anchorable) {
				anchorable.removeTargetConnection(this.host);
			}
		}
		
		this._detachListeners(this._sourceMappingInvalidationListeners);
		this._detachListeners(this._targetMappingInvalidationListeners);
		
		this._parentMapping = null;
		
		this.inherited(arguments);
	},

	/* 
	 * Returns the dojo shape associated with this graph policy element
	 */
	getDojoShape: function() {
		return this._hitzone ? this._hitzone : this._polyline;
	},
	
	setPolyline: function(polyline) {
		this._polyline = polyline;
		this._setVisibility(this._polyline, !this._unpositioned);
	},
	
	setHead: function(polyline) {
		if (dojo.isString(polyline)) {
			this._headTemplate = polyline;
		} else {
			this._head = polyline;
			this._setVisibility(this._head, !this._unpositioned);
		}
	},
	
	setTail: function(polyline) {
		if (dojo.isString(polyline)) {
			this._tailTemplate = polyline;
		} else {
			this._tail = polyline;
			this._setVisibility(this._tail, !this._unpositioned);
		}
	},
	
	getSmoothness: function(smoothness) {
		return this.host.sources.__metadata.smoothness;
	},
	
	setSmoothness: function(smoothness) {
		this.host.sources.__metadata.smoothness= smoothness; 
	},
	
	_getSmoothnessValue: function() {
		switch(this.host.sources.__metadata.smoothness) {
		case "none":
			return 0; 
		case "less":
			return 15; 
		case "normal":
			return 30; 
		case "more":
			return 45; 
		}
		
		return 0;
	},

	/*
	 * @deprecated - use setShape instead
	 */
	setFigure: function(figure) {
		this.setShape(figure);
	},
	
	setShape: function(figure) {
		this.inherited(arguments);
		
		if(dojox.gfx.svg) {
			// a thicker, hidden line for catching mouse hits to make the connection easier to select
			this._hitzone = figure.createPolyline();
			this._hitzone.setStroke({color: [255, 0, 0, 0], width: 10});
			/*
			 * in firefox 3.6, they implemented svg's pointer-events property:
			 * - http://www.w3.org/TR/SVG11/interact.html#PointerEventsProperty
			 * - https://bugzilla.mozilla.org/show_bug.cgi?id=563500 (bug about change from 3.5 > 3.6)
			 * 
			 * This requires us to set pointer-events explicitly to "stroke", or the zero opacity above
			 * will disqualify us from events under the default "visiblePainted" rules
			 */
		
			this._hitzone.rawNode.setAttribute("pointer-events", "stroke");
		}
		
		var mapping = this.host;
		var that = this;
		dojo.subscribe('gef/diagram/avoidObstructions', dojo.hitch(this, function(avoid) {
			that.getInternalGraphElement().setAvoidObstructions(avoid);
			mapping.invalidate();
		}));
	},
	
	/*
	 * Find the appropriate graph container for this edge - use the source mapping container
	 */
	getGraphContainer: function() {
		if (this._parentMapping) {
			return this._parentMapping.policies.graphElement.getGraphContainer();
		}
		else if (this.sourceMapping) {
			return this.sourceMapping.policies.graphElement.getGraphContainer();
		}
		else {
			return this.inherited(arguments);
		}
	},
	
	_createInternalGraphElement : function() {
		return this.host.viewer.graphConfigurator.createGraphEdge(this.host, 
				this.getGraphContainer().getInternalGraphElement(), 
				this.sourceMapping ? this.sourceMapping.policies.graphElement.getInternalGraphElement() : null, 
				this.targetMapping ? this.targetMapping.policies.graphElement.getInternalGraphElement() : null);
	},

	_getAttachedGraphElement: function(viewMappings, isSource) {
		if (viewMappings) {
			for (var i=0, viewMapping; viewMapping = viewMappings[i]; i++) {
				var anchorable = this.getAnchorable(viewMapping);
				if (anchorable && (viewMapping.root === this.host.root) && 
					viewMapping.parent.getDefaultSource() != viewMapping.getDefaultSource() &&
					// if source make sure this mapping doesn't already have this edge
					!this._hasThisEdge(isSource ? anchorable.ownedConnections : anchorable.targetConnections) ) {
					return viewMapping;
				}
			}
		}
		
		return null;
	},

	_hasThisEdge: function(connections) {
		for (var i=0, edgeMapping; edgeMapping=connections[i]; i++) {
			if (edgeMapping.hasSameModelSource(this.host) && this.host.declaredClass.localeCompare(edgeMapping.declaredClass) === 0) {
				return true;
			}
		}
		return false;
	},
	
	/*
	 * method that gets called for an extra logic required by graph elements for invalidation
	 */
	enforce: function() {
		var viewMapping;
		if (this._sourcesTarget && this._sourcesTarget.__mappings) {
			viewMapping = this._getAttachedGraphElement(this._sourcesTarget.__mappings);
			if(viewMapping) {
				this._setTargetMapping(viewMapping);
				delete this._sourcesTarget;
			} else {
				this.host.invalidate();
				return;
			}
		}
		
		if (this._sourcesSource && this._sourcesSource.__mappings) {
			viewMapping = this._getAttachedGraphElement(this._sourcesSource.__mappings, true);
			if(viewMapping) {
				this._setSourceMapping(viewMapping);
				delete this._sourcesSource;
			} else {
				this.host.invalidate();
				return;
			}
		}
		
		this._checkForReparent();
		
        this._refreshDecoration("_head", "_headTemplate");
        this._refreshDecoration("_tail", "_tailTemplate");
		
		// Note source and target must be set before source anchor and target anchor can be set
		this._enforceInternalTrace("sourceAnchor", "setSourceAnchor");
		this._enforceInternalTrace("targetAnchor", "setTargetAnchor");
		this._enforceInternalTrace("router", "setRouterType");
		this._enforceInternalTrace("bendpoints", "setBendpoints");
		this._enforceInternalTrace("smoothness", "setSmoothness");
		this._enforceInternalTrace("avoidObstructions", "setAvoidObstructions");
		
		// need to do this after we're sure initialization is complete - 
		// otherwise getInternalGraphElement may be null if source isn't resolved yet
		this.getInternalGraphElement().setAvoidObstructions(this.getAvoidObstructions());
		
		this.inherited(arguments);
	},
	
	_hasParent: function(mapping, parentMapping) {
		while (mapping.parent && mapping.parent != parentMapping) {
			mapping = mapping.parent;
		}
		
		if (mapping.parent == parentMapping) {
			return true;
		}
		
		return false;
	},
	
	/*
	 * Reparent the edge to the diagram if the source and target have different parents - otherwise, the
	 * shape z-order can obscure the edges from view.
	 * 
	 * It's only a problem if the source is level lower or even with the target level (otherwise, z-order will be ok)
	 */
	_checkForReparent: function() {
		if(this.sourceMapping && this.targetMapping && this._shape && 
		   this.sourceMapping.parent != this.targetMapping.parent) {
            var edgeParentShape = this._shape.parent;
            var newEdgeParentShape = edgeParentShape;
            
            if (this.sourceMapping.parent.parent == this.sourceMapping.root) {
            	return;
            }
            
            // find the new parent shape which is common to both target and source
            this._parentMapping = this.sourceMapping.parent.parent;
            while (!this._hasParent(this.targetMapping, this._parentMapping) && this._parentMapping.parent != this._parentMapping.root) {
            	this._parentMapping = this._parentMapping.parent;
            }
            
            newEdgeParentShape = this._parentMapping.policies.graphElement.getDojoShape().parent;
            
            if(edgeParentShape !== newEdgeParentShape) {
                if (dojo.isIE != 8) {
                	this._doReparent(edgeParentShape, newEdgeParentShape);
                } else {
                	// doing this immediately with IE8 causes the browser to terminate :(
	        		this._callbackTimeout = setTimeout( dojo.hitch(this, function() {
	        			if (this._callbackTimeout) {
	        				clearTimeout(this._callbackTimeout);
	        			}
	        			this._doReparent(edgeParentShape, newEdgeParentShape);
	             	}), 0);
                }
            }
            else {
            	this._shape.moveToFront();  // ensure the edge is in front
            }
        }	
	},
	
	_doReparent: function(oldParent, newParent) {
		oldParent.remove(this._shape);
		newParent.add(this._shape);
        this._shape.moveToFront();
        this.getInternalGraphElement().destroy();

        if (this.host.sources.__metadata.router) {
            this.getInternalGraphElement().setRouterType(this.host.sources.__metadata.router);
        }
        if (this.host.sources.__metadata.bendpoints) {
            this.getInternalGraphElement().setRouterConstraint(this.host.sources.__metadata.bendpoints, true);
        } 
	},
	
	_refreshDecoration: function(shape, temp) {
		if (this._polyline) {
			var styleable = this.host.policies.styleable;
			var isStyled = styleable && styleable.isLineStyled(this._polyline);
			if (isStyled) {
				styleable.removeLineShape(shape);
				styleable.removeFillShape(shape);
			}
			
			var rawTemplate = this[temp];
			if (rawTemplate) {
				// if edge already has a decoration, remove it
				if (this[shape]) {
					this[shape].removeShape();
				}
				// use template object to create new decoration 
				var template = dojo.fromJson(rawTemplate); 
				if (template.path || template.circle || template.rectangle) {
					var figure;
					if (template.path) {
						figure = this._polyline.getParent().createPath({path: template.path});
					} else if (template.circle) {
						figure = this._polyline.getParent().createCircle({cx: template.circle.cx, cy: template.circle.cy, r: template.circle.r});
					} else {
						figure = this._polyline.getParent().createRect({x: template.rectangle.x, y: template.rectangle.y, width: template.rectangle.width, height: template.rectangle.height });
					}
					if (figure) {
						if (template.strokeColor) {
							figure.setStroke({color: this._polyline.getStroke().color, width: template.strokeWidth});
							if (isStyled) {
								styleable.setLineShape(figure);
							}
						}
						if (template.fillColor) {
							figure.setFill(template.fillColor);
							if (isStyled) {
								styleable.setFillShape(figure);
							}
						}
						this[shape] = figure;
					}
				} else {
					this[shape] = null;
				}
				delete this[temp];
			}
		}
	},
	
	getState: function() {
		var state = {};
		
		var sourceAnchor = this._getInternalTraceData("sourceAnchor");
		if (sourceAnchor != null) {
			state.sourceAnchor = sourceAnchor;
		}
		var targetAnchor = this._getInternalTraceData("targetAnchor");
		if (targetAnchor != null) {
			state.targetAnchor = targetAnchor;
		}
		var router = this._getInternalTraceData("router");
		if (router != rmpw.diagrams.render.graph.BaseGraphConfigurator.prototype.OBLIQUE_ROUTING) {
			state.router = router;
		}
		var smoothness = this.getSmoothness();
		if (smoothness) {
			state.smoothness = smoothness;
		}
		var avoidObstructions = this._getInternalTraceData("avoidObstructions");
		if (avoidObstructions !== undefined) {
			state.avoidObstructions = avoidObstructions;
		}
		state.bendpoints = dojo.clone(this._getInternalTraceData("bendpoints"));
		return state;
	},
	
	setState: function(state) {
		if (!state) {
			return;
		}
		
		var stateChanged = false;
		if (state.sourceAnchor) {
			this.setSourceAnchor(state.sourceAnchor);
			stateChanged = true;
		}
		if (state.targetAnchor) {
			this.setTargetAnchor(state.targetAnchor);
			stateChanged = true;
		}
		if (state.router && state.router != rmpw.diagrams.render.graph.BaseGraphConfigurator.prototype.OBLIQUE_ROUTING) {
			this.setRouterType(state.router);
			stateChanged = true;
		}
		if (state.smoothness) {
			this.setSmoothness(state.smoothness);
			stateChanged = true;
		}
		if (state.avoidObstructions !== undefined) {  
			this.setAvoidObstructions(state.avoidObstructions);
			stateChanged = true;
		}
		if (state.bendpoints) {
			this.setBendpoints(state.bendpoints);
			stateChanged = true;
		}
		
		if (stateChanged) {
			this.getInternalGraphElement().invalidate();
		}
	},
	
	getAnimatedFigure: function() {
		if (this.isInitialized()) {
			return new rmpw.diagrams.render.graph.AnimatedEdge(this);
		}
		
		return null;
	},

	getRouterType: function() {
		return this.getInternalGraphElement().getRouterType();
	},
	
	setRouterType: function(router) {
		if (router !== undefined) {
			this.host.sources.__metadata.router = router;
		}
		
		this.getInternalGraphElement().setRouterType(router);
	},
	
	setBendpoints: function(bendpoints) {
		if (bendpoints !== undefined) {
			// nothing in rdf yet
			if (bendpoints === null) {
				bendpoints = [];
			} else if (dojo.isString(bendpoints)) {
				//convert back into array
				var bendpts = bendpoints.split(" ");
				bendpoints = [];
				for (var j = 0, lg = bendpts.length; j < lg; ) {
					bendpoints.push({sourceX: parseFloat(bendpts[j++]), sourceY: parseFloat(bendpts[j++]), 
						targetX: parseFloat(bendpts[j++]), targetY: parseFloat(bendpts[j++])});
				}
			}
			this.host.sources.__metadata.bendpoints = bendpoints;
		}
		
		this.getInternalGraphElement().setRouterConstraint(bendpoints, true);
		this.getInternalGraphElement().invalidate();
	},
	
	getAvoidObstructions: function() {
		return this.host.sources.__metadata.avoidObstructions;
	},
	
	setAvoidObstructions: function(avoidObstructions) {
		if (avoidObstructions !== undefined) {
			if (dojo.isString(avoidObstructions)) avoidObstructions = avoidObstructions == "true";
			this.host.sources.__metadata.avoidObstructions = avoidObstructions;
		}
	},
	
	setMaxBendpoints: function(maxBendpoints) {
		this.maxBendpoints = maxBendpoints;
	},
	
	getMaxBendpoints: function() {
		return this.maxBendpoints;
	},
	
	setSourceAnchor: function(sourceAnchor) {
		if (sourceAnchor !== undefined) {
			this.host.sources.__metadata.sourceAnchor = sourceAnchor;
		}
		
		var graphEl = this.getInternalGraphElement();
		if (graphEl.getSourceAnchor()) {
			if (sourceAnchor === null) {
				graphEl.getSourceAnchor().setRelativeReferencePoint(null);
			}
			else {
				graphEl.getSourceAnchor().setRelativeReferencePoint(graphEl.getSourceAnchor().parseTerminalString(sourceAnchor));
			}
		}
	},
	
	setTargetAnchor: function(targetAnchor) {
		if (targetAnchor !== undefined) {
			this.host.sources.__metadata.targetAnchor = targetAnchor;
		}
		
		var graphEl = this.getInternalGraphElement();
		if (graphEl.getTargetAnchor()) {
			if (targetAnchor === null) {
				graphEl.getTargetAnchor().setRelativeReferencePoint(null);
			}
			else {
				graphEl.getTargetAnchor().setRelativeReferencePoint(graphEl.getTargetAnchor().parseTerminalString(targetAnchor));
			}
		}
	},
	
	_setSourceMapping: function(sourceMapping) {
		this._detachListeners(this._sourceMappingInvalidationListeners);
				
		this.sourceMapping = sourceMapping;
		this.getInternalGraphElement().setSource(this.sourceMapping.policies.graphElement.getInternalGraphElement());
		var anchorable = this.getAnchorable(sourceMapping);
		if (anchorable) {
			this.getInternalGraphElement().setSourceAnchor(anchorable.getSourceConnectionAnchor(this.host));
			anchorable.addSourceConnection(this.host);
			this._addListeners(this._sourceMappingInvalidationListeners, this.sourceMapping);
		}
		else {
			console.debug("source isn't anchorable");
		}
	},
	
	setSource: function(source) {
		var self = this;
		this.host.getResolvedSemantic(source, function(realSource, awaken) {
			if(dojo.isArray(realSource)){
			    console.error("Array of sources not supported by GraphEdge, attempting to recover by using first element.");
				realSource = source[0];
			}
			
			if (self.sourceMapping ) {
				var anchorable = self.getAnchorable(self.sourceMapping);
				if (anchorable) {
					anchorable.removeSourceConnection(self.host);
				}
			}
				
			if (realSource) {
				var viewMapping = self._getAttachedGraphElement(realSource.__mappings, true);
				if (viewMapping) {
					self._setSourceMapping(viewMapping);
				}
				else {
					self._sourcesSource = realSource;
				}
				
				if (awaken) {
					self.host.awaken();	
				}
			}
		});
	},
	
	_setTargetMapping: function(targetMapping) {
		this._detachListeners(this._targetMappingInvalidationListeners);
		
		this.targetMapping = targetMapping;
		this.getInternalGraphElement().setTarget(this.targetMapping.policies.graphElement.getInternalGraphElement());
		var anchorable = this.getAnchorable(targetMapping);
		if (anchorable) {
			this.getInternalGraphElement().setTargetAnchor(anchorable.getTargetConnectionAnchor(this.host));
			anchorable.addTargetConnection(this.host);
			this._addListeners(this._targetMappingInvalidationListeners, this.targetMapping);
		}
		else {
			console.debug("target isn't anchorable");
		}
	},
	
	setTarget: function(target) {
		var self = this;
		this.host.getResolvedSemantic(target, function(realTarget, awaken) {
			if(dojo.isArray(realTarget)){
				console.error("Array of realTargets not supported by GraphEdge, attempting to recover by using first element.");
				realTarget = target[0];
			}
			
			if (self.targetMapping ) {
				var anchorable = self.getAnchorable(self.targetMapping);
				if (anchorable) {
					anchorable.removeTargetConnection(self.host);
				}
			}
			
			if (realTarget) {
				var viewMapping = self._getAttachedGraphElement(realTarget.__mappings);
				if (viewMapping) {
					self._setTargetMapping(viewMapping);
				}
				else {
					self._sourcesTarget = realTarget;
				}

				if (awaken) {
					self.host.awaken();	
				}
			}
		});
	},

	// listen to invalidation on mapping and any non-diagram container mappings
	_addListeners: function(listeners, mapping) {
		listeners = [];
		listeners.push(dojo.connect(mapping, "invalidate", this.host, "invalidate"));
		this._addParentListeners(listeners, mapping.parent);
	},
	
	// if a parent of this source/target is invalidated, we wait for it to be refreshed before invalidating
	// edge so that all parent nodes have their positions before the edge is routed
	_addParentListeners: function(listeners, mapping) {
		if (mapping && mapping.policies.graphContainer && !mapping.policies.graphContainer.isTopLevel()) {
			listeners.push(dojo.connect(mapping, "invalidate", dojo.hitch(this, function() {
				this.host.invalidate();					
			})));
			this._addParentListeners(listeners, mapping.parent);
		}
	},
	
	_detachListeners: function(listeners) {
		if (listeners) {
			dojo.forEach(listeners, dojo.disconnect);
			delete listeners;
		}
	},
	
	/*
	 * Gets the source connection anchor for this edge
	 */
	getSourceConnectionAnchor: function() {
		return this.getInternalGraphElement().getSourceAnchor();
	},
	
	/*
	 * Gets the target connection anchor for this edge
	 */
	getTargetConnectionAnchor: function() {
		return this.getInternalGraphElement().getTargetAnchor();
	},
	
	/*
	 * Returns boolean is true if this edge is orthogonal (straight) or false otherwise
	 */
	isOrthogonal: function() {
		return this.getInternalGraphElement().isOrthogonal();
	},
	
	/*
	 * Retrieve the polygon points - used for outline view and moveable anchor
	 */
	getPolygonShapes: function() {
		return [this.getPointList()];
	},
	
	/*
	 * Gets the points associated with this edge
	 */
	getPointList: function(smooth) {
		var pointList = this.getInternalGraphElement().getPointList();
		if ((smooth === undefined || smooth) && this._getSmoothnessValue()) {
			var util = new rmpw.diagrams.render.util.PointListUtilities();
			pointList = util.calcSmoothPolyline(pointList, this._getSmoothnessValue(), 15);
		}
		return pointList;
	},
	
	setPointList: function (points){
		var oldPoints = this.getDojoPoints();
		if (!this._oldPoints) {
			this._oldPoints = oldPoints;
		}
		this.updateShape();		
		this.getGraphManager().updateLocation(this.host);
	},
	
	_calculateAngle: function(p1x, p1y, p2x, p2y) {
		var dx = Math.abs(p2x - p1x);
		var dy = p2y - p1y;
		if(dx === 0 && dy === 0){ return 0; }
		var angle = Math.atan(dy/dx)/Math.PI*180;
		if(p2x - p1x < 0){
			angle = 180 - angle;
		}
		return angle;
	},
	
	/*
	 * Returns the bounds of the dojo shape
	 */
	getDojoPoints: function() {
		if (this._oldPoints) {
			return this._oldPoints;
		}
		
		var pts = this._polyline.shape.points;
		var relPts = new rmpw.diagrams.render.util.PointList();
		for (var i=0, pt; pt=pts[i]; i++) {
			relPts.addPoint_0(pt.x, pt.y);
		}
		return relPts;
	},
	
	updateShape: function () {
		try{
			var graphEl = this.getInternalGraphElement();
			if ((!this._polyline && !this._hitzone) || !graphEl.getTargetAnchor() || !graphEl.getSourceAnchor()) {
				return;
			}
			
			if (!this.isUsingDojoPosition()) {
				var pointList = this.getPointList();
				if (!this._getSmoothnessValue() && this.isOrthogonal()) {
					// orthogonal edge will have rounded corners (not customizable for now)
                    var util = new rmpw.diagrams.render.util.PointListUtilities();
					var rForBendpoint = [];
					var pl = util.calcRoundedCornersPolyline(pointList, 5, rForBendpoint, true);
					if (pl) {
						// just in case something goes wrong use the default polyline
						pointList = pl;
					}
				}
				var points = pointList.points;
				this._polyline.setShape(points);
				if (this._hitzone) {
					this._hitzone.setShape(points);
				}
		
				if(this._tail) {
					var firstPoint = points[0];
					var secondPoint = points[1];
					this._tail.setTransform({dx: firstPoint.x, dy: firstPoint.y});
					this._tail.applyTransform(dojox.gfx.matrix.rotategAt(
							this._calculateAngle(secondPoint.x, secondPoint.y, firstPoint.x, firstPoint.y), 0, 0));
				}
		
				if(this._head) {
					var len = points.length;
					if (len > 1) {
						var lastPoint = points[len - 1];
						var secondLastPoint = points[len - 2];
						this._head.setTransform({dx: lastPoint.x, dy: lastPoint.y});
						this._head.applyTransform(dojox.gfx.matrix.rotategAt(
								this._calculateAngle(secondLastPoint.x, secondLastPoint.y, lastPoint.x, lastPoint.y), 0, 0));
					}
				}
			}
		}
		finally {
			if (this._oldPoints) {
				delete this._oldPoints;
			}
		}
	},
	
	refresh: function() {
		this.inherited(arguments);
		
		// set point list just updates this._oldPoints
		this.setPointList(this.getPointList());

		if (this._unpositioned && this.sourceMapping && this.targetMapping) {
			delete this._unpositioned;
			this._setVisibility(this._polyline, true);
			this._setVisibility(this._head, true);
			this._setVisibility(this._tail, true);
		}
	},
	
	_setVisibility: function(shape, visible) {
		if (shape  && !dojo.isOffscreen) {
			if (visible) {
				shape.rawNode.style.visibility = '';
			} else {
				shape.rawNode.style.visibility = dojo.isIE === 8 ? "collapse"  : "hidden";
			}
		}
	}

});