/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010, 2013.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.diagrams.render.policies.GraphManager");

/**
 * Manages graph invalidation and location management.
 * 
 * @author Steve Shaw
 */
dojo.declare("rmpw.diagrams.render.policies.GraphManager", null, {
	_invalidNodeMappings: null,
	_invalidEdgeMappings: null,
	_invalidLabelMappings: null,
	
	ANIMATION_INCREMENT: 20,
	MIN_ANIMATION_DURATION: 30,
	MAX_ANIMATION_DURATION: 3000,
	
	BAGEXTENT: 150,
	SHAPE_BUFFER: 25,
	
	INTERSECT_ANY: 0,
	INTERSECT_BOUNDS: 1,
	INTERSECT_EXACT: 2, // includes INTERSECT_BOUNDS
	INTERSECT_IGNORE_INVALID: 4,
	INTERSECT_PEERS_ONLY: 8,
	
	constructor: function(diagram) {
		this._diagram = diagram;
		this._locationMap = {};
		this._bagMap = {};
		this._mappings = {};
		this._invalidIndex = {};

		// flat view of containment across all sub-graphs
		this._diagramLayer = diagram.root.diagramLayer;
		if (!dojo.isOffscreen && !this._diagramLayer.__focusableBoxes) {
			this._diagramLayer.__focusableBoxes = {};
			this._diagramLayer.__focusableMappings = {};
		}
	},
	
	destroy: function() {
	},
	
	/*
	 * Finds the particular bag associated with a modulated location
	 */
	_getBag: function(x, y) {
		var key = x + " " + y;
		
		var bag = this._bagMap[key];
		if (!bag) {
			bag = {location: {x: x, y: y}, weight: 0};
			this._bagMap[key] = bag;
		}
		
		return bag;
	},
	
	/*
	 * Gets all the bags associated with a particular bounding box
	 */
	_getBagsForBounds: function(box) {
		var bags = [];
		var rightBox = box.x + box.width;
		var leftBox = box.y + box.height;
		for (var x=Math.floor(((box.x - this.SHAPE_BUFFER) / this.BAGEXTENT)); x<Math.floor(((rightBox + this.SHAPE_BUFFER) / this.BAGEXTENT)) + 1; x++) {
			for (var y=Math.floor(((box.y - this.SHAPE_BUFFER) / this.BAGEXTENT)); y<Math.floor(((leftBox + this.SHAPE_BUFFER) / this.BAGEXTENT)) + 1; y++) {
				bags.push(this._getBag(x, y));
			}
		}
		
		return bags;
	},
	
	/*
	 * Remove all bags associated with a mapping element.
	 */
	_removeBags: function(mapping) {
		var oldBags = this._locationMap[mapping.id];
		if (oldBags) {
			var gm = this;
			dojo.forEach(oldBags, function(bag) {
				gm._removeFromBag(bag, mapping);
			});
			delete this._locationMap[mapping.id];
		}
	},
	
	/*
	 * Adds the mapping reference to a given bag
	 */
	_addToBag: function(bag, mapping) {
		if (!bag[mapping.id]) {
			bag[mapping.id] = true;
			this._locationMap[mapping.id].push(bag);
			if (!bag.weight) {
				bag.weight = 0;
			}
			bag.weight++;
		}
	},
	
	/*
	 * Removes the mapping reference from the given bag
	 */
	_removeFromBag: function(bag, mapping) {
		if (bag[mapping.id]) {
			delete bag[mapping.id];
			bag.weight--;
			if (bag.weight === 0) {
				// remove it from the master bag list
				var key = bag.x + " " + bag.y;
				delete this._bagMap[key];
			}
		}
	},
	
	/* 
	 * Sorts all bags by their weight from lowest to highest
	 */
	_sortBagsByWeight: function() {
		
		// _sortedBags is cached until updateLocation is called when it is reset
		if (!this._sortedBags) {
			this._sortedBags = [];
			for (var bagKey in this._bagMap) {
				this._sortedBags.push(this._bagMap[bagKey]);
			}
			
			this._sortedBags.sort( function(b1,b2) {
				if (b1.weight === b2.weight) {
					return 0;
				} else if (b1.weight < b2.weight) {
					return -1;
				} else {
					return 1;
				}
			});
		}
		
		return this._sortedBags;
	},
	
	/* 
	 * Finds an empty location in the diagram.  Uses the bag weight to find the emptiest bag and then searches 
	 * neighbours to find empty bag.
	 */
	findEmptyLocation: function() {
		var bags = this._sortBagsByWeight();
		
		for (var i=0, lg=bags.length; i<lg; i++) {
			var bag = bags[i];
			/* search neighbours for empty bag */
			var location = {x: bag.location.x * this.BAGEXTENT, y: bag.location.y * this.BAGEXTENT};
			var searchBox = { x: location.x > this.BAGEXTENT ? location.x - this.BAGEXTENT : this.BAGEXTENT, y: location.y > this.BAGEXTENT ? location.y - this.BAGEXTENT : this.BAGEXTENT, 
								width: this.BAGEXTENT * (location.x > this.BAGEXTENT  ? 3 : 2), height: this.BAGEXTENT * (location.y > this.BAGEXTENT  ? 3 : 2)};
			var searchBags = this._getBagsForBounds(searchBox);
			for (var j=0, lg2=searchBags.length; j<lg2; j++) {
				var sbag = searchBags[j];
				if (sbag.weight === 0) {
					location = {x: sbag.location.x * this.BAGEXTENT, y: sbag.location.y * this.BAGEXTENT};
					return {x: location.x + this.SHAPE_BUFFER, y: location.y + this.SHAPE_BUFFER};
				}
			}
		}
		
		// should only happen with an empty diagram
		return {x: this.SHAPE_BUFFER, y: this.SHAPE_BUFFER};
	},
	
	/*
	 * Update the location of this mapping object
	 */
	updateLocation: function(mapping) {
		// TODO (MG): shouldn't have to check this, only happens if mapping was destroyed before calling updateLocation on it
		if(mapping.policies) {
			var box = mapping.policies.graphElement.getTightBounds();
			
			// remove old entry from all bags
			this._removeBags(mapping);
			
			this._locationMap[mapping.id] = [];
			if (this._sortedBags) {
				delete this._sortedBags;
			}
			
			// determine which bags this mapping belongs to and associate them with the mapping.
			var newBags = this._getBagsForBounds(box);
			var gm = this;
			dojo.forEach(newBags, function(bag) {
				gm._addToBag(bag, mapping);
			});
			
			this._mappings[mapping.id] = mapping;
			
			// if this node is focusable, add to static map (for touch devices that don't have mouseenter/leave)
			if (mapping.__editController && mapping.__editController.policies &&
				mapping.__editController.policies.focusable && !dojo.isOffscreen) {
				this._diagramLayer.__focusableBoxes[mapping.id] = box;
				this._diagramLayer.__focusableMappings[mapping.id] = mapping;
			}
		}
	},
	
	getFocusableFromPoint: function(pt) {
		var x = pt.x, y = pt.y;
		var lastId;
		for (var id in this._diagramLayer.__focusableBoxes) {
			var box = dojo.clone(this._diagramLayer.__focusableBoxes[id]);
			this._diagramLayer.__focusableMappings[id].policies.graphElement.translateToAbsolute(box);
			if (box.contains_0(pt)) {
				// look for innermost focusable
				if (!lastId || this._isChildOf(id, lastId)) {
					lastId = id;
				}
			}
			pt.x = x;
			pt.y = y;
		}
		return lastId ? this._diagramLayer.__focusableMappings[lastId] : null;
	},
	
	getContainerFromPoint: function(pt, ignoreContainers) {
		var x = pt.x, y = pt.y;
		var lastId;
		for (var id in this._diagramLayer.__focusableBoxes) {
			// only containers
			if (this._diagramLayer.__focusableMappings[id].policies.graphContainer) {
				if (!this._focusableMatch(id, ignoreContainers)) {
					var box = dojo.clone(this._diagramLayer.__focusableBoxes[id]);
					this._diagramLayer.__focusableMappings[id].policies.graphElement.translateToAbsolute(box);
					if (box.contains_0(pt)) {
						// look for innermost container
						if (!lastId || this._isChildOf(id, lastId)) {
							lastId = id;
						}
					}
					pt.x = x;
					pt.y = y;
				}
			}
		}
		return lastId ? this._diagramLayer.__focusableMappings[lastId] : null;
	},
	
	removeFocusables: function(focusables) {
		for (var id in this._diagramLayer.__focusableBoxes) {
			if (this._focusableMatch(id, focusables)) {
				delete this._diagramLayer.__focusableBoxes[id];
			}
		}
	},
	
	_focusableMatch: function(id, focusables) {
		if (dojo.isArray(focusables)) {
			for(var i = 0, len = focusables.length; i < len; i++) {
				if (id == focusables[i].renderMapping.id) {
					return true;
				}
			}
			return false;
		}
		return id == focusables;
	},
	
	_isChildOf: function(childId, parentId) {
		var childMapping = this._diagramLayer.__focusableMappings[childId];
		var walker = childMapping;
		while(walker) {
			// === breaks
			if (walker.id == parentId) {
				return true;
			}
			walker = walker.parent;
		}
		return false;
		
	},
	
	/*
	 * Companion method to getNearMappings which retrieves all the corresponding near figures
	 */
	getNearFigures: function(mapping, mappingPoints, prefixes, intersectType, boundsBuffer) {
		var bags = mappingPoints ? this._getBagsForBounds(mappingPoints.getBounds()) : this._locationMap[mapping.id];
		var nearMappings = this._getMappingsFromBags(bags, mapping.id, prefixes);
		if (intersectType !== this.INTERSECT_ANY) {
			nearMappings = this._getNearMappingsForIntersectType(mapping, mappingPoints, nearMappings, intersectType, boundsBuffer);
		}

		var figures = [];
		for (var id in nearMappings) {
			figures.push(nearMappings[id].policies.graphElement.getInternalGraphElement());
		}

		return figures;
	},
	
	/*
	 * return map of mappings common to bags
	 */
	_getMappingsFromBags: function(bags, excludeID, prefixes) {
		var nearMappings = {};
		
		if (bags) {
			var mappings = this._mappings;
			dojo.forEach(bags, function(bag) {
				for (var id in bag) {
					if (id && id !== excludeID && id !== "location" && id !== "weight") {
						if (!prefixes) {
							nearMappings[id] = mappings[id];
						}
						else {
							for (var i=0, prefix; prefix = prefixes[i]; i++ ) {
								if (prefix in mappings[id].policies) {
									nearMappings[id] = mappings[id];
									break;
								}
							}
						}
					}
				}
			});
		}
		
		return nearMappings;
	},
	
	_relatedTo: function(checkMapping, relativeMapping) {
		return checkMapping.parent === relativeMapping || 
			(checkMapping.policies.graphEdge && 
			 checkMapping.policies.graphEdge.sourceMapping && 
			 checkMapping.policies.graphEdge.targetMapping &&
			(checkMapping.policies.graphEdge.sourceMapping === relativeMapping || checkMapping.policies.graphEdge.targetMapping === relativeMapping));
	},
	
	/*
	 * Companion method to getNearMappings which retrieves all the corresponding near figures
	 */
	_getNearMappingsForIntersectType: function(mapping, mappingPoints, nearMappings, intersectType, boundsBuffer) {
		var mappingBounds = boundsBuffer ? mappingPoints.getBounds().shrink(-boundsBuffer, -boundsBuffer) : mappingPoints.getBounds();
		var mappings = [];
		if (intersectType === this.INTERSECT_ANY) {
			return nearMappings;
		}
		
		var ibounds = (intersectType & this.INTERSECT_BOUNDS);
		var iexact = (intersectType & this.INTERSECT_EXACT);
		var iignoreinvalid = (intersectType & this.INTERSECT_IGNORE_INVALID);
		var ipeersOnly = (intersectType & this.INTERSECT_PEERS_ONLY);
		
		for (var id in nearMappings) {
			var nearMapping = nearMappings[id];
			
			if (mapping) {
				// ignore any mappings that are currently invalid or not in the right containment
				if ((iignoreinvalid && nearMapping.id in this._invalidIndex) ||
						ipeersOnly && nearMapping.parent !== mapping.parent) {
					continue;
				}
			
				// always ignore children and connected edges
				if (this._relatedTo(nearMapping, mapping) || 
						this._relatedTo(mapping, nearMapping)) {
					continue;
				}
			}
			
			var nearMappingBounds = nearMapping.policies.graphElement.getTightBounds();
			if ((ibounds || iexact) &&
					mappingBounds.intersectsWith(nearMappingBounds)) {
				if (ibounds) {
					mappings.push(nearMapping);
				}
				else {
					// do a more complex intersection check
					for (var i=0; i<mappingPoints.size - 1; i++) {
						var segPts = new rmpw.diagrams.render.util.PointList();
						segPts.addPoint_1(mappingPoints.getPoint(i));
						segPts.addPoint_1(mappingPoints.getPoint(i+1));
						var segBounds = segPts.getBounds();
						if (nearMappingBounds.intersectsWith(segBounds)) {
							mappings.push(nearMapping);
							break;
						}
					}
				}
			}
		}
		
		return mappings;
	},
	
	/*
	 * Finds which mappings are in proximity to this bounding box and potentially intersect
	 */
	getNearMappingsForBounds: function(bounds, prefixes, intersectType, boundsBuffer) {
		var bags = this._getBagsForBounds(bounds);
		var nearMappings = this._getMappingsFromBags(bags, null, prefixes);
		
		var polyRect = new rmpw.diagrams.render.util.PointList();
		polyRect.addPoint_1(new rmpw.diagrams.render.util.Point(bounds.x, bounds.y));
		polyRect.addPoint_1(new rmpw.diagrams.render.util.Point(bounds.x + bounds.width, bounds.y));
		polyRect.addPoint_1(new rmpw.diagrams.render.util.Point(bounds.x + bounds.width, bounds.y + bounds.height));
		polyRect.addPoint_1(new rmpw.diagrams.render.util.Point(bounds.x, bounds.y + bounds.height));
		polyRect.addPoint_1(new rmpw.diagrams.render.util.Point(bounds.x, bounds.y));
		return this._getNearMappingsForIntersectType(null, polyRect, nearMappings, intersectType, boundsBuffer);
	},
	
	/*
	 * Finds which mappings are in proximity to this mapping and potentially intersect
	 */
	getNearMappings: function(mapping, prefixes, intersectType, boundsBuffer) {
		var bags = this._locationMap[mapping.id];
		if (!bags) {
			bags = this._getBagsForBounds(mapping.policies.graphElement.getTightBounds());
		}
		var nearMappings = this._getMappingsFromBags(bags, mapping.id, prefixes);
		return this._getNearMappingsForIntersectType(mapping, mapping.policies.graphElement.getPolygonPoints(), nearMappings, intersectType, boundsBuffer);
	},
	
	/*
	 * Remove this mapping from the graph manager
	 */
	remove: function(mapping) {
		this._removeBags(mapping);
		delete this._locationMap[mapping.id];
		delete this._mappings[mapping.id];
		if (!dojo.isOffscreen) {
			delete this._diagramLayer.__focusableBoxes[mapping.id];
			delete this._diagramLayer.__focusableMappings[mapping.id];
		}
	},
	
	_addToInvalidMappings: function(mapping) {
		if (mapping.id in this._invalidIndex) {
			return;
		}
		
		this._invalidIndex[mapping.id] = true;
		
		if ("graphNode" in mapping.policies) {
			if (!this._invalidNodeMappings) {
				this._invalidNodeMappings = [];
			}
			this._invalidNodeMappings.push(mapping);
		} else if ("graphEdge" in mapping.policies) {
			if (!this._invalidEdgeMappings) {
				this._invalidEdgeMappings = [];
			}
			this._invalidEdgeMappings.push(mapping);
		} else if ("graphFloatingLabel" in mapping.policies) {
			if (!this._invalidLabelMappings) {
				this._invalidLabelMappings = [];
			}
			this._invalidLabelMappings.push(mapping);
		}
	},
	
	
	// do the refresh and keep track of all elements that were refreshed so we can
	// call the postRefreshHook.
	_doRefreshMappings: function(currentInvalidMappings, refreshedMappings) {
		var graphManager = this;
		dojo.forEach(currentInvalidMappings, function(mapping) {
			// TODO (MG): checking for mapping.policies is really checking to see if it's disposed.  When it is disposed,
			// it should be removed from this list instead before we even get here.
			if(mapping.policies) {
				graphManager.refreshMapping(mapping);
				refreshedMappings.push(mapping);
			}
		});
	},
	
	_refreshInvalidations: function(refreshedMappings, invalidMappingsProperty) {
		do {
			var currentInvalidMappings = [];
			dojo.forEach(this[invalidMappingsProperty], function(mapping) {
				currentInvalidMappings.push(mapping);
			});
			this[invalidMappingsProperty] = [];
			
			// do the refresh and keep track of all elements that were refreshed so we can
			// call the postRefreshHook.
			this._doRefreshMappings(currentInvalidMappings, refreshedMappings);
			
			// invalidation controller may have already done an onInvalidationComplete
			// containers use onInvalidationComplete to do a diagram refresh
			// so this could cause mappings to be invalidated in the middle of a diagram refresh
	/////		// if any secondary invalidations occurred - called the enforce logic
		///	this._diagram.viewer.getInvalidationController().processInvalidMappings();
		}
		while (this[invalidMappingsProperty] && this[invalidMappingsProperty].length > 0);
	},
	
	_refreshNodeInvalidations: function(refreshedMappings) {
		this._refreshInvalidations(refreshedMappings, "_invalidNodeMappings");
	},
	
	_refreshEdgeInvalidations: function(refreshedMappings) {
		this._refreshInvalidations(refreshedMappings, "_invalidEdgeMappings");
	},
	
	_refreshLabelInvalidations: function(refreshedMappings) {
		this._refreshInvalidations(refreshedMappings, "_invalidLabelMappings");
	},
	
	_doAnimation: function(refreshedMappings) {
		if (!this._diagram.viewer.allowAnimation || 
				(this._diagram.viewer.nonAnimatedMappings && this._diagram.viewer.nonAnimatedMappings.length === 0)) {
			return refreshedMappings;
		}
		
		var self = this;
		var animatedMappings = [];
		var nonAnimatedMappings = [];
		dojo.forEach(refreshedMappings, function(mapping) {
			if (mapping.__animatedFigure && self._canAnimate(mapping)) {
				animatedMappings.push(mapping);
			}
			else {
				nonAnimatedMappings.push(mapping);
			}
		});
		
		var adj = 0;
		if (this._diagram.viewer.nonAnimatedMappings) {
			adj = this._diagram.viewer.nonAnimatedMappings.length;
			this._diagram.viewer.nonAnimatedMappings = [];
		}
		
		if (animatedMappings.length === 0) {
			return refreshedMappings;
		}
		
		// calculate duration as a function of how many shapes are being animated - 
		// up to a limit of 8 seconds
		var duration = Math.max(this.MIN_ANIMATION_DURATION, 
				Math.min((this._getNodeCount(animatedMappings) + adj) * this.ANIMATION_INCREMENT, this.MAX_ANIMATION_DURATION));
		if (duration <= this.MIN_ANIMATION_DURATION) {
			return refreshedMappings;
		} else {
			var graphManager = this;
			graphManager._isAnimating = dojo.isOffscreen ? false : true;
			var anim = new dojo.Animation({
				curve: [0, 1], 
				duration: duration, 
				easing: null,
				beforeBegin: function(){
					for (var i=0, lg=animatedMappings.length; i<lg; i++) {
						if (animatedMappings[i].__animatedFigure) {
							animatedMappings[i].__animatedFigure.startingAnimation();
						}
					}
				},
				onAnimate: function(val){
					for (var i=0, lg=animatedMappings.length; i<lg; i++) {
						if (animatedMappings[i].__animatedFigure) {
							animatedMappings[i].__animatedFigure.runAnimationStep(val);
						}
					}
				},
				onEnd: function() {
					for (var i=0, lg=animatedMappings.length; i<lg; i++) {
						if (animatedMappings[i].__animatedFigure) {
							animatedMappings[i].__animatedFigure.endAnimation();
						}
						graphManager.postRefreshHook(animatedMappings[i]);
					}
					
					graphManager.onRefreshComplete();
				}
			});
			
			anim.play();
		}
		
		return nonAnimatedMappings;
	},
	
	_canAnimate: function(mapping) {
		if (this._diagram.viewer.nonAnimatedMappings && mapping.policies.graphNode) {
			for (var i=0, lg=this._diagram.viewer.nonAnimatedMappings.length; i<lg; i++) {
				if (mapping === this._diagram.viewer.nonAnimatedMappings[i].renderMapping) {
					return false;
				}
			}
		}
		return true;
	},
	
	_getNodeCount: function(animatedMappings) {
		var cnt = 0;
		for (var i=0, lg=animatedMappings.length; i<lg; i++) {
			if (animatedMappings[i].policies.graphNode) {
				cnt++;
			}
		}
		return cnt;
	},
	
	onRefreshComplete: function() {
		if (this._diagramLayer.outline.isUsingFixedDiagram() && (this._diagramLayer.LAYER_TYPE == "diagramLayer")) {
			this._diagram.viewer.setupSizeAfterLayoutComplete();
		}
		this._clear();
		delete this._isAnimating;
	},
	
	refreshDiagram: function() {
		if (this._isAnimating) {
			this._diagram.invalidate();
			return;
		}
		
		try {
			if (!this._invalidNodeMappings &&
				!this._invalidEdgeMappings &&
				!this._invalidLabelMappings) {
				// nothing invalid return
				return;
			}
			
			// loop through refreshMapping until no more invalidations occur
			var refreshedMappings = [];
			this._refreshNodeInvalidations(refreshedMappings);
			
			this._diagram.policies.graphContainer.getInternalGraphElement().routeInvalidEdges();
			
			this._refreshEdgeInvalidations(refreshedMappings);
			this._refreshLabelInvalidations(refreshedMappings);
			
			var nonAnimatedMappings = this._doAnimation(refreshedMappings);
			
			if (nonAnimatedMappings.length > 0) {
				var graphManager = this;
				dojo.forEach(nonAnimatedMappings, function(mapping) {
					graphManager.postRefreshHook(mapping);
				});
			}
			
			if (!this._isAnimating) {
				this.onRefreshComplete();
			}
		}
		catch (e) {
			this._clear();
		}
	},
	
	_clear: function() {
		this._invalidNodeMappings = null;
		this._invalidEdgeMappings = null;
		this._invalidLabelMappings = null;
		this._invalidIndex = {};
	},
	
	refreshMapping: function(mapping) {
		if (mapping.viewer.allowAnimation) {
			var animatedFigure = null;
			if (mapping.policies.graphEdge) {
				animatedFigure = mapping.policies.graphEdge.getAnimatedFigure();
			}
			else if (mapping.policies.graphNode) {
				animatedFigure = mapping.policies.graphNode.getAnimatedFigure();
			}
			
			mapping.__animatedFigure = animatedFigure;
		}
		
		mapping.policies.graphElement.refresh();
	},
	
	postRefreshHook: function(mapping) {
		if (mapping.__animatedFigure) {
			delete mapping.__animatedFigure;
		}
		if (mapping.policies) {
			mapping.policies.graphElement.updateShape();
		}
	}
});