/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010, 2013.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
dojo.provide("rmpw.diagrams.render.policies.GraphNode");

dojo.require("rmpw.diagrams.render.policies.GraphElement");
dojo.require("rmpw.diagrams.render.policies.GraphAnchorable");
dojo.require("rmpw.diagrams.render.graph.AnimatedNode");
dojo.require("rmpw.diagrams.render.util.Workarounds");
dojo.require("rmpw.diagrams.render.util.Rectangle");
dojo.require("rmpw.diagrams.render.util.PointListUtilities");

/**
 * Data needed to layout the graph node.  Graph nodes are anchorable by default
 * 
 * @author Steve Shaw
 */
dojo.declare("rmpw.diagrams.render.policies.GraphNode", [rmpw.diagrams.render.policies.GraphElement], {
	singleton: ["graphElement", "graphNode"],
	
	constructor: function() {
		// ensure graph node is anchorable
		if(!this.host.policies.anchorable) {
			this.host.createPolicy(rmpw.diagrams.render.policies.GraphAnchorable);
		}
		this._unpositioned = true;
		if (this.host.rootShapes.length>0 && !dojo.isOffscreen) {
		    this.host.rootShapes[0].rawNode.style.visibility = dojo.isIE === 8 ? "collapse"  : "hidden";
	    }
	},
	
	/*
	 * Allows for AddElement action to set an initial position in lieue of no notation model existing
	 */
	setInitialPosition: function(x, y) {
		this._initialPosition = {x: x, y: y};
	},
	
	_getInitialPosition: function() {
		if (this._initialPosition) {
			return this._initialPosition;
		}
		else {
			var viewportRect = this.getGraphContainer().getBounds();
			var bounds = this.getBounds();
			
			return { x: viewportRect.width / 2 - (bounds.width / 2), 
					 y: viewportRect.height / 2 - (bounds.height / 2)};
		}
	},
	
	_completeInternalTraces: function() {
		this.inherited(arguments);
		
		// determine which aspects of notation aren't traced yet and add internal traces to complete
		var initialPos = this._getInitialPosition();
		var internalTraces = [];
		
		this._createInternalTrace("x", initialPos.x, internalTraces);
		this._createInternalTrace("y", initialPos.y, internalTraces);
		
		if (internalTraces.length > 0) {
			this.host.connectTraces(this.host.sources.__metadata, internalTraces);
		}
	},
	
	setX : function(x) {
		if (x ) {
			this.host.sources.__metadata.x = dojo.isString(x) ? parseFloat(x) : x;
		}
	},
	
	setY : function(y) {
		if (y ) {
			this.host.sources.__metadata.y = dojo.isString(y) ? parseFloat(y) : y;
		}
	},
	
	_createInternalGraphElement: function() {
		var dojoB = this.getDojoBounds();
		return this.host.viewer.graphConfigurator.createGraphNode(this.host, 
				this.getGraphContainer().getInternalGraphElement(), {x: dojoB.x, y: dojoB.y, w: dojoB.width, h: dojoB.height});
	},
	
	getAnimatedFigure: function() {
		if (this.isInitialized()) {
			return new rmpw.diagrams.render.graph.AnimatedNode(this);
		}
		
		return null;
	},
	
	/*
	 * method that gets called for an extra logic required by graph elements for invalidation
	 */
	enforce: function() {
		
		this._enforceInternalTrace("x", "setX");
		this._enforceInternalTrace("y", "setY");
		
		this.inherited(arguments);
	},
	
	getState: function() {
		var state = {};
		
		state.x = this._getInternalTraceData("x");
		state.y = this._getInternalTraceData("y");
		return state;
	},
	
	setState: function(state) {
		var layoutValid = false;
		
		if (state.x) {
			this.setX(state.x);	
			this.getInternalGraphElement().invalidate();
			layoutValid = true;
		}
		if (state.y) {
			this.setY(state.y);
			this.getInternalGraphElement().invalidate();
			layoutValid = true;
		}
		
		var parentContainer = this.getGraphContainer();
		if (layoutValid && parentContainer.host.policies.graphLayout) {
			parentContainer.host.policies.graphLayout.validateLayout(this.host);
		}
	},
	
	/*
	 * Update the outline shape with the polyline points
	 */
	updateOutlineShape: function(parentShape) {
		this.inherited(arguments);
		this.outlineShape.setFill("#BDD7EF"); // TODO Get from figure
	},

	/*
	 * Get the bounds of the constraint.  Need to query the dojo shape for the width
	 * and height since the shape may be auto-sized.
	 */
	getConstraintBounds: function() {
		var bounds = this.getDojoBounds(false);
		
		var nodeX, nodeY, nodeWidth, nodeHeight;
		if (this.isUsingDojoPosition()) {
			nodeX = bounds.x;
			nodeY = bounds.y;
			nodeWidth = bounds.width;
			nodeHeight = bounds.height;
		}
		else {
			nodeX = this.host.sources.__metadata.x === undefined ? bounds.x : this.host.sources.__metadata.x;
			nodeY = this.host.sources.__metadata.y === undefined ? bounds.y : this.host.sources.__metadata.y;
			nodeWidth = this.host.sources.__metadata.width === undefined || this.host.sources.__metadata.width <= 0 ? bounds.width : this.host.sources.__metadata.width;
			nodeHeight = this.host.sources.__metadata.height === undefined || this.host.sources.__metadata.height <= 0 ? bounds.height : this.host.sources.__metadata.height;
		}
		
		var loc = { x: nodeX, y: nodeY, width: nodeWidth, height: nodeHeight };
		
		return new rmpw.diagrams.render.util.Rectangle(loc.x, loc.y, loc.width, loc.height);
	},
	
	/*
	 * Returns the bounds of the dojo shape
	 */
	getDojoBounds: function(cache) {
	    var bounds;
	    
		if (cache && this._oldBounds) {
			return this._oldBounds;
		}
		
		// if this shape's graph container is not the diagram 
		// just get transformed bounds up to that container
		var containerMapping = this.getGraphContainer().host;
		if (containerMapping && containerMapping.policies && containerMapping.policies.graphNode) {
			bounds = rmpw.diagrams.render.util.Workarounds.getTransformedBoundingBox(
					this.getDojoShape(), rmpw.diagrams.render.util.Workarounds.IGNORE_DECORATIONS,
					containerMapping.policies.graphNode.getDojoShape());
			return bounds;
		}
		
		// else calculate from diagram
		bounds = rmpw.diagrams.render.util.Workarounds.getTransformedBoundingBox(
				this.getDojoShape(), rmpw.diagrams.render.util.Workarounds.IGNORE_DECORATIONS);
		var relBounds = new rmpw.diagrams.render.util.Rectangle(bounds.x, bounds.y, bounds.width, bounds.height);
		this.translateToRelative(relBounds);
		return relBounds;
	},
	
	/*
	 * Finalizes any updates from the graph manager and ensures the dojo 
	 * shape has been updated accordingly
	 */
	updateShape: function () {
		try{
			if (!this._oldBounds) {
				// already updated - return
				return;
			}

			if (!this.isUsingDojoPosition()) {
				var figBounds = this.getBounds();
				this._shape.applyTransform({
					dx: figBounds.x - this._oldBounds.x, 
					dy: figBounds.y - this._oldBounds.y
				});
			}
		}

		finally {
			if (this._oldBounds) {
				delete this._oldBounds;
			}
			
			if (this._unpositioned) {
				delete this._unpositioned;
				if (this.host.rootShapes.length>0 && !dojo.isOffscreen){
				    this.host.rootShapes[0].rawNode.style.visibility = '';
				}
			}
		}
	},
	
	/*
	 * Sets the bounds of this floating label.  Usually called in response to
	 * the owning element and adjusts this floating label according to it's
	 * relative position to the owner.
	 */
	setBounds: function(rect) {
		var oldBounds = this.getDojoBounds();
		if (!this._oldBounds) {
			this._oldBounds = oldBounds;
		}
		
		this.resetPolygonShapes();
		this.getInternalGraphElement().setBounds(rect);
		this.updateShape();
		
		var graphManager = this.getGraphManager();
		graphManager.updateLocation(this.host);
	},
	
	/*
	 * Override since anchorable polyline is often too detailed to make use as a highlight.
	 * If users want the advanced polyline view, they should use GraphAnchorable policy to get this.
	 */
	getPolygonPoints: function() {
		return rmpw.diagrams.render.util.PointListUtilities.prototype.getRectanglePolyline(this.getBounds());
	},
	
	refresh: function() {
		this.inherited(arguments);
				
		var modelBounds = this.getConstraintBounds();
		this.setBounds(modelBounds);

		// notify when bounds changes
		if (!this._refreshedBnds  || 
				modelBounds.x != this._refreshedBnds.x || modelBounds.y != this._refreshedBnds.y || 
				modelBounds.width != this._refreshedBnds.width || modelBounds.height != this._refreshedBnds.height) {
			this.onBoundsChanged();
		}
		this._refreshedBnds = dojo.clone(modelBounds);
	}
});