/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.diagrams.render.policies.GridContainer");

dojo.require("rmpw.diagrams.render.util.Workarounds");

/**
 * @author Marc Gobeil
 */
dojo.declare("rmpw.diagrams.render.policies.GridContainer", null, {
	singleton: "gridContainer",
	
	setContainer: function(container) {
		if(!container) {
			throw "container cannot be null";
		}
		this._container = container;
	},
	
	setTopHeader: function(header) {
		this._topHeader = header;
	},
	
	setRows: function(rows) {
		this._rows = rows;
	},
		
	setColumns: function(cols) {
		this._cols = cols;
	},
	
	setCellSpacing: function(spacing) {
		this._cellSpacing = spacing;
	},
	
	enforce: function() {
		var cells = this._container.children,
		  topSizes;
		// TODO (MG): allow either rows or columns to be master
		if(!this._cols) {
			if(this._topHeader) {
				this._cols = this._topHeader.children.length;
			} else {
				this._cols = cells.length;
			}
		}
		this._rows = Math.ceil(cells.length / this._cols);

		if(!this._cellSpacing) {
			this._cellSpacing = 0;
		}

		var cellGrid = this._buildGrid(cells);
		var marginTop = 0;
		var marginLeft = 0;
		
		var topGrid;
		if(this._topHeader) {
			topGrid = this._buildGrid(this._topHeader.children);
			topSizes = this._calculateBandSizes(topGrid);
			marginTop += topSizes[this._heightIndex(0)];
		}
		// TODO (MG): other sides
		
		var sizes = this._calculateBandSizes(cellGrid, topSizes);
		this._assignCellTransforms(cellGrid, sizes, marginLeft, marginTop);
		
		if(this._topHeader) {		
			this._assignCellTransforms(topGrid, sizes, marginLeft);
		}
	},
	
	/**
	 * @param cells
	 * @returns
	 */
	_buildGrid: function(cells) {
		var grid = [];
		var i = 0;
		var cellCount = cells.length;
		while(i < cellCount) {
			var cell = cells[i];
			var bounds = rmpw.diagrams.render.util.Workarounds.getBoundingBox(cell);
			if(bounds) {
				grid[i] = {
					figure: cell,
					width: bounds.width + this._cellSpacing,
					height: bounds.height + this._cellSpacing
				};
			}
			i++;
		}
		return grid;
	},
	
	/**
	 * Retrieves a one dimensional array index for a two dimensional coordinate
	 * in row-major order. 
	 */
	_gridIndex: function(col, row) {
		return row * this._cols + col;
	},
	
	_prependHeaderRow: function(grid) {
		for(var i = 0; i < this._cols; i++) {
			var cell = this._topHeader.children[i];
			var bounds = rmpw.diagrams.render.util.Workarounds.getBoundingBox(cell);
			grid.unshift({
				figure: cell,
				width: bounds.width,
				height: bounds.height
			});
		}
	},

	/**
	 * Builds an array where first this._rows elements are the row heights,
	 * and the next this._cols elements are the column widths.
	 * 
	 * e.g. for a 3x2 grid [h,h,w,w,w]
	 * 
	 * @returns array of grid dimensions
	 */
	_calculateBandSizes: function(grid, topSizes) {
		var sizes = [],
		col;
		
		if(topSizes) {
			// initialize max widths to widths of top header columns
			for(col = 0; col < this._cols; col++) {
				var widthIndex = this._widthIndex(col);
				sizes[widthIndex] = topSizes[widthIndex];
			}
		}
		
		for(var row = 0; row < this._rows; row++) {
			var rowHeight = 0;
			for(col = 0; col < this._cols; col++) {
				var cell = grid[this._gridIndex(col, row)];
				if(cell) {
					// advance row height
					rowHeight = Math.max(rowHeight, cell.height);
					// advance column width
					var columnIndex = this._widthIndex(col);
					sizes[columnIndex] = Math.max(sizes[columnIndex] || 0, cell.width);
				}
			}
			var heightIndex = this._heightIndex(row);
			sizes[heightIndex] = Math.max(sizes[heightIndex] || 0, rowHeight);
		}
		return sizes;
	},
	
	_heightIndex: function(row) {
		return row;
	},
	
	_widthIndex: function(col) {
		return this._rows + col;
	},
	
	_assignCellTransforms: function(grid, sizes, initialX, initialY) {
		initialX = initialX || 0;
		initialY = initialY || 0;
		var x = initialX; 
		var y = initialY; 
		for(var row = 0; row < this._rows; row++) {
			x = initialX;
			for(var col = 0; col < this._cols; col++) {
				var cell = grid[this._gridIndex(col, row)];
				if(cell) {
					cell.figure.setTransform({dx: x, dy: y});
				}
				x += sizes[this._widthIndex(col)];
			}
			y += sizes[this._heightIndex(row)];
		}
	}
	
});