/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.diagrams.render.policies.StackLayout");

dojo.require("rmpw.diagrams.render.util.Workarounds");
dojo.require("rmpw.diagrams.render.util.ResizeObject");
dojo.require("rmpw.common.policy.Policy");

/**
 * Policy that stacks shapes either vertically or horizontally next to each other
 * with configurable spacing.
 * 
 * @author Marc Gobeil
 */
dojo.declare("rmpw.diagrams.render.policies.StackLayout", rmpw.common.policy.Policy, {
	
	/**
	 * @param shape whose children are to be stacked
	 */
	setContainer: function(container) {
		if(!container) {
			console.error("StackLayout: container cannot be null");
		}
		this._container = container;
	},
	
	/**
	 * @param direction "Horizontal" or "Vertical"
	 */
	setDirection: function(direction) {
		this._direction = direction;
	},
	
	setSpacing: function(spacing) {
		this._spacing = spacing;
	},
	
	setWidth: function(width) {
		if (width > 0) {
			this._width = width;
		}
	},
	
	setHeight: function(height) {
		if (height > 0) {
			this._height = height;
		}
	},
	
	setSorted: function(sorted) {
		this._isSorted = sorted;
	},
	
	enforce: function() {
	    var childStyle, bounds, child, i, style;
	    
		if(!this._container) {
			return;
		}
		
		// sort?
		if (this._isSorted) {
			this._container.children.sort(function(a,b) {
				if (a.shape.type == 'image' && b.shape.type == 'image') {
					return (a.shape.src < b.shape.src) ? -1 : ((a.shape.src > b.shape.src) ? 1 : 0);
				}
				return 0;
			});
		}
	
		var vertical = !(this._direction === "Horizontal");
		var spacing = this._spacing || 0;

		var cursor = 0;
		var paddingTop = 0;
		var paddingLeft = 0;
		
		// apply container's padding
		var containerStyle = this._container.customStyle;
		if(containerStyle) {
			if(containerStyle.paddingTop) {
				paddingTop = containerStyle.paddingTop;
				if(vertical) {
					cursor += paddingTop;
				}
			}
			if(containerStyle.paddingLeft) {
				paddingLeft = containerStyle.paddingLeft;
				if(!vertical) {
					cursor += paddingLeft;
				}
			}
		}
		
		var children = this._container.children;
		
		// calculate the dynamic width and height available for stretch
		var remainingWidth = this._width;
		var remainingHeight = this._height;
		if (remainingWidth || remainingHeight) {
			var nextBounds = null;
			var containerBounds = rmpw.diagrams.render.util.Workarounds.getTransformedBoundingBox(this._container);
			for(i = 0, child; child = children[i]; i++) {
				var currentBounds = nextBounds;
				if (!currentBounds) {
					currentBounds = rmpw.diagrams.render.util.Workarounds.getTransformedBoundingBox(child);
				}
					
				var nextChild = (i < children.length-1) ? children[i+1] : null;
				var nextDim;
				if (nextChild) {
					nextBounds = rmpw.diagrams.render.util.Workarounds.getTransformedBoundingBox(nextChild);
					nextDim = {width: nextBounds.x - currentBounds.x, height: nextBounds.y - currentBounds.y};
				}
				else {
					nextDim = {width: containerBounds.x + containerBounds.width - 1 - currentBounds.x, 
							   height: containerBounds.y + containerBounds.height - 1 - currentBounds.y};
				}
				
				childStyle = child.customStyle;
				if (vertical) {
					if ((!childStyle || (childStyle.yStretchRatio === undefined)) && remainingHeight) {
						remainingHeight -= nextDim.height;
					}
				}
				else {
					if ((!childStyle || (childStyle.xStretchRatio === undefined)) && remainingWidth) {
						remainingWidth -= nextDim.width;
					}
				}
			}
		}
		
		// calculate all child bounds ahead of time to drive alignment logic, since it's
		// dependent on the maximum size in the axis perpendicular to the stacking direction 
		var allBounds = [];
		var maxSize = 0;
		for(i = 0, child; child = children[i]; i++) {
			bounds = rmpw.diagrams.render.util.Workarounds.getBoundingBox(child);
			childStyle = child.customStyle;
			
			if (this._width) {
				if (childStyle && childStyle.xStretchRatio ) {
					bounds.width = childStyle.xStretchRatio * remainingWidth;
				}
				else {
					bounds.width = Math.min(this._width, bounds.width);
				}
			}
			
			if (this._height) {
				if (childStyle && childStyle.yStretchRatio) {
					bounds.height = childStyle.yStretchRatio * remainingHeight;
				}
				else {
					bounds.height = Math.min(this._height, bounds.height);
				}
			}
				
			allBounds[i] = bounds;
			if(bounds) {
				if(vertical) {
					maxSize = Math.max(maxSize, bounds.width);
				} else {
					maxSize = Math.max(maxSize, bounds.height);
				}
			}
		}
		
		// stack each child
		var dx, dy, textCorrection, transform;
		for(i = 0, child; child = children[i]; i++) {

			// acquire any custom alignment from the child
			var align = 0;
			if(child.customStyle) {
				style = child.customStyle;
				// TODO (MG): get generator to pass ordinals directly instead of strings
				if(vertical) {
					if(style.horizontalAlign) {
						if(style.horizontalAlign === "middle") {
							align = 1;
						} else if(style.horizontalAlign === "end") {
							align = 2;
						}
					}
				} else if(style.verticalAlign) {
					if(style.verticalAlign === "middle") {
						align = 1;
					} else if(style.verticalAlign === "end") {
						align = 2;
					}
				}
			}
			
			// position in stack
			bounds = allBounds[i];
			if(bounds) {
				if(vertical) {
					dx = paddingLeft;
					dy = cursor;
					if(align === 1) {
						dx += (maxSize - bounds.width) / 2;
					} else if(align === 2) {
						dx += maxSize - bounds.width;
					}
				} else {
					dx = cursor;
					dy = paddingTop;
					if(align === 1) {
						dy += (maxSize - bounds.height) / 2;
					} else if(align === 2) {
						dy += maxSize - bounds.height;
					}
				}
				transform = child.getTransform();
				if(!transform || dx !== transform.dx || dy !== transform.dy) {
					try {
						child.setTransform({dx: dx, dy: dy});
					}
					catch (e) {
						// Firefox seems to throw an exception on text in certain cases... catch to avoid error
					}
				}
				
				// do a resize if the custom style for stretching exists and if the width / height have been set
				// into the stack layout
				if (remainingWidth || remainingHeight) {
					style = child.customStyle;
					if (style && (style.xStretchRatio || style.yStretchRatio)) {
						var resizeUtil = new rmpw.diagrams.render.util.ResizeObject();
						try {
							resizeUtil.applySize(child, bounds, style.xStretchRatio !== undefined, style.yStretchRatio !== undefined);
						}
						catch (e) { 
							// some things aren't resizable (groups).  We can leave them alone since the position of the siblings will
							// be the equivalent of a resize.
						}
					}
				}
			
				// advance cursor
				if(vertical) {
					cursor += bounds.height;
				} else {
					cursor += bounds.width;
				}
				cursor += spacing;
			}
		}
	}
});