/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
dojo.provide("rmpw.common.policy.Host");

dojo.require("rmpw.common.datastruct.TreeNode");

/**
 * A host for policies that can be arranged in a tree.
 * 
 * @author Marc Gobeil
 */
dojo.declare("rmpw.common.policy.Host", rmpw.common.datastruct.TreeNode, {
	constructor: function() {
		this.policies = [];
	},
	
	/*
	 * Instantiates and add a policy to the host.  An optional key
	 * can be provided which will set the policy into the policies object
	 * accessible through that key value.
	 */
	createPolicy: function(policyType, key) {
		var policy = new policyType(this);
		var policyKey = key ? key : policy.singleton;
		var overridableKey = policy.overridable;
		var existed = false;
		if(policyKey) {
			if(dojo.isString(policyKey)) {
				existed = this._setPolicy(policyKey, policy);
			} else {
				for (var i=0, keyName; keyName = policyKey[i]; i++) {
					existed = existed || this._setPolicy(keyName, policy, overridableKey);
				}
			}
		}
		
		if (!existed) {
			this.policies.push(policy); // always push to maintain order of installation as well
		}
	
		return policy;
	},

	_setPolicy: function(policyKey, policy, overridableKey) {
		var existed = false;
		if (this.policies[policyKey]) {
			// if this policy already exists under this key but it can be overriden, just return
			if (overridableKey && dojo.indexOf(overridableKey, policyKey)!=-1) {
				return false;
			}
			existed = true;
			var idx = dojo.indexOf(this.policies, this.policies[policyKey]);
			this.policies[policyKey].destroy();
			this.policies[idx] = policy;
		}
		this.policies[policyKey] = policy;
		return existed;
	},

	destroyPolicy: function(policyKey) {
		if (this.policies[policyKey]) {
			var idx = dojo.indexOf(this.policies, this.policies[policyKey]);
			this.policies[policyKey].destroy();
			this.policies.splice(idx, 1);
			delete this.policies[policyKey];
		}
	},
	
	/**
	 * Optional method to create policies.  Not called by the Host base class,
	 * defined here for standardisation, but expected to be called at particular point
	 * in subclass's lifecycle.
	 * 
	 * @returns
	 */
	createPolicies: function() {
	},
	
	enforcePolicies: function() {
		try {
			for (var i=0, policy; policy=this.policies[i]; i++) {
				if (policy.enforce) {
					policy.enforce();
				}
			}
			this.onPoliciesEnforced();
		}
		catch (e) {
			throw e;
		}
	},
	
	onPoliciesEnforced: function() {
	},
	
	destroyPolicies: function() {
		// destroy policies
		if (this.policies) {
			var policies = this.policies;
			for (var i=0, policy; policy=policies[i]; i++) {
				if (policy.destroy) {
					policy.destroy();
				}
			}
			delete this.policies;
		}
	},
	
	destroy: function() {
		// destroy policies
		this.destroyPolicies();
		
		// call super
		this.inherited(arguments);
	}
});