/*******************************************************************************
 * Licensed Materials - Property of IBM
 *
 * com.ibm.xtools.rmps.web.core
 *
 * Copyright IBM Corporation 2010, 2013.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
(function() {
dojo.provide("rmps.core.review.Review");

dojo.require("rmps.core.resource.Resource");
dojo.require("dojox.html.entities");
dojo.require("rmps.core.review.util.ReviewUtil");
dojo.require("rmps.core.review.util.ReviewTypes");
dojo.require("rmps.core.base.BaseClientUtil");
dojo.require("dojo.date.stamp");
dojo.require("rmps.core.base.DataFeed");

var ReviewUtil = rmps.core.review.util.ReviewUtil;
var ReviewTypes = rmps.core.review.util.ReviewTypes;
var BaseClientUtil = rmps.core.base.BaseClientUtil;

var DC_NAMESPACE = "http://purl.org/dc/terms/";
var DC_QUALIFIER = "dcterms";

var RDF   = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";
var RCAM_REVIEW_NAMESPACE = "http://jazz.net/xmlns/rmps/reviews/1.0/";
var RCAM_REVIEW_QUALIFIER = "rmpsReview";
var CONTENT_TYPE_REVIEW = "application/rdf+xml";

var NAMESPACES = {
	"rmpsReview": RCAM_REVIEW_NAMESPACE,
	"rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
	"dcterms": DC_NAMESPACE,
	"dc": DC_NAMESPACE
};

/*
 * Used to wrap information for a reivew element. When constructed, an initial
 * review information is constructed, until we use parse method with the retrieved
 * document after using a GET method for a review resource URI.
 */
dojo.declare("rmps.core.review.Review", rmps.core.resource.Resource, {

	/* Main review object model that holds the reivew infomation in addition to the review results */
	model: null,
	myUserURI: null,

	_reviewResults: null,

	constructor: function(keywordArgs) {
		if (keywordArgs) {
			this.resourceUri = keywordArgs.resourceUri;
		}
		this.initForm();
	},

	parse: function(doc, ioArgs) {
		if (ioArgs) {
			var xhr = ioArgs.xhr;
			if (xhr.getResponseHeader("ETag")) {
				this.eTag = xhr.getResponseHeader("ETag");
			}
			var locationHeader = xhr.getResponseHeader("Location");
			if (locationHeader) {
				this.uri = locationHeader;
			} else {
				// Get the URI from the ioArgs.url removing any added parameters
				var index = ioArgs.url.indexOf("?");
				if (index > 0) {
					this.uri = ioArgs.url.substring(0, index);
				} else {
					this.uri = ioArgs.url;
				}
			}
		}
		this.model = BaseClientUtil.getModel(doc, { path: ["rdf:RDF", "rmpsReviews:Review"], numberOfOccurences: 1, identifier: this.uri, form: this.form}, "json");

		// Update resources to mark them as exist or not
		var resources = this.model.resources.dataSet;
		for (var index = 0; index < resources.length; index++) {
			var resource = resources[index];
			var deferred = ReviewUtil.resourceExists(resource.resourceUri);
			deferred.addCallback(this, function(exist){
				resource.exists = exist;
			});
		}
	},

	/*
	 * Return the value for an attribute in the reivew model object
	 *
	 * Example: this.getReviewAttribute("reviewName");
	 */
	getReviewAttribute: function(att) {
		return this.model[att];
	},

	/*
	 * Set a value for a given attribute in the review model
	 *
	 * Example: this. setReviewAttribute("reviewName", "MyReviewName");
	 */
	setReviewAttribute: function(att, value) {
		this.model[att] = value;
	},

	/*
	 * Get associated resources for the review
	 *
	 * Properties:
	 *      Optional: It is an array for the properties to be gotten in the resources. It is used to get the current associated resources for
	 *                  the review in addition to a summary for the approval/disapproval report for the resource
	 *                If someone would like to get a map for the resource URI only then, they could pass in an empty array
	 *                      Example: this.getReviewResources(null, []) ==> This will return a list of resourceUris associated with the review
	 * showResultsDetails
	 *      Optional: If true, get a given resource (from its resourceUri) for the review in addition to a summary for the approval/disapproval report
	 *                  for the resource. resourceUri is optional. If its value is undefined, then, we return the list of all esources
	 */
	getReviewResources: function(resourceUri, properties, showResultDetails) {
		if (resourceUri) {
			return this.getReviewResource(resourceUri, properties, showResultDetails);
		}

		var self = this, list, index;

		if (!properties /* && ! resourceUri */) {
			if (!showResultDetails) {
				/* Means: resourceUri, properties, and showResultDetails are undefined */
				return this.model.resources.dataSet;
			}

			/* Means: resourceUri, and properties are undefined */
			list = [];
			for (index = 0; index < resources.length; index++) {
				var val = resources[index];
				var currentValue = self.getReviewResourceWithProperties(val, properties);
				currentValue.resultDetails = self._reviewResults.getResultsDetailsForResource(val);
				list.push(currentValue);
			}
			return list;
		}

		/*
		 * At this place, the properties is null, so if someone decided to to
		 * resource based on properties, then we have to make sure to add resourceUri
		 * as it is the unique identifier.
		 */
		if (dojo.indexOf(properties, "resourceUri") == -1) {
			properties.push("resourceUri");
		}

		var resources, put;

		if (showResultDetails) {
			put = function(list, val) {
				var currentValue = self.getReviewResourceWithProperties(val, properties);
				currentValue.resultDetails = self._reviewResults.getResultsDetailsForResource(val);
				list.push(currentValue);
			};
			resources = this.getResourceResults(resourceUri);
		} else {
			put = function(val) {
				list.push(self.getReviewResourceWithProperties(val, properties));
			};
			resources = this.model.resources.dataSet;
		}

		list = [];
		for (index = 0; index < resources.length; index++) {
			put(resources[index]);
		}
		return list;
	},

	/*
	 * Get a single resource associated with the review
	 */
	getReviewResource: function(resourceUri, properties, showResultDetails) {
		if (showResultDetails) {
			return this.getResourceResults(resourceUri);
		}
		var resources = this.model.resources.dataSet;
		for (var index = 0; index < resources.length; index++) {
			var resource = resources[index];
			if (resource.resourceUri == resourceUri) {
				return this.getReviewResourceWithProperties(resource, properties);
			}
		}
	},

	/*
	 * Get a map for the specified properties for a give resource object.
	 *     For example:  this.getReviewResourceWithProperties(resourceUri, ["resourceUri, "modified"]);
	 */
	getReviewResourceWithProperties: function(resource, properties) {
		if (!properties) {
			return resource;
		}

		var get = properties.length == 1
			? function(value) {return value[properties[0]];}
			: function(value) {
				var object = {};
				for (var propertiesIndex = 0; propertiesIndex < properties.length; propertiesIndex++) {
					var property = properties[propertiesIndex];
					object[property] = value[property];
				}
				return object;
		};

		return get(resource);
	},

	/*
	 * Get detailed report for a given resource on the resource information and the users results
	 */
	getResourceResults: function(resourceUri) {
		var resourcesResults = {};
		var reviewResources = this.model.resources.dataSet;

		for (var index = 0; index < reviewResources.length; index++) {
			var current = resourcesResults[reviewResources[index].resourceUri] = reviewResources[index];
			current.userResults = [];
			current.resultDetails = this._reviewResults.getResultsDetailsForResource(resourceUri);

			/* Get the overall status for the resource */
			switch (current.resultDetails.percentageVotesCast) {
				case 100:
					current.resourceResultStatus = ReviewTypes.ARTIFACT_STATUS_TYPES.Reviewed;
					break;
				case 0:
					current.resourceResultStatus = ReviewTypes.ARTIFACT_STATUS_TYPES.NotStarted;
					break;
				default:
					current.resourceResultStatus = ReviewTypes.ARTIFACT_STATUS_TYPES.InProgress;
					break;
			}
		}

		var resourcesObject = {};
		var results = this.getUsersResults();

		for (index = 0; index < results.length; index++) {
			var result = results[index];
			var artifactResults = result.artifactResults.dataSet;

			/* Make sure to validate the existing results */
			artifactResults = this.validateArtifactResults(artifactResults);

			for (var subIndex = 0; subIndex < artifactResults.length; subIndex++) {
				var artifactResult = artifactResults[subIndex];

				var _current = resourcesResults[artifactResult.artifactId];

				if (_current) {
					/* We check for the NPE here to avoid such scenario:
					 *
					 * a) User submitted a result for a given resource
					 * b) The moderator pasued the review, and removed this result specifically
					 * c) Save, and continue the review. This means that the user result will contain the old value unless,
					 *    They submit new results
					 */
					resourcesResults[artifactResult.artifactId].userResults.push({
						participant:          this.getParticipant(result.userURI),
						resourceResultStatus: artifactResult.participantResultStatus,
						resourceURI:          result.resourceURI
					});
				}
			}
		}

		/*
		 * We have to loop through all resources even if we are looking for a single one,
		 * as estimating the results for a give resource requires looping though all results.
		 */
		if (resourceUri) {
			return resourcesResults[resourceUri];
		}

		var resources = [];
		for (var att in resourcesResults) {
			var resourcesResult = resourcesResults[att];
			/* Make sure to validate the existing results */
			resourcesResult.artifactResults = this.validateArtifactResults(resourcesResult.artifactResults);

			resources.push(resourcesResult);
		}
		return resources;
	},

	/*
	 * The call is required to maintain the required stability for the review editor, as it is true that handling the results
	 * being incorrect should come in other places, or the review is corrupted, or anything related, but we have to make sure
	 * that we are able even to handle the corrupted reviews, and make sure to fill in the missing resources, and discard the
	 * obsolete resources as well.
	 */
	validateArtifactResults: function(artifactResults) {
		var validatedArtifactResults = [];
		var resources = this.model.resources.dataSet;
		for (var index = 0; index < resources.length; index++) {
			var resource = resources[index];
			var artifact = resource.artifact;
			validateResource: {
				for (var subIndex = 0; subIndex < artifactResults.length; subIndex++) {
					var artifactResult = artifactResults[subIndex];
					if (artifactResult.artifactId == resource.artifact) {
						validatedArtifactResults.push(artifactResult);
						break validateResource;
					}
				}
				/* Well, this means that the resource artifact information is not there, so we need to add, and the
				 * status will be not started for sure */
				validatedArtifactResults.push({
					artifactId: artifact,
					participantResultStatus: ReviewTypes.ARTIFACT_STATUS_TYPES.NotStarted
				});
			}
		}
		return validatedArtifactResults;
	},

	/*
	 * Get the detailed results for all users.
	 *     showResultDetails:
	 *         Optional: If true, add the summary for each user results
	 */
	getUsersResults: function(showResultDetails, properties) {
		var results = [];
		var participants = this.model.participants.dataSet;
		for (var index = 0; index < participants.length; index++) {
			var participant = participants[index];
			/* Since we are looking for results, then, we have to only get the participants that have the required roles to submit results */
			if (participant.reviewRole == ReviewTypes.PARTICIPANT_ROLE_TYPES.RequiredReviewer
					|| participant.reviewRole == ReviewTypes.PARTICIPANT_ROLE_TYPES.Approver) {
				results.push(this.getUserResults(participants[index].contributor, showResultDetails, properties));
			}
		}
		return results;
	},

	getUserResults: function(userUri, showResultDetails, properties) {
		var participants = this.model.participants.dataSet;
		for (var index = 0; index < participants.length; index++) {
			var participant = participants[index];
			var contributor = participant.contributor;

			if (contributor == userUri) {
				var val = this._reviewResults.getResultsFor(contributor, false);
				if (!val) {
					val = this.fabricateResultsFor(contributor);
					this._reviewResults.getParticipantResults().push(val);
				}

				/* Required procedure to avoid getting corrupted results */
				val.artifactResults.dataSet = this.validateArtifactResults(val.artifactResults.dataSet);

				if (showResultDetails) {
					val.resultDetails = this._reviewResults.getResultsDetailsForUser(userUri);
				}
				return val;
			}
		}
	},

	/*
	 * If the review is not started, or if a user has not started making any review, we will have empty result storage for some user(s). so, we have to
	 * make sure to create ones that show the status is not started for all resouces for that user.
	 */
	fabricateResultsFor: function(userURI, reviewRole) {
		var result = {};
		result.associatedReview = this.uri;

		if (reviewRole) {
			result.reviewRole = reviewRole;
		} else {
			var participant = this.getParticipant(userURI);
			result.reviewRole = participant.reviewRole;
		}

		result.userURI = userURI;

		/* User URI is a unique identifier to get the review results. At this point,
		 * we use the user URI as a temporary place-holder for that identifier,
		 * so other places will know that this review result belong to that user. */
		result.resourceURI = userURI;

		result.isDone = false;

		var artifactResults = [];
		var resources = this.model.resources.dataSet;
		for (var index = 0; index < resources.length; index++) {
			artifactResults.push({
				artifactId: resources[index].resourceUri,
				participantResultStatus: ReviewTypes.PARTICIPANT_STATUS_TYPES.NotStarted
			});
		}

		/* The fabricated results must be identical to the model form */
		result.artifactResults = new rmps.core.base.DataFeed({
								dataSet: artifactResults,
								uniqueIdentifier: "artifactId"
							});
		return result;
	},

	/*
	 * Submit user result for a given resource.
	 *     userUri:
	 *         The user who's voting
	 *     resourceUri:
	 *         The resource that is being voted for
	 *     resultStatus:
	 *         The vote type: Rejected, Approved, etc
	 */
	voteForResource: function(userUri, resourceUri, resultStatus) {
		/* Required in order to make sure to create a new results instance fo the newly created users
		 *
		 * No need to check for null here
		 */

		var results = this.getUserResults(userUri);

		var isDone = true;
		var artifactResults = results.artifactResults.dataSet;
		for (var index = 0; index < artifactResults.length; index++) {
			var artifactResult = artifactResults[index];
			if (artifactResult.artifactId == resourceUri) {
				artifactResult.participantResultStatus = resultStatus;
				/* We do not break here as we still need to know the value for isDone flag */
			} else {
				var participantResultStatus = artifactResult.participantResultStatus;
				if (!participantResultStatus || participantResultStatus == ReviewTypes.ARTIFACT_STATUS_TYPES.NotStarted) {
					/* Well, we found a non-started element, then, isDone flag must be false */
					isDone = false;
				}
			}
		}

		var dataSet = this.model.resources.dataSet, current;
		for (index = 0; index < dataSet.length; index++) {
			current = dataSet[index];
			if (current.resourceUri == resourceUri) {
				/* Now, update the participant model object in order to notify all listeners */
				this.model.resources.updateDataEntry(resourceUri, "resourceResultStatus", resultStatus, current);

				// FIXME: 21068: TableData: Using the same fileName value causes the cells to be overlapped when making updates
				this.model.resources.updateDataEntry(resourceUri, "artifact", resourceUri, current);
			}
		}

		dataSet = this.model.participants.dataSet;
		for (index = 0; index < dataSet.length; index++) {
			current = dataSet[index];

			// FIXME: 21068: TableData: Using the same fileName value causes the cells to be overlapped when making updates
			this.model.participants.updateDataEntry(userUri, "result",        null, current);
			this.model.participants.updateDataEntry(userUri, "artifactInfo",  null, current);
			this.model.participants.updateDataEntry(userUri, "isDone",        isDone);
		}

		results.isDone = isDone;
	},

	/*
	 * Reset all associated results for all users for a given resource URI, and the result will be not started for all reviewers after that call.
	 */
	resetResourceResultsFor: function(resourceUri) {
		/* Update the acutal results */
		var list = this._reviewResults.resetResourceResultsFor(resourceUri);

		var dataSet = this.model.resources.dataSet;
		for (var index = 0; index < dataSet.length; index++) {
			var current = dataSet[index];
			if (current.resourceUri == resourceUri) {
				var userResults = current.userResults;

				/* At thie time, we know that all reviewers for this resource are not done, and their statuses are not started */
				for (var subIndex = 0; subIndex < userResults.length;subIndex++) {
					var userResult = userResults[subIndex];
					userResult.resourceResultStatus = ReviewTypes.ARTIFACT_STATUS_TYPES.NotStarted;
					userResult.participant.isDone = false;
				}
			}
		}
		return list;
	},

	/*
	 * Get participant object from the user URI.
	 */
	getParticipant: function(userUri) {
		var participants = this.model.participants.dataSet;
		for (var index = 0; index < participants.length; index++) {
			var participant = participants[index];
			if (participant.contributor == userUri) {
				return participant;
			}
		}
	},

	/*
	 * Return users list based on their roles
	 */
	getUsersBasedOnRoles: function(roles) {
		if (!roles || roles.length == 0) {
			return [];
		}

		var participants = this.model.participants.dataSet;

		var users = [];
		for (var index = 0; index < participants.length; index++) {
			var current = participants[index];
			var userRole = current.reviewRole;
			for (var roleIndex = 0; roleIndex < roles.length; roleIndex++) {
				if (userRole === roles[roleIndex]) {
					users.push(current);
					break;
				}
			}
		}
		return users;
	},

	/*
	 * Add a new resource fo the current review
	 */
	addNewResource: function(resource) {
		var resourceUri = resource.uri;

		var participantResults = this._reviewResults.getParticipantResults();

		for (var index = 0; index < participantResults.length; index++) {
			var participantResult = participantResults[index];

			/* Now, it cannot be done if it already, as a new resource has been added */
			participantResult.isDone = false;

			var artifactResults = participantResult.artifactResults;

			/* Safe check */
			if (!artifactResults) {
				artifactResults = participantResult.artifactResults = [];
			}

			/* Add the new resouce to the existing participants */
			artifactResults.dataSet.push({
				artifactId: resourceUri,
				participantResultStatus: ReviewTypes.PARTICIPANT_STATUS_TYPES.NotStarted
			});
		}

		var participants = this.model.participants.dataSet;

		for (index = 0; index < participants.length; index++) {
			/* Now, it cannot be done if it already, as a new resource has been added */
			participants[index].isDone = false;
		}

		this.model.resources.addDataEntry({
			resourceInfo: resource,
			artifact: resourceUri,
			artifactIndex: this.model.resources.dataSet.length,
			contributor: resource.creatorUri,
			format: resource.xmlNamespace + "#" + resource.typeName,
			modified: resource.creationTime,
			resourceUri: resourceUri,
			reviewArtifactInstructions: "",
			// we are just adding it so it exists
			exists: true,
			title: resource.shortName
		});
	},

	/*
	 * Provde a partial refresh for the participant references in all places.
	 *
	 * Example:
	 *     this.refreshParticipantFields(userUri, [{"field": "reviewRole", "value": myNewReviewrole}]);
	 */
	updateResource: function(resourceUri, updates) {
		var dataSet = this.model.resources.dataSet;
		for (var i = 0; i < dataSet.length; i++) {
			var data = dataSet[i];
			if (data.resourceUri == resourceUri) {
				for (var index = 0; index < updates.length; index++) {
					var current = updates[index];
					var field = current.field;
					var value = current.value;
					this.model.resources.updateDataEntry(resourceUri, field, current.value, data);
				}
			}
		}
	},

	/*
	 * Remove an existing resource in the review.
	 */
	removeResource: function(resourceUri) {
		var participantResults = this._reviewResults.getParticipantResults();

		for (var index = 0; index < participantResults.length; index++) {
			var participantResult = participantResults[index];

			/* No need to check for NPE */
			var artifactResults = participantResult.artifactResults.dataSet;
			var subIndex, artifactResult;

			for (subIndex = 0; subIndex < artifactResults.length; subIndex++) {
				artifactResult = artifactResults[subIndex];
				if (artifactResult.artifactId == resourceUri) {
					artifactResults.splice(subIndex, 1);
					break;
				}
			}

			/* Removing a resource would affect the isDone value if this resource is not reviewed by some reviewers. So, we have to make sure
			 * to update its value
			 */
			updateIsDone: {
				for (subIndex = 0; subIndex < artifactResults.length; subIndex++) {
					artifactResult = artifactResults[subIndex];
					if (artifactResult.participantResultStatus == ReviewTypes.PARTICIPANT_STATUS_TYPES.NotStarted) {
						/* Not Started status was found, so this reviewer has not done all results */
						artifactResult.isDone = false;
						break updateIsDone;
					}
				}
				/* We got here, this means that no resource URI with the status not started exists, so isDone must be true. */
				artifactResult.isDone = true;
			}
		}
		this.model.resources.removeDataEntry(resourceUri);
	},

	/*
	 * Add a new user and make sure to create initial results data for them.
	 */
	addNewParticipant: function(userURI, reviewRole) {
		this._reviewResults.getParticipantResults().push(this.fabricateResultsFor(userURI, reviewRole));
		this.model.participants.addDataEntry({
			contributor: userURI,
			isDone: false,
			reviewRole: reviewRole
		});
	},

	/*
	 * Remove an existing user, and its existing results
	 */
	removeParticipant: function(userUri) {
		this._reviewResults.removeResultsFor(userUri);
		this.model.participants.removeDataEntry(userUri);
	},

	/*
	 * Provde a partial refresh for the participant references in all places.
	 *
	 * Example:
	 *     this.refreshParticipantFields(userUri, [{"field": "reviewRole", "value": myNewReviewrole}]);
	 */
	updateParticipant: function(userUri, updates) {
		var dataSet = this.model.participants.dataSet;
		for (var i = 0; i < dataSet.length; i++) {
			var data = dataSet[i];
			if (data.contributor == userUri) {
				for (var index = 0; index < updates.length; index++) {
					var current = updates[index];
					var field = current.field;
					var value = current.value;
					var participant = this.model.participants.getItem(userUri);
					var existingValue = participant ? participant[field] : null;
					/* reviewRole is special case */
					if (field === "reviewRole") {
						if (value && value !== existingValue) {
							this.model.participants.updateDataEntry(userUri, field, value, data);

							/* Remove the current associated results in case the user is not a required reviewer anymore */
							if (value === ReviewTypes.PARTICIPANT_ROLE_TYPES.OptionalReviewer){
								this._reviewResults.removeResultsFor(userUri);
							} else {
								// Change the result statuses to match the new role
								var list = this._reviewResults.changeResourceResultsForUser(userUri,existingValue,value);
							}
						}
					} else {
						this.model.participants.updateDataEntry(userUri, field, value, data);
					}
				}
			}
		}
	},

	/*
	 * Get the overall results for the current review that shows the total approvals, disapprovals, percentage, in addition to a summary for each user's result
	 */
	getOverallReviewResult: function() {
		var report = {};
		var userResults = report.userResults = {};
		var totalObject = {};

		/* Make sure to iterate through the participants who can submit results */
		var participants = this.getUsersBasedOnRoles([ReviewTypes.PARTICIPANT_ROLE_TYPES.Approver, ReviewTypes.PARTICIPANT_ROLE_TYPES.RequiredReviewer]);

		for (var index = 0; index < participants.length; index++) {
			var contributor = participants[index].contributor;
			var current = this._reviewResults.getResultsDetailsForUser(contributor);
			userResults[contributor] = current;
			for (var att in current) {
				var currentAtt = current[att];
				if (!report[att]) {
					report[att] = 0;
				}
				report[att] += currentAtt;
			}
		}
		report.userResults = userResults;

		/* Make sure to divide the total percentage value by the reviewers number  */
		report.percentageVotesCast = report.percentageVotesCast/participants.length;

		return report;
	},

	/*
	 * Get the id from the review. It estimates it from the last number segment in the review URI.
	 */
	getReviewId: function() {
		if (!this.uri) {
			/* Return the resource URI directly, as this means that this review is just constructed, but not parsed (a new review) */
			return this.resourceUri;
		}
		var number = this.uri.match(/\/(\d+)$/);
		if (number != null && number[1]) {
			return number[1];
		}
		return null;
	},

	/*
	 * Returns true if the current user is the moderator.
	 */
	isModerator: function() {
		return this.getReviewAttribute("owner") == this.myUserURI;
	},

	getParticipantRoleType: function(documentRoleAttribute) {
		switch (documentRoleAttribute) {
			case "Approver":
				return ReviewTypes.PARTICIPANT_ROLE_TYPES.Approver;
			case "Reviewer":
				return ReviewTypes.PARTICIPANT_ROLE_TYPES.RequiredReviewer;
			case "OptionalReviewer":
				return ReviewTypes.PARTICIPANT_ROLE_TYPES.OptionalReviewer;
		}
	},

	/*
	 * Resources should be ordered based on their indices.
	 */
	resourcesComparator: function(a, b) {
		/* Safe check. both variables cannot be null here */
		if (!a) {return -1;}
		if (!b) {return 1;}

		/*
		 * No need to provide an NPE check here, as the artifact index cannot be undefined, and if it not, the comparison will
		 * fail anyway.
		 */
		var ida = a.artifactIndex;
		var idb = b.artifactIndex;
		return ida < idb ? -1 : ida === idb ? 0 : 1;
	},

	fragmentNode: function(qn, property, value, parsetype) {
		var fragmentNode = null;

		if (value == false || value == true) {
			value = value.toString();
		}
		var adoptedValue = value ? value.toString(): "";
		var xmlEncoded = adoptedValue ?  value : dojox.html.entities.encode(adoptedValue);
		if (parsetype) {
			fragmentNode = "<" + qn + ":" + property + " rdf:parseType=\"" + parsetype + "\">" + xmlEncoded + "</" + qn + ":" + property + ">\n";
		} else {
			fragmentNode = "<" + qn + ":" + property + ">" + xmlEncoded + "</" + qn + ":" + property + ">\n";
		}
		return fragmentNode;
	},

	rdfResourceNode: function(qn, property, value) {
		var returnValue = '<' + qn + ':' + property + ' rdf:resource="' + value + '"/>' + "\n";
		return returnValue;
	},

	getParticipantsXMLString: function() {
		var participantInfosStr = "";
		var results = this.model.participants.dataSet;
		for (var i = 0; i < results.length; i++) {
			var result = results[i];
			var isDone = result.isDone;
			var participantInfoStr = this.fragmentNode(RCAM_REVIEW_QUALIFIER, "isDone", result.isDone)
								   + this.fragmentNode(RCAM_REVIEW_QUALIFIER, "reviewRole", result.reviewRole)
								   + this.rdfResourceNode(DC_QUALIFIER, "contributor", result.contributor);
			participantInfoStr     = this.fragmentNode(RCAM_REVIEW_QUALIFIER, "ParticipantInfo", participantInfoStr);
			participantInfoStr     = this.fragmentNode(RCAM_REVIEW_QUALIFIER, "hasParticipant", participantInfoStr);
			participantInfosStr   += participantInfoStr;
		}
		return participantInfosStr;
	},

	getArtifactsXMLString: function() {
		var artifactInfosStr = "";
		var resources = this.getReviewResources();
		for (var i = 0; i < resources.length; i++) {
			var resource = resources[i];
			var lastModifiedDate = resource.modified;
			if (lastModifiedDate instanceof Date) {
				lastModifiedDate = dojo.date.stamp.toISOString(lastModifiedDate);
			} else {
				lastModifiedDate = "";
			}
			var artifactInfoStr = this.fragmentNode(RCAM_REVIEW_QUALIFIER, "artifactIndex", resource.artifactIndex)
								+ this.rdfResourceNode(RCAM_REVIEW_QUALIFIER, "artifact", resource.artifact)
								+ this.fragmentNode(DC_QUALIFIER, "modified", lastModifiedDate)
								+ this.rdfResourceNode(DC_QUALIFIER, "contributor", resource.contributor)
								+ this.rdfResourceNode(DC_QUALIFIER, "format", resource.format)
								+ this.rdfResourceNode(RCAM_REVIEW_QUALIFIER, "artifactId", resource.resourceUri)
								+ this.fragmentNode(RCAM_REVIEW_QUALIFIER, "reviewArtifactInstructions", ReviewUtil.encodeString(resource.reviewArtifactInstructions), "Literal")
								+ this.fragmentNode(DC_QUALIFIER, "title", resource.title);
			artifactInfoStr     = this.fragmentNode(RCAM_REVIEW_QUALIFIER, "ReviewArtifact", artifactInfoStr);
			artifactInfoStr     = this.fragmentNode(RCAM_REVIEW_QUALIFIER, "hasArtifactReview", artifactInfoStr);
			artifactInfosStr   += artifactInfoStr;
		}
		return artifactInfosStr;
	},

	toXMLString: function() {
		var titleContent = this.fragmentNode(RCAM_REVIEW_QUALIFIER, "reviewName", this.getReviewAttribute("reviewName"));
		var descriptionContent = this.fragmentNode(DC_QUALIFIER, "description", this.getReviewAttribute("description"));
		var instructions = ReviewUtil.encodeString(this.getReviewAttribute("reviewInstructions"));
		var instructionsContent = this.fragmentNode(RCAM_REVIEW_QUALIFIER, "reviewInstructions", instructions, "Literal");
		var ownerContent = this.rdfResourceNode(RCAM_REVIEW_QUALIFIER, "owner", this.getReviewAttribute("owner"));
		var reviewTypeContent = this.fragmentNode(RCAM_REVIEW_QUALIFIER, "formal", this.getReviewAttribute("isFormal"));

		var current = new Date();
		this.setReviewAttribute("reviewTimestamp", current);
		var reviewTimestamp = dojo.date.stamp.toISOString(current);

		var reviewTimestampContent = this.fragmentNode(RCAM_REVIEW_QUALIFIER, "reviewTimestamp", reviewTimestamp);

		var context = this.getReviewAttribute("reviewBaseline"), reviewContextContent;
		if (context) {
			reviewContextContent = this.rdfResourceNode(RCAM_REVIEW_QUALIFIER, "reviewContext", context);
		} else {
			reviewContextContent = "";
		}
		var reviewSummaryContent = "";
		var summary = this.getReviewAttribute("reviewSummary");
		if (summary) {
			reviewSummaryContent = this.fragmentNode(RCAM_REVIEW_QUALIFIER, "reviewSummary", ReviewUtil.encodeString(summary), "Literal");
		}

		var startDate = this.getReviewAttribute("startDate");
		if (startDate instanceof Date) {
			startDate = dojo.date.stamp.toISOString(startDate);
		}
		var reviewStartTimestampContent = this.fragmentNode(RCAM_REVIEW_QUALIFIER, "startDate", startDate);

		var closeDate = this.getReviewAttribute("closeDate");
		if (closeDate instanceof Date) {
			closeDate = dojo.date.stamp.toISOString(closeDate);
		}
		var reviewCloseTimestampContent = this.fragmentNode(RCAM_REVIEW_QUALIFIER, "closeDate", closeDate);

		var reviewStatus = this.fragmentNode(RCAM_REVIEW_QUALIFIER, "reviewStatus", this.getReviewAttribute("reviewStatus"));

		var reviewDueDate = this.getReviewAttribute("reviewDueDate");
		if (reviewDueDate instanceof Date) {
			reviewDueDate = dojo.date.stamp.toISOString(reviewDueDate);
		}
		var reviewDueDateContent = this.fragmentNode(RCAM_REVIEW_QUALIFIER, "reviewDueDate", reviewDueDate);

		var participantsContent = this.getParticipantsXMLString();
		var artifactsContent = this.getArtifactsXMLString();
		var reviewXML = "<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n"
					  + "<rdf:RDF"
					  + " xmlns:" + DC_QUALIFIER + '="' + DC_NAMESPACE + '"'
					  + " xmlns:" + RCAM_REVIEW_QUALIFIER + '="' + RCAM_REVIEW_NAMESPACE + '"'
					  + ' xmlns:rdf="' + RDF + '">\n'
					  + "  <" + RCAM_REVIEW_QUALIFIER+ ":Review>\n"
					  + titleContent
					  + descriptionContent
					  + reviewSummaryContent
					  + instructionsContent
					  + ownerContent
					  + reviewTypeContent
					  + reviewTimestampContent
					  + reviewStatus
					  + reviewDueDateContent
					  + reviewContextContent
					  + reviewStartTimestampContent
					  + reviewCloseTimestampContent
					  + participantsContent
					  + artifactsContent
					  + "  </" + RCAM_REVIEW_QUALIFIER + ":Review>\n"
					  + "</rdf:RDF>";
		return reviewXML;
	},

	initForm: function() {
		var self = this;

		var artifactsObjectForm = { path: ["rmpsReviews:ReviewArtifact"], form: [
			{name: "contributor", namespace: "http://purl.org/dc/terms", numberOfOccurences: 1, root: ["dcterms:contributor"], attribute: "rdf:resource"},
			{name: "artifact", namespace: "http://jazz.net/xmlns/rmps/reviews/1.0", numberOfOccurences: 1, root: ["rmpsReviews:artifact"], attribute: "rdf:resource"},
			{name: "modified", namespace: "http://purl.org/dc/terms", numberOfOccurences: 1, root: ["dcterms:modified"], node: "#text", typeFunction: function(val) {
				if (val instanceof Date) {
					return val;
				}
				return val ? dojo.date.stamp.fromISOString(val) : null;
			}},
			{name: "title", namespace: "http://purl.org/dc/terms", numberOfOccurences: 1, root: ["dcterms:title"], node: "#text", typeFunction: function(val) {
				return val ? val : "";
			}},
			{name: "format", namespace: "http://purl.org/dc/terms", numberOfOccurences: 1, root: ["dcterms:format"], attribute: "rdf:resource"},
			{name: "reviewArtifactInstructions", namespace: "http://jazz.net/xmlns/rmps/reviews/1.0", numberOfOccurences: 1, root: ["rmpsReviews:reviewArtifactInstructions"], node: "#text", typeFunction: function(val) {
				return val ? ReviewUtil.decodeString(val) : "";
			}},
			{name: "artifactIndex", namespace: "http://jazz.net/xmlns/rmps/reviews/1.0", numberOfOccurences: 1, root: ["rmpsReviews:artifactIndex"], node: "#text", typeFunction: function(val) {
				return val ? parseInt(val) : -1;
			}},
			{name: "resourceUri", namespace: "http://jazz.net/xmlns/rmps/reviews/1.0", prefix: "artifactId", numberOfOccurences: 1, root: ["rmpsReviews:artifactId"], attribute: "rdf:resource"}
		]};

		var participantobjectForm = { path: ["rmpsReviews:ParticipantInfo"], form: [
			{name: "contributor", namespace:"http://purl.org/dc/terms", root: ["dcterms:contributor"], numberOfOccurences: 1, attribute: "rdf:resource"},
			{name: "reviewRole", namespace: "http://jazz.net/xmlns/rmps/reviews/1.0", root: ["rmpsReviews:reviewRole"], numberOfOccurences:1, node: "#text", typeFunction: function(val) {
				return self.getParticipantRoleType(val);
			}},
			{name: "isDone", namespace: "http://jazz.net/xmlns/rmps/reviews/1.0", root: ["rmpsReviews:isDone"], node: "#text", numberOfOccurences: 1, typeFunction: function(val) {
				return val == "true";
			}}
		]};

		this.form = [
			{name: "reviewName", namespace: "http://jazz.net/xmlns/rmps/reviews/1.0", root: ["rmpsReviews:reviewName"], numberOfOccurences: 1, node: "#text"},
			{name: "description", namespace: "http://jazz.net/xmlns/rmps/reviews/1.0", root: ["rmpsReviews:reviewDescription"], numberOfOccurences: 1, node: "#text"},
			{name: "reviewBaseline", prefix: "reviewContext", namespace: "http://jazz.net/xmlns/rmps/reviews/1.0", root: ["rmpsReviews:reviewContext"], numberOfOccurences: 1, attribute: "rdf:resource"},
			{name: "owner", namespace: "http://jazz.net/xmlns/rmps/reviews/1.0", root: ["rmpsReviews:owner"], numberOfOccurences: 1, attribute: "rdf:resource", typeFunction: function(val) {
				if (!val && self.resourceUri == "rmps.app.newReviewEditor") {
					val = ReviewUtil.getCurrentUserUri();
				}
				return ReviewUtil.getCorrectUserUri(val);
			}},
			{name: "reviewInstructions", namespace: "http://jazz.net/xmlns/rmps/reviews/1.0", root: ["rmpsReviews:reviewInstructions"], numberOfOccurences: 1, node: "#text", typeFunction: function(val) {
				return val ? ReviewUtil.decodeString(val) : "";
			}},
			// no need here to do the <br/> replacement because we only display the reviewSummary we never update it in the ui after
			// it has been created
			{name: "reviewSummary", namespace: "http://jazz.net/xmlns/rmps/reviews/1.0", root: ["rmpsReviews:reviewSummary"], numberOfOccurences: 1, node: "#text", typeFunction: function(val) {
				return val ? ReviewUtil.decodeString(val) : "";
			}},
			// need to add a prefix because our attribute has a different
			// name than the corresponding rdf property
			{name: "isFormal", prefix: "formal", namespace: "http://jazz.net/xmlns/rmps/reviews/1.0", root: ["rmpsReviews:formal"], numberOfOccurences: 1, node: "#text", typeFunction: function(val) {
				return val === "true";
			}},
			{name: "reviewDueDate", namespace: "http://jazz.net/xmlns/rmps/reviews/1.0", root: ["rmpsReviews:reviewDueDate"], numberOfOccurences: 1, node: "#text", typeFunction: function(val) {
				if (val instanceof Date) {
					return val;
				}
				return val ? dojo.date.stamp.fromISOString(val) : null;
			}},
			{name: "reviewStatus", namespace: "http://jazz.net/xmlns/rmps/reviews/1.0", root: ["rmpsReviews:reviewStatus"], numberOfOccurences: 1, node: "#text", typeFunction: function(val) {
				return val ? val : ReviewTypes.REVIEW_STATE_TYPES.Draft;
			}},
			{name: "reviewTimestamp", namespace: "http://jazz.net/xmlns/rmps/reviews/1.0", root: ["rmpsReviews:reviewTimestamp"], numberOfOccurences: 1, node: "#text", typeFunction: function(val) {
				return val ? dojo.date.stamp.fromISOString(val) : val;
			}},
			{name: "startDate", namespace: "http://jazz.net/xmlns/rmps/reviews/1.0", root: ["rmpsReviews:startDate"], numberOfOccurences: 1, node: "#text", typeFunction: function(val) {
				if (val instanceof Date) {
					return val;
				}
				return val ? dojo.date.stamp.fromISOString(val) : null;
			}},
			{name: "closeDate", namespace: "http://jazz.net/xmlns/rmps/reviews/1.0", root: ["rmpsReviews:closeDate"], numberOfOccurences: 1, node: "#text", typeFunction: function(val) {
				if (val instanceof Date) {
					return val;
				}
				return val ? dojo.date.stamp.fromISOString(val) : null;
			}},
			{name: "resources", prefix: "hasArtifactReview",  namespace: "http://jazz.net/xmlns/rmps/reviews/1.0", comparatorFunction: this.resourcesComparator, dataFeedKey: "resourceUri", root: ["rmpsReviews:hasArtifactReview"], objectForm: artifactsObjectForm},
			{name: "participants", prefix: "hasParticipant", namespace: "http://jazz.net/xmlns/rmps/reviews/1.0", dataFeedKey: "contributor", root: ["rmpsReviews:hasParticipant"], objectForm: participantobjectForm}
		];
		this.model = BaseClientUtil.getInitialModel(this.form);
		// add default context
		var currentModelContext = this.model["reviewBaseline"];
		if (!currentModelContext || dojo.trim(reviewContext) === "") {
			this.model["reviewBaseline"] = jazz.app.context.get().baseline;
		}
	}
});

})();