/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010, 2012.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.diagrams.edit.tools.AcceleratorScope");

dojo.require("rmpw.common.policy.Policy");
dojo.require("rmpw.common.command.CompoundCommand");
dojo.require("rmpw.diagrams.edit.actions.CommandWrapper");
dojo.require("rmpw.diagrams.edit.KeyStrokeDescriptor");

/**
 * Central authority for invoking keyboard actions
 * 
 * @author Steve Shaw
 */
dojo.declare("rmpw.diagrams.edit.tools.AcceleratorScope", rmpw.common.policy.Policy, {
	singleton: "acceleratorScope",
	
	constructor: function() {
	},
	
	destroy: function() {
		this.inherited(arguments);
		this.disable();
	},
	
	disable: function() {
		if (this._connectKeyPress) {
			dojo.disconnect(this._connectKeyPress);
			delete this._connectKeyPress;
		}
	},
	
	/*
	 * Adds actions for handling accelerators.  It is assumed to be an array of bindings that are subclasses of
	 * rmpw.diagrams.edit.actionBindings.ActionBinding.  The actions are comma separated arguments that is
	 * parseable as an array.
	 */
	setActionBindings: function () {
		var keyboardActionBindings = arguments;
		for (var i=0, binding; binding=keyboardActionBindings[i]; i++) {
			this.registerBinding(binding);
		}
	},
	
	enable: function() {
		if (!this._connectKeyPress) {
			this._connectKeyPress = dojo.connect(document, "onkeypress", this, "onKeyPress");
		}
	},
	
	/*
	 * Register a binding so that any accelerators will be recognized by the diagram editor
	 */
	registerBinding: function(binding) {
		if (!this.keyStrokeMap) {
			this.keyStrokeMap = {};
		}
			
		var keyStrokeString = binding.getKeyStrokeDescriptor().toString();
		var actions = this.keyStrokeMap[keyStrokeString];
		if (actions) {
			actions.push(binding.getAction());
		}
		else {
			this.keyStrokeMap[keyStrokeString] = [binding.getAction()];
		}
	},
	
	/*
	 * Key handler that will execute the command associated with the key stroke.
	 */
	onKeyPress: function(e) {
		if (!this.keyStrokeMap) {
			return;
		}
		
		var args = {};
		args.viewer = this.host.viewer;
		
		var keyStroke = new rmpw.diagrams.edit.KeyStrokeDescriptor(e);
		var actions = this.keyStrokeMap[keyStroke.toString()];
		var action = null, 
		    i = 0, 
		    j = 0,
		    policy,
		    editController;
		if (actions) {
			for (i=0, action; action=actions[i]; i++) {
				if (action.isEnabled(args)) {
					if (action.run(args)) {
						dojo.stopEvent(e);
					}
					return;
				}
			}
		}
		
		// next check the current selection
		var selection = this.host.root.getSelectionController().getSelection();
		if (selection.length > 0) {
			var compoundCmd = null;
			args = {};
			dojo.mixin(args, e);
			
			for (i = 0, editController; editController = selection[i]; i++ ) {
				for (j=0, policy; policy=editController.policies[j]; j++) {
					if (policy.acceptKeyStroke && policy.acceptKeyStroke(keyStroke)) {
						action = policy.getAction() ? policy.getAction() : 
							this._getTunnelingAction(editController, keyStroke, policy.getTunnelingAction()); 
						if (action && action.isEnabled(args)) {
							if (policy.isRunAction && policy.isRunAction()) {
								if (action.run(args)) {
									dojo.stopEvent(e);
								}
								return;
							} else {
								args.targetController = editController;
								
								var cmd = action.getCommand(args);
								if (!cmd) {
									cmd = new rmpw.diagrams.edit.actions.CommandWrapper({}, action);
								}
								
								if (!compoundCmd) {
									compoundCmd = new rmpw.common.command.CompoundCommand();
								}
								compoundCmd.add(cmd, action.getExecutionArguments(args));
							}
						}
					}
				}
			}
			
			if (compoundCmd) {
				var commandStack = this.host.root.getCommandController().commandStack;
				commandStack.execute(compoundCmd, args);
				dojo.stopEvent(e);
			}
		}
	},
	
	// get action from child
	_getTunnelingAction: function(editController, keyStroke, actionId) {
		if (actionId) {
			for (var i = 0, child; child = editController.children[i]; i++ ) {
				for (var j=0, policy; policy=child.policies[j]; j++) {
					if (policy.acceptKeyStroke && policy.acceptKeyStroke(keyStroke)) {
						return policy.getAction();
					}
				}
			}
		}
	}

});