/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2011, 2013.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
(function() {
dojo.provide("rmpw.diagrams.edit.tools.ActionBarScope");

dojo.require("rmpw.common.policy.Policy");
dojo.require("rmpw.diagrams.edit.actions.Redo");
dojo.require("rmpw.diagrams.edit.actions.Undo");
dojo.require("dojo.i18n");
dojo.require("rmpw.widgets.Events");
dojo.require("rmpw.diagrams.edit.commands.ChangeColorCommand");
dojo.require("rmpw.diagrams.edit.commands.ChangeLineStyleCommand");
dojo.require("rmpw.diagrams.edit.commands.ChangeFontStyleCommand");
dojo.require("rmpw.widgets.pickers.ColorPicker");
dojo.require("rmpw.widgets.pickers.LineWidthPicker");
dojo.require("rmpw.widgets.pickers.LineStylePicker");
dojo.require("rmpw.widgets.pickers.TextPicker");
dojo.require("rmpw.diagrams.edit.tools.dialogs.LinePropertiesDialog");

dojo.requireLocalization("rmpw.diagrams.edit", "BehaviorMessages");
var behaviorMessages = dojo.i18n.getLocalization("rmpw.diagrams.edit", "BehaviorMessages");        

/**
 * Manages diagram toolbar. It adds static tools to it (that don't change with selection), 
 * and allows shapes to make their contributions (through actionbar policy, when the shape is selected).
 * 
 * @author Lidija Grahek
 */
dojo.declare("rmpw.diagrams.edit.tools.ActionBarScope", rmpw.common.policy.Policy, {
	singleton: "actionBarScope",
	
	constructor: function() {
		this._connections = [];
		this.selectionScope = this.findNearestParentByKey('selectionScope', true);	
		// see comment in handleSelection method
		this._connections.push(dojo.connect( this.selectionScope, "onSelectionChanged", this, "_onSelectionChanged" ));		
	},
	
	setToolbarWidget: function(toolbar) {
		this.toolbar = toolbar;
	},
	
	/**
	 * Adds the given item to the toolbar. 
	 */
	addTool: function(item){		
		this.toolbar.addChild(item);
	},
	
	/**
	 * Removes the given item from the toolbar (without destroying the item).
	 */
	removeTool: function(item){
		this.toolbar.removeChild(item);
	},
	
	/**
	 * Destroys the given item.
	 */
	destroyTool: function(item){
		item.destroyRecursive();
	},
	
	/**
	 * Adds given widgets to the toolbar.
	 */
	addTools: function(widgets) {
		for (var j = 0, widget; widget = widgets[j]; j++) {			
			this.addTool(widget);
		}
	},
	
	/**
	 * Removes the widgets from the toolbar and destroys them.
	 */
	destroyTools: function(widgets) {
	    var tool, i;
		if (widgets && widgets.length > 0) {
			for (i = 0; tool = widgets[i]; i++) {
				this.destroyTool(tool);
			}
		}
	},		
	
	/**
	 * Creates a button based on given arguments and adds it to the toolbar. 
	 */
	addButton: function( label, call, icon, show ){ 
		if( !show ){ 
		    show = false; 
		}

		var button = new dijit.form.Button({ 
				label: label, 
				showLabel: show, 
				onClick: call,
				iconClass: icon,
				style: "font-size:10px"});

		this.addTool( button );

		return button;
	},	
	
	/**
	 * Adds the static toolbar widgets (that don't change with selection). 
	 * Dynamic ones (depend on the currently selected shape) are contributed by actionbar 
	 * policies installed on shapes when needed.
	 */
	addStaticToolbarWidgets: function(args) {
		this.diagramLayer = args.diagramLayer;
		var controls = rmpw.jazz.ui.Controls.items;
		var tools = args.toolInfo;
        for(var item = 0; item < tools.length; item++) {
            if(tools[item] && tools[item].display === true) {
                switch( tools[item].name ) {
                    case controls.ZOOM:
                        this._addZoom(args.rootLayer);
                        break;

                    case controls.UNDO:
                        this.addButton(behaviorMessages.undo, dojo.hitch(this, "_undo"), "undoIcon");
                        break;

                    case controls.REDO:
                        this.addButton(behaviorMessages.redo, dojo.hitch(this, "_redo"), "redoIcon");
                        break;
                }
            }
        }
        
        // special handling for freeform diagram - to allow domain selection
        if (this.diagramLayer.rootLayer.viewer.editor.definition.id == "FreeFormDiagram") {
        	var isSketch = false;
        	var types = this.diagramLayer.resourceInfo.model.rdf_type;
        	for (var i=0, lg=types.length; i<lg; i++) {
        		if (types[i].rdf_about == "http://jazz.net/ns/dm/sketch#Diagram") {
        			isSketch = true;
        			break;
        		}
        	}
        	
        	if (isSketch) {
        		this._domainId = "http://jazz.net/ns/dm/domain/definition/docSystemDomain#";
            	dojo.publish("rmpw.diagrams.edit.ActionBarScope/domainId");
        	}
        	else {
        		var domainIdStore = new dojo.data.ItemFileReadStore({
    				url: net.jazz.ajax._contextRoot + '/domainRegistry/associatedDomains' + "?rmps.context=" + jazz.app.context.get().baseline + "&projectId=" + jazz.app.context.get().itemId
    			});
            	
            	if (!this.domainCombo) {
        			this.domainCombo = new dijit.form.ComboBox({
        				store: domainIdStore,
        				autoComplete: true,
        				searchAttr: "label",
        				onChange: dojo.hitch( this, "publishCurrentDomain" )
        			});
        			this.domainCombo.domNode.style.height = "20px";   
        		    this.domainCombo.domNode.style.fontSize = "10pt";
        			this.addTool(this.domainCombo);
        		}
        		
            	this._domainId = "http://all";
    	        domainIdStore.fetch({ 
    				query: {id: this._domainId }, 
    		        onComplete: dojo.hitch(this, function(items, request){  
    		        	if (items.length > 0) {
    		        		if (this.domainCombo) {
    			        		this.domainCombo.set("displayedValue", items[0].label[0]);
    		        		}
    		        	}
    		        }) 
    			}); 
        	}
        }
	},
	
	publishCurrentDomain: function() {
		this.domainCombo.store.fetch({ 
			query: {label: this.domainCombo.value }, 
	        onComplete: dojo.hitch(this, function(items, request){  
	        	if (items.length > 0) {
		        	this._domainId = items[0].value[0]; 
	        	}
	        }) 
		}); 
		   
		dojo.publish("rmpw.diagrams.edit.ActionBarScope/domainId");
	},
	
	getDomainId: function() {
		return this._domainId;  // may be undefined if domain combo doesn't exist 
	},
	
	setDomainLabel: function(domainId, actionBinding) {
		if (this.domainCombo) {
			this.domainCombo.store.fetch({ 
				query: {id: domainId }, 
		        onComplete: dojo.hitch(this, function(items, request){  
		        	if (items.length > 0) {
		        		actionBinding.setLabel(items[0].label[0]);
		        	}
		        }) 
			});
		} 
	},
	
	_addZoom : function(rootLayer) {
		// code in this methd is taken from rmps.diagram.ui.DiagramWidget 
		// (and was previously in rmpw.jazz.ui.Layout)
		this.slider = new rmpw.widgets.slider.HorizontalSlider({
				name: "slider",
				value: 1,
				minimum: 0.10,
				maximum: 4,
				intermediateChanges: true,
				style: "width:200px;vertical-align:middle;top:5px;",
				discreteValues: 40, // 20% - 400% at 20% intervals
				showButtons: false,
				handleSrc: dojo.moduleUrl( "rmps.graphics","images/etool16/SliderHandle.png" )
		});
 
		if ( !rmps.app.core.util.checkIfMobile() ){
			this.addTool( this.slider );
		}

		this.label = new rmps.graphics.ui.LabelWidget();
        this.label.setLabel("100%");
        this.addTool(this.label); 

		this.percent = this.addButton( "100%", dojo.hitch( this, "_resetZoom" ), "actualIcon", false );
		
		if (this.diagramLayer.outline) {
			this.zoomToFit = this.addButton( behaviorMessages.zoomToFit, dojo.hitch( this, "_zoomToFit" ), "zoomToFitIcon", false );
		}
					
		var percentLabel = this.label; 
		
		if( this.slider ){
			// catch when the slider changes the zoom
			this.slider.onChange = dojo.hitch(this, function(value) {
				rootLayer.setScale(value, null, null, true );			
				var val = Math.round( 100 * value );				
				var label = val + "%";				
				percentLabel.setLabel( label );
			});
			
			// catch when something else like the WheelZoom edit policy changes the zoom
			var viewport = this.host.root.viewer.viewport;
			this._connections.push(dojo.connect(viewport, "onViewportChanged", this, function() {
				var value = viewport.getZoom();
				var val = Math.round( 100 * value );				
				percentLabel.setLabel(val + "%");
            }));	
			this._connections.push(dojo.connect(viewport, "onViewportFinished", this, function() {
				if (this._centerView) {
					this._centerView = false;
					var self = this;
					setTimeout(function() {
						self.diagramLayer.outline.centerView();
					}, 0);
				}
            }));	
			this._connections.push(dojo.connect(rootLayer, "setScale", dojo.hitch(this, function(scale) {
				this.slider.set("value", scale);
			})));
		}		
	},
	
	_resetZoom: function(){
		this.slider.set( 'value', 1 );
		this.diagramLayer.setScale(1);
	},
	
	_zoomToFit: function(){
		var z = this.diagramLayer.outline.getScaleToFit();
		if (z!=this.slider.get( 'value')) {
			this.slider.set( 'value', z );
			this._centerView = true;
		} else {
			this.diagramLayer.outline.centerView();
		}
	},

	_redo: function() {
		var action = new rmpw.diagrams.edit.actions.Redo();
		action.run({viewer : this.host.root});
	},
		
	_undo: function() {
		var action = new rmpw.diagrams.edit.actions.Undo();
		action.run({viewer : this.host.root});
	},
	
	_lineProperties: function() {
		// put up dialog with appearance properties for routing, smoothness and avoid obstructions
		new rmpw.diagrams.edit.tools.dialogs.LinePropertiesDialog({selection: this.selectionScope.getSelection(), viewer: this.host.root});
	},

	_removeStyleTools: function() {
		this.removeTool( this.lineColor );
		this.removeTool( this.fillColor );
		this.removeTool( this.lineWidth );
		this.removeTool( this.lineStyle );
		this.removeTool( this.lineProperties );
		this.removeTool( this.fontColor );
		this.removeTool( this.fontStyle );
	},
	
	_onSelectionChanged: function() {
		var selection = this.selectionScope.getSelection();
		if (selection && selection.length > 0 && this._selectionIncludesStyleable(selection)) {
			if (!this.fillColor) {	  				
		        this.fillColor = new rmpw.widgets.pickers.ColorPicker({type:"FILL"});	
		        this.addTool( this.fillColor );
		        this.fillColor.startup();
		        this._connections.push(dojo.connect(this.fillColor, "colorChange", dojo.hitch(this, function() {
		        	this.setFillColor();
		        })));	
			} else {
				this.addTool(this.fillColor);					
			}

			if (!this.lineColor) { 
				this.lineColor = new rmpw.widgets.pickers.ColorPicker({});
				this.addTool( this.lineColor ); 
		        this.lineColor.startup();
		        this._connections.push(dojo.connect(this.lineColor, "colorChange", dojo.hitch(this, function() {
		        	this.setLineColor();
		        })));
			} else {
				this.addTool(this.lineColor);
			}

			if (!this.lineWidth) { 
		        this.lineWidth = new rmpw.widgets.pickers.LineWidthPicker({});
		        this.addTool( this.lineWidth );
		        this.lineWidth.startup();
		        this._connections.push(dojo.connect(this.lineWidth, "change", dojo.hitch(this, function() {
		        	this.setLineWidth();
		        })));
			} else {
				this.addTool(this.lineWidth);
			}
		        
			if (!this.lineStyle) { 
		        this.lineStyle = new rmpw.widgets.pickers.LineStylePicker({});
		        this.addTool( this.lineStyle );
		        this.lineStyle.startup();
		        this._connections.push(dojo.connect(this.lineStyle, "change", dojo.hitch(this, function() {
		        	this.setLineStyle();
		        })));
			} else {
				this.addTool(this.lineStyle);
			}
			
			if (!this.lineProperties) {
				this.lineProperties = this.addButton(behaviorMessages.lineProperties, dojo.hitch(this, "_lineProperties"), "linePropertiesIcon");
			}
			else {
				this.addTool(this.lineProperties);
			}

			if (!this.fontColor) { 
		        this.fontColor = new rmpw.widgets.pickers.ColorPicker({type:"TEXT"});	
		        this.addTool( this.fontColor );
		        this.fontColor.startup();
		        this._connections.push(dojo.connect(this.fontColor, "colorChange", dojo.hitch(this, function() {
		        	this.setFontColor();
		        })));
			} else {
				this.addTool(this.fontColor);
			}
			
		    if (!this.fontStyle) {    
		        this.fontStyle = new rmpw.widgets.pickers.TextPicker();	
		        this.addTool( this.fontStyle );
		        this.fontStyle.startup();
		        this._connections.push(dojo.connect(this.fontStyle, "setStyles", dojo.hitch(this, function() {
	        		this.setFontStyle();
		        })));
		        this._connections.push(dojo.connect(this.fontStyle, "openDropDown", dojo.hitch(this, function() {
		        	this._updateFontStyleDropDown();
		        })));       		        
			} else {
				this.addTool(this.fontStyle);					
			}
			
			this._updateEnablement(selection);
			this._updateSwatches(selection);
		} else {
			this._removeStyleTools();
		}
	},
	
	_selectionIncludesStyleable: function(selection) {
		for(var i = 0; editController = selection[i]; i++) {
			if (editController.renderMapping.policies.styleable) {
				return editController.renderMapping.policies.styleable;
			}
		}
		return null;
	},
	
	_updateEnablement: function(selection) {
		var enableFillColor = false;
		var enableLineColor = false;
		var enableLineTools = false;
		var enableFontTools = false;
		
		for(var i = 0; editController = selection[i]; i++) {
			var styleable = editController.renderMapping.policies.styleable;
			if (styleable) {
				if (!enableFillColor && styleable.enableFillColorTool()) {
					enableFillColor = true;
				}
				if (!enableLineColor && styleable.enableLineColorTool()) {
					enableLineColor = true;
				}
				if (!enableLineTools && styleable.enableLineStyleTools()) {
					enableLineTools = true;
				}
				if (!enableFontTools && styleable.enableFontTools()) {
					enableFontTools = true;
				}
			}
		}
		if (this.fillColor) {
			this.fillColor.set('disabled', !enableFillColor);
		}
		if (this.lineColor) {
			this.lineColor.set('disabled', !enableLineColor);
		}
		if (this.lineWidth) {
			this.lineWidth.set('disabled', !enableLineTools);
		}
		if (this.lineStyle) {
			this.lineStyle.set('disabled', !enableLineTools);
		}
		if (this.lineProperties) {
			this.lineProperties.set('disabled', !enableLineTools);
		}
		if (this.fontColor) {
			this.fontColor.set('disabled', !enableFontTools);
		}
		if (this.fontStyle) {
			this.fontStyle.set('disabled', !enableFontTools);
		}

	},

	_updateSwatches: function(selection) {
		for(var i = 0; editController = selection[i]; i++) {
			var styleable = editController.renderMapping.policies.styleable;
			if (styleable) {
				var selFillColor = styleable.getFillColor();
				if (this.fillColor && selFillColor) {
					this.fillColor.updateSwatch(selFillColor.toHex());
				}
				var selLineColor = styleable.getLineColor();
				if (this.lineColor && selLineColor) {
					this.lineColor.updateSwatch(selLineColor.toHex());
				}
				var selFontColor = styleable.getFontColor();
				if (this.fontColor && selFontColor) {
					this.fontColor.updateSwatch(selFontColor.toHex());
				}
			}
		}
	},

	_updateFontStyleDropDown: function() {
		var selection = this.selectionScope.getSelection();
		if (selection && selection.length > 0) {
			var styleable = this._selectionIncludesStyleable(selection);
			if (styleable) {
				this.fontStyle.initializeValues(styleable.getFont());
			}
		}
	},
	
	setLineColor: function() {
		var args = { viewer: this.host.root };
		args.editControllers = this.selectionScope.getSelection();
		args.lineColor = this.lineColor.color;
		args.opacity = this.lineColor.decimalValue;
		args.viewer.getCommandController().execute("rmpw.diagrams.edit.commands.ChangeColorCommand", args);
	},
	
	setFillColor: function() {
		var args = { viewer: this.host.root };
		args.editControllers = this.selectionScope.getSelection();
		args.fillColor = this.fillColor.color;
		args.opacity = this.fillColor.decimalValue;
		args.viewer.getCommandController().execute("rmpw.diagrams.edit.commands.ChangeColorCommand", args);
	},
	
	setLineWidth: function(){	
		var args = { viewer: this.host.root };
		args.editControllers = this.selectionScope.getSelection();
		args.width = this.lineWidth.width;
		args.viewer.getCommandController().execute("rmpw.diagrams.edit.commands.ChangeLineStyleCommand", args);
	},
	
	setLineStyle: function(){
		var args = { viewer: this.host.root };
		args.editControllers = this.selectionScope.getSelection();
		args.style = "Solid";
		switch( this.lineStyle.dash ){
			case "1":
				args.style = "Dash";
				break;	
			case "2": 
				args.style = "Dot";
				break;
		}	
		args.viewer.getCommandController().execute("rmpw.diagrams.edit.commands.ChangeLineStyleCommand", args);
	},
	
	setFontColor: function() {
		var args = { viewer: this.host.root };
		args.editControllers = this.selectionScope.getSelection();
		args.fontColor = this.fontColor.color;
		args.opacity = this.fontColor.decimalValue;
		args.viewer.getCommandController().execute("rmpw.diagrams.edit.commands.ChangeColorCommand", args);
	},
	
	setFontStyle: function(){		
		var args = { viewer: this.host.root };
		args.editControllers = this.selectionScope.getSelection();
		args.font = this.fontStyle.getFontObject();	
		args.viewer.getCommandController().execute("rmpw.diagrams.edit.commands.ChangeFontStyleCommand", args);
	},	
	
	/**
	 * Removes all tools from the toolbar other than those related to zoom. 
	 * Removed tools can be later re-added by calling reAddRemovedTools. 
	 */
	removeToolsExceptZoom: function() {
		this._removedTools = [];
		var i = 0;
		var toolbarItems = this.toolbar.getChildren();
		while (i < toolbarItems.length) {
			if (toolbarItems[i] !== this.slider
					&& toolbarItems[i] !== this.label 
					&& toolbarItems[i] !== this.percent
					&& toolbarItems[i] !== this.domainCombo) {
				// skip removal of zoom related tools
				this._removedTools.push(toolbarItems[i]);
				this.removeTool(toolbarItems[i]);
			}
			i++;
        }
	},
	
	/**
	 * Assumes that removeToolsExceptZoom was called before. 
	 * Appends all tools removed by that call back to the toolbar.
	 */
	reAddRemovedTools: function() {
		if (!this._removedTools) {
			return;
		}
		for (var i = 0; i < this._removedTools.length; i++) {
			this.addTool(this._removedTools[i]);
		}
		this._removedTools = [];
	},	
	
	destroy: function() {
		dojo.forEach(this._connections, function(conn){
			dojo.disconnect(conn);
		});
		// remove all tools from the toolbar
		this._removeStyleTools();
		this.lineColor = null;
		this.fillColor = null;
		this.lineWidth = null;
		this.lineStyle = null;
		this.lineProperties = null;
		this.fontColor = null;
		this.fontStyle = null;
		
		this.label = null;
		this.slider = null;
		this.percent = null;	
		this.domainCombo = null;
		while (this.toolbar.getChildren().length > 0) {
			this.destroyTool(this.toolbar.getChildren()[0]);
		}
	}
});

}());