/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
dojo.provide("rmpw.diagrams.edit.tools.Connectable");

dojo.require("rmpw.diagrams.edit.tools.Feedback");
dojo.require("rmpw.diagrams.edit.handles.ConnectionCreationHandle");
dojo.require("rmpw.diagrams.edit.handles.MoveableEndHandle");
dojo.require("rmpw.diagrams.edit.tools.figures.PolylineFeedback");
dojo.require("rmpw.diagrams.edit.tools.ContextActionBar");

/**
 * Policy that allows a particular element to be connectable be edges.  Required in order
 * to allow creation of edges or reorient of edges to the policy owner.
 * 
 * @author Steve Shaw
 */
dojo.declare("rmpw.diagrams.edit.tools.Connectable", [rmpw.diagrams.edit.tools.Feedback, rmpw.diagrams.edit.tools.ContextActionBar], {
	singleton: "connectable",
	
	foregroundColor: "#4C7399",
	showFeedbackDelay : null,
	eraseFeedbackDelay : null,
	orientation : 1,
	
	constructor: function() {
		this.feedbackFigureType = rmpw.diagrams.edit.handles.ConnectionCreationHandle;
		this.feedbackLayer = this.getFrontFeedbackLayer();
		this.setTargetShape( this.host.getDefaultTargetShape() );
	},
	
	initToolbar: function() {
		// do nothing for connectable since it's a special toolbar policy
		// invoked during connection
	},
	
	destroy: function() {
		this.inherited(arguments);	
		// erase feedback
		this.hideFeedback();
		
		// detach listeners
		this.detachListeners();
		this.removeFeedbackDelays();
	},
	
	/*
	 * remove any existing delays in progress
	 */
	removeFeedbackDelays: function() {
		if (this.eraseFeedbackDelay) {
			clearTimeout(this.eraseFeedbackDelay);
			this.eraseFeedbackDelay = null;
		}
		
		if (this.showFeedbackDelay) {
			clearTimeout(this.showFeedbackDelay);
			this.showFeedbackDelay = null;
		}
	},

	/*
	 * creates four connection handles and attaches listeners to them
	 */	
	showFeedback: function() {
		if (this.connectableAsSource() && !this.feedbackFigures) {
			this.feedbackFigures = [];
			this.handleListeners = [];
			var feedbackFig;			
			for (var i = 0; i < 4; i++) {
				this.orientation = i + 1;
				this.feedbackFigures[i] = feedbackFig = this._createFeedback();
							
				this.handleListeners.push(dojo.connect(feedbackFig, "onFirstMove", this, "onFirstMove"));
				this.handleListeners.push(dojo.connect(feedbackFig, "onMove", this, "onMove"));
				this.handleListeners.push(dojo.connect(feedbackFig, "onMoveStop", this, "onMoveStop"));
				this.handleListeners.push(dojo.connect(feedbackFig, "onCancel", this, "onCancel"));
				this.handleListeners.push(dojo.connect(feedbackFig.policies.focusable, "setFocus", this, "_focusChanged"));				
			}
			this._viewportListener = dojo.connect(this.host.root.viewport, "onViewportChanged", this, "refreshFeedback");		
		}
	},
	
	_createFeedback: function() {
		return  new rmpw.diagrams.edit.handles.ConnectionCreationHandle(
				{owner: this.feedbackLayer, target: this.figure, 
					 editController: this.host,
					 orientation: this.orientation,
					 isAnchorToMappingValid: dojo.hitch(this, function(targetMapping) {
						 return this.getEnabledBindings(targetMapping).length > 0;
					 })});
	},
	
	hideFeedback: function() {
		this._isHiding = true;
		if (this.feedbackFigures) {
			// delete all 4 feedback figures
			var ff = this.feedbackFigures;
			for (var i = 0; i < 4; i++) {
				var feedbackFigure = ff[i];
				
				if (feedbackFigure) {
					feedbackFigure.destroy();
				}
			}
		}

		this.feedbackFigure = null;
		this.feedbackFigures = null;
		dojo.disconnect(this._viewportListener);
		delete this._viewportListener;		
		
		if (this._startHandle) {
			this._startHandle.destroy();
			delete this._startHandle;
		}
		
		if (this.handleListeners !== null) {
			dojo.forEach(this.handleListeners, dojo.disconnect);
			this.handleListeners = null;
		}
		delete this._isHiding;
	},
	
	isShowingFeedback: function() {
		return this.feedbackFigures !== null;
	},
	
	onFirstMove: function() {
		// find out which feedback figure is targeted
		for (var i = 0; i < 4; i++) {
			if (arguments[0].host.shape.shape.src === this.feedbackFigures[i].handleShape.shape.src) {
				this.feedbackFigure = this.feedbackFigures[i];
				break;
			}
		}
		var bounds = rmpw.diagrams.render.util.Workarounds.getTransformedBoundingBox(this.figure);
		var start = {x: bounds.x + (bounds.width / 2), y: bounds.y + (bounds.height / 2)};
		this._startHandle = new rmpw.diagrams.edit.handles.MoveableEndHandle(
				{ owner: this.feedbackLayer, 
					x: start.x, y: start.y, color: this.foregroundColor, 
					editController: this.host, 
					invisible: true, source: true});
		var anchorable = this._startHandle.getAnchorable(this.host.renderMapping);
		if (anchorable) {
			this._startHandle.setDefaultAnchor(anchorable.getSourceConnectionAnchor());
		}
		var handles = [];
		handles.push(this._startHandle);
		handles.push(this.feedbackFigure);
		
		this.removeFeedbackDelays();
		
		var mapping = this.host.renderMapping;
		this.connectionCreationFeedback = 
			new rmpw.diagrams.edit.tools.figures.PolylineFeedback(
					handles, mapping, this.host.renderMapping.policies.graphElement.getGraphContainer(), 
					rmpw.diagrams.render.graph.BaseGraphConfigurator.prototype.OBLIQUE_ROUTING, 
					this.feedbackLayer, this.foregroundColor);
		this._isMoving = true;
		
		if (this.orientationListener) {
			dojo.disconnect(this.orientationListener);
		}
	},
	
	onMove: function() {
		this.connectionCreationFeedback.updateConnectionLines();
	},
	
	onCancel: function() {
		// deletes connectable feedback
		if (this.feedbackFigures && !this._isHiding) {
			this.hideFeedback();
			delete this.feedbackFigures;
		}
		//if user presses Esc key while moving
		if (this._isMoving) {
			delete this._isMoving;
			this._removeConnectionCreationFeedback();
		}
	},
	
	onMoveStop: function(mover) {
		delete this._isMoving;
		if (!this.feedbackFigure) {
			this._removeConnectionCreationFeedback();
			return;
		}
		var targetController = this.feedbackFigure.targetMapping;
		if (targetController) {
			var figurePoint = targetController.renderMapping.root.viewport.getSurfaceAbsoluteMousePosition(mover.lastX, mover.lastY);
			this._onCreateRelationshipRequested({
				sourceController: this.host,
				targetController: targetController,
				point: figurePoint
			});
		} else {
			this._removeConnectionCreationFeedback();
		}
	},
	
	/* 
	 * Determines if this element is connectable as a source element or not.  If not, then it won't display the
	 * edge handles.
	 */
	connectableAsSource: function() {
		var actionBindings = this.getActionBindings();
		if (!actionBindings || actionBindings.length === 0) {
			return false;
		}
		
		return true;
	},
	
	getEnabledBindings: function(targetMapping) {
		var actionBindings = this.getActionBindings();
		var enabledBindings = [];
		
		// only return action bindings that are enabled
		for (var i=0, ab; ab=actionBindings[i]; i++) {
			var abEnabled = ab._filterDisabled(targetMapping);
			if (abEnabled) {
				enabledBindings.push(ab);
			}
		}
		
		return enabledBindings;
	},
	
	/*
	 * Queries the command mappings to determine which relationships are creatable
	 */
	_onCreateRelationshipRequested: function(args){	
		
		// delete line feedback when container refreshes
		var graphElement = this.host.renderMapping.policies.graphElement;
		this._connectRefreshContainer = dojo.connect(graphElement.getGraphManager(), "onRefreshComplete", this, "_removeConnectionCreationFeedback");		

		var actionBindings = this.getEnabledBindings(args.targetController);
		if (actionBindings.length === 1) {
			var binding = actionBindings[0];
			if(binding._action) {
				binding._action.run(args);
			} else {
				console.error("No action bound to ", binding);
			}
		}
		else if (actionBindings.length > 1) {
			var controller = args.targetController.root.policies.contextActionBarScope;
			controller.showToolBarImmediateAt(args.targetController, actionBindings, args.point, controller.CENTER, args);
			
			// delete line feedback if toolbar is removed
			this._connectRemoveToolbar = dojo.connect(controller, "removeToolBarImmediate", this, "_removeConnectionCreationFeedback");		
		}
	},
	
	// deletes connection creation feedback 
	_removeConnectionCreationFeedback: function() {
		this.hideFeedback();
		if (this.connectionCreationFeedback) {
			this.connectionCreationFeedback.destroy();
			delete this.connectionCreationFeedback;
		}
		if (this._connectRefreshContainer) {
			dojo.disconnect(this._connectRefreshContainer);
			delete this._connectRefreshContainer;
		}
		if (this._connectRemoveToolbar) {
			dojo.disconnect(this._connectRemoveToolbar);
			delete this._connectRemoveToolbar;
		}
	},
	
	
	/*
	 * Reset feedback back to default - usually in response to escape, or zoom change
	 */
	refreshFeedback: function() {
		this.onCancel();
		if(this.feedbackFigures) {
			this.hideFeedback();
			this.showFeedback();
		}
	},
	
	/**
	 * @override
	 */
	attachListeners: function() {
		this.inherited(arguments);
		
		// attach to focus event
		if (this.host.policies.focusable) {
			this.listeners.push(dojo.connect(this.host.policies.focusable, "setFocus", this, "_focusChanged"));
		}
		else {
			// maybe Marc's dependency ordering would solve this issue
			this._connectCreateFocusable = dojo.connect(this.host, "createPolicy", this, dojo.hitch(this, function() {
				if (this.host.policies.focusable) {
					dojo.disconnect(this._connectCreateFocusable);
					delete this._connectCreateFocusable;
					this.listeners.push(dojo.connect(this.host.policies.focusable, "setFocus", this, "_focusChanged"));
				}
			}));
		}
		
	},
	
	activateHandle: function(force) {
		// don't do any feedback if connection creation is in progress
		if (this.connectionCreationFeedback) {
			return;
		}
		if (this.eraseFeedbackDelay !== null) {
			clearTimeout(this.eraseFeedbackDelay);
			this.eraseFeedbackDelay = null;
		}
		
		if (force) {
			this.showFeedback();
		} else {
			var feedbackPolicy = this;
			if (this.showFeedbackDelay === null && !this.feedbackFigures) {
				this.showFeedbackDelay = setTimeout( function() {feedbackPolicy.showFeedback();}, 400);	
			}
		}
	},
	
	deactivateHandle: function() {
		// don't do any feedback if connection creation is in progress or if it currently has focus
		if (this.connectionCreationFeedback || 
			(this.feedbackFigure && this.feedbackFigure.policies.focusable.getFocus())) {
			return;
		}
		
		if (this.showFeedbackDelay !== null) {
			clearTimeout(this.showFeedbackDelay);
			this.showFeedbackDelay = null;
		}
		
		var feedbackPolicy = this;
		this.feedbackFigure = null;
		if (this.eraseFeedbackDelay === null && this.feedbackFigures) {
			this.eraseFeedbackDelay = setTimeout( function() {feedbackPolicy.hideFeedback();}, 1000);	
		}
	},
	
	/*
	 * Process when focus changed to put up handles or not
	 */
	_focusChanged: function(focus, force) {
		if (this.connectableAsSource()) {
			if (focus) {
				this.activateHandle(force);
			}
			else {
				this.deactivateHandle();
			}
		}
	}
	
});
