/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2012, 2013.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
dojo.provide("rmpw.diagrams.edit.tools.Duplicatable");

dojo.require("rmpw.diagrams.edit.tools.Feedback");
dojo.require("rmpw.diagrams.edit.handles.DuplicationHandle");
dojo.require("rmpw.common.command.DuplicateElement");

/**
 * Policy that allows a Node to be duplicated with a handle.
 * 
 * @author John Swanke
 */
dojo.declare("rmpw.diagrams.edit.tools.Duplicatable", [rmpw.diagrams.edit.tools.Feedback], {
	singleton: "duplicatable",
	
	gridSpacer: 20,

	showHandleFeedbackDelay : null,
	eraseHandleFeedbackDelay : null,
	
	constructor: function() {
		this.feedbackLayer = this.getFrontFeedbackLayer();
		this.setTargetShape( this.host.getDefaultTargetShape() );
	},
	
	setOwnerChildrenProperty: function(ownerChildrenProperty) {
		this.ownerChildrenProperty = ownerChildrenProperty;
	},
	
	destroy: function() {
		this.inherited(arguments);	
		// erase feedback
		this.hideHandleFeedback();
		
		// detach listeners
		this.detachListeners();
		this.removeFeedbackDelays();
	},
	
	detachListeners: function() {
		if (this.handleListeners !== null) {
			dojo.forEach(this.handleListeners, dojo.disconnect);
			this.handleListeners = null;
		}
	},
	
	/*
	 * remove any existing delays in progress
	 */
	removeFeedbackDelays: function() {
		if (this.eraseHandleFeedbackDelay) {
			clearTimeout(this.eraseHandleFeedbackDelay);
			this.eraseHandleFeedbackDelay = null;
		}
		
		if (this.showHandleFeedbackDelay) {
			clearTimeout(this.showHandleFeedbackDelay);
			this.showHandleFeedbackDelay = null;
		}
	},

	/*
	 * creates four connection handles and attaches listeners to them
	 */	
	showHandleFeedback: function() {
		if (!this.feedbackFigure) {
			this.handleListeners = [];
			var feedbackFig;			
			this.feedbackFigure = feedbackFig = this._createHandleFeedback();
			this.handleListeners.push(dojo.connect(feedbackFig, "onFirstMove", this, "onFirstMove"));
			this.handleListeners.push(dojo.connect(feedbackFig, "onMove", this, "onMove"));
			this.handleListeners.push(dojo.connect(feedbackFig, "onMoveStop", this, "onMoveStop"));
			this._viewportListener = dojo.connect(this.host.root.viewport, "onViewportChanged", this, "refreshFeedback");		
		}
	},
	
	_createHandleFeedback: function() {
		return  new rmpw.diagrams.edit.handles.DuplicationHandle(
				{owner: this.feedbackLayer, editController: this.host});
	},
	
	hideHandleFeedback: function() {
		if (!this.lassoFeedbackFigure) {
			if (this.feedbackFigure) {
				this.feedbackFigure.destroy();
				delete this.feedbackFigure;
			}
		
			dojo.disconnect(this._viewportListener);
			delete this._viewportListener;		
			
			if (this.handleListeners !== null) {
				dojo.forEach(this.handleListeners, dojo.disconnect);
				this.handleListeners = null;
			}
		}
	},

	isShowingFeedback: function() {
		return this.feedbackFigure !== null;
	},
	
	/*
	 * Reset feedback back to default - usually in response to escape, or zoom change
	 */
	refreshFeedback: function() {
		if(this.feedbackFigure) {
			this.hideHandleFeedback();
			this.showHandleFeedback();
		}
	},
	
	/**
	 * @override
	 */
	attachListeners: function() {
		this.inherited(arguments);
		
		// attach to focus event
		if (this.host.policies.focusable) {
			this._attachFocusListeners();
		}
		else {
			// maybe Marc's dependency ordering would solve this issue
			this._connectCreateFocusable = dojo.connect(this.host, "createPolicy", this, dojo.hitch(this, function() {
				if (this.host.policies.focusable) {
					dojo.disconnect(this._connectCreateFocusable);
					delete this._connectCreateFocusable;
					this._attachFocusListeners();
				}
			}));
		}
		
	},
	
	_attachFocusListeners: function() {
		if (!this.host.parent || !this.host.parent.renderMapping.policies.graphEdge) {
			this.listeners.push(dojo.connect(this.host.policies.focusable, "setFocus", this, "_focusChanged"));
		}
	},
	
	activateHandle: function(force) {
		if (this.eraseHandleFeedbackDelay !== null) {
			clearTimeout(this.eraseHandleFeedbackDelay);
			this.eraseHandleFeedbackDelay = null;
		}
		
		if (force) {
			this.showHandleFeedback();
		} else {
			var feedbackPolicy = this;
			if (this.showHandleFeedbackDelay === null && !this.feedbackFigure) {
				this.showHandleFeedbackDelay = setTimeout( function() {feedbackPolicy.showHandleFeedback();}, 400);	
			}
		}
	},
	
	deactivateHandle: function() {
		// don't do any feedback if connection creation is in progress or if it currently has focus
		if (this.duplicationFeedbackFigure || 
			(this.feedbackFigure && this.feedbackFigure.policies.focusable.getFocus())) {
			return;
		}

		if (this.showHandleFeedbackDelay !== null) {
			clearTimeout(this.showHandleFeedbackDelay);
			this.showHandleFeedbackDelay = null;
		}
		
		var feedbackPolicy = this;
		if (this.eraseHandleFeedbackDelay === null && this.feedbackFigure) {
			this.eraseHandleFeedbackDelay = setTimeout( function() {feedbackPolicy.hideHandleFeedback();}, 2000);	
		}
	},
	
	/*
	 * Process when focus changed to put up handles or not
	 */
	_focusChanged: function(focus, force) {
		// can't dup containers--yet :) 
		if (this.host.children.length == 0) {
			if (focus) {
				this.activateHandle(force);
			} else {
				this.deactivateHandle();
			}
		}
	},

	onFirstMove: function(/* dojox.gfx.Mover */ mover) {
		this._container = this.host.renderMapping.policies.graphNode.getGraphContainer().host;
		this.lassoFeedbackFigure = this._createLassoFeedback();
		this._currentBounds = dojo.clone(this._originalBounds);
		this.duplicationFeedbackFigures = [];
	},
	
	onMove: function(/* dojox.gfx.Mover */ mover, /* Object */ shift) {
		if (this.lassoFeedbackFigure) {
			this._currentBounds.width += shift.dx;
			this._currentBounds.height += shift.dy;
			var width = this._currentBounds.width>this._originalBounds.width ? this._currentBounds.width : this._originalBounds.width;
			var height = this._currentBounds.height>this._originalBounds.height ? this._currentBounds.height : this._originalBounds.height;
			this.lassoFeedbackFigure.setShape({width: width, height: height});
			
			var size = this._calcMultiElementSize(width, height);
			if (!this._lastSize || this._lastSize.rows!= size.rows || this._lastSize.cols!= size.cols) {
				this._lastSize = size;
				var figs = 0;
				var x = this._cellSize.width;
				var y = 0;
				// add figures in a grid
				for (var row = 0; row < size.rows; row++) {
					for (var col = 0; col < size.cols; col++) {
						if (row==0 && col==0) continue;
						var dupFigure;
						if (figs+1 > this.duplicationFeedbackFigures.length) {
							dupFigure = this._createDuplicationFeedback();
							this.duplicationFeedbackFigures.push(dupFigure);
						} else {
							dupFigure = this.duplicationFeedbackFigures[figs];
						}
						dupFigure.setTransform({dx: x, dy: y});
						x += this._cellSize.width;
						figs++;
					}
					x = 0;
					y += this._cellSize.height;
				}
				// remove excess dup feedback figures
				while (this.duplicationFeedbackFigures.length>figs) {
					this.duplicationFeedbackFigures.pop().removeShape();
				}
			}
		}
	},
	
	onMoveStop: function(/* dojox.gfx.Mover */ mover) {	
		if (!this._lastSize || this._lastSize.rows<=1 && this._lastSize.cols<=1) {
			// no duplication -- just remove feedback
			this.hideDuplicationFeedback();
		} else {
			// remove feedback after elements are created
			this._saveContainerIds();
			var graphElement = this._container.policies.graphElement;
			this._connectContainerRefresh = dojo.connect(graphElement.getGraphManager(), "onRefreshComplete", this, "hideDuplicationFeedback");
			this._containerStyle = this.host.viewer.viewer.container.style; 
			this._containerStyle.cursor = 'progress';
	
			var root = this.host.root;
			root.getCommandController().execute("rmpw.common.command.DuplicateElement", {
				container: this._container,
				duplicate: this.host,
				size: this._lastSize,
				cellSize: this._cellSize,
				matrix: this.host.renderMapping.root.viewport.contents.matrix,
				refPropName: this.refPropName,
				ownerChildrenProperty: this.ownerChildrenProperty
			});
		}
	},

	_calcMultiElementSize: function(width, height) {
		var x=0, y=0, cols = 0, rows = 0;
		do {
			cols = 0;
			do {
				x += this._cellSize.width;
				cols++;
			} while (x < width);
			x = 0;
			y += this._cellSize.height;
			rows++;
		} while (y < height);
		return {rows: rows, cols: cols};
	},
	
	_createLassoFeedback: function() {	
		var shape = this.host.policies.selectable.getTargetShape();
		this._originalBounds = rmpw.diagrams.render.util.Workarounds.getTransformedBoundingBox(shape);
		
		// determine cell size
		// if thinner then wider, position closer together
		var vSpacer = 20;
		var r = this._originalBounds.width / this._originalBounds.height;
		if (r>4) {
			vSpacer = 5;
		} else if (r>2) {
			vSpacer = 10;
		}
		// if taller then thinner, position further apart
		r = this._originalBounds.height / this._originalBounds.width;
		var hSpacer = 20;
		if (r>4) {
			hSpacer = 100;
		} else if (r>2) {
			hSpacer = 60;
		}
		// else if smaller then a 100x100 grid, increase spacer
		if (hSpacer==20 && vSpacer==20) {
			if (this._originalBounds.width < 120) {
				hSpacer= 120 - this._originalBounds.width;
			}
			if (this._originalBounds.height < 100) {
				vSpacer = 100 - this._originalBounds.height;
			}
		}
		
		var matrix = this.host.renderMapping.root.viewport.contents.matrix;
		this._cellSize = {width: ((this._originalBounds.width + hSpacer) * matrix.xx), height: ((this._originalBounds.height + vSpacer) *  matrix.yy)};
		var feedback = this.feedbackLayer.createRect({width: this._originalBounds.width, height: this._originalBounds.height});
		feedback.setTransform({dx: this._originalBounds.x, dy: this._originalBounds.y});
		feedback.setStroke( {color: "#888888", width: "1px", style: "dashed" } );
		return feedback;
	},
	
	_createDuplicationFeedback: function() {
		var feedback = this.feedbackLayer.createGroup();
		
		var shape = this.host.policies.selectable.getTargetShape();
		var bounds = rmpw.diagrams.render.util.Workarounds.getTransformedBoundingBox(shape);
		
		// Can only clone the shape for feedback if SVG is supported.  Also code below assumes that a transform
		// has been applied to shape.  If shape has been changed directly in applyModelChanges, then no transform
		// may exist on the shape.  For now, default to blue rectangle in these cases.
		if(dojox.gfx.svg && shape.matrix != null) {					
			// create ghost copy of figure
			var subfeedback = null;
			if(shape.children) {
				var children = shape.children;
				subfeedback = feedback.createGroup();
				for(var j = 0, len_j = children.length; j < len_j; j++) {
					subfeedback.rawNode.appendChild(children[j].rawNode.cloneNode(true));
				}
			} else {
				subfeedback = feedback.createShape(shape.shape);
				subfeedback.setFill(shape.fillStyle);
				subfeedback.setStroke(shape.strokeStyle);
			}

			// calculate feedback figure scale and shift
			// TODO: separate getTransformedBoundingBox we can get the original, keep it, then make a
			// 		 transformed copy without having to calculate the original bounds twice
			var originalBounds = rmpw.diagrams.render.util.Workarounds.getBoundingBox(shape);
			var scale = bounds.width / originalBounds.width;
			var transform = {dx: bounds.x, dy: bounds.y, xx: scale, yy: scale};
			subfeedback.setTransform(transform);
		} else {
			// semi-transparent blue rectangle
			var subfeedback = feedback.createRect({width: bounds.width, height: bounds.height});
			subfeedback.setFill([0, 0, 255, 0.1]);
			
			// shift feedback figure
			subfeedback.setTransform({dx: bounds.x, dy: bounds.y});
		}
		
		return feedback;
	},
	
	
	hideDuplicationFeedback: function() {
		if (this._connectContainerRefresh) {
			dojo.disconnect(this._connectContainerRefresh);
			delete this._connectContainerRefresh;
		}
		if (this.lassoFeedbackFigure) {
			this.lassoFeedbackFigure.removeShape();
			delete this.lassoFeedbackFigure;
		}
		if (this.duplicationFeedbackFigure) {
			this.duplicationFeedbackFigure.removeShape();
			delete this.duplicationFeedbackFigure;
		}
		if (this.duplicationFeedbackFigures) {
			while (this.duplicationFeedbackFigures.length>0) {
				this.duplicationFeedbackFigures.pop().removeShape();
			}
			delete this.duplicationFeedbackFigures;
		}
		if (this.feedbackFigure) {
			this.feedbackFigure.destroy();
			delete this.feedbackFigure;
		}
		if (this._viewportListener) {
			dojo.disconnect(this._viewportListener);
			delete this._viewportListener;		
		}
		if (this.handleListeners !== null) {
			dojo.forEach(this.handleListeners, dojo.disconnect);
			this.handleListeners = null;
		}
		if (this._savedContainerIds) {
			this._selectNewContainerIds();
		}
		if (this._containerStyle) {
			this._containerStyle.cursor = '';
			delete this._containerStyle;
		}
	},
	
	_saveContainerIds: function () {
		this._savedContainerIds = [];
		for (var i = 0, len = this._container.children.length; i < len; i++) {
			this._savedContainerIds[this._container.children[i].id] = true;
		}
	},
	
	_selectNewContainerIds: function () {
		var selectionController = this.host.root.getSelectionController();
		selectionController.deselectAll();
		selectionController.select(this.host);
		for (var i = 0, len = this._container.children.length; i < len; i++) {
			var child = this._container.children[i];
			if (!child.policies.graphEdge && !(child.id in this._savedContainerIds)) {
				selectionController.select(child.__editController);
			}
		}
		delete this._savedContainerIds;
	}
	
	
});
