/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010, 2013.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.diagrams.edit.tools.Feedback");

dojo.require("rmpw.common.policy.Policy");
dojo.require("rmpw.diagrams.edit.tools.figures.RectangularHighlight");

/**
 * Abstract edit policy for implementing visual feedback.
 * 
 * Implementors must:
 * - Implement show/erase target feedback methods
 * - Attach their own figure or key listeners to show and hide their feedback at appropriate times
 * 
 * @author Marc Gobeil
 */
dojo.declare("rmpw.diagrams.edit.tools.Feedback", rmpw.common.policy.Policy, {
	feedbackFigureType: rmpw.diagrams.edit.tools.figures.RectangularHighlight,
	
	/**
	 * Sets the target for the feedback policy
	 */
	setTargetShape: function(targetFigure) {
		if(this.listeners && this.listeners.length !== 0) {
			this.detachListeners();
		}
		
		if(dojo.isString(targetFigure)) {
			this.figure = this.host.renderMapping.template[targetFigure];
		} else {
			this.figure = targetFigure;
		}
		this.attachListeners();
	},
	
	getTargetShape: function() {
		return this.figure;
	},
	
	/**
	 * Sets the layer where the feedback will occur
	 */
	setFeedbackLayer: function(feedbackLayer) {
		this.feedbackLayer = feedbackLayer;
	},
		
	setFeedbackFigureType: function(feedbackFigureType) {
		this.feedbackFigureType = feedbackFigureType;
	},
	
	/**
	 * @abstract final - should not be subclassed.  Subclass _addListeners instead
	 */
	attachListeners: function() {
		if(this.listeners && this.listeners.length !== 0) {
			console.warn("listeners already connected");
			return;
		}
		this.listeners = [];
		
		this._addListeners();
	},
	
	_addListeners: function() {
		// TODO (MG): find more generic way to do this
		if (this.host.renderMapping.policies) {
			var graphElement = this.host.renderMapping.policies.graphElement;
			if(graphElement) {
				this.listeners.push(dojo.connect(graphElement, "updateShape", this, "refreshFeedback"));
			}
		}
	},
	
	detachListeners: function() {
		if(!this.listeners) {
			console.warn("listeners already detached");
			return;
		}
		dojo.forEach(this.listeners, dojo.disconnect);
		delete this.listeners;
	},
	
	destroy: function() {
		// erase feedback
		this.hideFeedback();
		
		// detach listeners
		this.detachListeners();
	},

	getFrontFeedbackLayer: function() {
		return this.host.renderMapping.root.feedbackLayer;
	},
	
	getRearFeedbackLayer: function() {
		var renderMapping = this.host.renderMapping;
		if(renderMapping.getGraphContainerPolicy) {
			return renderMapping.getGraphContainerPolicy().getRearFeedbackLayer();
		} else {
			return renderMapping.root.rearFeedbackLayer;
		}		
	},	

	activate: function() {
		this.attachListeners();
	},

	deactivate: function() {
		this.hideFeedback();
		this.detachListeners();
	},
	
	setFeedbackStyle: function(mode,strokeWidth) {
		this._mode = mode;
		this._strokeWidth = strokeWidth;
		if(this.feedbackFigure) {
			this.feedbackFigure.destroy();
			this.feedbackFigure = this._createFeedback(mode,strokeWidth);			
		}
	},
	
	showFeedback: function(mode,strokeWidth) {
		if(!this.feedbackFigure) {
			this.feedbackFigure = this._createFeedback((mode?mode:this._mode), (strokeWidth?strokeWidth:this._strokeWidth));
			this._viewportListener = dojo.connect(this.host.root.viewport, "onViewportChanged", this, "refreshFeedback");
			this._onresizeListener = dojo.connect("onresize", dojo.hitch(this, this.refreshFeedback));
			return true;
		}
		return false;
	},
	
	_createFeedback: function() {
		return new this.feedbackFigureType(this.feedbackLayer, this.figure, this.host);
	},
	
	hideFeedback: function() {
		if(this.feedbackFigure) {
			this.feedbackFigure.destroy();
			delete this.feedbackFigure;
			dojo.disconnect(this._viewportListener);
			delete this._viewportListener;
			dojo.disconnect(this._onresizeListener);
			delete this._onresizeListener;
		}
	},
	
	// remove any feedback that's still fading
	hideFeedbackNoFade: function() {
		// if hide feedback hasn't already been called we can surgically remove it
		if(this.feedbackFigure && this.feedbackFigure.shape) {
			this.feedbackLayer.remove(this.feedbackFigure.shape);
			this.hideFeedback();
		} else {			
			// else we need to remove them all
            for (var child; child=this.feedbackLayer.children[0];) {
				this.feedbackLayer.remove(child);
			}
		}
	},
	
	refreshFeedback: function() {
		if(this.feedbackFigure) {
			this.hideFeedback();
			this.showFeedback();
		}
	},
	
	isShowingFeedback: function() {
		return this.feedbackFigure? true : false;
	}
});