/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.diagrams.edit.tools.FocusScope");

/**
 * Keeps track of which element is focused on a diagram.  Only one element can be in focus at a time.
 * 
 * @author Steve Shaw
 */
dojo.declare("rmpw.diagrams.edit.tools.FocusScope", null, {
	singleton: "focusScope",
	topics: rmpw.diagrams.render.util.Events.EventTopics,
	
	constructor: function(viewer) {
		this._viewer = viewer;
		this._focusedElement = null;
		this._captureElement = null;
		this._contentCell = dojo.byId("content-cell");
		var fc = this;
		this._subscribeFocusNode = dojo.subscribe("focusNode", function(widget) {
			if (widget !== fc._contentCell) {
				fc.setFocus(null);
				viewer.getAcceleratorController().disable();
			}
			else {
				// set the top level container as focused
				if (fc.getFocus() === null) {
					fc.onDefaultFocus();
				}
			}
		}); 
		
		/* Want to squash focus effect when panning */
		this._connectBeginPan = dojo.connect(viewer.viewport, "beginPan", this, "suspendFocus");
		this._connectEndPan = dojo.connect(viewer.viewport, "endPan", this, "restoreFocus");
	},
	

	destroy: function() {
		dojo.unsubscribe(this._subscribeFocusNode);
		if (this._connectBeginPan) {
			dojo.disconnect(this._connectBeginPan);
			delete this._connectBeginPan;
		}
		if (this._connectEndPan) {
			dojo.disconnect(this._connectEndPan);
			delete this._connectEndPan;
		}
		this.releaseCapture();
	},
	
	onDefaultFocus: function() {
		// do nothing - used for listeners
	},
	
	/*
	 * Returns the mapping element currently in focus - can be null
	 */
	getFocus: function() {
		return this._focusedElement;
	},
	
	disable: function() {
		this.setFocus(null);
		this.releaseCapture();
		this._disabled = true;
	},
	
	enable: function(noFocus) {
		delete this._disabled;
		if (!noFocus) {
			this.onDefaultFocus();
		}
	},
	
	suspendFocus: function() {
		// suspend focus
		if (this._focusedElement && this._focusedElement.policies.focusable.suspendable()) {
			this._suspendedFocusedElement = this._focusedElement;
			this._suspendedFocusedElement.policies.focusable.hideFeedbackNoFade();
		}
		this.disable();
	},
	
	restoreFocus: function() {
		this.enable(true);
		// resume focus
		if (this._suspendedFocusedElement) {
			if (!this._focusedElement) {
				this.setFocus(this._suspendedFocusedElement);
			}
			delete this._suspendedFocusedElement;
		}
	},
	
	/*
	 * Returns the current element under focus capture.  This is useful for
	 * continuous gestures that are asynchronous w.r.t. UI (drag / move etc.).
	 * The convenience of this is that it allows other items to be focusable
	 * while the particular item is under capture.  For instance if you're dragging
	 * an item onto another, the target object could be focused while the element
	 * is still under capture.
	 */
	getCaptureFocus: function(capture) {
		if (!this._captureElement) {
			if (this._focusedElement) {
				this.setCapture(this._focusedElement);
			}
		}
		
		return this._captureElement;
	},
	
	setCapture: function(element) {
		if (element) {
			this._captureElement = element;
			this._connectCommandStack = dojo.connect(this._viewer.getCommandController().commandStack, "postExecuteHook", dojo.hitch(this, function() {
				this.releaseCapture();
			}));
			this._connectDestroy = dojo.connect(this._captureElement, "destroy", dojo.hitch(this, function() {
				this.releaseCapture();
			}));
			if (this._viewer.policies.selectionScope) {
				this._connectSelect = dojo.connect(this._viewer.policies.selectionScope, "onSelectionChanged", dojo.hitch(this, function() {
					this.releaseCapture();
				}));
			}
		}
	},
	
	releaseCapture: function() {
		this._captureElement = null;
		if (this._connectSelect) {
			dojo.disconnect(this._connectSelect);
			delete this._connectSelect;
		}
		if (this._connectCommandStack) {
			dojo.disconnect(this._connectCommandStack);
			delete this._connectCommandStack;
		}
		if (this._connectDestroy) {
			dojo.disconnect(this._connectDestroy);
			delete this._connectDestroy;
		}
	},
	
	/*
	 * Sets the current focus to a particular mapping element
	 */
	setFocus: function(focusedElement, force, touchedElement) {
		if (this._disabled && !force) {
			return;
		}
		if (focusedElement !== this._focusedElement) {
			if (this._focusedElement && this._focusedElement.policies.focusable && !this._touchedElement) {
				this._focusedElement.policies.focusable.setFocus(false);
			}
			this._focusedElement = focusedElement;
			
			// ensure content is focused as well
			if (this._focusedElement) {
				dijit.focus(this._contentCell); // ensure drawing surface has focus
				this._viewer.getAcceleratorController().enable();
			}
		}
		if (this._focusedElement && this._focusedElement.policies.focusable) {
			this._focusedElement.policies.focusable.setFocus(true, force);
		}
		
		// don't unfocus a touched element until another element is touched
		if (this._touchedElement && (touchedElement || !focusedElement)) {
			this._touchedElement.policies.focusable.setFocus(false);
		}
		if (touchedElement) {
			this._touchedElement = focusedElement;
		} else if (!focusedElement) {
			this._touchedElement = undefined;
		}
	},
	
	/*
	 * Resets the current focus to the first focusable parent, 
	 * or null if there is no focusable parent.
	 */
	resetFocus: function() {
		if (this._focusedElement) {
			// find a parent that is focusable
			var parent = this._focusedElement.parent;
			var policies = this._focusedElement.renderMapping.policies;
			if (parent &&
				policies &&
				policies.graphEdge &&
				policies.graphEdge.sourceMapping === parent.renderMapping) {
                // for edge, if its parent is the same as its source, consider the parent of the source 
				// (so the source does not get the focus)
				parent = parent.parent;		
			}
			while (parent) {
				if (parent.policies && parent.policies.focusable) {
					this.setFocus(parent);
					return;
				}
				parent = parent.parent;
			}
		}
		this.setFocus(null);
	}
});