/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2012.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
dojo.provide("rmpw.diagrams.edit.tools.LabelEditable");

dojo.require("rmpw.diagrams.edit.tools.Feedback");
dojo.require("rmpw.diagrams.edit.handles.LabelEditHandle");

/**
 * Policy that allows a Label to be Editable with a rename handle.
 * 
 * @author John Swanke
 */
dojo.declare("rmpw.diagrams.edit.tools.LabelEditable", [rmpw.diagrams.edit.tools.Feedback], {
	singleton: "labeleditable",
	
	showFeedbackDelay : null,
	eraseFeedbackDelay : null,
	
	constructor: function() {
		this.feedbackLayer = this.getFrontFeedbackLayer();
		this.setTargetShape( this.host.getDefaultTargetShape() );
	},
	
	destroy: function() {
		this.inherited(arguments);	
		// erase feedback
		this.hideFeedback();
		
		// detach listeners
		this.detachListeners();
		this.removeFeedbackDelays();
	},
	
	detachListeners: function() {
		if (this.handleListeners !== null) {
			dojo.forEach(this.handleListeners, dojo.disconnect);
			this.handleListeners = null;
		}
	},
	
	setShape: function(label) {
		if(dojo.isString(label)) {
			label = this.host.renderMapping.template[label];
		}
		this._label = label;
	},
	
	setOnEditLabel: function(action) {
		this._action = action;
	},
	
	/*
	 * remove any existing delays in progress
	 */
	removeFeedbackDelays: function() {
		if (this.eraseFeedbackDelay) {
			clearTimeout(this.eraseFeedbackDelay);
			this.eraseFeedbackDelay = null;
		}
		
		if (this.showFeedbackDelay) {
			clearTimeout(this.showFeedbackDelay);
			this.showFeedbackDelay = null;
		}
	},

	/*
	 * creates four connection handles and attaches listeners to them
	 */	
	showFeedback: function() {
		if (!this.feedbackFigures) {
			if (!this._labels) {
				this._labels = [];
				this._labels[0] = this._label;
			}
			
			this.feedbackFigures = [];
			this.handleListeners = [];
			var feedbackFig;			
			for (var i = 0; i < this._labels.length; i++) {
				var label = this._labels[i];
				this.feedbackFigures[i] = feedbackFig = this._createFeedback(label);
				this.handleListeners.push(feedbackFig.handleShape.connect("onmousedown", this, "_onMouseDown"));
				this.handleListeners.push(feedbackFig.handleShape.connect("onmouseup", this, "_onMouseUp"));
			}
			this._viewportListener = dojo.connect(this.host.root.viewport, "onViewportChanged", this, "refreshFeedback");		
		}
	},
	
	_onMouseDown: function(e) {
		this._saveStrt = {x: e.clientX, y: e.clientY};
	},
	
	_onMouseUp: function(e) {
		// only execute action if user hasn't clicked and dragged on this shape
		if ( this._saveStrt && this._saveStrt.x === e.clientX && 
				this._saveStrt.y === e.clientY && this._action) {
			this.hideFeedback();
			this._action.run({});
		}
	},
	
	_createFeedback: function(label) {
		return  new rmpw.diagrams.edit.handles.LabelEditHandle(
				{owner: this.feedbackLayer, target: label, editController: this.host});
	},
	
	hideFeedback: function() {
		if (this.feedbackFigures) {
			for (var i = 0; i < this.feedbackFigures.length; i++) {
				if (this.feedbackFigures[i]) {
					this.feedbackFigures[i].destroy();
					delete this.feedbackFigures[i];
				}
			}
		}
		this.feedbackFigure = null;
		this.feedbackFigures = null;
		dojo.disconnect(this._viewportListener);
		delete this._viewportListener;		
		
		if (this.handleListeners !== null) {
			dojo.forEach(this.handleListeners, dojo.disconnect);
			this.handleListeners = null;
		}
	},
	
	isShowingFeedback: function() {
		return this.feedbackFigures !== null;
	},
	
	onCancel: function() {
		this.hideFeedback();
	},
	
	/*
	 * Reset feedback back to default - usually in response to escape, or zoom change
	 */
	refreshFeedback: function() {
		if(this.feedbackFigures) {
			this.hideFeedback();
			this.showFeedback();
		}
	},
	
	/**
	 * @override
	 */
	attachListeners: function() {
		this.inherited(arguments);
		
		// attach to focus event
		if (this.host.policies.focusable) {
			this._attachFocusListeners();
		}
		else {
			// maybe Marc's dependency ordering would solve this issue
			this._connectCreateFocusable = dojo.connect(this.host, "createPolicy", this, dojo.hitch(this, function() {
				if (this.host.policies.focusable) {
					dojo.disconnect(this._connectCreateFocusable);
					delete this._connectCreateFocusable;
					this._attachFocusListeners();
				}
			}));
		}
		
	},
	
	_attachFocusListeners: function() {
		this.listeners.push(dojo.connect(this.host.policies.focusable, "setFocus", this, "_focusChanged"));
		// if on an edge, also listen to its focus
		if (this.host.parent && this.host.parent.renderMapping.policies.graphEdge) {
			this.listeners.push(dojo.connect(this.host.parent.policies.focusable, "setFocus", this, "_focusChanged"));			
		}
	},
	
	activateHandle: function(force) {
		if (this.eraseFeedbackDelay !== null) {
			clearTimeout(this.eraseFeedbackDelay);
			this.eraseFeedbackDelay = null;
		}
		
		if (force) {
			this.showFeedback();
		} else {
			var feedbackPolicy = this;
			if (this.showFeedbackDelay === null && !this.feedbackFigures) {
				this.showFeedbackDelay = setTimeout( function() {feedbackPolicy.showFeedback();}, 400);	
			}
		}
	},
	
	deactivateHandle: function() {
		// don't do any feedback if connection creation is in progress or if it currently has focus
		if (this.connectionCreationFeedback || 
			(this.feedbackFigure && this.feedbackFigure.policies.focusable.getFocus())) {
			return;
		}
		
		if (this.showFeedbackDelay !== null) {
			clearTimeout(this.showFeedbackDelay);
			this.showFeedbackDelay = null;
		}
		
		var feedbackPolicy = this;
		this.feedbackFigure = null;
		if (this.eraseFeedbackDelay === null && this.feedbackFigures) {
			this.eraseFeedbackDelay = setTimeout( function() {feedbackPolicy.hideFeedback();}, 2000);	
		}
	},
	
	/*
	 * Process when focus changed to put up handles or not
	 */
	_focusChanged: function(focus, force) {
		if (focus) {
			this.activateHandle(force);
		} else {
			this.deactivateHandle();
		}
	}
	
});
