/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.diagrams.edit.tools.Moveable");

dojo.require("rmpw.diagrams.render.util.Workarounds");
dojo.require("rmpw.diagrams.edit.tools.MoveableBase");
dojo.require("rmpw.diagrams.render.util.LabelHelper");

/**
 * TODO: there's a bug when if you initiate a move, but the machine is being slow and doesn't create
 * the feedback figure until the mouse has dragged outside of where the figure's bounds,
 * the move won't initiate until the mouse returns to its bounds.
 * 
 * @author Marc Gobeil / Steve Shaw
 */
dojo.declare("rmpw.diagrams.edit.tools.Moveable", rmpw.diagrams.edit.tools.MoveableBase, {
	singleton: "moveable",
	
	constructor: function() {
		this.feedbackLayer = this.host.renderMapping.root.feedbackLayer; 
		
		if(!this.host.onFigureMoving) {
			this.host.onFigureMoving = function() { };
		}
		
		if (!this.host.__setUpdatePosition) {
			this.host.__setUpdatePosition = dojo.hitch(this, function(changeSet, x, y) {
				
				var xTrace = this.host.renderMapping.traces[this.host.renderMapping.policies.graphElement._getInternalTraceName("x")];
				var yTrace = this.host.renderMapping.traces[this.host.renderMapping.policies.graphElement._getInternalTraceName("y")];
				if (xTrace && yTrace) {
					var change = changeSet.createChange(xTrace.getOwner());
					change.set(xTrace.propertyName, x);
					change.set(yTrace.propertyName, y);
					return true;
				}
					
				return false;
			});
		}
		
		this.setTargetShape(this.host.getDefaultTargetShape());
		this._connectMove.push(dojo.connect(this._moveable, "onMoving", this.host, "onFigureMoving"));
		
		this._reparentable = this.host.parent.renderMapping.policies.graphEdge === undefined;	
		
		// force live feedback if we're in a container - otherwise, the selection flashing is too distracting
		if (this.host.parent.parent != this.host.root) {
			this.setLiveFeedback(true);
		}
	},
	
	onCancel: function() {
		if (this.host) {
			if (this.host.root.policies.focusScope.getCaptureFocus() === this.host) {
				this.host.root.policies.focusScope.releaseCapture();
			}
			
			// stop looking for reparenting
			this._disableReparenting();
			
			if (!this._connectContainerRefresh && this._hidden) {
				if(this.feedbackFigure) {
					// reset feedback
					this.feedbackFigure.removeShape();
					this.feedbackFigure = null;
				}
				
				//if wysiwyg, restore host
				if (this._liveFeedback) {
					this._showDraggedShapes();
				}
				
				if (this._containerStyle) {
					this._containerStyle.cursor = '';
					delete this._containerStyle;
				}
			}
		}
	},
	
	detachListeners: function(){
		
		/* Detaching moveable listeners is tricky - for re-attachment. Instead we're not
		 * detaching - just preventing the moving code from running, and on re-attachment
		 * we're letting it run again. Detach is equivalent to no action and doesn't stop the
		 * event.
		 */
		
		this._moveable.detachListeners();
	},
	
	reAttach: function(){
		
		this._moveable.reAttach();
	},
	
	setLiveFeedback: function(liveFeedback) {
		this._liveFeedback = liveFeedback;
		
		// if live feedback, only animate surrounding shapes on a move
		if (liveFeedback) {
			this.host.viewer.nonAnimatedMappings = [];
		}
	},
	
	setOnChangePosition: function(changePositionAction) {
		this.host.__setUpdatePosition = dojo.hitch(this, function(changeSet, x, y) {
			changePositionAction.run({changeSet: changeSet, x: x, y: y});
		});
	},
	
	/*
	 * Performs the move
	 */
	doMove: function(dx, dy, selection) {
	    var moveAction;
		dojo.publish("/gfx/move/start");
		
		// TODO (MG): should not be special cases, standardize action interface for move result
		if (this.host.renderMapping.policies.graphNode) {
			this._doNodeMove.apply(this, arguments);
		}
		else if (this.host.renderMapping.policies.graphFloatingLabel) {
			this._doLabelMove.apply(this, arguments);
		}
		else {
			moveAction = this.host.policies.moveAction;
			moveAction.run({ dx: dx, dy: dy });
		}
		
		// if not moved in the diagram invalidate that container to get the new children size
		if (this.host.renderMapping.policies.graphNode) {
			var container = this.host.renderMapping.policies.graphNode.getGraphContainer();
			if (!container.isTopLevel()) {
				container.host.invalidate();
			}
		}
		
		dojo.publish("/gfx/move/stop");
	},
	
	/*
	 * Performs the move for a graph node
	 */
	_doNodeMove: function(dx, dy, selection) {
		var nodes = [];
		nodes.push(this.host);
		var root = this.host.root;
		var args = {
				shift: { dx: dx, dy: dy }, 
				nodes: selection ? root.getSelectionController().getSelection() : nodes
		};
		if (this._liveFeedback) {
			args.onCommitBlocked = dojo.hitch(this, this._removeFeedback);
		}
		root.getCommandController().execute("rmpw.diagrams.edit.commands.MoveCommand", args);
	},
	
	/*
	 * Performs the move for graph floating label
	 */
	_doLabelMove: function(dx, dy, selection) {
		var nodes = [];
		nodes.push(this.host);
		var root = this.host.root;
		var args = {
				location: this._calculateLabelLocation(dx, dy),
				nodes: selection ? root.getSelectionController().getSelection() : nodes
		};
		if (this._liveFeedback) {
			args.onCommitBlocked = dojo.hitch(this, this._removeFeedback);
		}
		root.getCommandController().execute("rmpw.diagrams.edit.commands.MoveCommand", args);
	},	
	
	_calculateLabelLocation: function(dx, dy){
		var renderMapping = this.host.renderMapping;
		var figureBounds = renderMapping.policies.graphElement.getBounds();
		figureBounds.width =  figureBounds.width;
		figureBounds.height =  figureBounds.height;
		figureBounds.x =  (figureBounds.x+dx);
		figureBounds.y =  (figureBounds.y+dy);
		var helper = new rmpw.diagrams.render.util.LabelHelper();
		return helper.offsetFromRelativeCoordinate(figureBounds, 
				renderMapping.policies.graphFloatingLabel.locator.getParentPointList(), 
				renderMapping.policies.graphFloatingLabel.locator.getReferencePoint());
	},
		
	onFirstMove: function(/* dojox.gfx.Mover */ mover) {
		
		// create move feedback
		this.feedbackFigure = this.createFeedbackFigure();
		this.startX = mover.lastX;
		this.startY = mover.lastY;
		
		// watch for reparenting
		this._enableReparenting();
		
		//if liveFeedback, hide host
		if (this._liveFeedback) {
			this._hideDraggedShapes();
		}		
	},
	
	onMove: function(/* dojox.gfx.Mover */ mover, /* Object */ shift)  {
		if (!this._actualFirstMove && (shift.dx!==0 || shift.dy!==0)) {
			this._actualFirstMove = true;
			this.host.root.policies.focusScope.disable();
			this.host.root.policies.focusScope.setCapture(this.host);
			
			//if liveFeedback, feedback on any selected shape
			if (this._liveFeedback) {
				this.hideSelectedFeedback(this.host);
			}
		}
		
		if (this.feedbackFigure) {
			this.feedbackFigure.applyLeftTransform(shift);
		}
		if (this._movedShapeBox) {
			this._movedShapeBox.translate(shift.dx, shift.dy);
		}
		
		// watch for reparenting
		this._highlightParent(mover);
	},
	
	onMoveStop: function(/* dojox.gfx.Mover */ mover) {
		if (this._actualFirstMove) {
			delete this._actualFirstMove;
			this.host.root.policies.focusScope.enable();
		}
		
		if (!this.feedbackFigure) {
			return;
		}
		
		// disable reparenting
		this._disableReparenting();
		
		var m = dojox.gfx.matrix;
		var viewportMatrix = m.invert(this.host.renderMapping.root.viewport.contents.matrix);
		var transformedStart = m.multiplyPoint(viewportMatrix, this.startX, this.startY);
		var transformedEnd = m.multiplyPoint(viewportMatrix, mover.lastX, mover.lastY);

		var dx = transformedEnd.x - transformedStart.x;
		var dy = transformedEnd.y - transformedStart.y;
		var moved = dx !== 0 || dy !== 0;
		
		var reparent = this._toParentContainer !== this._fromParentContainer;
		if (moved) {
			//if liveFeedback, leave feedback up until container is refreshed
			var graphElement = reparent ? this._toParentContainer.policies.graphElement : this.host.renderMapping.policies.graphElement;
			this._connectContainerRefresh = dojo.connect(graphElement.getGraphManager(), "onRefreshComplete", this, "_removeFeedback");
			this._containerStyle = this.host.viewer.viewer.container.style; 
			this._containerStyle.cursor = 'progress';

			// if live don't animate the move
			if (this._liveFeedback) {
				this.host.viewer.nonAnimatedMappings.push(this.host);
			}
		} else {
			this._removeFeedback();
		}
		
		if (moved) {
			if (reparent) {
				if (!this.doReparent(dx, dy)) {
					this.doMove(dx, dy, true);
				}
			} else {
				this.doMove(dx, dy, true);
			}
		}		
	},
	
	// hide feedback on this shape and any inner shape
	hideSelectedFeedback: function(host) {
		var selection = host.root.getSelectionController().getSelection();
		for (var i = 0, len_i = selection.length; i < len_i; i++) {
			this._hideSelectedFeedback(selection[i]);
		}
	},
	
	_hideSelectedFeedback: function(controller) {
		if (controller.policies.focusable) {
			controller.policies.focusable.hideFeedbackNoFade();			
		}
		for (var policy in controller.policies) {
			if (controller.policies[policy].hideFeedback) {
				controller.policies[policy].hideFeedback();
			}
		}
		
		if (controller.children) {
			for (var i = 0, len_i = controller.children.length; i < len_i; i++) {
				this._hideSelectedFeedback(controller.children[i]);				
			}
		}
	},
	
	
	
	_enableReparenting: function() {
		if (!this._reparentable){ return;}
		
		// remove now so that if another move is done before this move finishes, it won't get inserted into this container
		this._graphManager = this.host.renderMapping.policies.graphElement.getGraphManager();
		if (this._graphManager) {
			this._graphManager.removeFocusables(this.host.root.getSelectionController().getSelection());
		}
	
		// remember to check that it isn't moved to a smaller container
		this._thisBox = this.host.renderMapping.policies.graphElement.getBounds();

		// initialize "from"/"to" container attributes to current host's parent mapping
		var container = this.host.renderMapping.policies.graphNode.getGraphContainer();
		this._fromParentContainer = this._toParentContainer = container.host;
		
		// setup for moving shapes out of containers
		var movedShape = this.host.policies.selectable.getTargetShape();
		var bounds = rmpw.diagrams.render.util.Workarounds.getTransformedBoundingBox(movedShape);
		this._movedShapeBox =  new rmpw.diagrams.render.util.Rectangle(bounds.x, bounds.y, bounds.width, bounds.height);
		this._initialMovedShapeBox =  dojo.clone(this._movedShapeBox);
		if (!container.isTopLevel() && this._fromParentContainer.__editController.policies.selectable) {		
			var fromShape = this._fromParentContainer.__editController.policies.selectable.getTargetShape();
			bounds = rmpw.diagrams.render.util.Workarounds.getTransformedBoundingBox(fromShape);
			this._fromContainerBox =  new rmpw.diagrams.render.util.Rectangle(bounds.x, bounds.y, bounds.width, bounds.height);
		}
	},
	
	_highlightParent: function(mover) {
	    var toBox;
		if (!this._reparentable) {return;}
		// if "from" container is not the diagram get "from" container's focusable policy and show focus feedback in blue
		if (!this._fromParentContainer.policies.graphContainer.isTopLevel() && !this._fromFocusable) {
			this._fromFocusable = this._fromParentContainer.__editController.policies.focusable;
			this._fromFocusable.showFeedback();
		}
		
		// can't use setFocus mechanism because mouse never leaves dragged feedback shape
		var pt = this.host.renderMapping.root.viewport.getSurfaceAbsoluteMousePosition(mover.lastX, mover.lastY);
		var mapping = this._graphManager.getContainerFromPoint(pt, this.host.root.getSelectionController().getSelection());
		var container = mapping ? mapping : this.host.viewer.diagram;
		if (container !== this._toParentContainer) {
			if (mapping) {
				var toShape = mapping.__editController.policies.selectable.getTargetShape();
				var bounds = rmpw.diagrams.render.util.Workarounds.getTransformedBoundingBox(toShape);
				toBox =  new rmpw.diagrams.render.util.Rectangle(bounds.x, bounds.y, bounds.width, bounds.height);
			}
			
			// see if this to container is valid
			var isNewTo = container !== this._fromParentContainer; 
			if (isNewTo) {
				if (toBox) {
					// can't move a shape to its own inner shape
					if (this._initialMovedShapeBox.contains_4(toBox)) {
						return;
					}
				}
				// don't leave an inner shape until moved shape is 75% out of it
				if (this._movedShapeBox && this._fromContainerBox) {
					if (!toBox || toBox.contains_4(this._fromContainerBox)) {
						var intersection = dojo.clone(this._fromContainerBox).intersect(this._movedShapeBox);
						if ( intersection.width * intersection.height * 100 / this._movedShapeBox.width / this._movedShapeBox.height > 25) {
							return;
						}
					}
				}
			}
			
			// hide any "to" container focus
			if (this._toFocusable) {
				this._toFocusable.hideFeedbackNoFade();
				delete this._toFocusable;
			}
			var toContainer = this.host.viewer.diagram;
			if (toBox) {
				// prevent really big shapes from being accidently moved to tiny shapes
				if (this._thisBox.width / toBox.width < 8 && this._thisBox.height / toBox.height < 8) {
					// show any new "to" container focus
					if (isNewTo && mapping.__editController.policies.focusable) {
						this._toFocusable = mapping.__editController.policies.focusable;
						this._toFocusable.showFeedback("selection", 4);
					}
					toContainer = container;
				}
			} else {
				// highlight diagram
				if (isNewTo && toContainer.__editController.policies.focusable) {
					this._toFocusable = toContainer.__editController.policies.focusable;
					this._toFocusable.showFeedback("selection");
				}
			}
			this._toParentContainer = toContainer;
		}
	},
	
	_disableReparenting: function() {
	     // cleanup focus feedback 
		if (this._fromFocusable) {
			this._fromFocusable.hideFeedbackNoFade();
			delete this._fromFocusable;
		}
		if (this._toFocusable) {
			this._toFocusable.hideFeedbackNoFade();
			delete this._toFocusable;
		}
	},
	
	doReparent: function(dx, dy) {
		if (this._toParentContainer.__editController.policies.reparentable) {
			this._saveToContainerIds();
			this._selectionController = this.host.root.getSelectionController();
			this._toParentContainer.__editController.policies.reparentable.
				doReparent(this.host.root.getSelectionController().getSelection(), dx, dy);
			return true;
		}
		return false;
	},
	
	
	createFeedbackFigure: function() {
		var feedback = this.feedbackLayer.createGroup();
		
		var selection = this.host.root.getSelectionController().getSelection();
		for (var i = 0, len_i = selection.length; i < len_i; i++) {
			var shape = selection[i].policies.selectable.getTargetShape();
			var bounds = rmpw.diagrams.render.util.Workarounds.getTransformedBoundingBox(shape);
			var subfeedback;
			
			// Can only clone the shape for feedback if SVG is supported.  Also code below assumes that a transform
			// has been applied to shape.  If shape has been changed directly in applyModelChanges, then no transform
			// may exist on the shape.  For now, default to blue rectangle in these cases.
			if(dojox.gfx.svg && shape.matrix) {					
				// create ghost copy of figure
				subfeedback = null;
				if(shape.children) {
					var children = shape.children;
					subfeedback = feedback.createGroup();
					for(var j = 0, len_j = children.length; j < len_j; j++) {
						subfeedback.rawNode.appendChild(children[j].rawNode.cloneNode(true));
					}
				} else {
					subfeedback = feedback.createShape(shape.shape);
					subfeedback.setFill(shape.fillStyle);
					subfeedback.setStroke(shape.strokeStyle);
				}
				
				// if liveFeedback, feedback is not faded
				if (!this._liveFeedback) {
					subfeedback.rawNode.style.opacity = 0.4;
				}

				// calculate feedback figure scale and shift
				// TODO: separate getTransformedBoundingBox we can get the original, keep it, then make a
				//       transformed copy without having to calculate the original bounds twice
				var originalBounds = rmpw.diagrams.render.util.Workarounds.getBoundingBox(shape);
				var scale = bounds.width / originalBounds.width;
				var transform = {dx: bounds.x, dy: bounds.y, xx: scale, yy: scale};
				subfeedback.setTransform(transform);
			} else {
				// semi-transparent blue rectangle
				subfeedback = feedback.createRect({width: bounds.width, height: bounds.height});
				subfeedback.setFill([0, 0, 255, 0.1]);
				
				// shift feedback figure
				subfeedback.setTransform({dx: bounds.x, dy: bounds.y});
			}
		}
		
		return feedback;
	},
	
	_removeFeedback: function () {
		if(this.feedbackFigure) {
			this.feedbackFigure.removeShape();
			this.feedbackFigure = null;
			
			// if we weren't reparented we would still have a host
			if (this.host && this._liveFeedback) {
				this._showDraggedShapes();				
			} else if (this._savedContainerIds) {
				this._selectToContainerIds();
			}
		}
		if (this._containerStyle) {
			this._containerStyle.cursor = '';
			delete this._containerStyle;
		}
		if (this._connectContainerRefresh) {
			dojo.disconnect(this._connectContainerRefresh);
			delete this._connectContainerRefresh;
		}
	},
	
	_hideDraggedShapes: function () {
		this._hiddenShapes = [];
		var selection = this.host.root.getSelectionController().getSelection();
		for (var i = 0, len_i = selection.length; i < len_i; i++) {
			var selectable = selection[i].policies.selectable;
			var shape = selectable.getTargetShape();
			if (shape) {
				this._hidden = true;
				selectable.hideFeedback();
				dojo.style(shape.rawNode, "visibility", 'hidden');
				this._hiddenShapes.push(selection[i]);
			}
		}
	},
	
	_showDraggedShapes: function () {
		for (var i = 0, len_i = this._hiddenShapes.length; i < len_i; i++) {
			var selectable = this._hiddenShapes[i].policies.selectable;
			var shape = selectable.getTargetShape();
			if (shape) {
				delete this._hidden;
				if (selectable._selected) {
					selectable.showFeedback();
				}
				dojo.style(shape.rawNode, "visibility", '');
			}
		}
		delete this._hiddenShapes;
	},
	
	_saveToContainerIds: function () {
		this._savedContainerIds = [];
		for (var i = 0, len = this._toParentContainer.children.length; i < len; i++) {
			this._savedContainerIds[this._toParentContainer.children[i].id] = true;
		}
	},
	
	_selectToContainerIds: function () {
		var controller = this.host ? this.host.root.getSelectionController() : this._selectionController;
		if (controller) {
			for (var i = 0, len = this._toParentContainer.children.length; i < len; i++) {
				var child = this._toParentContainer.children[i];
				if (!child.policies.graphEdge && !(child.id in this._savedContainerIds)) {
					controller.select(child.__editController);
				}
			}
		}
		delete this._savedContainerIds;
	}

});