/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.diagrams.edit.tools.PushMoveableNeighbours");

dojo.require("rmpw.diagrams.render.util.Rectangle");
dojo.require("rmpw.diagrams.render.util.Point");
dojo.require("rmpw.diagrams.render.util.LineSeg");
dojo.require("rmpw.diagrams.layout.client.figures.MoveablePolygonAnchor");


dojo.declare("rmpw.diagrams.edit.tools.PushMoveableNeighboursScope", rmpw.common.policy.Policy, {
	singleton: "pushMoveableScope",
	
	constructor: function(editController) {
		this._pushedElements = {};
		this._pushQueue = {};
		var id = null;
		
		// process secondary push of floating labels after shape push is done
		var graphManager = editController.renderMapping.policies.graphElement.getGraphManager();
		
		this._connectContainerRefresh = dojo.connect(graphManager, "onRefreshComplete", dojo.hitch(this, function() {
			if (this._pushQueue.length > 0) {
				// save current push queue
				var pushQueue = {};
				for (id in this._pushQueue) {
					pushQueue[id] = this._pushQueue[id];
				}
				this._pushQueue = {};
				
				// enforce PushMoveableNeighbours policy on secondary elements
				for (id in pushQueue) {
					var graphEl = pushQueue[id];
					graphEl.__secondaryRefresh = true;
					graphEl.host.invalidate();
					graphManager.updateLocation(graphEl.host);
				}
			} else {
				// if done, no need to remember pushed elements
				this._pushedElements = {};
			}
			
		}));
	},
	
	destroy: function() {
		this.inherited(arguments);
		dojo.disconnect(this._connectContainerRefresh);
	},
	
	queuePush: function(graphEl) {
		if (this._pushedElements[graphEl.host.id] === undefined) {
			this._pushQueue[graphEl.host.id] = graphEl;
		}
	},
	
	pushNeighbours: function(graphEl, isAllowAlignment) {
		if (graphEl._unpositioned) {
			return;
		}
		
		this._pushedElements[graphEl.host.id] = graphEl;
		delete graphEl.__secondaryRefresh;
		
		var nodeMapping = graphEl.host;
		var graphManager = nodeMapping.policies.graphElement.getGraphManager();
		var buffer = graphManager.SHAPE_BUFFER;
		
		var types = [];
		if (graphEl.host.policies.graphNode) {
			types = ["graphNode"];
			if (isAllowAlignment) {
			    buffer = graphManager.SHAPE_BUFFER / 2;
			}
		}
		else if (graphEl.host.policies.graphFloatingLabel) {
			types = ["graphFloatingLabel"];
			buffer = 3; // minimize buffer for labels
		}
		
		var sourceAnchor = new rmpw.diagrams.layout.client.figures.MoveablePolygonAnchor(nodeMapping.policies.graphElement.getInternalGraphElement());
		var neighbours = graphManager.getNearMappings(nodeMapping, types, graphManager.INTERSECT_BOUNDS, (isAllowAlignment ? -5 : buffer));
		if (neighbours.length > 0) {
			dojo.forEach(neighbours, dojo.hitch(this, function(mapping) {
				if (nodeMapping.id!=mapping.id && this._pushedElements[mapping.id] === undefined) {
					var editController = mapping.__editController;
					if (editController) {
						if (editController.policies.moveable) {
							
							var targetAnchor = new rmpw.diagrams.layout.client.figures.MoveablePolygonAnchor(mapping.policies.graphElement.getInternalGraphElement());
							
							var sourceRef = sourceAnchor.getReferencePoint();
							var targetRef = targetAnchor.getReferencePoint();
							var sourcePoint = sourceAnchor.getLocation(targetRef);
							var targetPoint = targetAnchor.getLocation(sourceRef);
							
							// calculate the new target point taking the buffer into account
							var lineToSourceBorder = new rmpw.diagrams.render.util.LineSeg(sourceRef, sourcePoint);
							
							// if target is almost inside or inside the source, then calculate the buffer differently
							if (sourceRef.getDistance(targetRef) < sourceRef.getDistance(sourcePoint)) {
								lineToSourceBorder = new rmpw.diagrams.render.util.LineSeg(sourceRef, targetPoint);
								var dist = targetPoint.getDistance(sourcePoint);
								buffer = isAllowAlignment ? dist : buffer + dist;
							}
							else if (sourceRef.getDistance(sourcePoint) > sourceRef.getDistance(targetPoint) && isAllowAlignment) {
								// project the line to Source Border by the buffer or if allow alignment if less the buffer
								buffer = 0;
							}
							
							var bufferPoint = new rmpw.diagrams.render.util.Point();
							lineToSourceBorder.pointOn(lineToSourceBorder.length() + buffer, "ORIGIN", bufferPoint);
							
							// only move the target if it's less then the buffer distance
							if (sourceRef.getDistance(bufferPoint) > sourceRef.getDistance(targetPoint)) {
								editController.policies.moveable.doMove(bufferPoint.x - targetPoint.x, bufferPoint.y - targetPoint.y, false);
							}
						}
					}
				}
			}));
		}
	}
});

/**
 * Policy to push moveable neighbours away with a given buffer zone around this node
 * 
 * @author Steve Shaw
 */
dojo.declare("rmpw.diagrams.edit.tools.PushMoveableNeighbours", rmpw.common.policy.Policy, {
	singleton: "pushable",
	
	constructor: function(editController) {
		var gc = this.getGraphContainer();
		if (!gc.host.__editController.policies.pushMoveableScope) {
			gc.host.__editController.createPolicy(rmpw.diagrams.edit.tools.PushMoveableNeighboursScope);
		}
		this._graphElement = editController.renderMapping.policies.graphElement;
		this._connectRefresh = dojo.connect(this._graphElement, "refresh", this, "checkNeighbours");
		this._isAllowAlignment = false;
	},
	
	setIsAllowAlignment: function(isAllowAlignment) {
		this._isAllowAlignment = isAllowAlignment;
	},
	
	/**
	 * Enforces the policy.
	 */
	enforce: function() {
		if (!this._isAllowAlignment) {
			this.checkNeighbours();
		}
	},
	
	getGraphContainer: function() {
		return this.host.renderMapping.policies.graphElement.getGraphContainer();
	},
	
	destroy: function() {
		this.inherited(arguments);
		dojo.disconnect(this._connectRefresh);
	},
	
	checkNeighbours: function() {
		// only check neighbour position on explicit command execution.  Undo / Redo should be left alone.
		var cc = this.host.root.getCommandController();
		if (cc.isUndoRedoExecuting()) {
			return;
		}
			
		if (!this._isAllowAlignment) {
			if (!cc.isExecuting() && this.host.renderMapping.policies.graphNode) {
				return;
			}
		}
		
		var gc = this.getGraphContainer();
		
		var graphEl = this.host.renderMapping.policies.graphElement;
		if (this.host.renderMapping.policies.graphNode || graphEl.__secondaryRefresh) {
			gc.host.__editController.policies.pushMoveableScope.pushNeighbours(graphEl, this._isAllowAlignment);
		} else if (this.host.renderMapping.policies.graphFloatingLabel) { // do labels asynchronously
			gc.host.__editController.policies.pushMoveableScope.queuePush(graphEl);
		}
	}
});