/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010, 2012.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.diagrams.edit.tools.SelectionScope");

/**
 * Keeps track of which elements are selected on a diagram.
 * 
 * @author Marc Gobeil
 */
dojo.declare("rmpw.diagrams.edit.tools.SelectionScope", null, {
	singleton: "selectionScope",
	
	constructor: function() {
		this._selected = [];
		this._subscribeTabChange = dojo.subscribe("/TabbedEditor/onTabChange", this, "deselectAll");
	},
	
	destroy: function() {
		dojo.unsubscribe(this._subscribeTabChange);	
	},
	
	selectExclusive: function(editControllers, e) {
		editControllers = dojo.isArray(editControllers) ? editControllers : [editControllers];
		
		var newSelect = [];
		for (var i = 0, editController; editController = editControllers[i]; i++) {
			var index = this._getSelectionIndex(editController);
			if (index !== -1) {
				// if it was already selected, remove it from the list so it's not sent a deselection event
				this._selected.splice(index, 1);
			} 
			
			newSelect.push(editController);
		}
		
		// notify of deselection
		dojo.forEach(this._selected, dojo.hitch(this, function(editController) {
			this._fireDeselected(editController, e);
		}));
		
		// set the selection
		this._selected.length = 0;
		for (i = 0, editController= undefined; editController = editControllers[i]; i++) {
			this._selected.push(editController);
		}
		
		// notify of new selection
		// only notify the edit part if it wasn't already selected
		dojo.forEach(newSelect, dojo.hitch(this, function(editController) {
			this._fireSelected(editController, e);
		}));
				
		this.onSelectionChanged(e);
	},
	
	toggleSelect: function(editController) {
		if (editController.policies.selectable.isSelected()) {
			this.deselect(editController);
		}
		else {
			this.select(editController);
		}
	},
	
	select: function(editControllers, e) {
		// if just the diagram is selected, make this an exclusive selection
		if (this._selected[0] && this._selected[0].renderMapping.policies.graphContainer && 
				this._selected[0].renderMapping.policies.graphContainer.isTopLevel()) {
			this.selectExclusive(editControllers, e);
			return;
		}
		
		var changed = false;
		editControllers = dojo.isArray(editControllers) ? editControllers : [editControllers];
		for (var i = 0, editController; editController = editControllers[i]; i++) {
			var index = this._getSelectionIndex(editController);
			if(index === -1) {
				this._selected.push(editController);
				
				// send notifications
				this._fireSelected(editController, e);
				changed = true;
			}
		}
		if (changed) {
			this.onSelectionChanged(e);
		}
	},
	
	deselect: function(editControllers) {
		var changed = false;
		editControllers = dojo.isArray(editControllers) ? editControllers : [editControllers];
		for (var i = 0, editController; editController = editControllers[i]; i++) {
			var index = this._getSelectionIndex(editController);
			if(index !== -1) {
				var removed = this._selected.splice(index, 1);
				
				// send notifications
				this._fireDeselected(removed[0]);
				changed = true;
			}
		}
		if (changed) {
			this.onSelectionChanged();
		}
	},
	
	deselectAll: function() {
		dojo.forEach(this._selected, this._fireDeselected, this);
		this._selected.length = 0;
		this.onSelectionChanged();
	},
		
	getSelection: function() {
		return this._selected;
	},
	
	getPrimarySelection: function() {
		var primarySelection = null;
		if(this._selected.length > 0) {
			primarySelection = this._selected[0];
		}
		return primarySelection;
	},
	
	saveSelectionToMemento: function() {
        //returns a shallow copy of this._selected as a memento
        var _selected = dojo.map(this._selected, function(x) {return x;});
        return {
            getSavedSelection: function(){ return _selected; } 
        };
	},
	
	restoreSelectionFromMemento: function( memento ){
		this.deselectAll();
		this.select(memento.getSavedSelection());
	},
	
	onSelectionChanged: function(e) {
		// do nothing, for listeners
	},
	
	_getSelectionIndex: function(editController) {
		return dojo.indexOf(this._selected, editController);
	},
	
	_fireSelected: function(/* EditMapping */ editController, e) {
		var selectable = editController.policies.selectable;
		if(selectable) {
			selectable.onSelectionChanged(true, e);
		}
	},
	
	_fireDeselected: function(/* EditMapping */ editController, e) {
		var selectable = editController.policies.selectable;
		if(selectable) {
			selectable.onSelectionChanged(false, e);
		}
	}
});