/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.common.transform.Expression");
dojo.require("rmpw.common.transform.BoundVariable");

/**
 * Allows for evaluation of an expression based on trace variables.
 * 
 * Models the following situation in a QVT transform:
 * 
 * 		foo = :Bar {
 * 			property = traceVar1.concat(traceVar2) - 5
 * 		}
 * 
 * In this case, the expression is "traceVar1.concat(traceVar2) - 5".  Since expressions
 * are always defined "in-place" in QVT, they never need to be named.  In our
 * implementation, it's more convenient to name them so they can be defined in one place
 * and evaluated in multiple.  The previous block can then be thought of as:
 * 
 * 		exp1 = traceVar1.concat(traceVar2) - 5;
 *  	foo = :Bar {
 * 			property = exp1
 * 		}
 * 
 * To implement exp1 for use in incremental updates, we need to grab the current values
 * of each of the trace variables it depends on (traceVar1 and traceVar2), then encode the expression
 * as a javascript function that can be evaluated in the context of the model at the time of the
 * update.  This class implements exactly that, so that expression functions can be treated the same
 * as single valued trace variables, similar to how a QVT parser sees trace variables as simply
 * one of many subclasses of the more general Expression.
 * 
 * Example usage:
 * 
 * 		var exp1 = new rmpw.common.transform.Expression([traceVar1, traceVar2], function(traceVar1, traceVar2) {
 * 			return (traceVar1 + traceVar2) - 5;
 * 		});
 * 
 * Note:	as a consequence of its definition, part of Expression's contract is that it not evaluate
 * 			variables beyond the scope of the input domains of a QVT relation, implemented as this.sources
 * 			in the javascript Mappings. 
 * 
 * @author Steve Shaw / Marc Gobeil
 */
dojo.declare("rmpw.common.transform.Expression", rmpw.common.transform.BoundVariable, {
	
	/*
	 * Takes a list of traces that this expression is based on and allows them to be evaluated by the expressions
	 * @param traces the array of traces
	 * @param expression allows the trace data to be transformed to client data
	 * @param propertyName name of the property that is updated through the change set
	 * @param collection determines if this expression represents a collection
	 */
	constructor: function(traces, expression, propertyName, collection) {
		this._traces = traces;
		if (!expression) {
			// identity expression
			this._expression = function(value) { return value; };
		}
		else {
			this._expression = expression;
		}
		
		this._collection = collection;
		this.propertyName = propertyName;
	},
	
	/*
	 * Determines if this expression should re-synced with the runtime policies
	 */
	isSynced: function() {
		for(var i = 0, trace; trace = this._traces[i]; i++) {
			if (!trace.isSynced()) {
				return false;
			}
		}
		
		return true;
	},
	
	/*
	 * Sets the sync status of associated traces
	 */
	setSynced: function(bool) {
		for(var i = 0, trace; trace = this._traces[i]; i++) {
			trace.setSynced(bool);
		}
	},
	
	/*
	 * Determines if this variable is a collection
	 */
	isCollection: function() {
		return this._collection;
	},
	
	/**
	 * Returns the current value of the expression based on the current associated traces values
	 */
	get: function() {
		var values = dojo.map(this._traces, "return item.get();");
		return this._expression.apply(this, values);
	}
});