/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.common.transform.InvalidationController");
dojo.require("rmps.core.error");

/**
 * Manages general mapping invalidation
 * 
 * @author Steve Shaw
 */
dojo.declare("rmpw.common.transform.InvalidationController", null, {
	constructor: function(viewer) {
		this._nextID = 0;
		this._invalidIndex = {};
		this._invalidList = [];
	},
	
	destroy: function() {
	},
	
	getNextID: function() {
		return this._nextID++;
	},
	
	_addToInvalidMappings: function(mapping) {
		if (this.isInvalid(mapping)) {
			return false;
		} else {
			this._invalidIndex[mapping.id] = true;
			this._invalidList.push(mapping);
			return true;
		}
	},
	
	/*
	 * Checks to see if the given mapping is invalid and requires updating
	 */
	isInvalid: function(mapping) {
		return mapping.id in this._invalidIndex;
	},
	
	/*
	 * Checks to see if currently processing list
	 */
	isProcessingList: function() {
		return this._processingList;
	},
	
	/*
	 * Gets current list of invalid mappings
	 */
	getInvalidList: function() {
		return this._invalidList;
	},
	
	/* 
	 * Invalidate the diagram element and invokes an asynchronous method call through timeOut
	 * in order to do the validation in one call
	 */
	invalidate: function(mapping) {
		if (this._addToInvalidMappings(mapping)) {
			this._deferInvalidate();
		}
	},
	
	/*
	 * Cancels and invalidate request - i.e. mapping sleep
	 */
	cancelInvalidate: function(mapping) {
		delete this._invalidIndex[mapping.id];
		var idx = dojo.indexOf(this._invalidList, mapping);
		if (idx != -1) {
			this._invalidList.splice(idx, 1);
		}
	},
	
	_deferInvalidate: function() {
		if (this._callbackTimeout) {
			clearTimeout(this._callbackTimeout);
		}
		
		var controller = this;
		this._callbackTimeout = setTimeout( function() {
			controller.performInvalidation();
		}, 0);	
	},
	
	/*
	 * Performs the invalidation hook
	 */
	performInvalidation: function() {
		this._processingList = true;
		this.processInvalidMappings();
		this._processingList = false;
		this.onInvalidationComplete();
	},
	
	/*
	 * Performs the invalidation hook
	 */
	processInvalidMappings: function() {
		// clone invalid mapping list
		var invalidList = dojo.map(this._invalidList, this._identity),
		 	mapping;
		
		// clear, may receive new entries during the following iteration
		this._clear();
		
		// sort children first, then up to parents
		invalidList.sort(this._levelComparator);
		for(var i = 0; mapping = invalidList[i]; i++) {
			try {
				this.onInvalidate(mapping);
			} catch (e) {
			    var errorAction = "Invalidation of " + mapping.declaredClass;
				rmps.core.error.createError(e, errorAction).dispatch();
				console.error(errorAction + " failed: ", e);
			}
		}
	},
	
	_identity: function(item) {
		return item;
	},
	
	_levelComparator: function(a, b) {
		if (a.level == b.level) {
			return 0;
		}
		else if (b.level>a.level) {
			return 1;
		}
		else {
			return -1;
		}
	},
	
	/*
	 * Hook for completion of the validation process
	 */
	onInvalidationComplete: function() {
		
	},
	
	/*
	 * Hook for invalidation
	 */
	onInvalidate: function(mapping) {
		// make sure mapping is still valid
		if (mapping.policies) {
			mapping.applyModelChanges();
		}
	},
	
	_clear: function() {
		if (this._callbackTimeout) {
			clearTimeout(this._callbackTimeout);
		}
		this._invalidIndex = {};
		this._invalidList.length = 0;
	}
});