/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010, 2013.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.common.transform.Mapping");

dojo.require("rmpw.common.policy.Host");
dojo.require("rmpw.common.transform.Trace");
dojo.require("rmpw.common.transform.Expression");
dojo.require("rmpw.common.transform.MappingState");
dojo.require("rmpw.common.vcs.ChangeSet");
dojo.require("rmpw.common.util.IdentifierUtil");
dojo.require("rmpw.common.vcs.Change");

dojo.declare("rmpw.common.transform.Mapping", rmpw.common.policy.Host , /**
 * @author sshaw
 *
 */
{
		
	constructor: function(options) {
		// root elements of the semantic model
		this.sources = { __metadata: {} };
		// named elements of the view element trees
		this.template = {};
		// root elements of the the view element trees (one level below those passed in from the parent)
		this.rootShapes = [];
		
		this._options = options;
	},
	
	// required for offscreen
	idToSemanticMap: {},
	idUnresolved: {},
	
	/*
	 * Gets a map of semantic elements so that we can be sure that there is only a single 
	 * object for a particular semantic element.  This is stored relative to the repository so
	 * that if a repository is shared between editors, the data will match.
	 */
	getSemanticMap: function() {
		if (dojo.isOffscreen) {
			return this.idToSemanticMap;
		}
		if (!this.root.repository.__idToSemanticMap) {
			this.root.repository.__idToSemanticMap = {};
		}
		
		return this.root.repository.__idToSemanticMap;
	},
	
	fixSemanticMap: function(dmcodegen_defines, documentUri, repository) {
		for (var i=0, lg=dmcodegen_defines.length; i<lg; i++) {
			var subObj = dmcodegen_defines[i];
			if (subObj.rdf_about && !(subObj.rdf_about.indexOf(documentUri) > -1)) {
				// fix the identifier so that it's fully qualified with the document URI
				var old_id = subObj.rdf_about;
				subObj.rdf_about = rmpw.common.util.IdentifierUtil.normalizeURI(documentUri) + subObj.rdf_about;
				
				if (repository.__idToSemanticMap) {
					delete repository.__idToSemanticMap[old_id];
					repository.__idToSemanticMap[subObj.rdf_about] = subObj;
				}
			}
		}
	},
	
	/*
	 * Stores the map of unresolved elements - semantic references that don't have a mapping
	 */
	getUnresolved: function() {
		if (dojo.isOffscreen) {
			return this.idUnresolved;
		}
		if (!this.root.repository.__idUnresolved) {
			this.root.repository.__idUnresolved = {};
		}
		
		return this.root.repository.__idUnresolved;
	},
	
	/*
	 * Ensures that the source reference is a real resolved json object.
	 */
	getResolvedSemantic: function(modelEl, callback) {
		if (callback) {
			if (modelEl && modelEl.rdf_about) {
				var realSource = this.getSemanticMap()[modelEl.rdf_about];
				if (realSource) {
					callback(realSource, false);
				}
				else {
					var unresolvedEls = this.getUnresolved()[modelEl.rdf_about];
					if (!unresolvedEls) {
						unresolvedEls = [];
						this.getUnresolved()[modelEl.rdf_about] = unresolvedEls;
					}
					this.sleep();
					unresolvedEls.push(callback);
				}
			}
			else {
				callback(modelEl, false);
			}
		}
		else {
			if (modelEl && modelEl.rdf_about) {
				var realSource = this.getSemanticMap()[modelEl.rdf_about];
				return realSource ? realSource : modelEl;
			}
			else {
				return modelEl;
			}
		}
	},
	
	/*
	 * Back reference to the referred property that incarnated this mapping so that
	 * refresh of the owning collection can occur for undo of deletions
	 */
	setReferredProperty: function(referredPropertyName) {
		this._referredPropertyName = referredPropertyName;
	},
	
	getReferredProperty: function() {
		return this._referredPropertyName;
	},
	
	/**
	 * Called when the view mapping is attached to its tree, after source and target
	 * model root element setters have been called.
	 */
	setParent: function(parent) {
		this.inherited(arguments);

		// TODO: eliminate references to this.viewer, and then the property itself
		this.viewer = this.root;
		
		this.level = parent.level + 1;		
		this.id = this.root.getInvalidationController().getNextID();
		
		this.setAll(this._options);
		delete this._options;
		
		this.createTraces();
		this.createPolicies();
		
		if (!this.isSleeping()) {
			this.awaken();
		}
	},
	
	/*
	 * Checks if the mapping is in sleep mode and if so won't fully activate the mapping
	 */
	isSleeping: function() {
		return this._sleep;
	},
	
	/*
	 * Policies can call this if some information is incomplete or not sane in which case this 
	 * mapping won't be fully created and will be in an invisible "sleep" mode until awakened.
	 */
	sleep: function() {
		this.viewer.getInvalidationController().cancelInvalidate(this);
		this.detachListeners();
		this._sleep = true;
	},
	
	/*
	 * Wakes up the mapping and reactivates the policies
	 */
	awaken: function() {
		delete this._sleep;
		
		// set __mappings reverse references
		for(var sourceName in this.sources) {
			var source = this.sources[sourceName];
			
			if(source && dojo.isObject(source)) {
				if (source.__mappings) {
					source.__mappings.push(this);
				}
				else {
					source.__mappings = [this];
				}
				
				if (source.rdf_about) {
					var masterSource = this.getSemanticMap()[source.rdf_about];
					if (masterSource) {
						this.sources[sourceName] = masterSource;
						source = masterSource;
					}
					else {
						this.getSemanticMap()[source.rdf_about] = source;
					}
					
					var unresolvedEls = this.getUnresolved()[source.rdf_about];
					if (unresolvedEls) {
						delete this.getUnresolved()[source.rdf_about];
						
						for (var i=0, lg=unresolvedEls.length; i<lg; i++) {
							unresolvedEls[i](source, true);
						}
					}
				}
				
				if (source.__inactiveMappings) {
					for (var i=0, ms; ms=source.__inactiveMappings[i]; i++) {
						if (source.__inactiveMappings.length == 1) {
							delete source.__inactiveMappings;
						}
						else {
							source.__inactiveMappings.splice(i--, 1);
						}
						if (ms.restoreState(this) || !source.__inactiveMappings) {
							break;
						}
					}
				}
			}
			else {
				var key = source + "";
				key =  sourceName + ":" + key;
				
				// create map from source to mapping in the parent
				if (!parent.__sourceToMapping) {
					parent.__sourceToMapping = {};
				}
				parent.__sourceToMapping[key] = this;
			}
		}
		
		this.createFixedChildren();
		this.applyModelChanges();
		this.attachListeners();
		this.invalidate();
	},
	
	createPolicies: function() {		
    	// abstract method
	},	
	
	/**
	 * Create children that only depend on parameters that remain fixed for the lifetime of the parent
	 * mapping: source roots, template variables and constants
	 */
	createFixedChildren: function() {
		// abstract method
	},
		
	/*
	 * Connect to any relevant method calls to listen to incoming events.
	 */
	attachListeners: function() {
		if(this.listeners && this.listeners.length != 0) {
			console.debug("Exception: listeners already connected");
			return;
		}
		
		this.listeners = [];
		this.listeners.push(dojo.connect(this, "applyModelChanges", this, "enforcePolicies"));
	},
	
	detachListeners: function() {
		if(!this.listeners) {
			console.warn("Listeners already disconnected");
			return;
		}
	
		dojo.forEach(this.listeners, dojo.disconnect);
		delete this.listeners;
	},
	
	destroy: function() {
		if(!this.policies) {
			console.error("Render mapping was already deleted.");
			return;
		}
		
		// remove mapping state
		if (this.parent.removeMappingState) {
			this.parent.removeMappingState(this);
		}
		
		this.detachListeners();
		
		// remove __mapping reverse references
		if (this.sources) {
			var mappingState = new rmpw.common.transform.MappingState(this);
			
			for (var sourceName in this.sources) {
				if(sourceName.substr(0, 2) != '__') {
					var source = this.sources[sourceName];
					if(source && dojo.isObject(source)) {
						if (source.__inactiveMappings) {
							source.__inactiveMappings.push(mappingState);
						}
						else {
							source.__inactiveMappings = [mappingState];
						}
						
						var list = source.__mappings;
						var found = false;
						if(list) {
							if(list.length == 1 && list[0] == this) {
								delete source.__mappings;
								found = true;
							} else {
								var i = dojo.indexOf(list, this);
								if(i != -1) {
									list.splice(i, 1);
									found = true;
								}
							}
						}
						
						if(!found) {
							console.warn("Mapping backreference in ", sourceName, " not found for ", this.declaredClass, " during destroy");
						}
						if (source.rdf_about && !source.__mappings) {
							delete this.getSemanticMap()[source.rdf_about];
						}
					}
				}
			}
		}
		
		// store deleted source in case required to reference in some delete post processing
		// Specifically when removing edges as a side effect of node deletion, we need to find the
		// default source of the edge parent so that refresh can be done on undo - @see GraphAnchorable#destroy
		this.__deletedSource = this.getDefaultSource();
		delete this.sources;
		
		if (this.root.getInvalidationController) {
			this.root.getInvalidationController().cancelInvalidate(this);
		}
		
		// policies deleted here
		this.inherited(arguments);
		
		var rootShapes = this.rootShapes;
		for(var i = 0, shape; shape = rootShapes[i]; i++) {
			if (shape.getParent()) {
				shape.getParent().remove(shape);
			}
		}
		delete this.rootShapes;
		delete this.template;
	},
	
	invalidate: function() {
		this.viewer.getInvalidationController().invalidate(this);
	},
	
	applyModelChanges: function() {
		// process deletions, but leave additions to subclasses
		for(var traceName in this.getQueries()) {
			var trace = this.getQuery(traceName);
			if(trace.isCollection() && !trace.isSynced()) {
				if (trace.removed && trace.removed.length > 0) {
					this.processChildrenToRemove(trace, trace.removed);
				}
				else if (trace.removedReference && trace.removedReference.length > 0){
					this.processChildrenToRemoveReference(trace);
				}
			}
		}
		
		if (this._boundExpressions) {
			for(var i = 0, expressionSetter; expressionSetter = this._boundExpressions[i]; i++) {
				expressionSetter();
			}
		}
		
		// reset expressions after they are evaluated in case there are multiple expression which depend
		// on the same trace
		if (this.__resetExpressions) {
			for(var i = 0, expressionResetter; expressionResetter = this.__resetExpressions[i]; i++) {
				expressionResetter();
			}
			
			delete this.__resetExpressions;
		}
	},
		
	/*
	 * Utility method for setting an expression value into a policy method.
	 * This needs to be done through utility since the expression variable may not
	 * be named.
	 */
	bind: function(obj, setter, expression) {
		if (!this._boundExpressions) {
			this._boundExpressions = [];
		}
		
		this._boundExpressions.push(dojo.hitch(this, function() {
    		if (expression && !expression.isSynced()) {
    			obj[setter](expression.get());
    			
    			if (!this.__resetExpressions) {
    				this.__resetExpressions = [];
    			}
    			this.__resetExpressions.push(function() {
        			expression.reset();
    			});
    		}
		}));
	},
	
	/**
	 * overridden in generated domain implementations for property mappings
	 */ 
	createTraces: function() {
		this.traces = {};
		this.expressions = {};
	},
	
	/**
	 * @param {string} name of the trace variable
	 * for the rest, see Trace
	 */
	createTrace: function(name, a, b, c) {
		// TODO (MG): splice arguments instead?
		this.traces[name] = new rmpw.common.transform.Trace(this.sources, a, b, c);
		this.traces[name].__name = name;
	},
	
	/**
	 * Create an expression based on multiple trace variables
	 */
	createExpression: function(name, traceNames, impl, propertyName, collection) {
		var traces = dojo.map(traceNames, function(name) { return this.getQuery(name); }, this);
		this.expressions[name] = new rmpw.common.transform.Expression(traces, impl, propertyName, collection); 
	},
	
	/*
	 * Gets all queries associated with this mapping
	 */
	getQueries: function() {
		var queries = {};
		dojo.mixin(queries, this.traces);
		dojo.mixin(queries, this.expressions);
		return queries;
	},
	
	/*
	 * Given a name will find the associated query (typically a trace or expression)
	 * 
	 * TODO (MG): align the name of this function with QVT concepts
	 */
	getQuery: function(name) {
		return this.traces[name] || this.expressions[name];
	},
	
	getSourceByPath: function(path) {
		var getter = new Function("root", "return root." + path + ";");
		try {
			return getter(this.sources);
		} catch (e) {
			return null;
		}
	},

	getSourceOwnerByPath: function(path) {
		// TODO: can fail for array notation [], solution available in Trace.js
		var lastDot = path.lastIndexOf('.');
		if(lastDot != -1) {
			return this.getSourceByPath(path.substring(0, lastDot));
		} else {
			return null;
		}
	},

	/**
	 * Returns the first source reference, generally under the assumption
	 * there is only one.
	 * 
	 * TODO: add option to filter for type and flag to check/return
	 * 		whether there is more than one key in this.sources
	 *
	 * @returns 
	 */
	getDefaultSource: function() {
		var source = null;
		for(var src in this.sources) {
			if(src.substring(0, 2) != "__") { // ignore injected __graph source
				source = this.sources[src];
				break;
			}
		}
		return source;
	},

	/**
	 * Returns the source path names.
	 */
	getSourcePathNames: function() {
		var source = [];
		for(var src in this.sources) {
			if(src.substring(0, 2) != "__") { // ignore injected __graph source
				source.push(src);
			}
		}
		return source.join(",");
	},

	/**
	 * Checks to see if .
	 *
	 * @returns true if this mapping and mapping argument have the save source(s)
	 */
	hasSameModelSource: function(mapping) {
		for(var src in this.sources) {
			if(src.substring(0, 2) != "__") { // ignore injected __graph source
				if (this.sources[src] != mapping.sources[src]) {
					return false;
				}
			}
		}
		return true;
	},
	
	connectTraces: function(source, traces) {
		// validate trace
		for(var i = 0, traceName; traceName = traces[i]; i++) {
			if (this.getQuery(traceName) == null) {
				throw "Cannot find trace or expression named " + traceName;
			}
		}

		// connect trace callback
		var callback = dojo.hitch(this, function(e) {
			this._onTraceSourceChanged(traces, e);
		});
		
		if (!this.listeners) {
			this.listeners = [];
		}
		
		if (!dojo.isOffscreen) { // don't listen to events if rendering offscreen
			this.listeners.push(dojo.connect(source, "onChange", callback));	
		}
	},
	
	/**
	 * Invalidates a trace and the entire view mapping if a trace has changed, based
	 * on an event fired on the object that owns the property it pulls data from.
	 * 
	 * Usage:
	 * dojo.connect(<source object>, "onChange, dojo.hitch(this, function(e) {
	 * 		this.onTraceSourceChanged([<array of trace vars>], e);
	 * }));
	 * 
	 * @param traces list of trace variables potentially affected by the change event
	 * @param e the change event describing what changed on the object
	 * @returns
	 */
	_onTraceSourceChanged: function(traces, e) {
		var invalid = false;
		
		for(var i = 0, trace; trace = this.getQuery(traces[i]); i++) {
			if ((e.updated && trace.propertyName in e.updated) 
					|| (e.added && trace.propertyName in e.added)
					|| (e.deleted && trace.propertyName in e.deleted)) {
				trace.setSynced(false);
				invalid = true;
			}
			else if (e.pushed && trace.propertyName in e.pushed) {
				dojo.forEach(e.pushed[trace.propertyName], function(item) {
					trace.getAdded().push(item);
				});
				trace.setSynced(false);
				invalid = true;
			}
			else if (e.pushedReference && trace.propertyName in e.pushedReference) {
				dojo.forEach(e.pushedReference[trace.propertyName], function(item) {
					trace.getAdded().push(item);
				});
				trace.setSynced(false);
				invalid = true;
			}
			else if (e.popped && trace.propertyName in e.popped) {
				dojo.forEach(e.popped[trace.propertyName], function(item) {
					trace.removed.push(item);
				});
				trace.setSynced(false);
				invalid = true;
			}
			else if (e.poppedReference && trace.propertyName in e.poppedReference) {
				dojo.forEach(e.poppedReference[trace.propertyName], function(item) {
					trace.removedReference.push(item);
				});
				trace.setSynced(false);
				invalid = true;
			}
			else if (e.insertArray && trace.propertyName in e.insertArray) {
				dojo.forEach(e.insertArray[trace.propertyName], function(item) {
					trace.getAdded().push(item.value);
				});
				trace.setSynced(false);
				invalid = true;
			}
			else if (e.insertReferenceArray && trace.propertyName in e.insertReferenceArray) {
				dojo.forEach(e.insertReferenceArray[trace.propertyName], function(item) {
					trace.getAdded().push(item.value);
				});
				trace.setSynced(false);
				invalid = true;
			}
			else if (e.removeArray && trace.propertyName in e.removeArray) {
				dojo.forEach(e.removeArray[trace.propertyName], function(item) {
					trace.removed.push(item.value);
				});
				trace.setSynced(false);
				invalid = true;
			}
			else if (e.removeReferenceArray && trace.propertyName in e.removeReferenceArray) {
				dojo.forEach(e.removeReferenceArray[trace.propertyName], function(item) {
					trace.removedReference.push(item.value);
				});
				if (e.removeReferenceSourcePathNames[trace.propertyName]) {
					trace.removeReferenceSourcePathNames = e.removeReferenceSourcePathNames[trace.propertyName];
				}
				trace.setSynced(false);
				invalid = true;
			}
			// refresh
			if (e.pushedRefresh && trace.propertyName in e.pushedRefresh) {
				dojo.forEach(e.pushedRefresh[trace.propertyName], function(item) {
					trace.getAdded().push(item);
				});
				trace.setSynced(false);
				invalid = true;
			}
			
		}
		
		if (invalid) {
			this.invalidate();
		}
	},
	
	/*
	 * cache the index of the mapping source so that it can be quickly retrieved
	 */
	getMappingStateIndex: function() {
		if (this.parent) {
			var idx = dojo.indexOf(this.parent.children, this);
			if (idx != -1) {
				return idx;
			}
		}
		
		return 0;
	},
	
	getPolicyState: function() {
		var policyState = {};
		dojo.map(this.policies, function(policy) {
			dojo.mixin(policyState, policy.getState());
		});
		
		return policyState;
	},
	
	/*
	 * if any of the internal policy meta-data has changed - update the diagram mapping state
	 */
	updateMappingState: function() {
		if (!dojo.isOffscreen && !this.root.getCommandController().isUndoRedoExecuting()) {
			// update the mapping state if persisted
			var mappingState = this.getMappingState();
			if (mappingState) {
				var policyState = this.getPolicyState();
						
				if (this.mappingStateDifferent(dojo.fromJson(mappingState.dmdiagram_mappingStateValue), policyState) ) {
					var changeSet = this.root.repository.getLastChangeSet();
					var semanticChange = new rmpw.common.vcs.Change(mappingState);
					changeSet.add(semanticChange);
					semanticChange.set("dmdiagram_mappingStateValue", dojo.toJson(policyState));
					semanticChange.commitToTarget(mappingState);  // no events
				}
			}
		}
	},
	
	mappingStateDifferent: function(oldState, newState) {
		if (!oldState) {
			return true;
		}
		
		var newStateCount = 0;
		for(var prop in newState) {
			newProp = dojo.isObject(newState[prop]) ? dojo.toJson(newState[prop]) : newState[prop];
			oldProp = dojo.isObject(oldState[prop]) ? dojo.toJson(oldState[prop]) : oldState[prop];
			
			if (newProp != oldProp) {
				return true;
			}
			
			newStateCount++;
		}
		
		var oldStateCount = 0;
		for(prop in oldState) {
			oldStateCount++;
		}

		if (newStateCount != oldStateCount) {
			return true;
		}
		
		return false;
	},
	
	/*
	 * Checks to see if the mapping source is a mapping state (i.e. check for diagram)
	 */
	isSourceMappingState: function() {
		var mappingState = this.sources ? this.getDefaultSource() : null;
		if (mappingState && mappingState.dmdiagram___inView && this.parent === this.root) {
			return true;
		}
		
		return false;
	},
	
	getElementId: function(refEl) {
		var uri = refEl.rdf_about ? refEl.rdf_about : refEl.$ref;
		return rmpw.common.util.IdentifierUtil.getIdentifier(uri);
	},
	
	/*
	 * Finds the corresponding mapping state for this mapping
	 */
	getMappingState: function(forceCreate) {
		if (this._mappingState) {
			return this._mappingState;
		}
		
		var mappingState = null;
		if (this.isSourceMappingState()) {
			// if the source is the diagram, then return it
			return this.getDefaultSource();
		}
		else {
			// try to find it from the parent
			var parentMappingState = this.parent.getMappingState ? this.parent.getMappingState() : null;
			var mappingStateIndex = this.getMappingStateIndex();
			if (parentMappingState && mappingStateIndex !== undefined && mappingStateIndex !== null) {
				if (parentMappingState.dmdiagram_childMappingState && parentMappingState.dmdiagram_childMappingState.length > mappingStateIndex) {
					mappingState = parentMappingState.dmdiagram_childMappingState[mappingStateIndex];
					 
					 // make sure it matches the key values
					var mappingElement = this.getDefaultSource();
					if (mappingElement && this.getElementId(mappingState.dmdiagram_mappingStateElement) != this.getElementId(mappingElement)) {
						mappingState = null;
						 
						// if not do a linear search to find the first one that matches
						for (var i=0, lg=parentMappingState.dmdiagram_childMappingState.length; i<lg; i++) {
							if (this.getElementId(parentMappingState.dmdiagram_childMappingState[i].dmdiagram_mappingStateElement) == this.getElementId(mappingElement)) {
								mappingState = parentMappingState.dmdiagram_childMappingState[i];
								break;
							}
						}
					}
					
					this._mappingState = mappingState;
				}
			}
		}
		
		if (forceCreate && !mappingState && this.parent.addMappingState) {
			// add it if it doesn't exist
			this.parent.addMappingState(this);
			
			// try again - only allow one chance, don't allow recursion
			return this.getMappingState();
		}
		
		return mappingState;
	},
	
	/*
	 * adds a new mapping state if one doesn't exist or restores the state of a mapping based on an existing mapping state
	 */
	addMappingState: function(child) {
		// otherwise restore the old one
		var childMappingState = child.getMappingState();

		// restore state
		if (childMappingState && childMappingState.dmdiagram_mappingStateValue) {
			var mappingStateValue = dojo.fromJson(childMappingState.dmdiagram_mappingStateValue);
			for (var i=0, policy; policy=child.policies[i]; i++) {
				policy.setState(mappingStateValue);
			}
			
			child._connectPoliciesEnforced = dojo.connect(child, "onPoliciesEnforced", child, "updateMappingState");
		}
		else {
			var mappingState = this.getMappingState(true);
			if (mappingState) {
				if (!mappingState.dmdiagram_childMappingState) {
					mappingState.dmdiagram_childMappingState = [];
				}
				
				// add new mapping state to parent collection if it doesn't exist
				if (mappingState.dmdiagram_childMappingState) {
					if (!child._mappingState) {
						// create a new mapping state
						if (!dojo.isOffscreen && !this.root.getCommandController().isUndoRedoExecuting()) {
							var changeSet = this.root.repository.getLastChangeSet();
							var semanticChange = new rmpw.common.vcs.Change(mappingState);
							changeSet.add(semanticChange);
							var element = child.getDefaultSource();
							var parentUri = null;
							if (element.__parentUri && rmpw.common.util.IdentifierUtil.isIdentifierFixed(element.__parentUri)) {
								parentUri = element.__parentUri;
								delete element.__parentUri;
							}
							else {
								parentUri = mappingState.dmdiagram_mappingStateElement ? mappingState.dmdiagram_mappingStateElement.rdf_about : null;
							}
							var childMappingState = {
									rdf_about: rmpw.common.util.IdentifierUtil.guid(),
									dmdiagram_mappingStateElement: { rdf_about: rmpw.common.util.IdentifierUtil.getGUIDWithParent(parentUri, element.rdf_about) }
								};
							
							semanticChange.push("dmdiagram_childMappingState", childMappingState);
							
							var diagramModel = this.root.diagram.getDefaultSource();
							var diagramChange = new rmpw.common.vcs.Change(diagramModel);
							changeSet.add(diagramChange);
							diagramChange.push("dmcodegen_defines", childMappingState);
							
							semanticChange.commitToTarget(mappingState); // no events
							diagramChange.commitToTarget(diagramModel); // no events
							
							child._connectPoliciesEnforced = dojo.connect(child, "onPoliciesEnforced", child, "updateMappingState");
						}
					 }
				}
			}	
		}
	},
	
	/*
	 * removes a mapping state from it's parent
	 */
	removeMappingState: function(child) {
		// remove the corresponding mapping state
		if (!dojo.isOffscreen && !this.root.getCommandController().isUndoRedoExecuting()) {
			var mappingState = this.getMappingState();
			var diagramModel = this.root.diagram.getDefaultSource();
			var childMappingState = child.getMappingState();
			
			if (diagramModel && mappingState && childMappingState && mappingState.dmdiagram_childMappingState) {
				var changeSet = this.root.repository.getLastChangeSet();
				var semanticChange = new rmpw.common.vcs.Change(mappingState);
				changeSet.add(semanticChange);
				semanticChange.removeArray("dmdiagram_childMappingState", childMappingState);
				semanticChange.commitToTarget(mappingState); // no events
				
				var diagramChange = new rmpw.common.vcs.Change(diagramModel);
				changeSet.add(diagramChange);
				diagramChange.removeArray("dmcodegen_defines", childMappingState);
				
				diagramChange.commitToTarget(diagramModel); // no events
			}
		}
		
		if (child._connectPoliciesEnforced) {
			dojo.disconnect(child._connectPoliciesEnforced);
			delete child._connectPoliciesEnforced;
		}
	},
	
	addChild: function(child) {
		this.inherited(arguments);
		if (this.onChange) { // allow traces on children collection
			var e = { pushed: { children: [child] }};
			this.onChange(e);
		}
		
		this.addMappingState(child);
	},
	
	removeChild: function(child) {
		var removed = this.inherited(arguments);
		if (removed && this.onChange) { // allow traces on children collection
			var e = { popped: { children: [child] }};
			this.onChange(e);
		}
	},
	
	/*
	 * Deletes all associated mappings for the views being removed
	 */
	processChildrenToRemove: function(trace, views) {
		for(var i=0, lg=views.length; i<lg; i++) {
			if (dojo.isObject(views[i])) {
				if (views[i].__mappings) {
					var self = this;
					dojo.forEach(views[i].__mappings.slice(), function(mapping) {
						// only destroy this child if it was removed from this parent
						if (mapping.parent === self) {
							mapping.destroy();
						}
					});
				}
				// if __mappings is empty, mapping.destroy() deletes it on its way out
			}
			else {
				var key = views[i] + "";
				var props = trace.propertyName.split('_');
				key = props[props.length-1] + ":" + key;
				if (this.__sourceToMapping[key]) {
					this.__sourceToMapping[key].destroy();
					delete this.__sourceToMapping[key];
				}
			}
		}
	},
	
	/*
	 * Deletes the all mappings in the mapping array
	 */
	processMappingChildrenToRemove: function(mappings) {
		if (mappings) {
			dojo.forEach(mappings.slice(), function(mapping) {
				mapping.destroy();
			});
		}
	},

	/*
	 * Only removes associated mappings that are in the containment hierarchy of this mapping since
	 * the views are references (not composition in the model).
	 */
	processChildrenToRemoveReference: function(trace) {
		var views = trace.removedReference;
		var sourcePathNames = trace.removeReferenceSourcePathNames;
		for(var i=0, lg=views.length; i<lg; i++) {
			var mappings = views[i].__mappings;
			
			if (mappings) {
				var mappingsToDelete = [];
				var self = this;
				dojo.forEach(mappings, function(mapping) {
					// check that the mapping being deleted has this mapping in it's parent tree
					var parent = mapping.parent;
					while (parent != null && parent != self) {
						parent = parent.parent;
					}
					
					if (parent == self &&  (!sourcePathNames || sourcePathNames == mapping.getSourcePathNames())) {
						// delete this mapping
						mappingsToDelete.push(mapping);
					}
				});
				
				dojo.forEach(mappingsToDelete, function(mapping) {
					mappings.splice(dojo.indexOf(mappings, mapping), 1);
					mapping.destroy();
				});
				
				if (mappings.length == 0) {
					delete views[i].__mappings;
				}
			}
		}
	},
	
	getChildFromId: function(id) {
		// TODO: Should use an index instead of traversing the tree.
		if (id == this.id) {
			return this;
		}
		var child = null;
		if (this.children) {
			dojo.some(this.children, function(c) {
				if (c.id === id) {
					child = c;
					return true;
				}
				child = c.getChildFromId(id);
				if (child != null)
					return true;
				return false;
			});
		}
		return child;
	},
	
	isTypeOf: function(obj, type) {
		return this.root.commandPackage.isTypeOf(obj, type);
	},
	
	set: function(property, value) {
		// for now call legacy function naming
		this["set" + property.substring(0, 1).toUpperCase() + property.substring(1)](value);
	},
	
	setAll: function(properties) {
		var postSet = [];
		for(var key in properties) {
			try {
				this.set(key, properties[key]);
			}
			catch (e) {
				// try setting these afterwards in case uninitialized dependencies are causing exception
				postSet.push({key: key, value: properties[key]});
			}
		}
		
		for (var i=0, lg=postSet.length; i<lg; i++) {
			this.set(postSet[i].key, postSet[i].value);
		}
	}	
});