/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.common.transform.Trace");
dojo.require("rmpw.common.transform.BoundVariable");

/**
 * Acts as a wire from a source property that fires with its latest value when it changes.
 * 
 * @author Marc Gobeil
 */
dojo.declare("rmpw.common.transform.Trace", rmpw.common.transform.BoundVariable, {
	/**
	 * Supports two signatures:
	 * 
	 * @param {object} scope the root scope of the path
	 * @param {string} path to the source property, relative to the scope
	 * @param {boolean} collection true if the source property is a collection, default false
	 * 
	 * @param {object} scope the root scope of the path
	 * @param {function} function to retrieve source value, with scope as this
	 * @param {string} name of the property to be sensitive to changes on when listener is fired (usually last item in source model path)
	 * @param {boolean} collection true if the source property is a collection, default false
	 */
	constructor: function(root) {
		this.synced = false;
		this.root = root;
		var path = arguments[1];
		if(path.indexOf('[') != -1) {
			// replace array notation with dot notation to make things uniform for the following component split on '.'
			path = path.replace('[', '.').replace(']', '');
		}
		// do the compiled equivalent of eval("root." + path), and guard against npe's along the way
		this.propertyName = path.substring(path.lastIndexOf(".") + 1);
		this.path = path.split(".");
		this._collection = !!arguments[2];
	},
	
	/**
	 * Returns the current value of the property at the end of the path.
	 */
	get: function() {
		var path = this.path;
		var value = this.root;
		for(var i = 0, prop; prop = path[i]; i++) {
			var nextValue = value[prop];
			if(nextValue === undefined || nextValue === null) {
				value = null;
				break;
			}
			value = nextValue;
		}
		return value;
	},
	
	/**
	 * Get a reference to the object that directly owns the final property in the path.
	 */
	getOwner: function() {
		var path = this.path;
		var value = this.root;
		for(var i = 0, prop; prop = path[i]; i++) {
			if(value[prop] !== undefined) {
				if(path[i + 1] !== undefined) {
					value = value[prop];
				} else {
					break;
				}
			} else {
				value = null;
				break;
			}
		}
		return value;
	},
	
	/*
	 * Determines if this expression should re-synced with the runtime policies
	 */
	isSynced: function() {
		return this.synced;
	},
	
	/*
	 * Sets this trace as being synchronized with it's data or not.
	 */
	setSynced: function(bool) {
		this.synced = bool;
	},
	
	/*
	 * Determines if this variable is a collection
	 */
	isCollection: function() {
		return this._collection;
	}
});