/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2012, 2013.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
(function() {
dojo.provide("rmpw.forms.ui.ChangeMixin");

dojo.require("rmpw.common.vcs.Repository");
dojo.require("rmpw.common.vcs.ChangeSet");
dojo.require("rmpw.common.vcs.ChangeSetWrapper");
dojo.require("rmpw.common.util.ChangeUtil");
dojo.require("rmpw.text.internal.widgets.RichTextWidget");

/**
 * Used as a mixin for different classes to set the default change, and change set.
 * 
 */
dojo.declare("rmpw.forms.ui.ChangeMixin", null, {
	target: null,
	changeSet: null,
	repository: null,
	namespaces: null,
	onSaveCompleteListener: null,
	
	docBodyProp_inRichTextDoc: "dmxhtml_body",
	
	postCreate: function(){
		this.inherited(arguments);
		
		this.textWidgets = []; /*needed to track rich text editor (in rich text or other documents) in order to customize change creation*/
		this.textFeatureNames = [];/*feature names corresponding to widgets in textWidgets*/
		this.updatedTextWidgets = [];
		// find out rich text editor widget(s), if present, as we will need to access it later
		// for customizing change creation
		var control;
		for (var key in this.template) {
			control = this.template[key];			
			if (control instanceof rmpw.text.internal.widgets.RichTextWidget &&
					(dojo.indexOf(this.textWidgets, control) === -1)) {
				this.textWidgets.push(control);
				this.textFeatureNames.push(control.featureName);
			}			
		}
	},
	
	addTypes: function(types) {
		if(!types){
			return;
		}
		
		var target= this.getTarget();
		if(!target){
			return;
		}
		if(!dojo.isArray(types)){
			types= [types];
		}
		if(!target["rdf_type"]){
			target["rdf_type"]= [];
		}
		target["rdf_type"]= target["rdf_type"].concat(types);
	},
	
	getTypes: function(){
		var target= this.getTarget();
		if(!target){
			return [];
		};
		var types= target["rdf_type"];
		if(!types){
			return [];
		}
		return dojo.map(types, function(entry){return entry.rdf_about});
	},
	
	setNamespaces: function(namespaces){
		this.namespaces= namespaces;	
	},
	
	getNamespaces: function(){
		return this.namespaces;
	},
	
	setChangeDetails: function(target, repository, isDiagramPopup){
		this.setRepository(repository);
		this.setTarget(target, isDiagramPopup);
	},
		
	setRepository: function(repository){
		if(!repository){
			repository = new rmpw.common.vcs.Repository();
		}
		this.repository = repository;
	},
	
	getRepository: function(){
		return this.repository;
	},
	
	isImmediate: function() {
		if (this.immediate !== undefined) {
			return this.immediate;
		}
		
		if (this._resource) {
			this.immediate = this.target == this._resource.formObject;
		}
		
		return this.immediate;
	},
	
    createChange: function(property, value, doChange) {
	   	var change = this.createUncommittedChange(property, value, doChange);
	    	
	   	// Change is undefined in case no actual change is there.
	   	// Only commit change right away if this is the top level form.
	   	if(change && this.isImmediate()) {	
			this.getRepository().commit(this.changeSet);   		
	    }
    },
	
	createUncommittedChange: function(property, value, doChange) {
		var textIndex = dojo.indexOf(this.textFeatureNames, property);
		if (textIndex === -1) {
			// regular change processing for properties other than rich text
			return this.doCreateUncommittedChange(property, value);
		}
		// following is specific for rich text body (either in rich text or another document)
		if (doChange) {
			// means user hit "Save", or "OK" on popup, create and commit the change now			
			return this.doCreateUncommittedChange(property, value);	
		}
		// check if the change for this text feature has already been registered
		if (dojo.indexOf(this.updatedTextWidgets, textIndex) > -1) {
			// check if the resource is locked, if not (e.g. user explicitly unlocked it after making edits) 
			// we need to commit this change to ensure locking happens
			var store = this._resource.page.editor.editorInput.store;
			if (store._isPersisted && !store.locked) {
				return this.doCreateUncommittedChange(property, value);
			}
			return null;
		}
		
		// ignore the case when there is no actual change			
		if (property === this.docBodyProp_inRichTextDoc) {
			if (value !== null) {
				// setting initial value in rich text document
				return;
			}
		} else {
			var currentValue = this.target[property];
			if (rmpw.common.util.ChangeUtil.valueEqual(currentValue, value)) {
				// nothing changed
				return;
			}
		}
			 						
		// the first real update to the rich text body happened;
    	// don't create a change on every update as that would involve parsing the whole body (for the case of rich text document) 
		// and storing its whole json representation (happens on every key stroke); 
		// instead, just indicate that there is a change to the document body, 
		// the change set will be created when and if the user hits Save (see preProcessOnSave)
		this.updatedTextWidgets.push(textIndex);
		if (property !== this.docBodyProp_inRichTextDoc) {
			// when not in rich text doc, ensure that locking happens on first change 
			// (happens on commit to repository)
			return this.doCreateUncommittedChange(property, value);				
		}
		this._resource.page.editor.addDirty("model"); // would have happened on commit to repository				
	},
    
    doCreateUncommittedChange: function(property, value) {
		var currentValue = this.target[property];
		
		if(rmpw.common.util.ChangeUtil.valueEqual(currentValue, value)){
    		return;
    	}
		
		// special case for identifier - since it may be set to the short version in the form and we don't want
		// the form to get dirtied when it's fixed
		if (property == "rdf_about" && this.isIdentifierFixed) {
			return;
		}
		
    	var change= this.changeSet.createChange(this.target)
    	
    	//Required for cases such as when a text box accepted value is set to integer, as in this case,
    	// the value will be NaN. So, mak sure to detect this case, and also avoid setting values like
    	// empty string, false boolean, or number "0" to undefined.
    	// It is better to put this check in this method as it is used by all places, and having
    	// a possiblity to set a property to "NaN" will created a unprocessable json to be sent to the 
    	// server, and causes the whole update to fail. 
    	if(typeof value!=="boolean"&& !value&& !dojo.isString(value)&& value!== 0){
    		value= undefined;
    	}
    	
    	change.set(property, value);
    	
    	return change;
    },
	
	preProcessOnSave: function() {
		// this method is specific for rich text body, either in rich text 
		// or another document (in which case there could be more that one on the form);
		for (var i = 0; i < this.updatedTextWidgets.length; i++) {
			var featureName = this.textFeatureNames[this.updatedTextWidgets[i]];
			var textWidget = this.textWidgets[this.updatedTextWidgets[i]];
			this.createChange(
				featureName, 
				featureName !== this.docBodyProp_inRichTextDoc ? textWidget.get('value') : textWidget.doGetReferenceValueAttr(), 
				true);
		}
		this.updatedTextWidgets.length = 0;
	},	
    
    applyChange: function(invoker, immediate){
    	//No need to apply redundant check on the invoker being a function. This is the responsability for the callers
    	if(this.immediate || immediate){
    		var changeSet = new rmpw.common.vcs.ChangeSet();
	    	invoker(changeSet.createChange(this.target));
	    	this.getRepository().commit(changeSet);	
    	}else{
	    	invoker(this.changeSet.createChange(this.target));
    	}
    },
    
    getTarget: function(){
    	return this.target;
    },
    
    get: function(property){
    	return this.getTarget()[property];
    },
    
    setTarget: function(target){
    	if(this.target){
    		return; //one time setter
    	}
    	this.target = target || {};
		this._resetChangeSet();
		
		// ensure that my change set is cleared after successful update
		if (!this.isInPopup && !this.onSaveCompleteListener && this.target.rdf_about !== "http://new" 
				&& this._resource /*should always be when !this.isInPopup */) {
			this.onSaveCompleteListener = dojo.connect(this._resource.page.editor.editorInput.store, "onSaveComplete", dojo.hitch(this, this._resetChangeSet));
		}		
    },
	
	_resetChangeSet: function() {
    	if (!this.isImmediate()) {
    		this.changeSet = new rmpw.common.vcs.ChangeSet(this.target);
    	} else {
	    	this.changeSet = new rmpw.common.vcs.ChangeSetWrapper(this.target);
    	}
	}
});

})();