/*
 *+------------------------------------------------------------------------+
 *| Licensed Materials - Property of IBM                                   |
 *|                                                                        |
 *| com.ibm.xtools.rmps.web.analysis                                       |
 *|                                                                        |
 *| Copyright IBM Corporation 2011, 2012.  All Rights Reserved.            |
 *|                                                                        |
 *| US Government Users Restricted Rights - Use, duplication or disclosure |
 *| restricted by GSA ADP Schedule Contract with IBM Corp.                 |
 *+------------------------------------------------------------------------+
 */
(function() {
dojo.provide("dm.app.analysis.ui.internal.application.AnalysisPage");

/** The widget for creating the analysis page and registering
 *  its actions. All actions used in the Analysis capability
 *  MUST extend the DM editor framework and have both an
 *  editor widget class and an editor input class. By extending
 *  the DM framework, the menu will automatically show recently
 *  viewed resources (editors) and unsaved resources.
 *  
 *  This widget should be extended by the analysis applications
 *  in order to register their actions. Each such application 
 *  should use the dojo aspect advise mechanism to extend the 
 *  method for registering actions. 
 *  
 *  @author Kevin Cornell
 */

dojo.require("dm.app.analysis.ui.internal.application.AnalysisActions");
dojo.require("dm.app.analysis.ui.internal.application.AnalysisPageController");

dojo.require("rmps.ui.editor.AbstractPage");
dojo.require("rmps.ui.editor.EditorRegistry");
dojo.require("net.jazz.ajax.ui.PlatformUI");
dojo.require("dm.app.analysis.ui.internal.application.AnalysisPageEditorTracker");


var PAGE_REMOVAL_ID = "com.ibm.xtools.rmps.jfs.web.removeAnalysisPage";

var AnalysisActions = dm.app.analysis.ui.internal.application.AnalysisActions;
var PlatformUI      = net.jazz.ajax.ui.PlatformUI;

/* The analysis application (page) is designed for extensibility by other 
 * DM applications, such as the impact analysis and reports. The extension
 * points allow addition actions, page contents, side navigation sections, etc.
 */
dojo.declare("dm.app.analysis.ui.internal.application.AnalysisPage", rmps.ui.editor.AbstractPage, {

	"class": "dm-app-analysis",
	
	// private properties
	_actionRegistry: null,
	_mainPageController: null,
	_section: null,
	
	createPageContents: function() {
		this.inherited(arguments);
	},
		
	/** Return the name of the analysis section being
	 *  configured. The return value will be one of the
	 *  constants:  AnalysisActions.SECTION_*.
	 */
	getCurrentSection: function() {
		return this._section;
	},
	
	/** This method is provided so that the analysis applications
	 *  can register their actions. This method will be called
	 *  multiple times, once for each analysis section so that the
	 *  order of the extension data is deterministic. The section
	 *  name is set before each call and each application extending
	 *  this page must only do its registration when the current
	 *  section corresponds to their section (e.g., getCurrentSection()).
	 *  
	 *  The applications should extend this method using the
	 *  dojo aspect advise mechanism. In the application extension
	 *  class (e.g., MyPageInitializer), add the method declaration
	 *  as follows:
	 *  
	 *  dojox.lang.aspect.advise(dm.app.analysis.ui.internal.application.AnalysisPage, 
	 *						"registerExtendedActions", 
	 *      {
	 *      after: function() {
	 *          var page = dojox.lang.aspect.getContext().instance;
	 *          if (page.getCurrentSection() === AnalysisActions.SECTION_MySection) {
	 *             var controller = new MyController(page);
	 *             page.registerActionEditor("actionName", controller, 
	 *                 "path.of.class.MyEditor",
	 *                 "path.of.class.MyEditorInput");
	 *          }
	 *      }
	 *  });
	 *
	 *  In order to connect the application extension so that it can be called,
	 *  the following extension should be defined in the plugin.xml of the 
	 *  analysis application.
	 *  
	 *  <extension point="com.ibm.xtools.rmps.web.app.webUiInitializers">
	 *	  <initializer jsModule="path.of.class.MyPageInitializer">
	 *		<resourceBinding jsModule="dm.app.analysis.ui.internal.application.AnalysisPage" />
	 *	  </initializer>
	 *  </extension>
	 *
	 *  For each action defined by your application extension, you must add an entry
	 *  to the cix.rmps.web.analysis/plugin.xml file in the <page> target as follows:
	 *      <action id="dm.app.analysis.sectionName_actionName"/>
	 *  
	 *  For example, the reports action to create a report would be defined as:
	 *      <action id="dm.app.analysis.reports_createReport"/>
	 */
	registerExtendedActions: function(){ /* extensions must implement this method */ },
	
	
	/** Register the default analysis page actions and all
	 *  of the extensions to this analysis page. All actions
	 *  must be registered with the corresponding action name,
	 *  controller, the path of the editor class and the path
	 *  of the editor input class.
	 *  
	 *  Each extension will be called multiple times and must
	 *  check the current section to determine if they should
	 *  contribute their actions.
	 */
	subscribeActions: function() {
		// Get the action registry.
		this._actionRegistry = PlatformUI.getWorkbench().getActionRegistry();

		// First register the common actions for analysis
		this._mainPageController = new dm.app.analysis.ui.internal.application.AnalysisPageController(this);
		
		// Register the default action to open the welcome page.
		this.registerActionEditor( 
				AnalysisActions.ACTION_DEFAULT, 
				this._mainPageController,
				"dm.app.analysis.ui.internal.pages.AnalysisWelcome",
				"dm.app.analysis.ui.internal.pages.AnalysisWelcomeInput");
		this.registerActionEditor( 
				AnalysisActions.ACTION_WELCOME, 
				this._mainPageController,
				"dm.app.analysis.ui.internal.pages.AnalysisWelcome",
				"dm.app.analysis.ui.internal.pages.AnalysisWelcomeInput");
		
		// Register the actions for the Reports application
		this._section = AnalysisActions.SECTION_REPORTS;
		this.registerExtendedActions();
		
		// Register the actions for the Traceability application
		this._section = AnalysisActions.SECTION_TRACEABILITY;
		this.registerExtendedActions();
		
		// Register the actions for the Impact Analysis application
		this._section = AnalysisActions.SECTION_IMPACTANALYSIS;
		this.registerExtendedActions();
		
		// Register the actions for the Validation application
		this._section = AnalysisActions.SECTION_VALIDATION;
		this.registerExtendedActions();
		
		this._section = null;
	},
		
	
	/** Register one action with its corresponding
	 *  controller, editor class and editore input
	 *  class.
	 *  
	 *  @param actionName the action name with no prefix
	 *  @param controller contains a method to handle the action
	 *  @param editorClass the path of the editor js class
	 *  @param editorInputClass the path of the editor input js class
	 */
	registerActionEditor: function(actionName, controller, editorClass, editorInputClass) {
		// If the controller has a method for this action
		var method = controller[actionName];
		if (method) {
			// Register the editor class against its corresponding editor input class
			rmps.ui.editor.EditorRegistry.register(editorClass, function(editorInput) {
				return editorInput.declaredClass === editorInputClass;
			});
			
			// Register the editor input class against the qualified action
			var action = AnalysisActions.ACTION_PREFIX + actionName;
			rmps.ui.editor.EditorRegistry.registerEditorInput(editorInputClass, action); 

			// Register the action for this page
			this._actionRegistry.registerAction(action, controller, actionName);
		}
	},

	
	getPageRemovalId: function(){
		return PAGE_REMOVAL_ID;
	},
		
	/** Return the page ID to be used as the prefix for registered actions. */
	getPageId: function(){
		return AnalysisActions.PAGE_ID;
	},
	
	/** Return the action URI for the default action ... the Analysis welcome page. */
	getDefaultUri: function(){
		return AnalysisActions.getActionUri(AnalysisActions.ACTION_WELCOME);
	},
	
	getEditorTrackerClass: function () {
		// summary:
		//	Override of PageInitializer method, to create an editor tracker class with an MRU of 10
		return dm.app.analysis.ui.internal.application.AnalysisPageEditorTracker;
	}
	
});


}());
