/*
 *+------------------------------------------------------------------------+
 *| Licensed Materials - Property of IBM                                   |
 *|                                                                        |
 *| com.ibm.xtools.rmps.web.reports                                        |
 *|                                                                        |
 *| Copyright IBM Corporation 2012.  All Rights Reserved.                  |
 *|                                                                        |
 *| US Government Users Restricted Rights - Use, duplication or disclosure |
 *| restricted by GSA ADP Schedule Contract with IBM Corp.                 |
 *+------------------------------------------------------------------------+
 */

(function() {
dojo.provide("dm.app.reports.ui.internal.pages.CreateReport");

/** Editor for creating a report configurations. Report configurations can
 *  either be personal or shared. Each report has a description, a parent
 *  folder and an associated report template. The folder determines whether
 *  the report is private or shared (folders are private or shared) because
 *  the JFS reporting framework does not save the "shared" field for a report.
 *  
 *  After a report is created and assigned a folder and template, the template
 *  assignment cannot be changed. If a folder in the parent hierarchy is
 *  deleted or the associated template is deleted, this report is also
 *  automatically deleted.
 *  
 *  @author Kevin Cornell
 */

dojo.require("rmps.app.ui.internal.editors.BaseEditor");
dojo.require("dm.app.reports.ui.internal.pages.ReportBrowserContent");
dojo.require("dm.app.reports.ui.internal.pages.TemplateBrowserContent");

dojo.require("dm.app.reports.ui.internal.application.ReportActions");
dojo.require("dm.app.reports.ui.internal.util.ReportServiceClient");
dojo.require("dm.app.reports.ui.internal.util.ReportUtil");

dojo.require("com.ibm.team.repository.web.transport.ServiceResponseHandler");

dojo.require("jazz.ui.MessageBox");

dojo.require("dojo.string");
dojo.require("dojo.i18n");
dojo.requireLocalization("dm.app.reports.ui.internal", "ReportMessages");
var messages  = dojo.i18n.getLocalization("dm.app.reports.ui.internal", "ReportMessages");

var ReportActions = dm.app.reports.ui.internal.application.ReportActions;
var ReportServiceClient = dm.app.reports.ui.internal.util.ReportServiceClient;
var ReportUtil = dm.app.reports.ui.internal.util.ReportUtil;
var ServiceResponseHandler = com.ibm.team.repository.web.transport.ServiceResponseHandler;


// Define the editor help UA key
var PAGE_UAKEY = "com.ibm.xtools.rmps.web.reports.page_CreateReport001";

//Define the fields of this editor that can be modified.
var DIRTY_GROUP_ID       = "dm.app.reports.CreateReport.dirty";
var FIELD_name           = "name";
var FIELD_description    = "description";
var FIELD_processArea    = "processAreaUUID";
var FIELD_folderUUID     = "folderUUID";	// selected folder determines private or shared report
var FIELD_reportUUID     = "reportUUID";	// selected report template

// Define the fields of the report to copy during initialization
var COPY_FIELDS = [
	FIELD_name, 
    FIELD_description, 
    FIELD_processArea, 
    FIELD_folderUUID,
    FIELD_reportUUID ];


dojo.declare("dm.app.reports.ui.internal.pages.CreateReport", rmps.app.ui.internal.editors.BaseEditor, {
	
	"class": "reports-CreateReport",
	
	// base editor properties
	isRefreshable: false,
	isReadOnly: false,
	
	// Keep an initial copy of the report data and a new copy
	_queryUUID: null,
	_oldReport: null,
	_newReport: null,
	
	// Define properties used by this editor
	_pageContent: null,
	
	createContentWidget: function() {
		// Specify the help for this editor
		this.uaKey = PAGE_UAKEY;

		// When this editor is finally open, do a resize to fix up the two browser sections
		var pageHelper = rmps.ui.editor.PageHelper.getInstance(ReportActions.PAGE_ID);
		dojo.connect(pageHelper, "onEditorOpened", dojo.hitch(this, function(editor) {
			if (editor === this) {
				// Wait for half a second until the editor opens properly
				// otherwise the resize does not work properly.
				setTimeout(dojo.hitch(this, function() {
					var size = dojo.position( this.containerNode );
					this.resize(size);
				}), 500);
			}
		}));			
		
		// Create the contents for the browser
		this._pageContent = new dm.app.reports.ui.internal.pages.CreateReport._Content({editor: this});
		return this._pageContent;
	},
	
	/** Add the cancel button. Save button is added because "isReadOnly" is false.
	 * 
	 * @param toolbar: the base editor page toolbar.
	 */
	addButtons: function(toolbar) {
		toolbar.addWidget(new rmps.ui.SimpleButton({
			label: messages.cancel,
			onClick: dojo.hitch(this, this._handleCancel)
		}));
		
		this.inherited(arguments);
	},

	/** When the editor is activated, initialize the corresponding
	 *  report data and update the editor controls.
	 */
	activate: function() {
		this.inherited(arguments);
		
		// If the editor is being opened for the first time
		// for this report
		if (!this.isDirty()) {
			// Initialize the report data
			this._oldReport = {};
			this._oldReport[FIELD_name]        = "";
			this._oldReport[FIELD_description] = "";
			this._oldReport[FIELD_processArea] = ReportUtil.getProjectId();
			this._oldReport[FIELD_folderUUID]  = null;
			this._oldReport[FIELD_reportUUID]  = null;

			// Create a copy of the template for modifications
			// Check explicitly for defined fields or else "false" boolean
			// fields will not get copied.
			this._newReport = {};
			for (var index = 0, len = COPY_FIELDS.length; index < len; ++index) {
				var fieldName = COPY_FIELDS[index];
				if (this._oldReport[fieldName] !== undefined) {
					this._newReport[fieldName] = this._oldReport[fieldName];
				}
			}
			
			// Update the page UI with the existing report data
			this._pageContent.initReportFields(this._oldReport);
		}		
	},

	/** Update the report (query) field with the new value.
	 *  If the new value differs from the current value then
	 *  update the dirty indicator for this field.
	 *  
	 *  @param fieldName the name of the template field to update
	 *  @param newValue  the new value for the template field
	 */
	_updateReportField: function(fieldName, newValue) {
		// If the report has been created or initialized
		if (this._oldReport) {
			// Get the current value of the corresponding field
			var oldValue = "";
			if (this._oldReport[fieldName] !== null) {
				oldValue = this._oldReport[fieldName];
			}
			
			// Determine if the current and new values are different
			var isEqual = oldValue === newValue;

			// If values are equal, clear the field's dirty indicator
			// else set the field's dirty indicator
			if (isEqual) {
				this.removeDirty(fieldName, DIRTY_GROUP_ID);
			} else {
				this.addDirty(fieldName, DIRTY_GROUP_ID);
			}
			
			// Update the new copy of the template.
			this._newReport[fieldName] = newValue;
			
			// Enable the save button only if the editor
			// is dirty and the name, folder and template
			// are all defined.
			this.saveButton.setEnabled(this.isDirty() &&
					this._newReport[FIELD_name] &&
					this._newReport[FIELD_folderUUID] &&
					this._newReport[FIELD_reportUUID]);
		}
	},
	
	
	/** Perform the saving of the report (query).
	 * 
	 */
	doSave: function(taskMonitor) {
		this.inherited(arguments);
		
		// At this point the modified report should contain
		// all of the data needed to create a new report.
		var serviceObject = {
			self: this,
			success: function(reportId) {
				this.self._saveComplete(reportId);
			},
					
			failure: function(error) {
				this.self.serviceError(messages.errorCreatingReport, error);
			}
		};
		var srh = new ServiceResponseHandler(serviceObject, "success", "failure");
		ReportServiceClient.postSaveQueryDescriptor(srh, this._newReport);
		
		// The TaskMonitor cannot be used to handled the deferred work from the
		// request because it overwrites the specific error message & details
		// with a generic "Bad Request" message that means nothing.
	},
	
	/** The saving of the report template was successful.
	 * 
	 * @param reportId the value returned from the POST request
	 * 
	 */
	_saveComplete: function(reportId) {
		// Remove the dirty indicator
		this.removeDirty(null, DIRTY_GROUP_ID);
		
		// Get the ID of the report just saved
		this._queryUUID = reportId;
		
		// Close this editor and remove it from the recently and unsaved editor lists.
		var pageHelper = rmps.ui.editor.PageHelper.getInstance(ReportActions.PAGE_ID);
		pageHelper.closeEditor(this, true, true);
		
		// Force the report editor to be opened for the report just created.
		// Also define the "created" flag to indicate an editor replacement.
		var name = this._newReport[FIELD_name];
		var href = ReportActions.getActionUri(ReportActions.ACTION_EDIT_REPORT) +
			"&queryUUID=" + this._queryUUID + 
			"&name=" + encodeURIComponent(name) +
			"&created=true";
		window.location.href = href;
	},

	/** When Cancel is clicked in the create report editor, go
	 *  back to the previous page.
	 */
	_handleCancel: function() {
		// Remove the dirty indicator
		this.removeDirty(null, DIRTY_GROUP_ID);
		
		// Close this editor and remove it from the recently and unsaved editor lists.
		var pageHelper = rmps.ui.editor.PageHelper.getInstance(ReportActions.PAGE_ID);
		pageHelper.closeEditor(this, true, true);

		window.history.back();
	},
		
	/** Report an error from a reporting service request.
	 * 
	 * @param errorMessage the main error message
	 * @param error the service error (details for error message)
	 */
	serviceError: function(errorMessage, error) {
		var details = null;
		if (error.message) {
			details = [ {message: error.message} ];
		} else if (error.errorClass) {
			details = [ {message: error.errorClass} ];
		}
		this.showMessage({
			type: jazz.ui.MessageBox.STATUS_TYPES.ERROR,
			summary: errorMessage,
			details: details,
			closable: true
		});
	}


});

/*----------------------------------------------------------------------*/

/** Define the content for the report template editor.
 * 
 */
dojo.declare("dm.app.reports.ui.internal.pages.CreateReport._Content", [dijit._Widget, dijit._Templated], {

	"class": "createReport",

	messages: messages, 
	
	widgetsInTemplate: true,
	templateString: 
		'<div>' +
			// Main section: new report name and description
			'<div class="mainSection" dojoAttachPoint="_mainSection" >' +
				'<div class="reportNameRow">' +
					'<div class="reportNameLabel">${messages.reportName}' + ReportUtil.MANDATORY_HTML + '</div>' +
					'<div class="reportNameContainer">' + 
						'<input class="nameInput" dojoAttachPoint="_nameInput" />' +
					'</div>' +
				'</div>' +
				'<div class="descriptionLabel">${messages.reportDescription}</div>' +
				'<div class="descriptionContainer">' +
					'<textarea class="descriptionInput" dojoAttachPoint="_descriptionInput" rows="5" ></textarea>' +
				'</div>' +
			'</div>' +

			// Folders section: list folders and existing reports
			'<div class="foldersSection" dojoAttachPoint="_foldersSection">' +
				'<div class="foldersHeading">${messages.foldersHeading}' + ReportUtil.MANDATORY_HTML + '</div>' +
				'<div class="foldersInfo">${messages.foldersInfo}</div>' +
				'<div class="foldersContainer" dojoAttachPoint="_foldersContainer">' +
					'<div class="reportBrowserHook" dojoAttachPoint="_reportBrowserHook"></div>' +
				'</div>' +
			'</div>' +

			// Templates section: list existing templates
			'<div class="templatesSection" dojoAttachPoint="_templatesSection">' +
				'<div class="templatesHeading">${messages.templatesHeading}' + ReportUtil.MANDATORY_HTML + '</div>' +
				'<div class="templatesInfo">${messages.templatesInfo}</div>' +
				'<div class="templatesContainer" dojoAttachPoint="_templatesContainer">' +
					'<div class="templateBrowserHook" dojoAttachPoint="_templateBrowserHook"></div>' +
				'</div>' +
			'</div>' +
		'</div>',
	
	// init properties
	editor: null,
	
	// Widgets for the report browser and template browser
	_reportBrowser: null,
	_templateBrowser: null,
	
	// Remember the current folder and template selected
	_folder: null,
	_template: null,
	
	constructor: function(args) {
		this.messages = messages;
	},

	
	postCreate: function() {
		this.inherited(arguments);
		
		// Hook in the Report Browser content to select a folder
		this._reportBrowser = new dm.app.reports.ui.internal.pages.ReportBrowserContent({editor: this.editor, inBrowser: false });
		this._reportBrowserHook.appendChild(this._reportBrowser.domNode);
		
		// Hook in the Template Browser content to select a report template
		this._templateBrowser = new dm.app.reports.ui.internal.pages.TemplateBrowserContent({editor: this.editor, inBrowser: false });
		this._templateBrowserHook.appendChild(this._templateBrowser.domNode);
		
		// Hook in methods to handle changes in various input fields
		if (!this.editor.isReadOnly) {
			// Report name
			ReportUtil.connectTextInput(this._nameInput, this, function() {
				var newValue = dojo.trim(this._nameInput.value);
				this.editor._updateReportField(FIELD_name, newValue);
			});

			// Report description
			ReportUtil.connectTextInput(this._descriptionInput, this, function() {
				var newValue = this._descriptionInput.value;
				this.editor._updateReportField(FIELD_description, newValue);
			});

			// Folder selected in the report browser
			dojo.connect(this._reportBrowser, "selectFolder", this, function(folder) {
				this._folder = folder;
				if (this._isTemplateFolderValid()) {
					var newValue = folder ? folder.folderUUID : null;
					this.editor._updateReportField(FIELD_folderUUID, newValue);
					if (this._template) {
						this.editor._updateReportField(FIELD_reportUUID, this._template.reportUUID);
					}
				} else {
					this.editor._updateReportField(FIELD_folderUUID, null);		// disable the save button			
				}
			});

			// Template selected in the template browser
			dojo.connect(this._templateBrowser, "selectTemplate", this, function(template) {
				this._template = template;
				if (this._isTemplateFolderValid()) {
					var newValue = template ? template.reportUUID : null;
					this.editor._updateReportField(FIELD_reportUUID, newValue);
					if (this._folder) {
						this.editor._updateReportField(FIELD_folderUUID, this._folder.folderUUID);
					}

				} else {
					this.editor._updateReportField(FIELD_reportUUID, null);		// disable the save button	
				}
			});
		}
	},
	
	/** Update the fields on the page based on the given
	 *  report.
	 *  
	 *  @param report the report data
	 */
	initReportFields: function(report) {
		
		// Set the report a name and description
		this._nameInput.value        = report[FIELD_name];
		this._descriptionInput.value = report[FIELD_description];
	},
	
	/** Determine if the template and folder combination
	 *  is valid. A report with a private template cannot
	 *  be created in a shared folder or else another
	 *  user will not be able to run the report.
	 */
	_isTemplateFolderValid: function() {
		var isValid = true;
		if (this._folder && this._template && 
			this._folder.shared && !this._template.shared) {
			isValid = false;
			alert(messages.templateFolderConflict);
		}
		return isValid;
	},
	
	/** Handle the resizing of the report create window.
	 *  The report browser and template browser will
	 *  perform the necessary resizing to their content. 
	 */
	resize: function(size) {
		this.inherited(arguments);
		if (!size) {
			return;
		}
		if (size.h) {
			var mainSectionPos = dojo.position(this._mainSection);
			
			// Resize the report browser content
			var reportBrowserPos = dojo.position(this._foldersContainer);
			var newHeight = size.h - reportBrowserPos.y + mainSectionPos.y - 30; // allow for accumulated top margins
			if (newHeight > 200) {
				this._reportBrowser.resize({h: newHeight});
			}

			// Resize the template browser content
			var templateBrowserPos = dojo.position(this._templatesContainer);
			newHeight = size.h - templateBrowserPos.y + mainSectionPos.y - 30; // allow for accumulated top margins
			if (newHeight > 200) {
				this._templateBrowser.resize({h: newHeight});
			}
		}
	}


});


})();
