/*
 *+------------------------------------------------------------------------+
 *| Licensed Materials - Property of IBM                                   |
 *|                                                                        |
 *| com.ibm.xtools.rmps.web.reports                                        |
 *|                                                                        |
 *| Copyright IBM Corporation 2011, 2012.  All Rights Reserved.            |
 *|                                                                        |
 *| US Government Users Restricted Rights - Use, duplication or disclosure |
 *| restricted by GSA ADP Schedule Contract with IBM Corp.                 |
 *+------------------------------------------------------------------------+
 */

(function() {
dojo.provide("dm.app.reports.ui.internal.pages.ReportBrowserItem");

/** Define the widget that displays items, reports and folders,
 *  in the tree control of the report browser (ReportBrowserContent).
 *  Each tree item has a corresponding reporting resource and displays
 *  an icon, the resource name and its description. Each item also 
 *  contains the tree "twistie" for expand/collapse and a toolbar with
 *  action buttons.
 *  
 *  @author Kevin Cornell
 */

dojo.require("dm.app.reports.ui.internal.util.ReportUtil");
dojo.require("dm.app.reports.ui.internal.util.Icons");

dojo.require("jazz.ui.tree.AbstractItem");
dojo.require("jazz.util.html");

dojo.require("rmps.ui.util.ToolbarFocusUtil");
var ToolbarFocusUtil = rmps.ui.util.ToolbarFocusUtil;

var ReportUtil = dm.app.reports.ui.internal.util.ReportUtil;
var Icons = dm.app.reports.ui.internal.util.Icons;
var HTML = jazz.util.html;

/** Define the widget for items in the folder/reports tree */
dojo.declare("dm.app.reports.ui.internal.pages.ReportBrowserItem", [jazz.ui.tree.AbstractItem, dijit._Templated], {

	// Constructor parameters used to configure the item (must be passed)
	resource: null,
	isFolder: true,
	toolbar:  null,
	labelAction: null,
	
	// Private controls within the item
	_table: null,
	_row: null,
	_label: null,
	_description: null,
	_actions: null,
	
	// Remember if this item is selected
	_isSelected: false,
	
	templateString: "<div dojoAttachPoint='_treeItem'  class='treeItem'></div>",
		
	postCreate: function() {
		this.inherited(arguments);
		
		// Create a table with a single row for the item
		this._table = document.createElement("table");
		this._treeItem.appendChild(this._table);

		// Create a row with 3 columns: label (with icon), description and itemActions
		this._row = this._table.insertRow(0);
		dojo.style(this._row, "backgroundColor", ReportUtil.COLOR_NORMAL );
		this._isSelected = false;
		
		// Column 1: label with appropriate icon (value set in update())
		this._label = document.createElement("td");
		this._label.className = "label";
		this._row.appendChild(this._label);
		
		// Column 2: the folder or report description (value set in update()
		this._description = document.createElement("td");
		this._description.className = "description";		
		this._row.appendChild(this._description);
		
		// Column 3: Action bar to add, edit or delete the item
		this._actions = document.createElement("td");
		this._actions.className = "itemActions";
		this._actions.appendChild(this.toolbar.domNode);
		this._row.appendChild(this._actions);
		this.toolbar.startup();
	},	

	/** Update the contents of the given tree item.
	 *  This method can be called if the name or description
	 *  of the associated resource has changed. 
	 */
	update: function() {
		// Only the name and description can be updated
		var icon = (this.resource._isFolder ? Icons.FOLDER : Icons.REPORT);
		if (this.labelAction) {
			this._label.innerHTML = "<td><img style='margin-right: 5px;' src='" + icon + "'><a href='" + this.labelAction + "'>" + HTML.escape(this.resource.name) + "</a>";
		} else {
			this._label.innerHTML = "<td><img style='margin-right: 5px;' src='" + icon + "'>" + HTML.escape(this.resource.name);
		}

		// Do not show the description for the root system folders
		var description = this.resource.description || " ";
		if (this.resource._isFolder && this.resource.system) {
			description = " ";
		}
		this._description.innerHTML = "<td>" + HTML.escape(description);
		
		// Ensure the item container can get the focus and add listeners
		// to show and hide the toolbar actions.
		this.getItemContainer().tabIndex = 0;
		ToolbarFocusUtil.addToolbarFocusListeners( { 
			actionToolbar: this.toolbar, 
			containerNode: this.getItemContainer(), 
			actionClassName: ".jazz-ui-SimpleToolbar-button",
			focusCallback:  dojo.hitch(this, this.focusCallback)
		});
	},
	
	/** When the elements in the tree item obtain the focus, the
	 *  listeners will call this function. This enables the background
	 *  color of the item to be modified.
	 *
	 * @param the item container (e.g., containerNode in addToolbarFocusListeners())
	 * @param hasFocus true = tree item has the focus, false = item lost the focus
	 */ 
	focusCallback: function(itemContainer, hasFocus) {
		this.highlight( (hasFocus ? ReportUtil.COLOR_IN_FOCUS : ReportUtil.COLOR_NORMAL));
	},
	
	/** When the enter key is pressed for this item, go to the
	 *  corresponding editor (e.g., invoke the editor action). 
	 *  Only tree items representing reports will have an editor action.
	 */
	onEnter: function() {
		if (this.labelAction) {
			window.location.href = this.labelAction;
		}
	},
	
	/** Highlight the entire item with the appropriate background color.
	 *  If the item is selected, then color requests for in focus and
	 *  normal are ignored so that the selected highlight color is not
	 *  changed. However, if the item is selected and the unselect color
	 *  is passed, the background is reset to normal.
	 * 
	 * @param color the highlight color (defined in ReportUtil)
	 */
	highlight: function(color) {
		// If the item is selected
		if (this._isSelected) {
			// If the item is being un-selected
			// change the color back to normal,
			if (color === ReportUtil.COLOR_UNSELECT) {
				this._isSelected = false;
				color = ReportUtil.COLOR_NORMAL;
				dojo.style(this._row, "backgroundColor", color);
			}
		} 
		
		// else if this item is being selected
		else if (color === ReportUtil.COLOR_SELECTED) {
			// Change its color to indicate selected
			this._isSelected = true;
			dojo.style(this._row, "backgroundColor", color);
		}
		
		// else if the item has just received the focus
		else if (color === ReportUtil.COLOR_IN_FOCUS) {
			// Change the color to indicate the item has focus
			dojo.style(this._row, "backgroundColor", color);
		}
		
		// else the item has lost the focus 
		else {
			// Change the item color back to normal
			dojo.style(this._row, "backgroundColor", color);
		}
		
	},
	
	/** Get the container for the item data. 
	 *  For Firefox or IE, using the row for the
	 *  container works but for Google Chrome, the
	 *  table itself must be used or else it is
	 *  impossible to select an action at the right
	 *  side of the row. The action bar disappears 
	 *  because Chrome does not consider blank columns
	 *  or spacing to be part of the row. 
	 * 
	 * @return the table that contains the label and description
	 */
	getItemContainer: function() {
		return this.getFocusNode();
	},
	
	/** Get the reporting resource represented by
	 *  this item.
	 * 
	 * @return the folder or the report
	 */
	getResource: function() {
		return this.resource;
	},

	/** Get UUID of the reporting resource represented by
	 *  this item.
	 * 
	 * @return the folder UUID or the report UUID
	 */
	getResourceId: function() {
		return (this.resource._isFolder ? this.resource.folderUUID : this.resource.queryUUID);
	},
	
	/** Destroy this item and all of its children.
	 */
	 destroy: function() {
		if (this.toolbar) {
			// Destroy the action toolbar to remove all listeners
			this.toolbar.destroy();
		delete this.toolbar;

			// Destroy all children (removing them causes them to be destroyed)
			var childItems = this.getChildren();
			var itemsToRemove = [];
			dojo.forEach(childItems, function(item, i) { itemsToRemove.push(item); });
			var thisItem = this;
			dojo.forEach(itemsToRemove, function(item, i) { thisItem.removeChild(item); });

			this.inherited(arguments);
		}
	 }
	
});


})();
