/*
 *+------------------------------------------------------------------------+
 *| Licensed Materials - Property of IBM                                   |
 *|                                                                        |
 *| com.ibm.xtools.rmps.web.reports                                        |
 *|                                                                        |
 *| Copyright IBM Corporation 2012.  All Rights Reserved.                  |
 *|                                                                        |
 *| US Government Users Restricted Rights - Use, duplication or disclosure |
 *| restricted by GSA ADP Schedule Contract with IBM Corp.                 |
 *+------------------------------------------------------------------------+
 */

(function() {
dojo.provide("dm.app.reports.ui.internal.pages.ReportEditor");

/** Editor to modify a report configuration. 
 *  The JFS reporting framework does not save the "shared" property
 *  so a report is shared if its parent folder is shared.
 * 
 * @author Kevin Cornell
 */

dojo.require("rmps.app.ui.internal.editors.TabbedEditor");
dojo.require("jazz.ui.Breadcrumbs");

dojo.require("dm.app.reports.ui.internal.application.ReportActions");
dojo.require("dm.app.reports.ui.internal.util.ReportServiceClient");
dojo.require("dm.app.reports.ui.internal.util.ReportSimpleButton");
dojo.require("dm.app.reports.ui.internal.util.ReportUtil");

dojo.require("jazz.app.context");
dojo.require("rmps.core.util.ContextUtils");
var ContextUtils = rmps.core.util.ContextUtils;

dojo.require("com.ibm.team.repository.web.transport.ServiceResponseHandler");

dojo.require("rmps.ui.SimpleButton");
dojo.require("jazz.ui.MessageBox");
dojo.require("net.jazz.ajax.ui.PlatformUI");

dojo.require("dojo.string");
dojo.require("dojo.i18n");
dojo.requireLocalization("dm.app.reports.ui.internal", "ReportMessages");
var messages  = dojo.i18n.getLocalization("dm.app.reports.ui.internal", "ReportMessages");

var PlatformUI				= net.jazz.ajax.ui.PlatformUI;
var ReportActions			= dm.app.reports.ui.internal.application.ReportActions;
var ReportServiceClient		= dm.app.reports.ui.internal.util.ReportServiceClient;
var ReportUtil				= dm.app.reports.ui.internal.util.ReportUtil;
var ServiceResponseHandler	= com.ibm.team.repository.web.transport.ServiceResponseHandler;

//Define the editor help UA key
var PAGE_UAKEY = "com.ibm.xtools.rmps.web.reports.page_ReportEditor001";

var DIRTY_GROUP_ID = "dm.app.reports.ReportEditor.dirty";

// Define the fields of a report that can be modified.
var FIELD_description    = "description";
var FIELD_name           = "name";

//Define the fields of a report that can be read but not modified.
var FIELD_folderUUID     = "folderUUID";			// the parent folder of the report configuration
var FIELD_processArea    = "processAreaUUID";		// the project ID
var FIELD_queryUUID      = "queryUUID";				// the ID of this report configuration
var FIELD_reportUUID     = "reportUUID";			// the corresponding report template

// When a report (QueryDescriptor) is read, the parameter fields are parameterNames
// and parameterValues but when saving the descriptor, the field names must be
// parameterName and parameterValue. Both fields are arrays of strings.
var FIELD_parameterName  = "parameterName";			// the array of parameter names
var FIELD_parameterValue = "parameterValue";		// the array of parameter values

// This editor keeps 2 copies of the report data: the current (old)
// report and the modified (new) report. When an existing report
// is read, various fields must be copied from the current report to
// the modified report and a Dojo clone cannot be performed because the
// old report has some non-primitive fields that causes the POST operation
// (which uses the modified report) to fail.
// Define the report fields that must be copied to the modified report.
var COPY_FIELDS = [
	FIELD_description, 
	FIELD_name, 
	FIELD_folderUUID,
	FIELD_processArea, 
	FIELD_queryUUID, 
	FIELD_reportUUID,
	FIELD_parameterName,
	FIELD_parameterValue ];

//Define the system variables that can be used in
//metadata URLs defined in RPE templates
var SYSVAR_ProjectName   = "projectName";
var SYSVAR_WorkspaceName = "workspaceName";
var SYSVAR_SnapshotName  = "snapshotName";
var SYSVAR_ContextUri    = "context";

dojo.declare("dm.app.reports.ui.internal.pages.ReportEditor", rmps.app.ui.internal.editors.TabbedEditor, {
	
	"class": "reports-ReportEditor",
	
	// base editor properties
	isRefreshable: false,
	isReadOnly: false,
		
	// Define the report (query) ID and the current/modified report data
	_queryUUID:  null,
	_oldReport: null,
	_newReport: null,
	_template: null,
	_folder: null,
	
	// Define a object to hold the report parameters and the 
	// values for each parameter. The default parameter values
	// are set and are overridden by the parameter values defined
	// in the report itself. When a parameter value is updated in 
	// the report, the paramValues[] entry in this object is also
	// updated accordingly. The default parameter values are only
	// saved in the parameterDTO objects.
	
	// The following fields are defined in the corresponding "params" object: 
	//     parameterDTO[] the parameter DTO objects (some may be groups)
	//     paramNames[]   the parameter names (unique parameters extracted from DTOs)
	//     paramValues[]  the current parameter values (array in sync with paramNames) 
	//     sysVarNames[]  the system variable names (project name, context URI and workspace/snapshot name)
	//     sysVarValues[] the system variable values (array in sync with sysVarNames) 
	//     sysVarMerged   a boolean indicating the system variables are merged into param arrays
	_params: null,
	_paramsLoaded: false,
	_sysVarsLoaded: false,
	
	// Remember if a report is already running
	_runButton: null,
	_isRunning: false,

	_breadCrumbs: null,
	_closeButton: null,

	postCreate: function() {	
		// Add the "Reports" bread crumb at the top of the editor
		this._breadCrumbs = new jazz.ui.Breadcrumbs({segments: [{label: messages.reportBrowserName, 
																 href: ReportActions.getActionUri(ReportActions.ACTION_VIEW_REPORTS)}]});
		dojo.place(this._breadCrumbs.domNode, this.headerContainerNode, "first");

		this.inherited(arguments);
	},
		
	/** Add the tabs to define the content widgets for the report configuration editor.
	 * 
	 * @return the widget that displays the content
	 */
	addTabs: function() {
		// Specify the help for this editor
		this.uaKey = PAGE_UAKEY;
		
		// Add the tab definitions to the editor
		var tabs = this.getEditorInput().getTabs();
		var index, length, tab;
		for (index = 0, length = tabs.length; index < length; index++) {
			tab = tabs[index];
			this.addTab(tab.id, tab.label, tab.widget, {
				editor: this,
				isReadOnly: this.isReadOnly,
				definition: tab
			});
		}
		this.inherited(arguments);
	},
	
	/** Indicate that the editor name can be changed. The
	 *  editor name is connected to the report name.
	 */
	isEditableTitle: function() {
		return true;
	},
	
	/** Handle changes to the editor name. Update the 
	 *  name in the report and mark the report as 
	 *  dirty.
	 *  
	 *  @param newName the new report name
	 */
	onTitleChange: function(newName){
		var newValue = dojo.trim(newName);
		this.updateReportField(FIELD_name, newValue);
	},

	/** When the editor is activated, obtain the corresponding
	 *  report, report template and parent folder. When all three
	 *  requests are complete, and update the editor tabs.
	 */
	activate: function() {
		this.inherited(arguments);
		
		// If the editor is completely loaded (e.g., we are switching tabs)
		if (this._checkLoadComplete()) {
			// Update the selected tab with the report information
			// (_updateSelectedTab() will have been called by _checkLoadComplete())
		}
		
		// else if the report is not in the process of loading
		// Load the report, its folder, its template and the parameters
		else if (!this._queryUUID) {
			// Disable the Run button while the report is loading.
			this._runButton.setEnabled(false);

			// If the editor being opened is for a report that was just created
			if (this.getEditorInput().getParams().created) {
				// Redisplay the saved message.
				this.showMessage({
					type: jazz.ui.MessageBox.STATUS_TYPES.OK,
					summary: messages.saveSuccessful,
					details: null,
					closable: true
				});
			}

			// Remember if an existing template is being edited
			var editorParams = this.getEditorInput().getParams();
			this._queryUUID = editorParams.queryUUID;
			
			// If no report ID was defined, display an error
			if (!this._queryUUID) {
				this.showMessage({
					type: jazz.ui.MessageBox.STATUS_TYPES.ERROR,
					summary: messages.errorNoReportID,
					details: [ {message: messages.errorNoReportIDDetails} ],
					closable: true
				});
			} else {
				// Issue a request to obtain the report contents.
				// It will subsequently issue requests for the folder
				// and the template.
				this._newReport = null;
				var serviceObject = {
						self: this,
						success: function(report) {
							this.self._loadReport(report, true);
						},
								
						failure: function(error) {
							this.self.serviceError(messages.errorLoadingReports, error);
						}
					};
				var srh = new ServiceResponseHandler(serviceObject, "success", "failure");
				ReportServiceClient.getQueryDescriptor(srh, {queryUUID: this._queryUUID});
			}
		}
		
	},
	
	/** Add the cancel button. Save button is added because "isReadOnly" is false.
	 * 
	 * @param toolbar: the base editor page toolbar.
	 */
	addButtons: function(toolbar) {
		this._closeButton = new dm.app.reports.ui.internal.util.ReportSimpleButton({
			label: messages.close,
			onClick: dojo.hitch(this, this._handleCancel)
		});
		toolbar.addWidget(this._closeButton);
		
		this.inherited(arguments);
		
		this._runButton = toolbar.addWidget(new rmps.ui.SimpleButton({
			label: messages.run,
			onClick: dojo.hitch(this, this._handleRun)
		}));
		
	},
	
	_updateCloseButtonLabel: function() {
		var newLabel = this.isDirty() ? messages.cancel : messages.close;
		this._closeButton.setLabel(newLabel);
	},
	
	/** Perform the saving of the report.
	 * 
	 */
	doSave: function(taskMonitor) {
		this.inherited(arguments);
		
		// At this point the modified report should contain
		// all of the data needed to update the query descriptor
		// resource (e.g., the report configuration).
		var serviceObject = {
				self: this,
				success: function(reportUUID) {
					this.self._saveComplete(reportUUID);
				},
						
				failure: function(error) {
					this.self.serviceError(messages.errorSavingReport, error);
				}
			};
		var srh = new ServiceResponseHandler(serviceObject, "success", "failure");
		ReportServiceClient.postSaveQueryDescriptor(srh, this._newReport);
		
		// The TaskMonitor cannot be used to handled the deferred work from the
		// request because it overwrites the specific error message & details
		// with a generic "Bad Request" message that means nothing.
	},
	
	/** The saving of the report was successful.
	 * 
	 * @param reportUUID the value returned from the POST request
	 * 
	 */
	_saveComplete: function(reportUUID) {
		// Update the editor input params in case the name changed
		var params  = this.getEditorInput().getParams();
		params.name = this._newReport[FIELD_name];
		this.getEditorInput().setParams(params);
		
		// Reload the modified report.
		// Disable the Run button while the report is loading.
		this._runButton.setEnabled(false);
		this._loadReport(this._newReport, false);
	},

	/** When Cancel is clicked in the report editor, go
	 *  back to the page for viewing all reports.
	 */
	_handleCancel: function() {
		// Remove the dirty indicator
		this.removeDirty(null, DIRTY_GROUP_ID);

		// Close the editor and remove it from the recently and unsaved editor lists.
		var pageHelper = rmps.ui.editor.PageHelper.getInstance(ReportActions.PAGE_ID);
		pageHelper.closeEditor(this, true, true);

		// The back button remembers each tab visited so go to the report browser.
		window.location.href = ReportActions.getActionUri(ReportActions.ACTION_VIEW_REPORTS);
	},
	
	/** Determine if a report is currently running.
	 * 
	 * @return true if a report is being generated
	 */	
	reportIsRunning: function() {
		return this._isRunning;
	},
	
	/** Update the status of the "is running" indicator.
	 *  This will enable or disable the Run button.
	 *  
	 *  @param isRunning true - a report is being generated
	 *							or an export is being performed
	 */
	setIsRunning: function(isRunning) {
		this._isRunning = isRunning;
		this._updateRunButton();
	},
	
	/** Update the Run button by examining the 
	 *  "is running" indicator and by examining
	 *  the parameters, ensuring all required 
	 *  parameters are defined.
	 */
	_updateRunButton: function() {
		// The run button is enabled if a report is not running
		var enableRun = !this.reportIsRunning();
		
		// If a required parameter is not defined,
		// to disable the run button
		if (enableRun && this._paramsLoaded) {
			// For each parameter definition
			var len = this._params.parameterDTO.length;
			for (var i = 0; enableRun && i < len; i++) {
				// If the parameter is required
				var paramDefn = this._params.parameterDTO[i];
				if (paramDefn.required) {
					// See if the parameter value is defined
					var paramLen = this._params.paramNames.length;
					for (var j = 0; j < paramLen; j++) {
						if (this._params.paramNames[j] === paramDefn.name) {
							// If a required parameter does not have
							// a value, disable the run button.
							enableRun = this._params.paramValues[j] &&
										this._params.paramValues[j] !== "''";
							break;
						}
					}
				}
			}
			
			// Update the label of the parameters tab to indicate
			// whether a require parameter has no value and needs
			// attention from the user.
			var parameterTabId = this.getEditorInput().getParametersTabId();
			var tab = this.getTab(parameterTabId);
			var paramTabLabel = messages.reportTabParams +
				(enableRun ? "" : ReportUtil.MANDATORY_HTML);
			tab.label = paramTabLabel;
			if (tab.linkNode) {
				tab.linkNode.innerHTML = paramTabLabel;
			}
		}
		
		// Enable or disable the run button
		this._runButton.setEnabled(enableRun);
	},
	
	/** When the Run button is pressed, switch to the output
	 *  tab and execute the "run" operation.
	 */
	_handleRun: function() {
		// Select the output tab.
		var outputTabId = this.getEditorInput().getOutputTabId();
		this.selectTab(outputTabId);

		// Update the selected tab with the report information
		this._updateSelectedTab();
		
		// Request the tab to run the report again
		var tab = this.getTab(outputTabId);
		if (tab && tab.page && !dojo.isString(tab.page)) {
			tab.page.runReport();
		}	
	},
	
	/** Determine if the report, its folder, template and all
	 *  of its parameters and variables have been loaded. If
	 *  loading is complete, then the selected tab is updated.
	 *  
	 *  @return true = loading of report is complete
	 */
	_checkLoadComplete: function() {
		if (this._newReport && this._folder && this._template && 
				this._paramsLoaded && this._sysVarsLoaded) {
			// Merge the system variables into the param arrays if needed
			if (!this._params.sysVarMerged) {
				// For each system variable
				var sysVarLen = this._params.sysVarNames.length;
				for (var i = 0; i < sysVarLen; i++) {
					// Determine if a parameter corresponds to this system variable
					var sysVarName = this._params.sysVarNames[i];
					var index = -1;
					var paramLen = this._params.paramNames.length;
					for (var j = 0; index < 0 && j < paramLen; j++) {
						if (sysVarName === this._params.paramNames[j]) {
							index = j;
						}
					}
					
					// If the system variable was also defined as a parameter
					if (index >= 0) {
						// Set the parameter value to the system variable value
						this._params.paramValues[index] = this._params.sysVarValues[i];
					} 
					
					// else append the system variable to the parameter arrays
					else {
						this._params.paramNames.push(sysVarName);
						this._params.paramValues.push(this._params.sysVarValues[i]);
					}
				}
				
				// Merging is complete
				this._params.sysVarMerged = true;
			}
			
			// Hide the loading message and update the selected tab.
			PlatformUI.getWorkbench().hideMessageArea();			
			this._updateSelectedTab();
			return true;
		}
		return false;
	},

	/** Initialize the editor with the data from the specified 
	 *  report. The issue requests to obtain the related data
	 *  including the associated report template and the parent
	 *  folder.
	 *  
	 *  @param report the report begin modified
	 *  @param wasRead  true = report was read (not reloaded)
	 */
	_loadReport: function(report, wasRead) {
		// Show the report loading message
		PlatformUI.getWorkbench().showMessage(messages.loadingReport);
		
		// Initialize the object to hold the parameter and variable data
		// and initialize the system variables
		this._params = {
				parameterDTO: [],
				paramNames:   [],
				paramValues:  [],
				sysVarNames:  [],
				sysVarValues: [],
				sysVarMerged: false
		};
		this._initSystemVariables();
		
		// Remember the existing report data
		this._oldReport = report;
		
		// If the description field is not set, force it to be the empty string.
		// If a report is saved with an empty string as the description, when
		// it is subsequently read, the description field is null, which will cause
		// the save request to fail (invalid request parameter).
		this._oldReport[FIELD_description] = this._oldReport[FIELD_description] || "";
		
		// The parameters fields are different for a report that is
		// read and a report that is being saved. If the report being
		// loaded is being read, convert the fields of the existing report 
		// to match the fields needed for the save. Otherwise the parameter
		// fields are already defined correctly for saving.
		if (wasRead) {
			this._oldReport[FIELD_parameterName]  = this._oldReport.parameterNames  || [];
			this._oldReport[FIELD_parameterValue] = this._oldReport.parameterValues || [];
		}
				
		// Create a copy of the report for modifications
		// Check explicitly for defined fields or else "false" boolean
		// fields will not get copied.
		this._newReport = {};
		for (var index = 0, len = COPY_FIELDS.length; index < len; ++index) {
			var fieldName = COPY_FIELDS[index];
			if (this._oldReport[fieldName] !== undefined) {
				var fieldValue = this._oldReport[fieldName];
				if (dojo.isArray(fieldValue)) {
					this._newReport[fieldName] = new Array(fieldValue.length);
					for (var j=0; j < fieldValue.length; j++) {
						this._newReport[fieldName][j] = fieldValue[j];
					}
				} else {
					this._newReport[fieldName] = fieldValue;
				}
			}
		}

		// The modified report data is not dirty yet (same as current data).
		this.removeDirty(null, DIRTY_GROUP_ID);
		this._updateCloseButtonLabel();
		
		// Issue a request to obtain the report's parent folder
		this._folder = null;
		var folderServiceObject = {
				self: this,
				success: function(folder) {
					this.self._loadFolder(folder);
				},
						
				failure: function(error) {
					this.self.serviceError(messages.errorLoadingFolders, error);
				}
			};
		var folderResponse = new ServiceResponseHandler(folderServiceObject, "success", "failure");
		ReportServiceClient.getFolderDescriptor(folderResponse, {folderUUID: this._newReport[FIELD_folderUUID]});
		
		// Issue a request to obtain the report's associated template
		this._template = null;
		var templateServiceObject = {
				self: this,
				success: function(template) {
					this.self._loadTemplate(template);
				},
						
				failure: function(error) {
					this.self.serviceError(messages.errorLoadingTemplates, error);
				}
			};
		var templateResponse = new ServiceResponseHandler(templateServiceObject, "success", "failure");
		ReportServiceClient.getReportDescriptor(templateResponse, {reportUUID: this._newReport[FIELD_reportUUID]});
		
		// Issue a request to obtain the report's associated parameters
		this._paramsLoaded = false;
		var paramServiceObject = {
				self: this,
				success: function(reportParamsArray) {
					this.self._loadParameters(reportParamsArray);
				},
						
				failure: function(error) {
					this.self.serviceError(messages.errorLoadingParameters, error);
				}
			};
		var paramResponse = new ServiceResponseHandler(paramServiceObject, "success", "failure");
		ReportServiceClient.getParameters(paramResponse, {queryUUID: this._newReport[FIELD_queryUUID]});
	},

	/** Initialize the parent folder data
	 * 
	 * @param folder the report's parent folder
	 */
	_loadFolder: function(folder) {
		this._folder = folder;
		this._checkLoadComplete();
	},

	/** Initialize the corresponding report template data
	 * 
	 * @param template the report template
	 */
	_loadTemplate: function(template) {
		this._template = template;
		this._checkLoadComplete();
	},

	/** Initialize the corresponding report parameter data
	 * 
	 * @param reportParamsArray the report parameter DTO objects
	 */
	_loadParameters: function(reportParamsArray) {
		// Process the parameters defined in the corresponding template
		dojo.forEach(reportParamsArray, dojo.hitch(this, function(paramDTO) {
			// Save the parameter definition object
			this._params.parameterDTO.push(paramDTO);
			
			// If the parameter is a group
			if (paramDTO._eQualifiedClassName === "com.ibm.team.reports.dto:ReportParameterGroupDTO") {
				// Process each of the parameters defined in the group
				dojo.forEach(paramDTO.parameters, dojo.hitch(this, function(groupParamDTO) {
					// Handle each parameter definition in the parameter group.
					this._handleParamDTO(groupParamDTO);
				}));
			}
			
			// If the parameter is a normal parameter (not a group)
			else if (paramDTO._eQualifiedClassName === "com.ibm.team.reports.dto:ReportParameterDTO") {
				// Handle the parameter definition.
				this._handleParamDTO(paramDTO);
			}
		}));
		
		// The parameters are  now loaded
		this._paramsLoaded = true;
		this._updateRunButton();
		this._checkLoadComplete();
		
	},
	
	/** Handle the processing of one parameter definition object. 
	 *  Add the parameter name to the parameter name array
	 *  and add the parameter default value to the value array.
	 *  In the parameter data object (_params) the name array 
	 *  and the value array MUST be kept in sync.
	 *  
	 *  Likewise in the report object (_newReport), the parameter
	 *  name and value arrays must be kept in sync, but order of
	 *  the entries in the parameter data object does not need 
	 *  to match the order in the report object.
	 *  
	 * @param paramDTO the parameter definition object
	 */
	_handleParamDTO: function(paramDTO) {
		// If the parameter has not been added to the list yet
		var paramIndex = -1;
		var paramsLen = this._params.paramNames.length;
		for (var i = 0; i < paramsLen; i++) {
			if (paramDTO.name === this._params.paramNames[i]) {
				paramIndex = i;
				break;
			}
		}
		if (paramIndex === -1) {
			// Extract the default value from the parameter
			// definition. 
			var paramValue = paramDTO.defaultValue || "";
			
			// Override the default value if the
			// report itself has a value for the parameter.
			var len = this._newReport[FIELD_parameterName].length;
			for (var j = 0; j < len; j++) {
				if (this._newReport[FIELD_parameterName][j] === paramDTO.name) {
					paramValue = this._newReport[FIELD_parameterValue][j];
					break;
				}
			}

			// Save the parameter name and value in the
			// "params" object.
			this._params.paramNames.push(paramDTO.name);
			this._params.paramValues.push(paramValue);
		}
	},

	/** Update the report field with the new value.
	 *  If the new value differs from the current value then
	 *  update the dirty indicator for this field.
	 *  
	 *  @param fieldName the name of the report field to update
	 *  @param newValue  the new value for the report field
	 *  @return true if the new value differs from the existing value
	 *               (e.g., the editor is now dirty)
	 */
	updateReportField: function(fieldName, newValue) {
		// If the report has been initialized
		var isEqual = true;
		if (this._oldReport) {
			// Get the current value of the corresponding field
			var oldValue = "";
			if (this._oldReport[fieldName] !== null) {
				oldValue = this._oldReport[fieldName];
			}
			
			// Determine if the current and new values are different
			// For an array, examine all entries and stop if one differs
			isEqual = oldValue === newValue;
			if (!isEqual && dojo.isArray(newValue) && oldValue && newValue && oldValue.length === newValue.length) {
				isEqual = true; // assume empty arrays are equal
				isEqual = dojo.every(newValue, function(item, idx) {
					return item === oldValue[idx];
				});
			}

			// If values are equal, clear the field's dirty indicator
			// else set the field's dirty indicator
			if (isEqual) {
				this.removeDirty(fieldName, DIRTY_GROUP_ID);
			} else {
				this.addDirty(fieldName, DIRTY_GROUP_ID);
			}
			this._updateCloseButtonLabel();
			
			// Update the new copy of the report. The new value
			// must always be set because the user could have 
			// previously set a new value and then reverted.
			this._newReport[fieldName] = newValue;
		}
		return !isEqual;
	},

	/** Update a report parameter with the new value.
	 *  If the new value differs from the current value then
	 *  update the dirty indicator for this parameter.
	 *  
	 *  @param paramName the name of the report parameter to update
	 *  @param newValue  the new value for the report parameter
	 *  @return true if the new value differs from the existing value
	 *               (e.g., the editor is now dirty)
	 */
	updateReportParameter: function(paramName, newValue) {
		// If the parameter is not a system variable
		var isEqual = true;
		if (!this.isSystemVariable(paramName)) {
			// Determine if the parameter was previous set
			
			// Get the current value of the corresponding parameter
			var oldValue = "";
			var oldLen = this._oldReport[FIELD_parameterName].length;
			for (var oldIndex = 0; oldIndex < oldLen; oldIndex++) {
				if (this._oldReport[FIELD_parameterName][oldIndex] === paramName) {
					oldValue = this._oldReport[FIELD_parameterValue][oldIndex];
					break;
				}
			}
			
			// Determine if the current and new values are different
			// For an array, examine all entries and stop if one differs
			isEqual = oldValue === newValue;
			if (!isEqual && dojo.isArray(newValue) && oldValue && newValue && oldValue.length === newValue.length) {
				isEqual = true; // assume empty arrays are equal
				isEqual = dojo.every(newValue, function(item, idx) {
					return item === oldValue[idx];
				});
			}

			// If values are equal, clear the parameter's dirty indicator
			// else set the parameters's dirty indicator
			var paramFieldName = "param_" + paramName;
			if (isEqual) {
				this.removeDirty(paramFieldName, DIRTY_GROUP_ID);
			} else {
				this.addDirty(paramFieldName, DIRTY_GROUP_ID);
			}
			this._updateCloseButtonLabel();
			
			// Update the parameter in the new copy of the report.
			// The new value must always be set because the user 
			// could have previously set a new value and then reverted.
			var newLen = this._newReport[FIELD_parameterName].length;
			var newIndex = -1;
			for (var i = 0; i < newLen; i++) {
				if (this._newReport[FIELD_parameterName][i] === paramName) {
					newIndex = i;
					break;
				}
			}
			if (newIndex >= 0) {
				this._newReport[FIELD_parameterValue][newIndex] = newValue;
			} else {
				this._newReport[FIELD_parameterName].push(paramName);
				this._newReport[FIELD_parameterValue].push(newValue);
			}
			
			// Update the current parameter value in the "params" object
			var paramsLen = this._params.paramNames.length;
			for (var j = 0; j < paramsLen; j++) {
				if (this._params.paramNames[j] === paramName) {
					this._params.paramValues[j] = newValue;
					break;
				}
			}
			
			// Enable or disable the run button
			this._updateRunButton();
		}
		return !isEqual;
	},
	
	/** Determine if the given parameter corresponds to a
	 *  system variable.
	 *  
	 *  @param paramName the parameter name
	 *  @return true = system variable
	 */
	isSystemVariable: function(paramName) {
		var sysVarLen = this._params.sysVarNames.length;
		for (var i = 0; i < sysVarLen; i++) {
			if (paramName === this._params.sysVarNames[i]) {
				return true;
			}
		}
		return false;
	},
	
	/** Update the selected tab with the current report information.
	 */
	_updateSelectedTab: function() {
		// Get the selected tab
		var tabId = this.getSelectedTabId();
		var tab = this.getTab(tabId);
		
		// Call the report editor tab to initialize its controls
		if (tab && tab.page && !dojo.isString(tab.page)) {
			tab.page.initReportData(this._newReport, this._folder, this._template, this._params);
		}	
	},
		
	/*----------------------------------------*/

	/** Initialize the system variables that can be
	 *  used in the metadata URLs for template parameters.
	 */	
	_initSystemVariables: function() {
		this._sysVarsLoaded = false;
		
		// Always set the project name variable.
		var projectArea = jazz.app.context.get();
		this._params.sysVarNames[0]  = SYSVAR_ProjectName;
		this._params.sysVarValues[0] = projectArea.name;
		
		// Always set the DM context URI variable, if there is a context.
		var contextUri = ContextUtils.getCurrentContextURI();
		if (contextUri) {
			this._params.sysVarNames[1]  = SYSVAR_ContextUri;
			this._params.sysVarValues[1] = contextUri;
			
			// Set either the workspace name or snapshot name variable
			if (ContextUtils.isWorkspace(contextUri)) {
				this._params.sysVarNames[2] = SYSVAR_WorkspaceName;
			} else {
				this._params.sysVarNames[2] = SYSVAR_SnapshotName;
			}
			this._params.sysVarValues[2] = "";
			var deferred = ContextUtils.getContextDeferred(contextUri);
			deferred.addCallback(dojo.hitch(this, function(context){
				this._params.sysVarValues[2] = context.name;
				this._sysVarsLoaded = true;
			}));
		}
		
		// else (no context) system variable loading is complete
		else {
			this._sysVarsLoaded = true;
		}
	},

	
	/** Report an error from a reporting service request.
	 * 
	 * @param errorMessage the main error message
	 * @param error the service error (details for error message)
	 */
	serviceError: function(errorMessage, error) {
		var details = null;
		if (error.message) {
			details = [ {message: error.message} ];
		} else if (error.errorClass) {
			details = [ {message: error.errorClass} ];
		}
		this.showMessage({
			type: jazz.ui.MessageBox.STATUS_TYPES.ERROR,
			summary: errorMessage,
			details: details,
			closable: true
		});
	}


});

})();
