/*
 *+------------------------------------------------------------------------+
 *| Licensed Materials - Property of IBM                                   |
 *|                                                                        |
 *| com.ibm.xtools.rmps.web.reports                                        |
 *|                                                                        |
 *| Copyright IBM Corporation 2012.  All Rights Reserved.                  |
 *|                                                                        |
 *| US Government Users Restricted Rights - Use, duplication or disclosure |
 *| restricted by GSA ADP Schedule Contract with IBM Corp.                 |
 *+------------------------------------------------------------------------+
 */

(function() {
dojo.provide("dm.app.reports.ui.internal.pages.ReportEditorOverview");

/** The Overview tab of the editor for modifying a report configuration.
 *  This tab allows the user to change the report name and description,
 *  but not the associated template. The user can also move the report
 *  to a different folder with restrictions (folder defines whether the
 *  report is private or shared). A report with a private template
 *  cannot be moved to a shared folder or else other users would be
 *  unable to run that report. A shared report (in a shared folder)
 *  cannot be moved to a private folder or else other users would no 
 *  longer be able to run that report.
 *  
 *  @author Kevin Cornell
 */

dojo.require("dm.app.reports.ui.internal.pages.ReportBrowserContent");

dojo.require("dm.app.reports.ui.internal.application.ReportActions");
dojo.require("dm.app.reports.ui.internal.util.ReportUtil");
dojo.require("dm.app.reports.ui.internal.util.Icons");

dojo.require("jazz.util.html");

dojo.require("dojo.string");
dojo.require("dojo.i18n");
dojo.requireLocalization("dm.app.reports.ui.internal", "ReportMessages");
var messages  = dojo.i18n.getLocalization("dm.app.reports.ui.internal", "ReportMessages");

var ReportActions = dm.app.reports.ui.internal.application.ReportActions;
var ReportUtil = dm.app.reports.ui.internal.util.ReportUtil;
var Icons = dm.app.reports.ui.internal.util.Icons;
var HTML = jazz.util.html;

// Define the report fields used by this tab.
var FIELD_description = "description";
var FIELD_folderUUID  = "folderUUID";

dojo.declare("dm.app.reports.ui.internal.pages.ReportEditorOverview", [dijit._Widget, dijit._Templated], {

	"class": "reportEditorOverview",

	messages: messages, 
	
	widgetsInTemplate: true,
	templateString: 
		'<div>' +
			// Main section: description (report name is the editor title)
			'<div class="mainSection" dojoAttachPoint="_mainSection" >' +
				'<div class="descriptionLabel">${messages.reportDescription}</div>' +
				'<div class="descriptionContainer">' +
					'<textarea class="descriptionInput" dojoAttachPoint="_descriptionInput" rows="5" ></textarea>' +
				'</div>' +
			'</div>' +

			// Template Info section: contains the name of the report template, its description
			// and its shared indicator
			'<div class="templateInfoSection">' +
				'<div class="templateInfoSectionTitle">${messages.templateInfoSectionTitle}</div>' +
				'<table >' +
					'<tr>' +
						'<td class="templateName"        dojoAttachPoint="_templateName"></td>' +
						'<td class="templateDescription" dojoAttachPoint="_templateDescription"></td>' +
						'<td class="templateShared"      dojoAttachPoint="_templateShared"></td>' +
				'</tr>' +
				'</table>' +
			'</div>' +
			
			// Folder selection section (host the report browser content)
			'<div class="folderInfoSection">' +
				'<div class="folderInfoSectionTitle">${messages.folderInfoSectionTitle}</div>' +
				'<button class="resetFolderButton" dojoAttachPoint="_resetFolder" >${messages.resetFolder}</button>' +
				'<div class="foldersContainer" dojoAttachPoint="_foldersContainer">' +
					'<div class="reportBrowserHook" dojoAttachPoint="_reportBrowserHook"></div>' +
				'</div>' +
			'</div>' +

		'</div>',
	
	// init properties
	editor: null,
	
	// Widget for the report browser
	_reportBrowser: null,
	
	// Remember the current folder and template (which should not change)
	_folder: null,
	_template: null,
	
	// Keep a copy of the initial folder for doing a reset
	_initialFolder: null,
	
	constructor: function(args) {
		this.messages = messages;
	},

	
	postCreate: function() {
		this.inherited(arguments);
		
		// Hook in the Report Browser to select a folder
		this._reportBrowser = new dm.app.reports.ui.internal.pages.ReportBrowserContent({editor: this.editor, inBrowser: false });
		this._reportBrowserHook.appendChild(this._reportBrowser.domNode);
		
		// Hook in methods to handle changes in various input fields
		if (!this.editor.isReadOnly) {
			// Report description
			ReportUtil.connectTextInput(this._descriptionInput, this, function() {
				var newValue = this._descriptionInput.value;
				this.editor.updateReportField(FIELD_description, newValue);
			});
			
			// Folder selection reset button
			this._resetFolder.disabled = true;
			this.connect(this._resetFolder, "onclick", function(event) {
				dojo.stopEvent(event);
				if (this._initialFolder && this._reportBrowser) {
					this._reportBrowser.selectFolder(this._initialFolder);					
				}
			});

			// Folder selected in the report browser
			dojo.connect(this._reportBrowser, "selectFolder", this, function(folder) {
				// Only change the report data if a folder was selected
				// and it is valid given the report's associated template
				this._folder = folder;
				if (folder && this._isTemplateFolderValid()) {
					// Update the folder ID
					var dirty = this.editor.updateReportField(FIELD_folderUUID, folder.folderUUID);
					this._resetFolder.disabled = !dirty;
				} else if (this._initialFolder) {
					this._reportBrowser.selectFolder(this._initialFolder);					
				}
			});
		}
	},
	
	/** Initialize the page controls based on the given
	 *  report data.
	 *  
	 *  @param report the report (query) descriptor
	 *  @param folder the parent folder descriptor
	 *  @param template the report template descriptor
	 *  @param params contains parameter definitions and values
	 */
	initReportData: function(report, folder, template, params) {
		// Save the current folder and template
		if (!this._folder) {
			this._folder = folder;
		}
		this._template = template;

		// Refresh the report browser if the initial folder changed
		// and reselect the initial folder
		var selectFolder = !this._initialFolder;
		if (this._initialFolder && this._initialFolder.folderUUID !== folder.folderUUID) {
			this._reportBrowser.refresh();
			selectFolder = true;
		}

		// Remember the initial folder for the reset button
		// (initialFolder must be set before calling reportBrowser.selectFolder())
		this._initialFolder = {
			name:       folder.name,
			folderUUID: folder.folderUUID,
			parentUUID: folder.parentUUID,
			shared:     folder.shared
		};
		
		// Select the initial folder if needed
		if (selectFolder) {
			this._reportBrowser.selectFolder(folder);	// now select the current folder			
		}
		
		// Set the report description
		var description = report[FIELD_description];
		this._descriptionInput.value = description || "";
		
		// Initialize the template information
		var editTemplateUri = ReportActions.getActionUri(ReportActions.ACTION_EDIT_TEMPLATE);
		editTemplateUri = editTemplateUri + "&reportUUID=" + template.reportUUID + "&name=" + encodeURIComponent(template.name);
		this._templateName.innerHTML = "<td><img style='margin-right: 5px;' src='" + Icons.TEMPLATE + "'><a href='" + editTemplateUri + "'>" + HTML.escape(template.name) + "</a>";
		description = template.description || "";
		this._templateDescription.innerHTML = "<td>" + HTML.escape(description);
		var sharedMsg = template.shared ? messages.resourceShared : messages.resourcePrivate;
		this._templateShared.innerHTML = "<td>" + HTML.escape(sharedMsg);
	},
	
	/** Determine if the template and folder combination
	 *  is valid. A report with a private template cannot
	 *  be moved to a shared folder or else another
	 *  user will not be able to run the report.
	 *  
	 *  Also, a shared report (initial parent folder
	 *  is shared) cannot be moved to a private
	 *  folder or else other users who have used the
	 *  report will no longer be able to access it.
	 */
	_isTemplateFolderValid: function() {
		var isValid = true;
		if (this._folder && this._template) {
			if (this._folder.shared && !this._template.shared) {
				isValid = false;
				alert(messages.templateFolderConflict);
			} 
			else if (!this._folder.shared && this._initialFolder.shared) {
				isValid = false;
				alert(messages.templateAlreadyShared);
			}			
		}
		return isValid;
	}


});


})();
