/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2013, 2014.  All Rights Reserved.
 * 
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

(function(){
dojo.provide("com.ibm.team.vvc.shared.web.ui.util.VvcCache");

dojo.require("dojo.store.Memory");
dojo.require("dojo.store.Cache");
dojo.require("com.ibm.team.vvc.shared.web.ui.util.VvcUtils");
var VvcUtils = com.ibm.team.vvc.shared.web.ui.util.VvcUtils;

dojo.require("com.ibm.team.vvc.shared.web.ui.notifications.notifications.VvcNotificationFactory");
var ncFactory = com.ibm.team.vvc.shared.web.ui.notifications.notifications.VvcNotificationFactory;

dojo.requireLocalization("com.ibm.team.vvc.shared.web.ui.util", "VvcCacheMessages");
var Messages = dojo.i18n.getLocalization("com.ibm.team.vvc.shared.web.ui.util", "VvcCacheMessages");

/**
 * <p>
 * The VVC cache is a centralized way of getting information about a
 * configuration. Ask it for a configuration by ID, and it will look up all
 * information it knows about already. If it does not know about the
 * configuration, it will ask the server. The VvcCache is actually two
 * separate caches, one responsible for the local data, and one responsible
 * for getting the remote data. Eventually, the entire VVC web UI should be
 * using the VVC cache to get information about configurations.
 * </p>
 * <p>
 * The cache is designed to work with configurations, but can easily be
 * extended to provide information about nearly anything within VVC, if you
 * implement the correct API. The cache has two modes of getting data, a
 * "single" mode, and a "bulk" mode. Single mode is designed for one-off
 * fetches of data from the cache, and is quite uninteresting in that it
 * behaves like any other cache.
 * </p>
 * <p>
 * Bulk mode is where the VvcCache gets interesting. To start a bulk-mode
 * transaction, you simply call "startBulkTransaction" and wait for the
 * deferred to return. <b>Once the deferred returns, the cache is locked and
 * it is ready for a bulk-mode operation, <u>although no other bulk
 * transactions can occur at this time.</u></b> There is no special way to request
 * data when in bulk mode, you proceed as if it were a single request. The
 * cache will automatically aggregate any requests it does not know about.
 * When "endBulkModeTransaction" is called, it will POST a request to the
 * front service asking for all the URI's which were collected during the
 * bulk-mode transaction. It will then resolve the deferreds individually,
 * as if it were a single transaction.
 * </p>
 * <p>
 * To support bulk-mode transactions, there should be a bulk-mode service on
 * the server side which is expecting to receive a list of URI's to resolve.
 * The data coming back from the server should be a well-formed JSON object
 * with key-value pairs. Keys should be the URI's which the cache is
 * requesting, while values should be the corresponding objects. Single-mode
 * transactions should behave the same way.
 * </p>
 * <p>
 * The local cache is a simple HTML5 session store. This allows the cache to
 * survive between page loads; changing pages does not clear the cache. To
 * prevent the cache from going stale, it will automatically kick objects
 * out based on a time limit. Each cache entry has its own time limit,
 * derived either from the cache's default limit, or from some other source.
 * This is achieved through a time stamp. If the current time exceeds this
 * time stamp, the local cache will treat any gets to that object as a cache
 * miss and re-fetch them from the server.
 * </p>
 * <p>
 * To achieve this level of simplicity, the cache is heavily dependent on
 * <code>dojo.Deferred</code>s and callbacks. If you ask the cache for
 * something, it will return a deferred for the request, which will resolve
 * with the item, when the cache knows about the item. Even for a cache hit,
 * a deferred object will be returned.  This forces developers to use the
 * preferred asynchronous pattern.
 * </p>
 * <p>
 * When a resource ID is requested, and it is a cache miss, a lock is created 
 * on that resource ID.  While that lock exists, any subsequent requests for
 * that same ID will return the same deferred that the original request is 
 * waiting on.  So, while a request is in flight, via bulk or single-get
 * mode, any number of subsequent requests for that ID can be made, and only
 * the result of the in-flight request will be returned for each of those 
 * requests.  This saves on a lot of chatter from the cache to the server.
 * </p>
 * <p>
 * For reference usage, examine
 * <code>com.ibm.team.vvc.server.core.ui.explorer.VvcExplorer</code>.
 * </p>
 */
var _localStorageMap = [];
dojo.declare("com.ibm.team.vvc.shared.web.ui.util.VvcCache", [], {
	
	_localStore: null,
	_idProperty: null,
	_target: null,
	_bulkTarget: null,
	_transaction: null,
	_transactionData: null,
	_locks: null,
	_idLock: null,
	
	constructor: function __VvcCache(args){
		this._localStore = this._getLocalStorage(args);
		this._idProperty = args && args.idProperty ? args.idProperty : "uri";
		this._target = args && args.target ? args.target : VvcUtils.getVvcUri() + "/client/configuration";
		this._bulkTarget = args && args.bulkTarget ? args.bulkTarget : VvcUtils.getVvcUri() + "/client/configurations";
		this._locks = [];
		this._idLock = [];
	},
	
	/**
	 * Rejects any pending deferreds in the cache and explicitly removes all the references
	 * being held by this object.  The reference to the localCache will also be nullified,
	 * but the local cache is NOT explicitly destroyed.  The localCache object is shared
	 * amongst instances of the VvcCache with the same key.  If the localCache needs to be
	 * cleaned up, then natural garbage collection will take care of it once there are no
	 * more references.
	 */
	destroy: function __destroy(){
		this.inherited(arguments);
		function rejectDeferreds(/* array */ deferredsToReject){
			dojo.forEach(deferredsToReject, dojo.hitch(this, function __rejectAllLocks(def){
				def.reject("This cache is being destroyed");
			}));
		}
		
		rejectDeferreds(this._idLocks);
		rejectDeferreds(this._locks);
		
		this._localStore = null;
		this._idProperty = null;
		this._target = null;
		this._bulkTarget = null;
		this._transaction = null;
		this._transactionData = null;
		this._locks = null;
		this._idLock = null;
	},
	
	/**
	 * Gets the proper local storage object for the arguments.  In order to ensure
	 * multiple cache's are kept up-to-date, we only create one local storage per
	 * key for all caches currently loaded.  For example, when loading the explorer,
	 * there are many individual caches instantiated for the ConfigurationUiNode objects
	 * and the explorer itself.  To keep them all in sync and ensure maximum data
	 * sharing, each individual cache is sharing a single local storage.  Since
	 * JavaScript is single-threaded, we do not need to worry about locking
	 * the local storage.
	 * 
	 * @param args The arguments
	 * @return A local storage object
	 */
	_getLocalStorage: function __getLocalStorage(args){
		// Handle temp storage areas immediately
		if(args.temporary){
			return new com.ibm.team.vvc.shared.web.ui.util.VvcLocalCache(args);
		}
		
		// Attempt to get the storage out of the map
		var key = args.storageKey
		var storage = _localStorageMap[key];
		if(!storage){
			// Initialize the local storage
			_localStorageMap[key] = new com.ibm.team.vvc.shared.web.ui.util.VvcLocalCache(args);
		}
		
		return _localStorageMap[key];
	},
	
	/**
	 * Gets a bunch of object from the store based on their id.  If the object
	 * does not exist, a query will goto the server to fetch it.
	 * 
	 * @param ids An array of ID's to get.
	 * @param queries An optional array of queries for the matching ID's.  If this
	 * 					is not specified, then the default query will be built on
	 * 					demand.  If it is specified, then the supplied queries will
	 * 					be used.  A 1:1 pairing of ids and queries is required,  
	 * @return Always returns a deferred.  When all the ID's have been resolved,
	 * 			the deferred will also resolve with a map of input ID's to their
	 * 			results.
	 */
	getAll: function __getAll(ids, queries){
		
		if(queries && ids.length !== queries.length){
			throw "VvcCache.getAll() requires a 1:1 relationship between the 'ids' and 'queries' arrays!";
		}
		
		var resultDeferred = new dojo.Deferred();
		var listOfDeferreds = [];
		var results = {};
		
		dojo.when(this.startBulkTransaction(), dojo.hitch(this, function __lockCache(){
			dojo.forEach(ids, dojo.hitch(this, function __forEach(id, index, ids){
				var getDef = this.get(id, queries ? queries[index] : undefined);
				getDef.then(dojo.hitch(this, function __getDefThen(result){
					results[id] = result;
				}), dojo.hitch(this, function __getDefError(error){
					if(!results.error){
						results.error = [];
					}
					results.error[id] = error;
				}));
				listOfDeferreds.push(getDef);
			}));
			
			var defList = new dojo.DeferredList(listOfDeferreds);
			defList.then(dojo.hitch(this, function __defListResolved(){
				resultDeferred.resolve(results);
			}));
			
			this.endBulkTransaction();
		}));
			
		return resultDeferred;
	},
	
	/**
	 * Gets an object from the store based on its id.  If the object
	 * does not exist, a query will goto the server to fetch it.
	 * 
	 * @return Always returns a deferred.  When the cache has a result (
	 * 			from the local store or from the server) the deferred will
	 * 			resolve with that result.
	 */
	get: function __get(id, query){
		var resultsDeferred;
		
		var cachedResult = this._localStore.get(id);
		if(cachedResult){
			resultsDeferred = new dojo.Deferred();
			resultsDeferred.resolve(cachedResult);
		} else {
			/*
			 * Puts a lock on that ID and return the pending deferred.
			 * This ensures only one request per resource per transaction
			 * is pending at any one given time.  Even if a bulk transaction
			 * is under way while a request for the new resource arrives, the
			 * result of the bulk transaction will be returned as the result
			 * for the new request.  This prevents unnecessary requests to
			 * the server.
			 */
			if(this._idLock[id]){
				return this._idLock[id];
			}
			
			if(!query){
				query = {};
			}
			if(query[this._idProperty] === undefined){
				query[this._idProperty] = id;
			}
			
			if (djConfig.isDebug) {
				query.debug = "true";
			}
			
			if(this._isBulkTransaction){
				// Bulk POST mode
				resultsDeferred = this._addBulkRequest(query);
			} else {
				// Single GET mode
				resultsDeferred =  this.query(id, query);
			}
			
			this._idLock[id] = resultsDeferred;
			resultsDeferred.then(dojo.hitch(this, function __unlockResource(){
				// Unlock the ID
				this._idLock[id] = null;
				delete this._idLock[id];
			}));
		}
		return resultsDeferred;
	},
	
	/**
	 * Executes a query against the remote data store.  
	 * 
	 * Based on the store#query method.
	 * 
	 * @param The query options.
	 * @return An array or promise.
	 * @see dojo.data.JsonRest
	 */
	query: function __query(id, query){
		var resultDef;
		if(this._isBulkTransaction){
			resultDef = this._addBulkRequest(query);
		} else {
			resultDef = new dojo.Deferred()
			var args = {
				url: this._target + "?" + dojo.objectToQuery(query),
				headers: {
					"Accept" : "application/json",
					"Content-Type" : "application/json"
				},
				handleAs: "json",
				preventCache: true
			};
			var xhrDeferred = jazz.client.xhrGet(args);
			xhrDeferred.then(dojo.hitch(this, function __processQueryResults(results){
				if(results){
					if(results[id]){
						this.put(results[id], id);
						resultDef.resolve(results[id]);
					} else {
						// Allow for server-side macros
						this.put(results, id);
						resultDef.resolve(results);
					}
				}
			}), dojo.hitch(this, function __processQueryResultsError(error){
				resultDef.reject(error);
			}));
		}
		return resultDef;
	},
	
	/**
	 * Puts an item in the local cache only.  
	 * 
	 * Based on the dojo.store.Memory#put method.
	 * 
	 * @param item The item to put in the cache.
	 * @param properties The properties 
	 */
	put: function __put(item, id, options){
		this._localStore.put(item, id, options);
	},
	
	/**
	 * Removes an item from the cache.  If no item is specified, 
	 * then the entire cache is cleared.
	 * 
	 * @param id The ID of the item to remove.
	 * 
	 * @return The removed item.
	 */
	clear: function __clear(id){
		return this._localStore.clear(id);
	},
		
	/**
	 * Starts a bulk transaction.  The cache will collect all 
	 * local cache misses and dispatch them at once, resolving 
	 * the individual deferreds one at a time, but only using 
	 * one HTTP request.
	 */
	startBulkTransaction: function __startBulkTransaction(){
		var lock = new dojo.Deferred();
		var locked;
		
		if(this._locks.length > 0){
			locked = this._locks[0];
		}else {
			// Create a dummy lock and release it.
			locked = new dojo.Deferred();
			locked.resolve();
		}
		
		this._locks.unshift(lock);
		dojo.when(locked, dojo.hitch(this, function __acquireLock(){
			this._isBulkTransaction = true;
			this._transactionData = [];
			this._transaction = new dojo.Deferred();
			this._transaction.then(dojo.hitch(this, function __startTransaction(){
				if(this._transactionData.length === 0){
					// Abort if there's nothing to get.
					this._isBulkTransaction = false;
					lock.resolve(true);
					this._locks.splice(this._locks.length - 1, 1);
					return;
				}
				var queries = [];
				dojo.forEach(this._transactionData, dojo.hitch(this, function __buildQueries(item, index, items){
					queries.push(item.requestQuery);
				}));
				var formData = {"queries" : queries};
				var args = {
						url: this._bulkTarget,
						headers: {
							"Accept" : "application/json",
							"Content-Type" : "application/json"
						},
						handleAs: "json",
						postData:  dojo.toJson(formData),
						preventCache: true
				};
				var post = jazz.client.xhrPost(args);
				post.then(dojo.hitch(this, function __processResults(results){
					dojo.forEach(this._transactionData, dojo.hitch(this, function __processResult(transaction, index, transactions){
						var uri = transaction.requestQuery[this._idProperty];
						if(results[uri]){
							this._isBulkTransaction = false;
							var data = results[uri];
							this.put(data, uri);
							transaction.requestDeferred.resolve(data);
						} else {
							transaction.requestDeferred.resolve(null);
						}
					}));
					this._isBulkTransaction = false;
					lock.resolve(true);
					this._locks.splice(this._locks.length - 1, 1);
				}), dojo.hitch(this, function __processResultError(error){
					// On error release the current lock.
					this._isBulkTransaction = false;
					dojo.forEach(this._transactionData, dojo.hitch(this, function __rejectTransactions(transaction, index, transactions){
						transaction.requestDeferred.reject(error);
					}));
					this._locks.splice(this._locks.length - 1, 1);
				}));
			}), dojo.hitch(this, function __acquireLockError(error){
				// On error, end the bulk transaction
				this.endBulkTransaction(error);
			}));
		}));
		return locked;
	},
	
	/**
	 * Ends a bulk transaction and fires the request to the server.
	 */
	endBulkTransaction: function __endBulkTransaction(error){
		if(error){
			this._transaction.reject(error);
		} else {
			this._transaction.resolve()
		}
	},
	
	_addBulkRequest: function __addBulkRequest(query){
		var deferred = new dojo.Deferred();
		var args = {
			requestDeferred: deferred,
			requestQuery: query
		};
		this._transactionData.push(args);
		return deferred;
	},
});

dojo.declare("com.ibm.team.vvc.shared.web.ui.util.VvcLocalCache", [], {

	_storageKey: null,
	_data: null,
	_timeout: 60000,	// Timeout in MS.  Default is 1 minute
	_temp: null,
	_errorTime: null,
	
	constructor: function __VvcLocalCache(args){			
		this._storageKey = (args && args.storageKey) ? args.storageKey : "VvcSesseionCache-"; 
		if(args){
			this._temp = args.temporary && true;
			if(args.timeout){
				this._timeout = args.timeout;
			}
		}
		
		if(!this._temp){
			var value = sessionStorage.getItem(this._storageKey);
			if (value) {
				this._data = dojo.fromJson(value);
			}
		}
				
		if(!this._data){
			this._data = {};
		}

	},
	
	get: function __get(id){
		var item = this._data[id];
		if(item){
			var now = new Date();
			if(item.expires > now.getTime()){
				return item.data;
			}
		}
		return null;
	},
	
	put: function __put(data, id, options){
		var now = new Date();
		var expires = now.getTime() + (options && options.timeout ? options.timeout : this._timeout);
		var item = {
			"id": id,
			"data": data,
			"expires": expires
		};
		this._data[id] = item;
		this.flush();
	},
	
	clear: function __clear(id){
		if(id){
			if(this._data[id]){
				delete this._data[id];
			}
		} else {
			for(var id in this._data){
				delete this._data[id];
			}
			this._data = {};
		}
		this.flush();
	},
	
	flush: function __flush(){
		if(!this._temp){
			if(this._data == {}){
				sessionStorage.removeItem(this._storageKey);
			} else {
				var json = dojo.toJson(this._data);
				try {
					sessionStorage.setItem(this._storageKey, json);
				} catch (e) {
					var now = new Date().getTime();
					if (!this._errorTime || (now - this._errorTime > this._timeout)) {
						ncFactory.createWarning({
							title: Messages.flushErrorTitle,
							description: Messages.flushErrorDescription,
							solution: Messages.flushErrorSolution
						}).dispatch();
					}
					this._errorTime = now;
				}
			}
		}
	}

});
})();
