/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.diagrams.render.util.LabelHelper");
dojo.require("rmpw.diagrams.render.util.Rectangle");
dojo.require("rmpw.diagrams.render.util.Point");
dojo.require("rmpw.diagrams.render.util.Dimension");
dojo.require("rmpw.diagrams.render.util.PointList");
dojo.require("rmpw.diagrams.render.util.PointListUtilities");
/**
 * 
 * Helper function for Label views
 * 
 * @author Susanta Datta
 */

dojo.declare("rmpw.diagrams.render.util.LabelHelper", null, {
	
	/**
	 * Calculates the label offset from the reference point given the label bounds and a points list.  
	 * 
	 * @param label the <code>IFigure</code> to calculate the offset for
	 * @param bounds the <code>Rectangle</code> that is the bounding box of the label.
	 * @param points the <code>PointList</code> that contains that the label offset is relative to.
	 * @param ref the <code>Point</code> that is the reference point that the offset
	 * is based on.
	 * @return a <code>Point</code> which represents a value offset from the <code>ref</code>
	 * point oriented based on the nearest line segment.
	 */
	offsetFromRelativeCoordinate: function(bounds, points, /*Point*/ ref) {
		var rect = new rmpw.diagrams.render.util.Rectangle(bounds.x, bounds.y, bounds.width, bounds.height);
		
		//Componsate for the fact that we are using the 
		// figure center
		rect.x += rect.width /2;
		rect.y += rect.height /2;
		var normalPoint = this.normalizeRelativePointToPointOnLine(points, ref,
				new rmpw.diagrams.render.util.Point(rect.x - ref.x, rect.y - ref.y));
		
		return normalPoint;
	},
	

	/**
	 * Calculates the relative coordinate that is equivalent to the offset from the reference 
	 * point, that can be used to set the label location.
	 * 
	 * @param label the <code>IFigure</code> to calculate the relative coordinate for 
	 * @param points the <code>PointList</code> that contains that the label offset is relative to.
	 * @param ref a <code>Point</code> located on the parent which the offset value
	 * is relative to.
	 * @param offset a <code>Point</code> which represents a value offset from the <code>ref</code>
	 * point oriented based on the nearest line segment.
	 * @return a <code>Point</code> that is the relative coordinate of the label that can be
	 * used to set it's location.
	 */
	relativeCoordinateFromOffset: function(rect,  /*Point*/ ref, /*Point*/ offset, /*PointList*/ points) {
		var location = this.calculatePointRelativeToPointOnLine(points, ref, offset);
		location = location.translate(-1 * rect.width /2, -1 * rect.height /2);
		return location;
	},
	

	
	/**
	 * Returns a point located relative to the line by the given offset.
	 * 
	 * @param ptLst the point
	 * @param ptOnLine
	 * @param offset
	 * @return the relative point given the line angle
	 */
	calculatePointRelativeToPointOnLine: function(ptLst, /*Point*/ ptOnLine, /*Point*/ offset) {
		// Calculate slope of line
		if (ptLst.size === 1) {
			// This is a node...
			return ptLst.getPoint(0).getCopy().translate_1(offset);
		} else if (ptLst.size >= 2){
			// This is an edge...
			var util = new rmpw.diagrams.render.util.PointListUtilities();
			var index = util.findNearestLineSegIndexOfPoint(ptLst, ptOnLine);		
			var segment = util.getLineSegments(ptLst).pop(index - 1);
			var relativeOffset = null;
			if (segment) {
				var p = ptOnLine.getCopy();
				if (segment.isHorizontal()) {
					if (segment.origin.x > segment.terminus.x) {
						return p.translate(-offset.x, -offset.y);
					} else {
						return p.translate_1(offset);	
					}
				} else if (segment.isVertical()) {
					if (segment.origin.y > segment.terminus.y) {
						return p.translate_1(offset.getCopy().scale(-1, 1).transpose());
						//return relativeOffset;						
					} else {
						return p.translate_1(offset.getCopy().scale(1, -1).transpose());
					}
				} else {				
					var slope = segment.slope();
					var theta = Math.atan(slope);
					var normalizedOffset = new rmpw.diagrams.render.util.Point(offset.x, offset.y);
					if (segment.origin.x > segment.terminus.x) {
						normalizedOffset = offset.getCopy().scale(-1, -1);
					}

					var calculatedOffset = new rmpw.diagrams.render.util.Point(normalizedOffset.x
						* Math.cos(theta) - normalizedOffset.y
						* Math.sin(theta), normalizedOffset.x * Math.sin(theta)
						+ normalizedOffset.y * Math.cos(theta));
					return ptOnLine.getCopy().translate_1(calculatedOffset);
				}
			}
		}					
		return null;
	},
	
	/**
	 * Calculates the normalized offset from a point on a <code>Connection</code>'s point list to an point.
	 * 
	 * @param ptLst
	 * @param ptOnLine
	 * @param offset
	 * @return the normalized offset
	 */
	normalizeRelativePointToPointOnLine:function ( ptLst, /*Point*/ ptOnLine, /*Point*/ offset) {
		// Calculate slope of line
		if (ptLst.size === 1) {
			// This is a node...
			return offset;
		} else if (ptLst.size >= 2){
			// This is a edge...
			var util = new rmpw.diagrams.render.util.PointListUtilities();
			var index = util.findNearestLineSegIndexOfPoint(ptLst, ptOnLine);		
			var segment = util.getLineSegments(ptLst).pop(index - 1);
			var normalOffset = null;
			if (segment) {
				if (segment.isHorizontal()) {
					if (segment.origin.x > segment.terminus.x) {
						normalOffset = offset.getCopy().scale(-1, -1);
						return normalOffset;
					} else {
						normalOffset = offset;
						return normalOffset;						
					}					
				} else if (segment.isVertical()) {
					if (segment.origin.y < segment.terminus.y) {
						normalOffset = offset.scale(-1, 1).transpose();
						return normalOffset;
					} else {
						normalOffset = offset.scale(1, -1).transpose();
						return normalOffset;
					}
				} else {
					var p = ptOnLine.getCopy();
					p.translate_1(offset);
					normalOffset = this.getOrthogonalDistances(segment, ptOnLine, p);	
					return normalOffset;										
				}
			}
		}					
		return null;

	},
	
	/**
	 * Calculates distances from a <code>Point</code> on a <code>LineSeg</code> to 
	 * another <code>Point</code>.  The sign of the distances indicate direction.
	 * 
	 * @param lineSeg
	 * @param ptOnLine
	 * @param refPoint
	 * @return the distance from <code>Point</code> on a <code>LineSeg</code> to another <code>Point</code>
	 */
	getOrthogonalDistances: function(lineSeg, /*Point*/ ptOnLine, /*Point*/ refPoint) {
		var parallelSeg = lineSeg.getParallelLineSegThroughPoint(refPoint);
		var p1 = parallelSeg.perpIntersect(ptOnLine.x, ptOnLine.y);
		var dx = p1.getDistance(refPoint) * ((p1.x > refPoint.x) ? -1 : 1);
		var dy = p1.getDistance(ptOnLine) * ((p1.y < ptOnLine.y) ? -1 : 1);
		var orth = new rmpw.diagrams.render.util.Point(dx, dy);	
		// Reflection in the y axis		
		if (lineSeg.origin.x > lineSeg.terminus.x){
			orth = orth.scale(-1, -1);
		}
		return orth;
	}
});
