/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
dojo.provide("rmpw.diagrams.render.util.LineSeg");

dojo.require("rmpw.diagrams.render.util.Point");

dojo.declare("rmpw.diagrams.render.util.LineSeg", null, {
	
	constructor: function(start, end) {
		this.origin = new rmpw.diagrams.render.util.Point(start.x, start.y);
		this.terminus = new rmpw.diagrams.render.util.Point(end.x, end.y);
	},
	
	/**
	 * Creates a segment using (fromX, fromY) as either the 
	 * first point of the segment (start == Origin) or the midpoint of the 
	 * segment (start == Midpoint), and using slope as its new slope and len as
	 * the new length.  xdir indicates which direction the segment should
	 * go in the x-axis.
	 * 
	 *  start KeyPoint from which the other parameters are relative to
	 *  fromX  x value of start KeyPoint
	 *  fromY  y value of start KeyPoint
	 *  slope float slope of the line
	 *  len  length of the line
	 *  xdir direction
	 */
	constructor: function(
		keyPtstart,
		fromX,
		fromY,
		slope,
		len,
		xdir) {
		var dx, dy;
		var dx_float;
		var len_squared;
		
		//TODO find better way of eliminating overloaded constructors
	if(arguments.length === 2){
		this.origin = new rmpw.diagrams.render.util.Point(arguments[0].x, arguments[0].y);
		this.terminus = new rmpw.diagrams.render.util.Point(arguments[1].x, arguments[1].y);
	}else{
		this.origin = new rmpw.diagrams.render.util.Point(0, 0);
		this.terminus = new rmpw.diagrams.render.util.Point(0, 0);
		// Find the delta y and x needed to get to the end points.  See
		// pointOn() for explanation of these equations
		if (keyPtstart === "ORIGIN") {
			len_squared = len * len;
		} else // start == Midpoint
			{
			len_squared = len / 2.0 * len / 2.0;
		}

		var slope_squared = slope * slope;
		dx_float = Math.sqrt(len_squared / (slope_squared + 1.0));

		// Set which direction the segment should go in the x direction.
		// The y direction will get set automatically based on slope
		// and the dx.

		dx_float *= xdir;
		dx = (dx_float + 0.5);

		dy = ((slope * dx_float) + 0.5);

		if (keyPtstart === "ORIGIN") {
			this.origin.x = fromX;
			this.origin.y = fromY;
		} else // start == DirectedLine::Midpoint
			{
			this.origin.x = fromX - dx;
			this.origin.y = fromY - dy;
		}

		this.terminus.x = fromX + dx;
		this.terminus.y = fromY + dy;
	  }
	},
	
	performScale: function (amount) {
		this.origin.performScale(amount);
		this.terminus.performScale(amount);
		return this;
	},
	perpSlope: function () {
		var m = this.slope();
		if (m === 0.0){
			return 9999;
		}else{
			return - (1.0 / m);
	    }
	},
	length: function () {
		return this.origin.getDistance(this.terminus);
	},
	
	transform: function (surfaceMatrix) {
		this.origin.transform(surfaceMatrix);
		this.terminus.transform(surfaceMatrix);
		return this;
	},
	translate: function(dx, dy) {
		this.origin.translate(dx, dy);
		this.terminus.translate(dx, dy);
		return this;
	},
	equals$: function (seg){
		if (!(seg && seg.origin && seg.terminus && seg.getCopy)){
			return false;
        }
		return this.origin.getCopy().equals$(seg.origin.getCopy()) && this.terminus.getCopy().equals$(seg.terminus.getCopy());
	},

	hashCode$: function (){
		return $hashCode_2(this.origin.getCopy()) ^ $hashCode_2(this.terminus.getCopy());
	},
	
	perpIntersect: function (startX, startY){
		var fx, m, ptResult;
		ptResult = new rmpw.diagrams.render.util.Point();
		m = this.slope();
		fx = (m * startY - m * this.origin.y + m * m * this.origin.x + startX) / (m * m + 1);
		
		if (Math.abs(m) <= 0.0001) {
			ptResult.y = this.origin.y;
		}
		else {
			ptResult.y = startY + (startX - fx) / m;
		}
		ptResult.x = fx;
		return ptResult;
	},
	
	/**
	 * Calculates the projection of the given point onto the line segment.
	 * 
	 *  xCoord the x coordinate of the point.
	 *  yCoord the y coordinate of the point.
	 * return  value of the calculated projection.
	 */
	projection: function ( xCoord, yCoord) {
		var dirx = this.terminus.x - this.origin.x;
		var diry = this.terminus.y - this.origin.y;

		/* get q - p */
		var qpx = xCoord - this.origin.x;
		var qpy = yCoord - this.origin.y;
		var dotprod = qpx * dirx + qpy * diry;

		/* avoid divide by 0 - check if point1 equals point2.  If so,
		   there is no segment - return a value which indicates projection
		   falls outside the segment. */
		if (dirx === 0 && diry === 0){
			return -1.0;
        }
		/* length (magnitude) of d is sqrt(dirx^2 + diry^2).  Don't
		   bother taking square root since we want the length squared. */
		return (dotprod / (dirx * dirx + diry * diry));
	},

	distanceToPoint: function (xCoord, yCoord) {
		
		var proj = this.projection(xCoord, yCoord);

		if (proj > 0 && proj < 1) {
			var pt = this.perpIntersect(xCoord, yCoord);
			return Math.round(pt.getDistance(new rmpw.diagrams.render.util.Point(xCoord, yCoord)));
		}

		var d1 = Math.round(this.origin.getDistance(new rmpw.diagrams.render.util.Point(xCoord, yCoord)));
		var d2 = Math.round(this.terminus.getDistance(new rmpw.diagrams.render.util.Point(xCoord, yCoord)));
		
		return (d1 < d2 ? d1 : d2);
	},
	
	setTerminus: function (terminus){
		this.terminus = terminus.getCopy();
	},

	slope: function (){
		if (this.origin.x === this.terminus.x){
			return 9999;
		}
		return (this.terminus.y - this.origin.y) / (this.terminus.x - this.origin.x);
	},
	
	/**
	 * Returns the coefficients of the generalized equation of the line passing through
	 * points (x1,y1) and (x2,y2)
	 */
	getLineEquation: function(x1, y1, x2, y2) {
		var equation = [];
		for (var i=0; i<3; i++){
			equation[i]=0;
		}
		if (x1 === x2 && y1 === y2){
			return equation;
		}
		if (x1 === x2) {
			equation[0]=1;
			equation[1]=0;
			equation[2]=x1;
			return equation;
		}
		
		equation[0]=(y1-y2)/(x2-x1);
		equation[1]=1.0;
		equation[2]=y2+equation[0]*x2;
		return equation;
	},

	/**
	 * Returns array with 3 numbers in it, which are the coefficients of the
	 * generalized line equation of the line corresponding to this line segment
	 * a*x+b*y=c is the equation => result[0]=a, result[1]=b, result[2]=c
	 */
	getEquation: function() {
		return this.getLineEquation(this.origin.x, this.origin.y, this.terminus.x, this.terminus.y);
	},
	
	/**
	 * Returns intersection points of two lines that contain this line segment and
	 * the argumet line segment.
	 */
	getLinesIntersections: function(line) {
		var intersections = new rmpw.diagrams.render.util.PointList();
		var temp = this.getEquation();
		var a1 = temp[0];
		var b1 = temp[1];
		var c1 = temp[2];

		temp = line.getEquation();
		var a2 = temp[0];
		var b2 = temp[1];
		var c2 = temp[2];
		// Cramer's rule for the system of linear equations
		var det = a1*b2 - b1*a2;
		if (det === 0) {
			if (a1===a2 && b1===b2 && c1===c2) {
				var points = [];
				points.push(this.origin);
				points.push(this.terminus);
				points.push(line.origin);
				points.push(line.terminus);
				
				points.sort( function(a,b) {
					if (a.equals$(b)) {
						return 0;
					} else if (a.x < b.x
							|| (a.x === b.x && a.y < b.y)) {
						return -1;
					} else {
						return 1;
					}
				});
				
				// if lines are the same, then instead of infinite number of intersections
				// we will put the 2 points out of 4 points - ends of 2 segments. Look above.
				intersections.addPoint_1(points[1]);
				intersections.addPoint_1(points[2]);
			}
		}
		else {
			intersections.addPoint_1(new rmpw.diagrams.render.util.Point((c1*b2-b1*c2)/det, (a1*c2-c1*a2)/det));
		}
		return intersections;
	},
	
	getLineIntersectionsWithLineSegs: function(points) {
		if(points.size <= 0) {
			return null;
		}
		
		var intersections = new rmpw.diagrams.render.util.PointList();
		if (points.size < 2) {
			if (this.containsPoint(points.getFirstPoint(), 1)) {
				intersections.addPoint_1(points.getFirstPoint());
			}
		} else {	
			for (var i=0; i<points.size-1; i++) {
				var seg = new rmpw.diagrams.render.util.LineSeg(points.getPoint(i), points.getPoint(i+1));
				var currentIntersections = this.getLinesIntersections(seg);
				for (var j=0; j<currentIntersections.size; j++) {
					var intersection = currentIntersections.getPoint(j);
					if (seg.containsPoint(intersection, 1)) {
						intersections.addPoint_1(currentIntersections.getPoint(j));
					}
				}
			}
		}
		return intersections;
	},
	
	/**
	 * Calculates intersection points of the line of the line segment and ellipse 
	 */
	getLineIntersectionsWithEllipse: function(ellipseBounds) {
		var intersections = new rmpw.diagrams.render.util.PointList();
		var preciseOrigin = this.origin.getCopy();
		var preciseTerminus = this.terminus.getCopy();
		var preciseEllipseBounds = ellipseBounds.getCopy();
		if (preciseEllipseBounds.width === 0 || preciseEllipseBounds.height === 0){
			return intersections;
		}
		var ellipsePreciseCenter = preciseEllipseBounds.getCenter();
		var xl1 = preciseOrigin.x - ellipsePreciseCenter.x;
		var xl2 = preciseTerminus.x - ellipsePreciseCenter.x;
		var yl1 = preciseOrigin.y - ellipsePreciseCenter.y;
		var yl2 = preciseTerminus.y - ellipsePreciseCenter.y;
		var equation = this.getLineEquation(xl1, yl1, xl2, yl2);
		
		if (equation.length<3 || (equation[0] === 0 && equation[1] === 0)){
			return intersections;
		}
		var a = equation[0];
		var b = equation[1];
		var c = equation[2];
		var w = preciseEllipseBounds.width;
		var h = preciseEllipseBounds.height;
		
		// Ellipse with a center at the origin has an equation:
		// (h*x)^2+(w*y)^2=(h*w/2)^2
		// Line equation: a*x+b*y=c
		
		if (b===0) {
			// b==0 is a special case since in general case we will express
			// y in terms of x, i.e. we need to divide by b, which should not
			// be 0
			// b==0 => a*x=c +> x=c/a;
			var x = c/a;
			// y^2 = (h/2)^2-((h*c)/(a*w))^2
			var y = Math.pow(h/2,2)-Math.pow((h*c)/(a*w),2);
			if (y<0){
				return intersections;
			}
			intersections.addPoint_0(x+ellipsePreciseCenter.x, Math.sqrt(y)+ellipsePreciseCenter.y);
			intersections.addPoint_0(x+ellipsePreciseCenter.x, -Math.sqrt(y)+ellipsePreciseCenter.y);
		}
		else {
			// y = (c-a*x)/b => we get quadratic equation for x
			// x^2*(h^2+(w*a/b)^2)-x*(2*w^2*a*c)/(b^2)+((w*c/b)^2-(h*w/2)^2)=0 or
			// x^2*xA+x*xB+xC=0
			var xA = Math.pow(h,2)+Math.pow((w*a)/b,2);
			var xB = (-2)*Math.pow(w,2)*a*c/Math.pow(b,2);
			var xC = Math.pow(w*c/b,2)-Math.pow(h*w/2,2);
			var xD = Math.pow(xB,2)-4*xA*xC;
			
			if (xD<0){
				return intersections;
			}
			var x1 = (-xB+Math.sqrt(xD))/(2*xA);
			var x2 = (-xB-Math.sqrt(xD))/(2*xA);
			intersections.addPoint_0(x1+ellipsePreciseCenter.x, (c-a*x1)/b+ellipsePreciseCenter.y);
			intersections.addPoint_0(x2+ellipsePreciseCenter.x, (c-a*x2)/b+ellipsePreciseCenter.y);
		}
		
		return intersections;
	},
	
	/**
	 * Checks if this line segment contains the given point within a tolerance value.
	 */
	containsPoint: function(aPoint, tolerance) {
		/*
		 * We need perform the calculations in double numbers to avoid possible integer
		 * overflows in Point#getDistance2() method
		 */
		var lengthOfSegment = Math.sqrt((this.origin.x - this.terminus.x)*(this.origin.x - this.terminus.x) + (this.origin.y - this.terminus.y)*(this.origin.y - this.terminus.y));
		var lengthFromOriginToPoint = Math.sqrt((this.origin.x - aPoint.x)*(this.origin.x - aPoint.x) + (this.origin.y - aPoint.y)*(this.origin.y - aPoint.y));
		var lengthFromTerminusToPoint = Math.sqrt((this.terminus.x - aPoint.x)*(this.terminus.x - aPoint.x) + (this.terminus.y - aPoint.y)*(this.terminus.y - aPoint.y));
		
		return lengthFromTerminusToPoint + lengthFromOriginToPoint - lengthOfSegment <= tolerance;
	},
	/**
	 * Locates a point at a given height and distance along the line segment.
	 * 
     *                   B (the point we are looking for)
	 *					 +
	 *					 |
	 *			  dist   |h		  this segment
	 *		P1-----------+------------------->
	 *					 A
	 *	
	 *	 get point A (on picture above)
	 *
	 *  pctDist distance along the line
	 *  theHeight height above the line
	 *  asOriented indicating relative position of the point to be located
	 *  value that was located on the line.
	 */
	locatePoint: function(
		pctDist,
		theHeight,
		asOriented) {

		var xdir;
		var dist = (pctDist * this.length());
		var pt = new rmpw.diagrams.render.util.Point();
		this.pointOn(dist, "ORIGIN", pt); // (x,y) now = A

		// get linesegment AB
		// first determine the direction AB should go in the x axis.  Don't ask-
		// just have faith.

		if (this.origin.y > this.terminus.y
			|| (this.origin.y === this.terminus.y && this.origin.x < this.terminus.x)) {
			xdir = (asOriented === "POSITIVE" ? -1 : 1);
		} else {
			xdir = (asOriented === "POSITIVE" ? 1 : -1);
		}

		var linesegAB =
			new rmpw.diagrams.render.util.LineSeg("ORIGIN", pt.x, pt.y, this.perpSlope(), theHeight, xdir);

		return (new rmpw.diagrams.render.util.Point(linesegAB.terminus.x, linesegAB.terminus.y));
	},
	/**
	 * Gets the point on the line segment at the given distance away from the 
	 * key point.
	 * 
	 *  theDistance distance along the line
	 *  fromKeyPoint to calculate the distance from
	 *  ptResult where the resulting calculating value is stored.
	 * return true if point can be calculated, 
	 * false otherwise.
	 */
	pointOn: function (
		theDistance,
		fromKeyPoint,
		ptResult) {
		var m, dx_float;
		var dx, dy, startX = 0, startY = 0, otherX = 0, otherY = 0;

		// Set the point to offset from and the other point used to determine
		// which direction dx and dy should be applied to get a point on the
		// line.

		if (fromKeyPoint === "ORIGIN") {
			startX = this.origin.x;
			startY = this.origin.y;
			otherX = this.terminus.x;
			otherY = this.terminus.y;
		}
		else if (fromKeyPoint === "TERMINUS") {
			startX = this.terminus.x;
			startY = this.terminus.y;
			otherX = this.origin.x;
			otherY = this.origin.y;
		}
		else if (fromKeyPoint === "MIDPOINT") {
			startX = (this.origin.x + this.terminus.x) / 2;
			startY = (this.origin.y + this.terminus.y) / 2;
			otherX = this.terminus.x;
			otherY = this.terminus.y;
		}
		else {
			return false;
		}

		m = this.slope(); // get the slope of this line

		// Find dx and dy - the delta x and y to get from the endpoint to the
		// point on the line at the specified distance away.
		// The following is based on solving 2 equations with 2 unknowns:
		//   dy/dx = m	  (m is slope of line)
		//   dy^2 + dx^2 = dist^2
		//
		var d_squared = theDistance * theDistance;
		var m_squared = m * m;

		// Add .5 so result is rounded to nearest integer when cast
		dx_float = Math.sqrt(d_squared / (m_squared + 1.0));
		dx = (dx_float + 0.5);
		dy = (Math.sqrt(d_squared * m_squared / (m_squared + 1.0)) + 0.5);

		/* negative distance means we want point off the line */
		if (theDistance < 0) {
			dx = -dx;
			dy = -dy;
		}

		ptResult.x = ((startX > otherX) ? startX - dx : startX + dx);
		ptResult.y = ((startY > otherY) ? startY - dy : startY + dy);
		var in_line;
		if (startX > otherX){
			in_line = ptResult.x >= otherX;
		}else{
			in_line = ptResult.x <= otherX;
		}
		if (in_line) {
			if (startY > otherY){
				in_line = ptResult.y >= otherY;
			}else{
				in_line = ptResult.y <= otherY;
		    }
		}
		return in_line;
	},
	/**
	 * Determines if this a horizontal segment
	 * 
	 * @return <code>boolean</code> <code>true</code> if horizontal, 
	 * <code>false</code> otherwise.
	 */
	 isHorizontal: function() {
		return (this.origin.y === this.terminus.y);
	},

	/**
	 * Determines if this a vertical segment
	 * 
	 * @return <code>boolean</code> <code>true</code> if vertical, 
	 * <code>false</code> otherwise.
	 */
	isVertical: function() {
		return (this.origin.x === this.terminus.x);
	},
	/**
	 * Returns out a positive or negative value (Positive / Negative) depending
	 * on the orientation of the given point to the line.
	 * 
	 *      Point on this side: Positive.				 
	 *
	 *      P1------------------------------>
	 *					this segment
	 *
	 *                 Point on this side: Negative.
	 * 
	 * @param rel <code>Point</code> to test the relative position against this line.
	 * @return <code>Sign</code> value indicating the relative position of the given point.
	 */
	positionRelativeTo: function(rel) {
		var ptRelRay = new rmpw.diagrams.render.util.Ray(this.origin, rel);

		var val = this.getTrigValues(ptRelRay);
		if (val) {
			var dNewAngle = Math.atan2(-val.sinTheta, -val.cosTheta);
			if (dNewAngle > 0){
				return "POSITIVE";
		    }
		}

		return "NEGATIVE";
	},
	/**
	 * Gets the trig values associated with the angle from this line segment
	 * to the given vector.
	 * 
	 * @param ptToVector <code>Ray</code> value to calculate trig values of.
	 * @return <code>TrigValues</code> object representing the trigonometry values
	 * for the angle of the passed in <code>Ray</code> relative to <code>this</code>
	 * or null if calculation is not possible, 
	 */
	getTrigValues: function (ptToVector) {
		var dFromLength = this.length();
		var dToLength = ptToVector.length();

		var ptFromVector = new rmpw.diagrams.render.util.Ray(this.origin, this.terminus);

		if (dFromLength <= 0 || dToLength <= 0) {
			return null;
		}

		// 1. find angle for ptToVector relative to the origin.
		var dAlpha;
		var dCosAlpha, dSinAlpha;

		dCosAlpha = ptFromVector.x / dFromLength;
		dSinAlpha = ptFromVector.y / dFromLength;
		dAlpha = Math.atan2(dSinAlpha, dCosAlpha);

		// 2. inverse the angle to get the rotation
		dCosAlpha = Math.cos(-dAlpha);
		dSinAlpha = Math.sin(-dAlpha);

		// 3. rotate vector 2 by angle above so that it's angle relative to vector 1 can
		// be calculated
		var dRotateX = (ptToVector.x * dCosAlpha) - (ptToVector.y * dSinAlpha);
		var dRotateY = (ptToVector.x * dSinAlpha) + (ptToVector.y * dCosAlpha);

		// 4. Now calculate the Theta trig values
		var val = {cosTheta: dRotateX / dToLength, sinTheta: dRotateY / dToLength};
//		val.cosTheta = dRotateX / dToLength;
//		val.sinTheta = dRotateY / dToLength;

		return val;
	},
	
	
	/**
	 * Returns a new <code>LineSeg</code> that is parallel to this by the given distance.  
	 * Orientation is relative to the start and end.  Negative implies to the 
	 * left and Position implies to the right.
	 * 
	 * @param ptLoc <code>Point</code> value to constrain the line to.
	 * @return <code>LineSeg</code> line that was calculated going through the given point
	 */
	getParallelLineSegThroughPoint : function(ptLoc) {
		if (this.isHorizontal()) {
            return new rmpw.diagrams.render.util.LineSeg(
                    new rmpw.diagrams.render.util.Point(this.origin.x, ptLoc.y), 
                    new rmpw.diagrams.render.util.Point(this.terminus.x, ptLoc.y));
		}
        else if (this.isVertical()) {
            return new rmpw.diagrams.render.util.LineSeg(
                    new rmpw.diagrams.render.util.Point(ptLoc.x, this.origin.y), 
                    new rmpw.diagrams.render.util.Point(ptLoc.x, this.terminus.y));
        }
        else {
            var ptProj = this.perpIntersect(ptLoc.x, ptLoc.y);
            var nHeight = Math.round(ptProj.getDistance(ptLoc));
            var position = this.positionRelativeTo(ptLoc); 
	
            return new rmpw.diagrams.render.util.LineSeg(
			     this.locatePoint(0.0, nHeight, position),
			     this.locatePoint(1.0, nHeight, position));
        }
	}
	
	

});