/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/


dojo.provide("rmpw.diagrams.render.util.PointListUtilities");
dojo.require("rmpw.diagrams.render.util.Point");
dojo.require("rmpw.diagrams.render.util.PointList");
dojo.require("rmpw.diagrams.render.util.LineSeg");

dojo.declare("rmpw.diagrams.render.util.PointListUtilities", null, {

	/**
	 * Method findNearestLineSegIndexOfPoint.
	 * Calculate the nearest line segment index distance wise to the given point.
	 * 
	 * @param points PointList to calculate the nearest line segment of.
	 * @param ptCoord the <code>Point</code> to test containment of.
	 * @return int Index of line segment that is nearest in the polyline to the given point.
	 * The index is 1 based where 1 represents the first segment.
	 */
	findNearestLineSegIndexOfPoint: function(points,ptCoord) {
		var mySegments = this.getLineSegments(points);
		var nNextIndex = 0;
		var nMinIndex = 0;
		var minDistance = 32766;//BIGDISTANCE;
		var nextDistance = 0;
		for(var i=0; i<mySegments.length; i++){
			var aSegment = mySegments[i];
			nNextIndex++;

			nextDistance = aSegment.distanceToPoint(ptCoord.x, ptCoord.y);

			if (nextDistance < minDistance) {
				minDistance = nextDistance;
				nMinIndex = nNextIndex;
			}
		}

		return nMinIndex;
	},
	/**
	 * Method getLineSegments.
	 * Converts the points of this polyline into a list of LineSeg
	 *      objects
	 * 
	 * @param points PointList to get LineSeg equivalents of.
	 * @return List of LineSeg objects.
	 */
	getLineSegments: function(points) {
		var lines = [];
		if (points.size <= 1)
			return lines;
		for (var i = 0; i < points.size - 1; i++) {
			lines.push(new rmpw.diagrams.render.util.LineSeg(points.getPoint(i), points.getPoint(i + 1)));
		}

		return lines;
	},
	
	/*
	 * gets the polyline equivalent of a rectangle
	 */
	getRectanglePolyline: function(rect) {
		var polyRect = new rmpw.diagrams.render.util.PointList();
		polyRect.addPoint_1(new rmpw.diagrams.render.util.Point(rect.x, rect.y));
		polyRect.addPoint_1(new rmpw.diagrams.render.util.Point(rect.x + rect.width, rect.y));
		polyRect.addPoint_1(new rmpw.diagrams.render.util.Point(rect.x + rect.width, rect.y + rect.height));
		polyRect.addPoint_1(new rmpw.diagrams.render.util.Point(rect.x, rect.y + rect.height));
		polyRect.addPoint_1(new rmpw.diagrams.render.util.Point(rect.x, rect.y));
		return polyRect;
	},

	/**
	 * Finds a point relative to the pointList passed in based on the parameters passed in.
	 * 
	 * @param pointList the <code>PointList</code> 
	 * @param fromLine distance off the line 
	 * @param fromEnd distance from the starting point of the line (i.e. distance from source end along the line)
	 * @return Point
	 */

	calculatePointRelativeToLine: function(pointList,fromLine,fromEnd, isPercentage){
		var fractionDistance = 0;
		if (isPercentage) {
			fractionDistance = fromEnd / 100.0;
		} else {
			fractionDistance = fromEnd / pointList.size;
		}
		var locateInfo = this.locateSegment(
				this.getLineSegments(pointList),
				fractionDistance);
		if (locateInfo.theSegment) {
			var inSegPercDist = 0;
			var seg = locateInfo.theSegment;
			if (seg.length() > 0) {
				inSegPercDist = locateInfo.remainingDist /	seg.length();
			}			
			// relative position is coded as the the sign of the height
			var location =
				seg.locatePoint(
					inSegPercDist,
					Math.abs(fromLine),
					(fromLine > 0
						? "POSITIVE"
						: "NEGATIVE"));
			return location;
		}
		return null;
		},
		/**
		 * Find the line segment of this polyline that includes the given
		 * percentage of the length of the polyline from the given keypoint.
		 * Return a pointer to the segment and the portion of the length
		 * of the return segment to complete the distance of the total length.
		 * The pctDist must be between 0.0 and 1.0 if the keypoint is either
		 * the origin or the terminus; pctDist must be between 0.0 and 0.5
		 * if the keypoint is the midpoint.
		 * 
		 * @param mySegments
		 * @param pctDist
		 * @param fromKeyPoint the constant value indicating the key point
		 *      Origin, Terminus, Midpoint
		 * @param locateInfo LocateInfo where the calculated info is stored for return.
		 * @return boolean true if the segment could be located, false otherwise.
		 */
		locateSegment: function(
			mySegments,
			pctDist) {
			var locateInfo = {theSegment:null, remainingDist:null};
			var thePctDist = pctDist;
			if (pctDist < 0.0) {
				thePctDist = 0.0;
			} else if (1.0 < pctDist) {
				thePctDist = 1.0;
			}

			var theLength = this.length(mySegments);
			var remainingLength = Math.round(thePctDist * theLength);
			var nextLength = 0;

			locateInfo.theSegment = null;
			for(var i=0; i<mySegments.length; i++){
				var aSegment = mySegments[i];
				nextLength = Math.round(aSegment.length());
				if (nextLength >= remainingLength) {
					locateInfo.theSegment = aSegment;
					break;
				} else {
					remainingLength -= nextLength;
				}
			}
			locateInfo.remainingDist = remainingLength;
			return locateInfo;
		},
		/**
		 * Method length.
		 * Static utility function to calculate the length of a series of
		 * line segments.
		 * @param mySegments List of line segments to calculate the length of.
		 * @return long length of the line segments
		 */
		length: function(segments) {
			var theLength = 0;
			for(var i=0; i<segments.length; i++){
				theLength += Math.round(segments[i].length());
			}

			return theLength;
		},
		
		/**
		 * Method calcSmoothPolyline.
		 * Calculate the smooth polyline approximation of this polyline based on 
		 *      a smooth factor.
		 * 
		 * @param points the <code>PointList</code> that is used to calculate the
		 * smooth point list from.  
		 * @param nSmoothFactor the <code>int</code> smooth factor to smooth the line with 
		 *          0 - None, 15 - some, 30 - lots
		 * @param nBezierSteps the <code>int</code> number of line steps used to approximate the smooth curve
		 * @return PolylinePointList List of PolylinePoint representing the smooth 
		 *      polyline.
		 */
		calcSmoothPolyline: function(points, nSmoothFactor, nBezierSteps) {
			var theBezierPoints =
				this.calcBezier(points, nSmoothFactor, 0, points.size - 1);
			return this.calcApproxPolylineFromBezier( theBezierPoints, nBezierSteps);
		},

		MIN_LINE_LENGTH: 5,
		DEFAULT_BEZIERLINES: 16,
		MAX_BEZIERLINES: 32,
		SMOOTH_FACTOR_LESS: 15,
	    SMOOTH_FACTOR_NORMAL: 30,
	    SMOOTH_FACTOR_MORE: 50,
	    
		/**
		 * Calculate the actual bezier points of this polyline based on a smooth factor.
		 * 
		 * @param points PointList to calculate the bezier approximation from
		 * @param nSmoothFactor the <code>int</code> smooth factor to smooth the line with 
		 *          0 - None, 15 - some, 30 - lots
		 * @param nStartIndex int Index to start the calculation at
		 * @param nEndIndex int Index to end the calculation at
		 * @return PolylinePointList List of PolylinePoint representing the bezier.
		 */
		calcBezier: function(points, nSmoothFactor, nStartIndex, nEndIndex) {
			var ptPrevControl = null;
			var theBezierPoints = new rmpw.diagrams.render.util.PointList();
			// parse through the line segments and create control points based on a smoothing factor.
			var theSegments = this.getLineSegments(points);

			for (var i = 0, lg=theSegments.length; i < lg; i++) {
				var pLineSeg = theSegments[i];

				var dLineLength = pLineSeg.length();
				var dControlLength = dLineLength * nSmoothFactor / 100;

				var bAddToBezier = false;
				if (i >= nStartIndex && i <= nEndIndex) {
					bAddToBezier = true;
				} else if (i > nEndIndex) {
					return theBezierPoints;
				}
				if (dLineLength > this.MIN_LINE_LENGTH) {
					var ptStartControl = new rmpw.diagrams.render.util.Point();
					var ptTerminusControl = new rmpw.diagrams.render.util.Point();
					var ptStart = new rmpw.diagrams.render.util.Point(pLineSeg.origin.x, pLineSeg.origin.y);
					var ptTerminus = new rmpw.diagrams.render.util.Point(pLineSeg.terminus.x, pLineSeg.terminus.y);

					if (theBezierPoints.size === 0) {
						if (bAddToBezier) {
							theBezierPoints.addPoint_1(ptStart);
						}
					}

					if (ptPrevControl) {
						var prevControlSeg =
							new rmpw.diagrams.render.util.LineSeg(ptPrevControl, ptStart);
						ptStartControl = new rmpw.diagrams.render.util.Point();
						prevControlSeg.pointOn(
							Math.round(prevControlSeg.length() + dControlLength),
							"ORIGIN",
							ptStartControl);
					} else {
						ptStartControl = new rmpw.diagrams.render.util.Point();
						pLineSeg.pointOn(
							Math.round(dControlLength),
							"ORIGIN",
							ptStartControl);
					}
					if (bAddToBezier) {
						theBezierPoints.addPoint_1(ptStartControl);
					}

					// Calculate the next terminating control point
					var pNextSeg = null;
					if (i + 1 < theSegments.length) {
						pNextSeg = theSegments[i + 1];
						while (pNextSeg !== null
							&& pNextSeg.length() < this.MIN_LINE_LENGTH) {
							i++;
							if (i + 1 < theSegments.length){
								pNextSeg = theSegments[i + 1];
							}else{
								pNextSeg = null;
							}
						}
					}

					if (pNextSeg !== null) {
						// compute two vectors and calculate the angle between them.
						var ptVector1 =
							new rmpw.diagrams.render.util.Ray(pLineSeg.origin, pLineSeg.terminus);
						var ptVector2 =
							new rmpw.diagrams.render.util.Ray(pNextSeg.origin, pNextSeg.terminus);
						var dNewAngle = 0.0;

						var val = pLineSeg.getTrigValues(ptVector2);

						dNewAngle = Math.atan2(-val.sinTheta, -val.cosTheta);

						if (dNewAngle > 0) {
							dNewAngle = (Math.PI - dNewAngle) / -2;
						} else {
							dNewAngle = (-Math.PI - dNewAngle) / -2;
						}
						var ptVector1Prime = new rmpw.diagrams.render.util.Point(ptVector1.x
								* Math.cos(dNewAngle) - ptVector1.y
								* Math.sin(dNewAngle), ptVector1.x * Math.sin(dNewAngle)
								+ ptVector1.y * Math.cos(dNewAngle));
						
						var nextControlSeg =
							new rmpw.diagrams.render.util.LineSeg(
								new rmpw.diagrams.render.util.Point(0, 0),
								new rmpw.diagrams.render.util.Point(ptVector1Prime.x, ptVector1Prime.y));
						var ptProjection = new rmpw.diagrams.render.util.Point();
						nextControlSeg.pointOn(
							Math.round(dControlLength), "ORIGIN", ptProjection);

						ptTerminusControl =
							new rmpw.diagrams.render.util.Point(
								pLineSeg.terminus.x - ptProjection.x,
								pLineSeg.terminus.y - ptProjection.y);
					} else {
						pLineSeg.pointOn(
							Math.round(dLineLength - dControlLength), "ORIGIN", ptTerminusControl);
					}

					ptPrevControl = ptTerminusControl.getCopy();
					if (bAddToBezier) {
						theBezierPoints.addPoint_1(ptTerminusControl);
						theBezierPoints.addPoint_1(ptTerminus);
					}
				}
			}

			return theBezierPoints;
		},

		/**
		 * Utility function that takes a set of bezier points and calculates a polyline 
		 * approximation.
		 * 
		 * @param points the <code>PointList</code> to calculate the bezier from.
		 * @param nBezierSteps the <code>int</code> number of line steps that will be used
		 *      to approximate each bezier curve
		 * @return PolylinePointList List of PolylinePoint representing the smooth 
		 *      polyline.
		 */
		calcApproxPolylineFromBezier: function(points, nBezierSteps) {
			var thePolyPoints = new rmpw.diagrams.render.util.PointList();

			var ptCtl1, ptCtl2, ptCtl3, ptCtl4;
			var bStart = true;

			if (points.size < 4) {
				return thePolyPoints;
			}
			ptCtl1 = new rmpw.diagrams.render.util.Point();
			for (var i = 0; i < points.size - 3; i = i + 3) {
				if (bStart) {
					ptCtl1 = points.getPoint(i).getCopy();
					bStart = false;
				} else {
					thePolyPoints.removePoint(thePolyPoints.size - 1);
				}

				ptCtl2 = points.getPoint(i + 1).getCopy();
				ptCtl3 = points.getPoint(i + 2).getCopy();
				ptCtl4 = points.getPoint(i + 3).getCopy();

				if (!this.BezierToLines(thePolyPoints, ptCtl1, ptCtl2, ptCtl3, ptCtl4, nBezierSteps)) {
					return null;
				}
				ptCtl1 = new rmpw.diagrams.render.util.Point(ptCtl4.x, ptCtl4.y);
			}

			// now construct the PolyLine based on the points
			thePolyPoints.setPoint(points.getPoint(0), 0);
			thePolyPoints.setPoint(
				points.getPoint(points.size - 1),
				thePolyPoints.size - 1);

			return thePolyPoints;
		},

		/**
		 * Method BezierToLines.
		 * Utility function that takes a set of bezier points and calculates a polyline 
		 * approximation.
		 * 
		 * @param thePolyPoints
		 * @param ptCtl1
		 * @param ptCtl2
		 * @param ptCtl3
		 * @param ptCtl4
		 * @param nSteps
		 * @return boolean
		 */
		BezierToLines: function(thePolyPoints, ptCtl1, ptCtl2, ptCtl3, ptCtl4, nSteps) {
			var bRC = true;

			var nTotalPoints;
			/* total number of coordinate pairs in working arrays */
			var nLinePoints;
			/* total number of bezier endpoints in working arrays */
			var lpWorkX = []; /* pointer to X coordinate working array */
			var lpWorkY = []; /* pointer to Y coordinate working array */
			var i, j;

			nSteps = Math.min(this.MAX_BEZIERLINES, nSteps);

			/* enter original bezier X coordinates into X working array */
			lpWorkX[0] = ptCtl1.x;
			lpWorkX[1] = ptCtl2.x;
			lpWorkX[2] = ptCtl3.x;
			lpWorkX[3] = ptCtl4.x;

			/* enter original bezier Y coordinates into Y working array */
			lpWorkY[0] = ptCtl1.y;
			lpWorkY[1] = ptCtl2.y;
			lpWorkY[2] = ptCtl3.y;
			lpWorkY[3] = ptCtl4.y;

			/* initially 2 bezier endpoints and 4 coordinate pairs total */
			nLinePoints = 2;
			nTotalPoints = 4;

			/* generate enough bezier endpoints to satisfy requested number of steps */

			while (nLinePoints <= nSteps) {
				/* spread coordinate pairs in working array */

				for (i = nTotalPoints - 1; i > 0; i--) {
					lpWorkX[2 * i] = lpWorkX[i];
					lpWorkY[2 * i] = lpWorkY[i];
				}

				nTotalPoints = (2 * nTotalPoints) - 1;

				/* average consecutive coordinates and put average between coordinates */

				for (i = nTotalPoints - 2; i > 0; i -= 2) {
					lpWorkX[i] = (lpWorkX[i - 1] + lpWorkX[i + 1]) / 2;
					lpWorkY[i] = (lpWorkY[i - 1] + lpWorkY[i + 1]) / 2;
				}

				/* now average averages and store over control points */
				/* but do not overwrite bezier endpoints */

				for (i = nTotalPoints - 3; i > 0; i -= 2) {
					/* only control points */

					if ((i % 6) !== 0) {
						lpWorkX[i] = (lpWorkX[i - 1] + lpWorkX[i + 1]) / 2;
						lpWorkY[i] = (lpWorkY[i - 1] + lpWorkY[i + 1]) / 2;
					}
				}

				/* lastly overwrite midpoint of control points with */
				/* average of previous calculation, yeilds midpoint on bezier */

				for (i = nTotalPoints - 4; i > 0; i -= 6) {
					lpWorkX[i] = (lpWorkX[i - 1] + lpWorkX[i + 1]) / 2;
					lpWorkY[i] = (lpWorkY[i - 1] + lpWorkY[i + 1]) / 2;
				}

				nLinePoints = (nTotalPoints / 3) + 1;
			}

			/* have enough bezier endpoints, move bezier endpoints to */
			/* polyline CGIPOINTs array */

			for (i = 0; i < nSteps; i++) {
				var ptTemp = new rmpw.diagrams.render.util.Point();

				j = (3 * i);
				ptTemp.x = lpWorkX[j];
				ptTemp.y = lpWorkY[j];

				thePolyPoints.addPoint_1(ptTemp);
			}

			return bRC;
		},
		
		DEFAULT_CORNER_APPROXIMATION_PTS : 15,
			
	    /**
	     * Replaces bendpoints (all points except the first and last) in points list by points defining 
	     * corresponding arcs. That is, each bendpoint is replaced with the start and end point of the arc,
	     * and if calculateApproxPoints is true, there will also be points in between that approximate the arc. 
	     * Arc size is r. Sometimes, segment between two bendpoints is shorter than r. In that case, arc size for 
	     * both bendpoints has to be reduced to be half of the segment length, so that both arcs can fit. 
	     * These reduced arc sizes (if any) are kept in rForBendpoint hashtable.  
	     * 
	     * @param points Initial list of connection points (PointList)
	     * @param r Size (radius) of the arc that should replace bendpoint
	     * @param rForBendpoint Keeps track of bendpoints whose arcs have to be smaller than r
	     * @param calculateAppoxPoints Indicates if arcs replacing bendpoints should be approximated
	     */
	    calcRoundedCornersPolyline : function(points, r, rForBendpoint, calculateAppoxPoints) {
			var newPoints = new rmpw.diagrams.render.util.PointList();
			// First, clean up the points list if needed. Each segment is defined by two end points,
			// so if it happens that the segment has points in between, remove them since there is no
			// use for them, they can just create problems.
			var k = 1;
			var x0, y0, x1, y1, x2, y2;
			while (k < points.size - 1) {
				x0 = points.getPoint(k-1).x;
				y0 = points.getPoint(k-1).y;
				x1 = points.getPoint(k).x;
				y1 = points.getPoint(k).y;
				x2 = points.getPoint(k+1).x;
				y2 = points.getPoint(k+1).y;
				if ((x0 === x1 && x1 === x2) || (y0 === y1 && y1 === y2)) {
					// (x1, y1) is not needed, remove it
					points.removePoint(k);
				} else {
					k++;
				}
			}
			newPoints.addPoint_1(points.getFirstPoint()); 
			var rDefault = r;
			for (var i = 1; i < points.size - 1; i++) {
				x0 = points.getPoint(i-1).x;
				y0 = points.getPoint(i-1).y;
				x1 = points.getPoint(i).x; // x of bendpoint to be replaced
				y1 = points.getPoint(i).y; // y of bendpoint to be replaced
				x2 = points.getPoint(i+1).x;
				y2 = points.getPoint(i+1).y;
				// there are 8 possibilities: four types of corners, each can be traversed in two directions
				var cornerCase;
				if (x0 === x1 && x2 > x1 && y0 < y1 && y2 === y1) {
					cornerCase = 1;
				} else if (x0 > x1 && x2 === x1 && y0 === y1 && y2 < y1) {
					cornerCase = 2;
				} else if (x0 < x1 && x2 === x1 && y0 === y1 && y2 < y1) {
					cornerCase = 3;
				} else if (x0 === x1 && x2 < x1 && y0 < y1 && y2 === y1) {
					cornerCase = 4;
				} else if (x0 > x1 && x2 === x1 && y0 === y1 && y2 > y1) {
					cornerCase = 5;
				} else if (x0 === x1 && x2 > x1 && y0 > y1 && y2 === y1) {
					cornerCase = 6;
				} else if (x0 === x1 && x2 < x1 && y0 > y1 && y2 === y1) {
					cornerCase = 7;
				} else if (x0 < x1 && x2 === x1 && y0 === y1 && y2 > y1) {
					cornerCase = 8;
				} else {
					return null; // not rectilinear routing - shouldn't happen
				}
				// It is possible that the distance between (x0, y0) and (x1, y1), or (x1, y1) and (x2, y2) 
				// is smaller than the desired arc width and heighr r. In that case, we have to shrink the arc
				// to fit whatever space we have. Add changed r in rForBendpoint so it can be used later.
				r = rDefault;
				var distance = Math.min(points.getPoint(i-1).getDistanceOrthogonal(points.getPoint(i)),
				points.getPoint(i).getDistanceOrthogonal(points.getPoint(i+1)));
				if (r > distance / 2) {
					r = distance / 2 - 1;
					rForBendpoint.push(r);
				} else {
					rForBendpoint.push(rDefault);
				}

				// Find the coordinates of the arc center, as well as the sign (+ or -) for the circle equasion
				var sign = 1;
				var p; var q; // coordinates of the arc center 

				switch (cornerCase) {
				case 1:
				case 2:
					p = x1 + r;
					q = y1 - r; 
					break;
				case 3:
				case 4:
					p = x1 - r;
					q = y1 - r;
					break;
				case 5:
				case 6:
					p = x1 + r;
					q = y1 + r;  
					sign = -1;
					break;
				default: // 7 and 8
					p = x1 - r;
					q = y1 + r;
					sign = -1;
					break;
				}
				// Find the first and last point of the arc, and add the first point to the result list
				var lastPoint = null; // last point in bendpoint approximation
				switch (cornerCase) {
				case 1:
				case 4:
				case 6:
				case 7:
					newPoints.addPoint_0(x1, q);
					lastPoint = new rmpw.diagrams.render.util.Point(p, y1);
					break;
				default:
					newPoints.addPoint_0(p, y1);
					lastPoint = new rmpw.diagrams.render.util.Point(x1, q);
					break;
				}
				// Find out if x will be decreasing or increasing while calculating approximation points.
				var incrementSign = 1;
				switch (cornerCase) {
				case 2:
				case 4:
				case 5:
				case 7:
					incrementSign = -1;
					break;
				}
				// If arcs need to be approximated: for given x find y so that (x, y)
				// is on the arc
				if (calculateAppoxPoints) {
					var x = newPoints.getLastPoint().x;
					var rSq = r*r; 
					var increment = incrementSign * r / this.DEFAULT_CORNER_APPROXIMATION_PTS;
					var nrOfIncrement = this.DEFAULT_CORNER_APPROXIMATION_PTS;
					if (increment === 0) {
						increment = incrementSign * 1;
						nrOfIncrement = r;
					}
					for (var j = 1; j < nrOfIncrement; j++) {
						x += increment;
						// calculate y with given x so that (x, y) is on the arc
						var y = Math.round((q + sign*Math.sqrt(rSq - (x - p)*(x - p))));
						newPoints.addPoint_0(x, y);
					}
				}
				// Add the last point to the result list
				newPoints.addPoint_1(lastPoint);
			}
			newPoints.addPoint_1(points.getLastPoint());
			return newPoints;
	    }		

});