/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2009.  All Rights Reserved.
 * 
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
dojo.provide("rmps.core.util.StringUtils");

rmps.core.util.StringUtils = {
	
	/**
	 * Truncates the given string value and adds an ellipsis at the location
	 * specified in the ellipsisLocation parameter.
	 * 
	 * @param value the string to ellipsisLocation
	 * @param length the length of the resulting string value
	 * @param ellipsisLocation the location of the ellipsis. One of: s|m|e (start, middle, end), default is 'e'
	 * @param trim true if the result should be trimmed of leading and trailing
	 *             whitespace (default), false otherwise
	 * @param ellipsisWord whether or not the ellipsis should occur at a word boundary, default is false
	 * 
	 * @return the truncated string
	 */
	truncate: function(value, length, ellipsisLocation, trim, ellipsisWord) {
		if (trim != false) {
			value = dojo.trim(value);
		}
		
		if (value.length <= length) {
			return value;
		}
		
		if (length == 0) {
			return '';
		}
		
		function trimValue(start, end, ellipsisEnd) {
			var result = '';
			if (ellipsisWord) {
				var word = '', i, c;
				if (ellipsisEnd) {
					for (i = start; i < end; ++i) {
						c = value.charAt(i);
						if (c.match(/\w/)) {
							word += c;
						} else {
							result += word;
							word = c;
						}
					}
					if (word.length > 0 && (i == value.length || !value.charAt(i).match(/\w/))) {
						result += word;
					}
				} else {
					for (i = end - 1; i >= start; --i) {
						c = value.charAt(i);
						if (c.match(/\w/)) {
							word = c + word;
						} else {
							result = word + result;
							word = c;
						}
					}
					if (word.length > 0 && (i == -1 || !value.charAt(i).match(/\w/))) {
						result = word + result;
					}
				}
			}
			if (!ellipsisWord || result.length == 0) {
				result = value.substring(start, end);
			}
			if (trim != false) {
				result = dojo.trim(result);
			}
			return result;
		}
		
		if (length <= 3) {
			return trimValue(0, length);
		}
		
		if (ellipsisLocation == null || !(ellipsisLocation == 's' || ellipsisLocation == 'm' || ellipsisLocation == 'e')) {
			ellipsisLocation = 'e';
		}
		
		// length > 3
		var truncated;
		var minLength = 3;
		switch (ellipsisLocation) {
			case 's':
				truncated = '...' + trimValue(value.length - length, value.length, false);
				break;
				
			case 'm':
				minLength = 4;
				var mid = Math.floor(length / 2);
				var offset = length % 2;
				truncated = trimValue(0, mid - offset, true) + '...' + trimValue(value.length - mid, value.length, false);
				break;
				
			case 'e':
				truncated = trimValue(0, length, true) + '...';
				break;
		}
		
		if (truncated.length <= minLength) {
			truncated = value.substring(0, minLength);
			if (trim != false) {
				truncated = dojo.trim(truncated);
			}
		}
		return truncated;
	},
	
	/*
	 *Get the size for a given text.
	 * Paramaters:
	 * 			1) node: The node that contains this text. Required in order
	 * 				to make sure that width calculation is based on its style attributes (fonts, etc)
	 * 			2) text: The text to get its size calcualted 
	 */
	calculateTextSize: function(text, node){
		/*clone the node, so we make sure no side effects would happen */
		var cloned= dojo.clone(node);
		var span = document.createElement("span");
		span.innerHTML = text;
		span.style.visibility= "hidden";
		
		/*Append the hidden span to the current node so it would inherit its style properties */
		cloned.appendChild(span);
		
		/*Required to force client height/width to get intialized for the created span */    
		dijit.popup.moveOffScreen(span);
		
		var size= {
			width: parseInt(span.offsetWidth, 10),
			height: parseInt(span.offsetHeight, 10)
		};
		
		/*Remove and destroy the temp span after getting the text size information */
		span.parentNode.removeChild(span);
		dojo.destroy(cloned);
		
		return size;
	}
};