/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2012. All Rights Reserved.
 *
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/
(function() {
dojo.provide("rmps.ui.util.ToolbarFocusUtil");

/** This utility provides methods for showing and hiding the actions
 *  in a toolbar when the nodes in a container element receive the
 *  keyboard focus or the container itself receives hover events.
 *
 *  The general DM style when many entries in a table are displayed is to
 *  only show the actions for an entry when the mouse hovers over the entry.
 *  This is usually done using a CSS :hover style on the container widget.
 *  However, hover events are not triggered when the keyboard is used to
 *  traverse the web page. As a result, actions are not displayed when
 *  using the keyboard. This utility solves that problem.
 *
 *  One public methods is available:
 *		addToolbarFocusListeners()    which must be called after the toolbar is created
 *
 *  In order to remove the listeners, when the corresponding container node is
 *  deleted, ensure the toolbar is destroyed. This utility adds a listener to
 *  the toolbar destroy() method and performs the listener cleanup at that time.
 */

rmps.ui.util.ToolbarFocusUtil = {
	
	/** Add keyboard focus listeners to the nodes in the container
	 *  hierarchy to show and hide the actions in the contained toolbar.
	 *  This method also adds mouse hover listeners to the container 
	 *  itself to show and hide the toolbar actions. 
	 *
	 *  In order to find the actions in the toolbar, all of the actions 
	 *  must have the same class name. Typically the visibility of the
	 *  actions are controlled by a CSS :hover style defined on the
	 *  container node (e.g., a table row). Pass the exact class name
	 *  for the actions as is defined in the CSS file, including the 
	 *  name prefix ".". The CSS :hover style does not need to be 
	 *  deleted but these listeners will override the CSS. 
	 *
	 *  During testing, it was discovered that if the listeners were
	 *  only added for focus events and not mouse hover events, after
	 *  the actions were hidden by the focus listener, the CSS :hover
	 *  style no longer worked. Therefore, the mouse hover listeners
	 *  are also defined on the main container node.
	 *
	 * @param actionToolbar the toolbar widget containing the actions
	 * @param containerNode the main container node (e.g., a table row) that has 
	 *						at least one visible element which can receive the 
	 *						keyboard focus (e.g., a resource link)
	 * @param actionClassname the class name from the CSS that identifies the actions
	 * @param focusCallback an optional callback function (containerNode, hasFocus)
	 */ 
	addToolbarFocusListeners: function(args) {
		// Extract the parameters from the arguments
		var actionToolbar = args.actionToolbar;
		var containerNode = args.containerNode;
		var focusCallback = args.focusCallback;
		var actionClassName = args.actionClassName;

		// Initialize the array for the listener connections
		actionToolbar._focusListeners = [];

		// Remember the class name of the actions in the toolbar.
		actionToolbar._focusActionClassName = actionClassName;
		
		// Remember the main container node for the entry.
		actionToolbar._focusContainer = containerNode;
		
		// Remember the container node if the callback function is defined
		if (dojo.isFunction(focusCallback)) {
			actionToolbar._focusCallback  = focusCallback;
		}
		
		// Recursively add focus listeners to the elements in the
		// container node hierarchy that can receive the keyboard focus.
		// For the top container node, also add mouse hover listeners.
		this._addFocusListenersToNode(actionToolbar._focusContainer, actionToolbar, true, true);
		actionToolbar._focusDoCheck = true;
		
		// Add a connection to the destroy method to do any cleanup.
		var removeListeners = dojo.partial(this._removeToolbarFocusListeners, actionToolbar);
		actionToolbar._focusListeners.push(dojo.connect(actionToolbar, "destroy", this, removeListeners));
	},
	
	/** Remove all of the focus listeners that were added by the
	 *  method addToolbarFocusListeners. This method must be called
	 *  before the toolbar widget is destroyed.
	 *
	 * @param actionToolbar the toolbar widget containing the actions
	 */
	_removeToolbarFocusListeners: function(actionToolbar) {
		// If the focus timeout still exists, clear it.
		if (actionToolbar._focusTimeout) {
			clearTimeout(actionToolbar._focusTimeout);
			delete actionToolbar._focusTimeout;
		}
		
		// If listeners were added for this toolbar, disconnect them.
		var focusListeners = actionToolbar._focusListeners; 
		if (dojo.isArray(focusListeners)) {
			var length = focusListeners.length;
			for (var i = 0; i < length; i++) {
				dojo.disconnect(focusListeners[i]);
			}
			delete actionToolbar._focusListeners;
			delete actionToolbar._focusActionClassName;
			delete actionToolbar._focusContainer;
		}
		
		// Remove the callback information
		if (actionToolbar._focusCallback) {
			delete actionToolbar._focusCallback;
		}
	},
	
   
    /** Recursively walk a node hierarchy and add focus listeners to every node 
     *  that can obtain the focus (e.g., tabIndex >= 0). The listener will
     *  show or hide the corresponding actions in the given action toolbar.
     *  Optionally add mouse hover listeners for this node which will show and
     *  hide the actions.
     *
     * @param node the root node of the hierarchy
     * @param actionToolbar the toolbar with the corresponding actions
     * @param addHover  true = add mouse hover listeners
     * @param forceAdd  true = always add listener to focusable node
     *					false = add listener only if none added yet 
     */
    _addFocusListenersToNode: function(node, actionToolbar, addHover, forceAdd) {
		// Define function calls to show and hide the actions in the toolbar.
		var showActions = dojo.partial(this._setActionToolbarVisibility, actionToolbar, true);
		var hideActions = dojo.partial(this._setActionToolbarVisibility, actionToolbar, false);
		
		// Add mouse hover listeners to the top container node if desired.
		if (addHover) {
			actionToolbar._focusListeners.push(dojo.connect(node, "onmouseover", this, showActions));
			actionToolbar._focusListeners.push(dojo.connect(node, "onmouseout",  this, hideActions));
		}
		
		// If the node is focusable, add focus listeners
		if ( (node.tabIndex >= 0 || node.className === "button" ) && (forceAdd || !node._focusSet)) {
		                           //jazz.ui.toolbar.Button adds class "button" to <a> tags 
			actionToolbar._focusListeners.push(dojo.connect(node, "onfocus", this, showActions));
			actionToolbar._focusListeners.push(dojo.connect(node, "onblur",  this, hideActions));
			node._focusSet = true;
		}

		// For each child of the given node
		var length = node.childNodes.length;
		for (var index = 0; index < length; index++) {
			// Recursively examine the node's children
			var child = node.childNodes[index];
			this._addFocusListenersToNode(child, actionToolbar, false, forceAdd);
		}
    },
    
    
    /** Show or hide the actions in the given toolbar when nodes
     *  in the corresponding entry obtain and lose keyboard focus. 
     *
     *  When focus is lost, the actions are not hidden immediately 
     *  because each focusable node in the entry obtains and loses  
     *  focus during keyboard traversal. If the actions are hidden 
     *  as soon as focus is lost, the browser is unable to pass the
     *  focus to any of the actions because they are not visible.
     *  Therefore, the hiding of the actions is delayed slightly.
     *
     * @param actionToolbar the toolbar actions for this entry
     * @param show  true = a node in the entry has obtained the focus
     */
	_setActionToolbarVisibility: function(actionToolbar, show) {
		// In some cases, during the initial call to addToolbarFocusListeners
		// some of the nodes in the entry hierarchy may not have
		// been created yet. This happens with an entry that contains
		// a rich text widget. The first time any node in the entry
		// gets focus, re-check for other focusable nodes that do
		// not have listeners yet.
		if (actionToolbar._focusDoCheck) {
			this._addFocusListenersToNode(actionToolbar._focusContainer, actionToolbar, false, false);
			delete actionToolbar._focusDoCheck;
		}
		
		// Extract the CSS class name for the toolbar actions.
		var actionClassName = actionToolbar._focusActionClassName;
		
		// If a node in the entry has received the focus
		if (show) {
			// If the focus is changing from one node in the entry to
			// another, cancel the timeout that hides the actions.
			if (actionToolbar._focusTimeout) {
				clearTimeout(actionToolbar._focusTimeout);
				delete actionToolbar._focusTimeout;
			}
			
			// Ensure all of the actions in the toolbar are visible
			dojo.query(actionClassName, actionToolbar.domNode.parentNode).forEach(function(action, index, nodeList) {
				dojo.style(action, "visibility", "visible");	
			});
			
			// If a callback is defined, call it to indicate the entry has focus.
			if (actionToolbar._focusCallback) {
				actionToolbar._focusCallback(actionToolbar._focusContainer, true);
			}
		}
		
		// else a node in the entry has lost the focus 
		else {
			// If the timeout to hide the actions does not exist yet
			if (!actionToolbar._focusTimeout) {
			
				// Create a timeout to delay the hiding of the toolbar actions.
				actionToolbar._focusTimeout = setTimeout(function() {
					// Hide all of the actions in the toolbar
					dojo.query(actionClassName, actionToolbar.domNode.parentNode).forEach(function(action, index, nodeList) {
						dojo.style(action, "visibility", "hidden");	
					});
					
					// Delete the timeout handle
					delete actionToolbar._focusTimeout;
			
					// If a callback is defined, call it to indicate the entry lost focus.
					if (actionToolbar._focusCallback) {
						actionToolbar._focusCallback(actionToolbar._focusContainer, false);
					}
				}, 10);
			}
		}
    }
	
};
}());