/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.diagrams.render.util.Workarounds");

dojo.require("rmpw.diagrams.render.util.metrics");
dojo.require("dojox.html.entities");

/*
 * A collection of functions that we really shouldn't need, but that patch up functionality
 * that should be available and working properly in dojo.gfx
 * 
 * TODO: should extend the dojo gfx class prototypes directly
 * 
 * Associated bugs:
 *      getBoundingBox appears broken for group objects
 *      http://trac.dojotoolkit.org/ticket/7019
 * 
 *      Bounding box calculations are broken for many shapes.
 *      http://trac.dojotoolkit.org/ticket/3145
 * 
 * @author Marc Gobeil
 */

(function(){
	var w = rmpw.diagrams.render.util.Workarounds;

/**
 * @param figure the figure to get the bounds of
 * @param accept a function that returns true if the figure should be included in the bounds calculation
 * @param recursiveCall indicates the method is called recursively
 */
w.getBoundingBox = function(/* dojox.gfx.Shape */ figure, /* function? */ accept, recursiveCall) {
	var retBounds = null,
	   style;
	if(!accept || accept(figure)) {
		if(figure.children) {
			// calculate group bounding box
			var groupBounds = null;
			var children = figure.children;
			for(var i = 0, len = children.length; i < len; i++) {
				var child = children[i];
				var bounds = this.getBoundingBox(child, accept, true);
				w.transformBoundingBox(child.matrix, bounds);
				
				if(bounds) {
					if(!groupBounds ) {
						// initialize group bounds with first child
						groupBounds = dojo.clone(bounds);
					} else {
						// union bounds with groupBounds
						if(bounds.x < groupBounds.x) {
							groupBounds.width += groupBounds.x - bounds.x;
							groupBounds.x = bounds.x;
						}
						if(bounds.y < groupBounds.y) {
							groupBounds.height += groupBounds.y - bounds.y;
							groupBounds.y = bounds.y;
						}
						if(bounds.x + bounds.width > groupBounds.x + groupBounds.width) {
							groupBounds.width += (bounds.x + bounds.width) - (groupBounds.x + groupBounds.width);
						}
						if(bounds.y + bounds.height > groupBounds.y + groupBounds.height) {
							groupBounds.height += (bounds.y + bounds.height) - (groupBounds.y + groupBounds.height);
						}
					}
				}
			}
			retBounds = groupBounds;
		} else if(figure.shape.type === "text") {			
			// calculate text dimensions off-screen in html, otherwise not supported in dojo.gfx
			
			// NOTE: calculating text box bounds is particularly expensive, always cache it. Also
			// retBounds is frequently modified during subsequent loops of this function, so have
			// to create defensive clones of cache BOTH ways.
			var cache = figure.__boundsCache; 
			if(cache && (cache.text === figure.shape.text)) {
				// use cache
				retBounds = dojo.clone(cache.bounds);
			} else {
				// calculate text bounding box
				retBounds = this.getTextBounds(figure);

				// populate cache
				figure.__boundsCache = {
					text: figure.shape.text,
					bounds: dojo.clone(retBounds)
				};
			}

			if (recursiveCall) {
				retBounds.x = figure.shape.x;
				retBounds.y = figure.shape.y;
				
				if (retBounds.y !== 0) {
					retBounds.y -= this.getTextBounds(figure) - this.BASELINE_FUDGEFACTOR;
				}
				else if (figure.parent.shape.type == "wrappingText") {
					retBounds.y -= retBounds.height;
				}
			}
		} else if(figure.shape.type === "path" || figure.shape.type === "polyline") {
			retBounds = (!dojo.isOffscreen && figure.rawNode.style.visibility==='hidden') ? null : dojo.clone(figure.getBoundingBox());		
		} else {
			// call built in bounding box function
			retBounds = dojo.clone(figure.getBoundingBox());
		}

		if(!retBounds) {
			// can happen in case of empty group
			retBounds = {x: 0, y: 0, width: 0, height: 0};
		}
		
		// apply custom styles that help implement layout policies
		style = figure.customStyle;
		if(style) {
			if(style.paddingLeft) {
				retBounds.x -= style.paddingLeft;
				retBounds.width += style.paddingLeft;
			}
			if(style.paddingRight) {
				retBounds.width += style.paddingRight;
			}
			if(style.paddingTop) {
				retBounds.y -= style.paddingTop;
				retBounds.height += style.paddingTop;
			}
			if(style.paddingBottom) {
				retBounds.height += style.paddingBottom;
			}
		}
	}
	return retBounds;
};

w.IGNORE_DECORATIONS = function(figure) {
	return figure.isDecoration ? !figure.isDecoration : true;
};
	
// workaround for transforms not seeming to work on groups in IE/VML
w.applyTransform = function(figure, args) {
	if(dojox.gfx.vml && figure.children) {
		var children = figure.children;
		for(var i = 0, len = children.length; i < len; i++) {
			this.applyTransform(children[i], args);
		}
	} else {
		figure.applyTransform(args);
	}
};

/**
 * workaround for getTransformedBoundingBox not seeming to work
 * 
 * Gets the bounding box for 
 * 
 * NOTE: assumes no rotation or shearing
 * 
 * @param figure figure to get the bounds of
 * @param accept a function that returns true if the figure should be included in the bounds 
 * @param ancestor the ancestor whose coordinate system to transform the bounds into, null will calculate it in the root
 *        NOTE: will throw an NPE if the given node is not in the parent chain of figure
 */
w.getTransformedBoundingBox = function(figure, accept, ancestor) {
	var boundingBox = w.getBoundingBox(figure, accept, false);
	if (!boundingBox ) {
		return { x: 0, y: 0, width: 0, height: 0 };
	}
	
	if(figure.parent) {
		for(var node = figure; node && node !== ancestor; node = node.parent) {
			w.transformBoundingBox(node.matrix, boundingBox);
		}
	}
	return boundingBox;
};

w.transformBoundingBox = function(transform, boundingBox) {
	if(transform && boundingBox) {
		boundingBox.x = boundingBox.x * transform.xx + transform.dx;
		boundingBox.y = boundingBox.y * transform.yy + transform.dy;
		boundingBox.width *= transform.xx;
		boundingBox.height *= transform.yy;
	}
};

w.getTextBounds = function(figure) {
	var style = null;
	if(figure.fontStyle) {
		style = {
			fontFamily: figure.fontStyle.family,
			fontSize: figure.fontStyle.size,
			fontWeight: figure.fontStyle.weight
		};
	}
	var textBounds = figure.shape.text ? rmpw.diagrams.render.util.metrics.getTextBox(figure.shape.text, style) : {w: 0, h: 0};
	return {x: 0, y: 0, width: textBounds.w, height: textBounds.h};
};

w.BASELINE_FUDGEFACTOR = 3;

/**
 * returns the height of text, frequently used to top-align it
 */
w.getTextBaselineHeight = function(text, style) {
	text = String(text);
	var s = null;
	if(style) {
		s = {
			fontFamily: style.family,
			fontSize: style.size,
			fontWeight: style.weight
		};
	}
	// -3 is a fudge factor for the overhang past the baseline 
	return rmpw.diagrams.render.util.metrics.getTextBox(dojox.html.entities.encode(text), s).h - this.BASELINE_FUDGEFACTOR;
};

/*
 * Translates a geometric data structure coordinates to absolute screen coodinates
 */
w.translateToAbsolute = function (t, dojoShape){
	if (!dojoShape){
		return null;
	}
	dojoShape = dojoShape.parent;
	while (dojoShape) {
		if (dojoShape.matrix) {
			t.transform(dojoShape.matrix);
		}
		dojoShape = dojoShape.parent;
	}
};

/*
 * Translates a geometric data structure coordinates to relative model coordinates
 */
w.translateToRelative = function (t, dojoShape, container){
	if (!dojoShape){
		return null;
	}
	
	if (!container){ dojoShape = dojoShape.parent;}
	var transforms = [];
	while (dojoShape) {
		if (dojoShape.matrix) {
			var m = dojox.gfx.matrix;
			var dojoShapeMatrix = m.invert(dojoShape.matrix);
			if (dojoShapeMatrix) {
				transforms.push(dojoShapeMatrix);
			}
		}
		dojoShape = dojoShape.parent;
	}
	
	// do the transforms in reverse
	for(var i = transforms.length - 1; i >= 0; i--){
		t.transform(transforms[i]);
	}

};

}());
