/*******************************************************************************
 * Licensed Materials - Property of IBM
 *
 * com.ibm.xtools.rmps.web.core
 *
 * Copyright IBM Corporation 2011, 2012.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/
(function() {
dojo.provide("com.ibm.team.vvc.shared.oslc.web.utils.BaseClientUtil");

dojo.require("dojox.xml.DomParser");
dojo.require("com.ibm.team.vvc.shared.oslc.web.utils.DataFeed");

com.ibm.team.vvc.shared.oslc.web.utils.BaseClientUtil = {

	/*
	 * @see rmps.core.base.BaseClient#getResponse
	 *
	 * Example:
	 *
	 * var nestedObjectForm1 = { path:["rmpsObject:ReviewArtifact"], form: [
	 *      {name: "contributor", root: ["dcterms:contributor"], attribute: "rdf:resource"},
	 *      {name: "id", root: ["rmpsObject:artifactId"], node: "#text"},
	 *  ]};
	 *
	 * var nestedObjectForm2 = { path:["rmpsObject:ReviewArtifact"], form: [
	 *      {name: "user", root: [["dcterms:user"], ["dcterms:contibutor"]], typeFunction: function(value){return value;}, attribute: "rdf:resource"},
	 *      {name: "id", root: ["rmpsObject:artifactId"], node: "#text"},
	 *  ]};
	 *
	 *  var mainObjectForm = [
	 *      {name: "name", root: ["rmpsObject:name"], node: "#text"},
	 *      {name: "nestedObjectForm1", root: ["rmpsObject:nestedObjectForm1"], comparatorFunction: myComparator, objectForm: nestedObjectForm1},
	 *      {name: "nestedObjectForm2", root: ["rmpsObject:nestedObjectForm2"], objectForm: nestedObjectForm2}
	 *  ];
	 *
	 *  var model = BaseClientUtil.getModel(doc, { path:["rdf:RDF", "rmpsObject:Object", numberOfOccurences:1], form: form});
	 *
	 * Paramaters:
	 *      response:   The server response that can be in any form such as xml/text/json
	 *      objectForm: The form for the object to be constructed.
	 *                      objectForm expects required arguments, and other optional arguments as follows:
	 *                          Required:
	 *                              name: The name for the attribute to be assigned to the model.
	 *                              root: This value is required if handleAs is not json. It can be an empty array [] is do not have a value. It is used to
	 *                                    specify the nested level for the nodes to search at. And it expects an array of strings to let the method know
	 *                                    how to reach the exepcted node.
	 *                          Optional:
	 *                              objectForm: An object form attribute can refer to another object form. This means that an attribute can expect its value
	 *                                          to be a model to be constructed as well. The objectForm here can have the same arguments like any object form.
	 *                              typeFunction: A callback function to be called when assigning the value for the attribute. So, if the caller want to add
	 *                                          some tailored calculations on the value, then they can provide this callback. The typeCallback must expect a value
	 *                                          and it is expected to return a normalized value.
	 *                              numberOfOccurences: If specified this means that we want a limited number of entries to be shown. Used mainly if we know
	 *                                          if the returned entries are only one. So, we have the value of numberOfOccurences:1 in the base form to refer
	 *                                          that we only have one node returned instead of returning an array with signle item for its case.
	 *                              comparatorFunction: A callback caomprator function to sort the returned array for given nodes.
	 *
	 *      handleAs:   The type of response. If value is "json", then the handling will assue that the response is json. The json construction is done by
	 *                  the method _getModelAsJson. An object form must provide namespace and prefix values for its entries in order to ensure that it will
	 *                  work for the json responses.
	 *                      Example:
	 *                        var usersObjectForm = {path: ["rmpsReviews:ParticipantInfo"], form: [
	 *                              {name: "name", namespace: "http://jazz.net/xmlns/rmps/myNamespace/1.0", root: ["rmpsReviews:name"], node: "#text", numberOfOccurences:1, typeFunction: function(val) {
	 *                                  return "The name is:" + val;
	 *                              }}
	 *                          ]};
	 *                          var form = {path: ["somePath"], form: [
	 *                              {name: "MyUsers", objectForm: usersObjectForm, namespace: "http://purl.org/dc/terms", prefix:"contributor" root: [["dcterms:user"], ["dcterms:contibutor"]], attribute: "rdf:resource"}
	 *                      ]};
	 *                      BaseClientUtil.getModel(doc, { path:["rdf:RDF", "rmpsObject:Object", identifier: "https://host:9443/context/service/1"], form: form}, "json");
	 *
	 *                  Note, that we have to provide an identifer that clarifies the starting point in the json object. For instance, a json object can be:
	 *                          https://host:9443/context/service/1
	 *                                  http://purl.org/dc/terms/contributor
	 *                                              0
	 *                                                  type: uri
	 *                                                  value: https://host:9443/context/service/1#contributor_1
	 *                                              1
	 *                                                  type: uri
	 *                                                  value: https://host:9443/context/service/1#contributor_2
	 *
	 *                          https://host:9443/context/service/1#contributor_1
	 *                                                  type: literal
	 *                                                  value: user1
	 *                          https://host:9443/context/service/1#contributor_2
	 *                                                  type: literal
	 *                                                  value: user2
	 *                  In the example, we used "https://host:9443/context/service/1" as the identifier, which means that it will act as the start
	 *                  to look into at that json response. In the previous example, MyUser attribute will have an array for the values contained under
	 *                  http://purl.org/dc/terms/contributor (that we constructed from namespace + prefix. When prefix is undefined, then it is constructed from
	 *                  namespace + name. Then, it will search in the main json response to get the main node for each entry (such as https://host:9443/context/service/1#contributor_1)
	 *                  It will use the usersObjectForm to get the values. To sum up, the constructed object for the previous example will be:
	 *                          MyUsers
	 *                              0
	 *                                  name: The name is: user1
	 *                              1
	 *                                  name: The name is: user2
	 *
	 * Note: Your object form can support both json and text/xml formats. It's optional if we would like to provide the namespaces and prefixes for the entries.
	 */
	getModel: function(response, objectForm, handleAs) {
		if (!objectForm) {
			return null;
		}

		if (handleAs&& handleAs === "json") {
			return this._getModelAsJson(response, objectForm);
		}

		/* If response is a text, then, we need to parse it as an xml document. Otheriwse, we already have the response parsed as XML. */
		var doc = dojo.isString(response) ? dojox.xml.DomParser.parse(response) : response;
		var path = objectForm.path;
		var objects = [];
		var baseNodes = this._getNodesForPath(dojo.clone(path), doc);

		if (!baseNodes) {
			console.error("Could not get main node for the provided path");
			return;
		}

		var formEntries = objectForm.form;
		var formNumberOfOccurences = parseInt(objectForm.numberOfOccurences);
		if (!formNumberOfOccurences|| formNumberOfOccurences > baseNodes.length) {
			/* Make sure to avoid number of occurence being larger than the curent numbe of nodes. */
			formNumberOfOccurences = baseNodes.length;
		}

		for (var index = 0; index < formNumberOfOccurences; index++) {
			var baseNode = baseNodes[index];
			var object = {};
			var dataFeedKey;
			for (var formIndex = 0; formIndex < formEntries.length; formIndex++) {
				var form = formEntries[formIndex];

				var formNode = this._getNodesForPath(dojo.clone(form.root), baseNode);
				if (formNode.length == 0) {
					dataFeedKey = form.dataFeedKey;
					/* Create empty data feed at least, so the callers will not be effected if the resource does not have enties for the current node. */
					if (dataFeedKey) {
						object[form.name] = new com.ibm.team.vvc.shared.oslc.web.utils.DataFeed({
								dataSet: formNode,
								uniqueIdentifier: dataFeedKey
							});
					} else {
						/* So, at least the attribute will be listed in the object. */
						object[form.name] = null;
					}
					continue;
				}

				var numberOfOccurences = parseInt(form.numberOfOccurences);
				if (!numberOfOccurences || numberOfOccurences>formNode.length) {
					/* Make sure to avoid number of occurence being larger than the curent numbe of nodes. */
					numberOfOccurences = formNode.length;
				}

				if (form.node) {
					var node = form.node;
					var vals = [];
					for (var nodeIndex = 0; nodeIndex < formNode.length && vals.length < numberOfOccurences; nodeIndex++) {
						var newVal = this._getNodesForPath([node], formNode[nodeIndex]);

						for (var valIndex = 0; valIndex < newVal.length; valIndex++) {
							vals.push(this._getNodeValue(newVal[valIndex], form));
						}

						var typeFunction = form.typeFunction;
						var val;
						if (form.numberOfOccurences == 1) {
							val = typeFunction ? typeFunction(vals[0]) : vals[0];
						} else if (vals.length < 1) {
							/* Make sure to call fo the type function if is is defined even for null values, so the callers could specify alternative values. */
							val = typeFunction ? null : typeFunction(null);
						} else {
							val = vals;
						}
						object[form.name] = val;
					}
				} else if (form.objectForm) {
					if (formNode.length == 1) {
						var models = this.getModel(formNode[0], form.objectForm);
						dataFeedKey = form.dataFeedKey;
						if (dataFeedKey) {
							/* If the caller specified a value for the data feed key, then create a new data feed instance. */
							object[form.name] = new com.ibm.team.vvc.shared.oslc.web.utils.DataFeed({
								dataSet: models,
								uniqueIdentifier: dataFeedKey
							});
						} else {
							object[form.name] = models;
						}
					} else {
						var models = [];
						for (var modelIndex = 0; modelIndex < formNode.length && models.length < numberOfOccurences; modelIndex++) {
							var subModel = this.getModel(formNode[modelIndex], form.objectForm);
							var currentLength = models.length;

							if (subModel) {
								if (dojo.isArray(subModel) && subModel.length > 0) {
									models = models.concat(subModel.slice(0, numberOfOccurences - currentLength));
								} else {
									models.push(subModel);
								}
							}
						}

						var comparatorFunction = form.comparatorFunction;

						/* Checking for comparatorFunction being function of not is the caller responsbility in order to avoid extra checks here.
						 *
						 * We only check if the caller specified a comparatorFunction or not.
						 */
						if (comparatorFunction) {
							models.sort(comparatorFunction);
						}

						var value;
						dataFeedKey = form.dataFeedKey;
						if (dataFeedKey) {
							/* If the caller specified a value for the data feed key, then create a new data feed instance. */
							value = new com.ibm.team.vvc.shared.oslc.web.utils.DataFeed({
								dataSet: models,
								uniqueIdentifier: dataFeedKey
							});
						} else {
							/* Should show null value instead of empty array */
							value = form.numberOfOccurence == 1 ? models[0] : models.length < 1 ? null : models;
						}

						object[form.name] = value;
					}
				} else if (form.attribute) {
					if (formNode.length == 1) {
						object[form.name] = formNode[0].getAttribute(form.attribute);
					} else {
						var attibutes = [];
						for (var attributesIndex = 0; attributesIndex < formNode.length && attributesIndex < numberOfOccurences; attributesIndex++) {
							attibutes.push(formNode[attributesIndex].getAttribute(form.attribute));
						}

						var typeFunction = form.typeFunction;
						var val;
						if (form.numberOfOccurences == 1) {
							val = typeFunction ? typeFunction(attibutes[0]) : attibutes[0];
						} else if (vals.length < 1) {
							/* Make sure to call fo the type function if is is defined even for null values, so the callers could specify alternative values. */
							val = typeFunction ? null : typeFunction(null);
						} else {
							val = attibutes;
						}

						object[form.name] = val;
					}
				}
			}
			objects.push(object);
		}
		/* Use objectForm.numberOfOccurences instead of formNumberOfOccurences, as formNumberOfOccurences would have the same length of the base sequence. */
		return objectForm.numberOfOccurences == 1 ? objects[0] : objects;
	},

	/*
	 * Compare between two models, and return a report for the differences that have been found in the two models. If the two models are identical, then,
	 * it returns null.
	 * Important to mention that it does not compare two arrays instances unless dataFeed is used, as comparing two arrays does not provide a way to know
	 * which is the indexer to be used to decide the items to be compaed in both arrays.
	 *
	 * Form:
	 *      Required: It is the form that was used to construct both models before.
	 *
	 * Path:
	 *      Optional: It refers to the starting path fo the report attributes.
	 *
	 */
	compareModels: function(model1, model2, form, path) {
		/* checking for objectForm being defined, is the clinet's resposnilbity */
		var report = {};
		if (!path) {
			path = "";
		}
		report = this._compareModels(model1, model2, form, path, report, null);
		for (var att in report) {
			/*
			 * We do this in order to return null in case no attributes exist, so the callers will no from the null that there are no differences.
			 * We make this for in loop as a way to know that at least there is one difference, and once, we make sure about that, we return the differences
			 * report.
			 */
			return report;
		}
		return null;
	},

	/*
	 * Return a shallow cloned instance for a given model. We need to provide a shallow clone instead of deep cloning in such case as the model can contain
	 * data feed that holds listeners and widgets infomation, thus, we will not be needing to clone such information. In addition, using direct cloning for
	 * such cases will be too slow as it will clones all widgets.
	 */
	shallowCloneModel: function(model) {
		if (!model) {
			return null;
		}
		var clonedModel = {};
		for (var att in model) {
			var current = model[att];
			if (dojo.isArray(current)) {
				clonedModel[att] = this.shallowCloneModel(current);
			} else if (current instanceof com.ibm.team.vvc.shared.oslc.web.utils.DataFeed) {
				clonedModel[att] = this._shallowCloneDataFeed(current);
			} else {
				clonedModel[att] = current;
			}
		}
		return clonedModel;
	},

	/*
	 * Required to fill in the basic empty information for the model object in case that no document has been received yet.
	 */
	getInitialModel: function(objectForm) {
		if (!objectForm) {
			return null;
		}
		return this._fillInMissingAttributes({}, objectForm);
	},

	_getModelAsJson: function(response, objectForm) {
		if (!objectForm) {
			return null;
		}
		var doc = response;
		var identifier = objectForm.identifier;
		var objects = [];
		var baseNodes = doc[identifier];

		if (!baseNodes) {
			console.error("Could not get main node for the provided identifier");
			return;
		}

		var formEntries = objectForm.form;
		var formNumberOfOccurences = parseInt(objectForm.numberOfOccurences);
		if (!formNumberOfOccurences|| formNumberOfOccurences>baseNodes.length) {
			/* Make sure to avoid number of occurence being larger than the curent number of nodes. */
			formNumberOfOccurences = baseNodes.length;
		}

		var createEmptyDetails = function(object, form, dataFeedKey, typeFunction, listOfNodes) {
			/* Create empty data feed at least, so the callers will not be effected if the resource does not have entries for the current node. */
			if (dataFeedKey) {
				object[form.name] = new com.ibm.team.vvc.shared.oslc.web.utils.DataFeed({
						dataSet: [],
						uniqueIdentifier: dataFeedKey
					});
			} else {
				/* So, at least the attribute will be listed in the object. */
				object[form.name] = typeFunction ? typeFunction(listOfNodes) : listOfNodes;
			}
		}
		var object = {};
		objects.push(object);

		for (var formIndex = 0; formIndex < formEntries.length; formIndex++) {
			var form = formEntries[formIndex];
			var name = form.name;
			var typeFunction = form.typeFunction;
			var namespace = form.namespace;
			var prefix = form.prefix;
			if (!prefix) {
				prefix = name;
			}
			var newObjectForm = form.objectForm;
			var currentName = namespace ? namespace + "/" + prefix : prefix;
			if (newObjectForm) {
				var listOfNodes = baseNodes[currentName];
				if (!listOfNodes) {
					/* Create empty data feed at least, so the callers will not be affected if the resource does not have enties for the current node. */
					createEmptyDetails(object, form, form.dataFeedKey, typeFunction, listOfNodes);
				} else {
					var models = [];
					/* listOfNodes is expected to be an array but use for in loop to ensure looping through both types */
					for (var item in listOfNodes) {
						var entry = listOfNodes[item];
						newObjectForm.identifier = entry.value;
						var values = this._getModelAsJson(doc, newObjectForm);
						if (dojo.isArray(values)) {
							models = models.concat(values);
						} else {
							models.push(values);
						}
					}
					var comparatorFunction = form.comparatorFunction;
					if (comparatorFunction) {
						models.sort(comparatorFunction);
					}

					var dataFeedKey = form.dataFeedKey;
					if (dataFeedKey) {
						/* If the caller specified a value for the data feed key, then create a new data feed instance. */
						object[form.name] = new com.ibm.team.vvc.shared.oslc.web.utils.DataFeed({
							dataSet: models,
							uniqueIdentifier: dataFeedKey
						});
					} else {
						object[form.name] = models;
					}
				}
			} else {
				var formNode = baseNodes[currentName];
				if (!formNode) {
					object[name] = typeFunction ? typeFunction(formNode) : formNode;
					continue;
				}
				if (formNode.length == 0) {
					createEmptyDetails(object, form, form.dataFeedKey, typeFunction, formNode);
					continue;
				}
				var numberOfOccurences = parseInt(form.numberOfOccurences);
				if (!numberOfOccurences|| numberOfOccurences>formNode.length) {
					/* Make sure to avoid number of occurence being larger than the curent number of nodes. */
					numberOfOccurences = formNode.length;
				}
				var getValue = function(newVal) {
					return typeFunction ? typeFunction(newVal) : newVal;
				};

				if (numberOfOccurences === 1) {
					var entry = formNode[0];
					object[form.name] = entry ? getValue(entry.value) : entry;
				} else {
					var vals = [];
					for (var nodeIndex = 0; nodeIndex < formNode.length && vals.length < numberOfOccurences; nodeIndex++) {
						vals.push(getValue(formNode[nodeIndex].value));
					}
					object[form.name] = vals;
				}
			}
		}

		/* Use objectForm.numberOfOccurences instead of formNumberOfOccurences, as formNumberOfOccurences would have the same length of the base sequence. */
		return objectForm.numberOfOccurences == 1 ? objects[0] : objects;
	},

	_fillInMissingAttributes: function(model, objectForm) {
		var getValue = function(value, typeFunction) {
			return typeFunction ? typeFunction(value) : value;
		};
		for (var index = 0; index < objectForm.length; index++) {
			var current = objectForm[index];
			var name = current.name;
			/* If there is a value set, then, continue, no need to be here. */
			if (model[name]) {
				continue;
			}

			var dataFeedKey = current.dataFeedKey;
			if (dataFeedKey) {
				model[name] = new com.ibm.team.vvc.shared.oslc.web.utils.DataFeed({
								dataSet: [],
								uniqueIdentifier: dataFeedKey
							});
			} else {
				var typeFunction = current.typeFunction;
				model[name] = parseInt(current.numberOfOccurences) <= 1 ? getValue(null, typeFunction) : getValue([], typeFunction);
			}
		}
		return model;
	},

	_shallowCloneDataFeed: function(dataFeed) {
		var dataSet = dataFeed.dataSet;
		var newDataSet = [];
		for (var index = 0; index < dataSet.length; index++) {
			var _current = dataSet[index];
			if (_current instanceof com.ibm.team.vvc.shared.oslc.web.utils.DataFeed) {
				newDataSet.push(this._shallowCloneDataFeed(dataFeed));
			} else {
				newDataSet.push(this.shallowCloneModel(_current));
			}
		}
		return new com.ibm.team.vvc.shared.oslc.web.utils.DataFeed({
			dataSet: newDataSet,
			uniqueIdentifier: dataFeed.uniqueIdentifier
		});
	},

	_compareModels: function(model1, model2, form, path, report, dataFeedKey) {
		/* If the path is an array then, we show the indexer as this array to string value. */
		return this.fillReportInformation(model1, model2, form, dojo.isArray(path) ? path.toString() : path, report, dataFeedKey);
	},

	fillReportInformation: function(model1, model2, form, path, report, dataFeedKey, segment) {
		var isDataFeed1 = model1 instanceof com.ibm.team.vvc.shared.oslc.web.utils.DataFeed;
		var isDataFeed2 = model2 instanceof com.ibm.team.vvc.shared.oslc.web.utils.DataFeed;

		/* Both must be instances of DataFeed; otheriwse, the comparison fails. */
		if (isDataFeed1 ^ isDataFeed2) {
			report[path] = {
				model1: "instanceof " + com.ibm.team.vvc.shared.oslc.web.utils.DataFeed.declaredClass + " is: " + isDataFeed1,
				model2: "instanceof " + com.ibm.team.vvc.shared.oslc.web.utils.DataFeed.declaredClass + " is: " + isDataFeed2
			};
		} else if (isDataFeed1 && isDataFeed2) {
			/* If the lengths are different, then the two data feeds are not identical, but we make sure to know where the differences are. */
			this._addDataFeedsDifferences(model1, model2, dataFeedKey, form, path, report, "model1", "model2");
			this._addDataFeedsDifferences(model2, model1, dataFeedKey, form, path, report, "model2", "model1");
		} else {
			var isArray1 = dojo.isArray(model1);
			var isArray2 = dojo.isArray(model2);

			if (isArray1 ^ isArray2) {
				report[_path] = {
					model1: "instanceof array is: " + isArray1,
					model2: "instanceof array is: " + isArray2
				};
			}

			if (isArray1 && isArray2) {
				var self = this;
				var getDifferences = function(m1, m2, att1, att2, expectedPath) {
					for (var arrayIndex = 0; arrayIndex < m1.length; arrayIndex++) {
						var current1 = m1[arrayIndex];
						var identifier = current1[dataFeedKey];
						var current2 = null;
						for (var subIndex = 0; subIndex < m2.length; subIndex++) {
							var _current = m2[subIndex];
							if (_current[dataFeedKey] == identifier) {
								current2 = _current;
								break;
							}
						}
						if (!current2) {
							var _path = self._assignOrGetPath(report, [path, segment, expectedPath]);
							_path[identifier] = {m1: current1, m2: current2};
						} else {
							for (var formIndex = 0; formIndex < form.length; formIndex++) {
								var current = form[formIndex];
								var name = current.name;
								var val1 = current1[name];
								var val2 = current2[name];
								if (val1 != val2) {
									/* It is intentional to avoid checking for NPE here, as the path should not be null by any means. */
									var _r = self._assignOrGetPath(report, [path, segment, current.root.toString(), identifier]);
									_r[att1] = val1;
									_r[att2] = val2;
								}
							}
						}
					}
				};
				getDifferences(model1, model2, "model1", "model2", path);
				getDifferences(model2, model1, "model2", "model1", path);
			} else {
				for (var formIndex = 0; formIndex < form.length; formIndex++) {
					var current = form[formIndex];

					/* It is intentional to avoid checking for NPE here, as the path should not be null by any means. */
					var _path = path ? path + ":" + current.root.toString() : current.root.toString();
					var name = current.name;
					var objectForm = current.objectForm;
					if (objectForm) {
						var subFom = objectForm.form;
						var newPath = path ? path + ":" + objectForm.path.toString() : objectForm.path.toString();
						this._compareModels(model1[name], model2[name], subFom, newPath, report, current.dataFeedKey, segment);
					} else {
						var value1 = model1[name];
						var value2 = model2[name];
						if (value1 != value2) {
							report[_path] = {model1: value1, model2: value2};
						}
					}
				}
			}
		}
		return report;
	},

	_assignOrGetPath: function(object, path) {
		var value;
		var val = object;
		for (var index = 0; index < path.length; index++) {
			var current = path[index];
			var temp = val;
			val = val[current];
			if (!val) {
				val = temp[current] = {};
			}
			value = val;
		}
		return value;
	},

	_addDataFeedsDifferences: function(model1, model2, dataFeedKey, form, path, report, att1, att2) {
		for (var dataFeedIndex = 0; dataFeedIndex < model1.dataSet.length; dataFeedIndex++) {
			var data1 = model1.dataSet[dataFeedIndex];
			var data2 = model2.getItem(data1[dataFeedKey]); /* At this point, dataFeedKey cannot be null, otherwise, the form was corrupted. */

			if (!data2) {
				var _report = report[path];
				if (!_report) {
					_report = report[path] = {};
				}
				var newPath = _report[data1[dataFeedKey]] = {};
				newPath[att1] = data1;
				newPath[att2] = data2;
				continue;
			}

			var addToReport = function(val1, val2, assignee, segment) {
				var _report = report[path];
				if (!_report) {
					_report = report[path] = {};
				}
				var newPath = _report[assignee] = {};
				var _newPath = newPath[segment] = {};
				_newPath[att1] = val1;
				_newPath[att2] = val2;
			};
			for (var formIndex = 0; formIndex < form.length; formIndex++) {
				var subForm = form[formIndex];
				var name = subForm.name;
				var value1 = data1[name];
				/* If the current name == dataFeedKey and the current item is not there, then data2 will be null.
				 * We continue at this place even if value1 would value in case item2 is there, just to make sure that the loop would go fine
				 * for rest fields.
				 */
				var value2 = data2 ? data2[name] : null;

				var isDataFeed1 = value1 instanceof com.ibm.team.vvc.shared.oslc.web.utils.DataFeed;
				var isDataFeed2 = value2 instanceof com.ibm.team.vvc.shared.oslc.web.utils.DataFeed;

				/* Both must be instances of DataFeed; oitheriwse, the comparison fails. */
				if (isDataFeed1 ^ isDataFeed2) {
					addToReport(value1, value2, data1[dataFeedKey], subForm.root);
				} else if (isDataFeed1 && isDataFeed2) {
					var newForm = subForm.objectForm.form;

					/* so, we keep track of what this difference item belong to (its parent) */
					var assignee = data1[dataFeedKey];
					/* If the lengths are different, then the two data feeds are not identical, but we make sure to know where the differences are. */
					this.fillReportInformation(value1.dataSet, value2.dataSet, newForm, path, report, subForm.dataFeedKey, assignee);
					this.fillReportInformation(value2.dataSet, value1.dataSet, newForm, path, report, subForm.dataFeedKey, assignee);
				} else if (value1 != value2) { /* Now, apply the normal comparison. */
					addToReport(value1, value2, data1[dataFeedKey], subForm.root);
				}
			}
		}
	},

	_getNodeValue: function(currentVal, form) {
		var value = currentVal.nodeValue;
		if (dojo.isFunction(form.typeFunction)) {
			value = form.typeFunction(value);
		}
		return value;
	},

	_getNodesForPath: function(segments, node) {
		if (!segments|| segments.length == 0) {
			return [node];
		}
		var list = [];
		return this._getNodesForPathSegments(segments.reverse(), node, list);
	},

	_getNodesForPathSegments: function(segments, node, list) {
		if (!segments|| segments.length == 0) {
			return list;
		}
		var segment = segments.pop();
		if (dojo.isArray(segment)) {
			/* Get nodes for the current segment */
			this._getNodesForPathSegments(dojo.clone(segment), node, list);

			/* Get nodes for the rest segments */
			this._getNodesForPathSegments(dojo.clone(segments), node, list);
		} else {
			var childNodes = node.childNodes;
			for (var index = 0; index < childNodes.length; index++) {
				var current = childNodes[index];

				if (current.nodeName == segment) {
					if (segments.length == 0) {
						list.push(current);
					} else {
						this._getNodesForPathSegments(segments, current, list);
					}
				}
			}
		}
		return list;
	}
};

})();