/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010, 2013.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.common.vcs.Change");

dojo.require("rmpw.common.util.ChangeUtil");

/**
 * TODO: assure there's no overlap between deleted and overwritten, or simulate sequential overwriting behavior between them
 * 
 * @author Marc Gobeil
 */
dojo.declare("rmpw.common.vcs.Change", null, {
	/**
	 * Creates a new change of the given target object.  When committed, this Change
	 * will apply all the changes recorded to it with the Change.set(property, value) method.
	 * 
	 * 
	 * There are three valid call patterns:
	 * 
	 * // change to occur to obj in isolation (mainly for testing)
	 * new Change(obj);
	 * 
	 * // change to occur to obj as part of a change set (mainly for testing)
	 * new Change(obj, changeSet);		
	 * 
	 * // change to occur to the object at the given path within the model of the repository to which the change set is attached
	 * new Change("path.to.obj", changeSet);
	 */
	constructor: function(/* Object */ target) {
		if(!target) {
			throw "target is not set";
		}
	
		// target
		this.target = target;
		
		// forward changes	// [exists before, exists after] > [new data to store, old data to back up, action on rollback]
		this.updated = {};	// [true,	true	] > [key/value, original, 	restore	]
		this.added = {};	// [false,	true	] > [key/value, nothing, 	delete	]
		this.deleted = {};	// [true,	false	] > [key only, 	original,	restore	]
		this.arrayPush = {};// [false,	true	] > [key/value,	nothing,	delete	]
		this.arrayPushReference = {};
		this.arrayPushRefresh = {};
		this.arrayInsert = {}; 
		this.arrayInsertReference = {};
		this.arrayRemove = {};
		this.arrayRemoveReference = {};
		this.arrayRemoveReferenceSourcePathNames = {};
		this.arrayOrderChanges = {};
		
		// inverse changes
		this.original = null;
		
		// support chaining
		return this;
	},
	
	/**
	 * Set a bulk of changes in one call.
	 */
	setAll : function(allSets) {
		for ( var prop in allSets) {
			this.set(prop, allSets[prop]);
		}
	},
		
	/**
	 * Records a property change on the target object.
	 * To delete a property, set its value to undefined.
	 */
	set: function(/* String */ property, /* Object|null|undefined */ value) {
		if (value instanceof Array && !dojo.isString(value[0])) {
			return this.setReferenceArray(property, value);
		} else if(rmpw.common.util.ChangeUtil.objectEqual(this.target[property], value)) {
			delete this.deleted[property];
			delete this.updated[property];
			delete this.added[property];
			return this;
		}

		if(property in this.target) {
			if(value === undefined || value === null || value === "") {
				this.deleted[property] = true;
				delete this.updated[property];
				delete this.added[property];
			} else {
				this.updated[property] = value;
				delete this.deleted[property];
				delete this.added[property];
			}
		} else if (value !== null && value !== ""){
			this.added[property] = value;
			delete this.updated[property];
			delete this.deleted[property];
		}
		
		// support chaining
		return this;
	},
	
	/**
	 * Records a property change on the target object.
	 * To add or remove a resource reference.
	 */
	setReferenceArray: function(/* String */ arrayProperty, /* array*/ newArray) {
			if (newArray instanceof Array){
				// TODO add support of ordered collection.
				var originalArray = this.target[arrayProperty];
				// push the new ones 
				for(var i = 0; i < newArray.length; i++) {
					var value = newArray[i];
					var found = false;
					if (originalArray){
					for(var j = 0; j < originalArray.length; j++) {
						if (value === originalArray[j]){
							found = true;
							break;
						}
					}
					}
					if (!found){
						this.pushReference(arrayProperty, value);
					}
				}
				
				// Remove the deleted ones. 
				if (originalArray){
				for(var i = originalArray.length-1; i >= 0 ; i--) {
					var target = originalArray[i];
					var found = false;
					for(var j = newArray.length-1; j >= 0; j--) {
						if (target === newArray[j]){
							found = true;
							break;
						}
					}
					if (!found){
						this.removeReferenceArray(arrayProperty, target);
					}
				}
				
				var newArrayLength= newArray.length;
				
				//Required to ensure that the order commands will be executed in sequence when the collection order is changed after each execution
				var tempArray= originalArray.slice();
				
				for(var i=0;i<newArrayLength; i++){
					var value= newArray[i];
					var originalIndex= dojo.indexOf(tempArray, newArray[i]);
					
					//If NOT originalIndex>=originalIndex, means that this order change was already captured. For instance Item1, and Item3. When we are at item1, we will be able to
					//address the required order change. But when we get to item3, we have to make sure not to add it to the order changes, as it will be redundant
					//If originalIndex<newArrayLength means that this item has been added, so no need for the order, as adding it will ensure appending it to the end according
					//to its new order
					if(originalIndex<newArrayLength&& originalIndex>-1&&i!== originalIndex){
						var temp= tempArray[originalIndex];
						tempArray[originalIndex]= tempArray[i];
						tempArray[i]= temp;
						
						this.arrayMoveElement(arrayProperty, originalIndex, i);							
					}
				}
			}
			}
		// support chaining
		return this;
	},
	
	_push: function(arrayPush, property, value) {
		var pushValues = arrayPush[property]; 
		if(!pushValues) {
			pushValues = [];
			arrayPush[property] = pushValues;
		}
		pushValues.push(value);
		
		// support chaining
		return this;
	},
	
	push: function(property, value) {
		if (!this._revert(this.arrayRemoveReference, property, value)) {
			return this._push(this.arrayPush, property, value);
		}
		return true;
	},
	
	pushReference: function(property, value) {
		if (!this._revert(this.arrayRemoveReference, property, value)) {
			this._push(this.arrayPushReference, property, value);
		}
		return this;
	},
	
	pushRefresh: function(property, value) {
		return this._push(this.arrayPushRefresh, property, value);
	},
	
	arrayMoveElement: function(arrayProperty, originalIndex, newIndex) {
		//If newIndex!== originalIndex, means that there is an order change
		if(newIndex!== originalIndex){
			var arrayOrderChanges = this.arrayOrderChanges[arrayProperty]; 
			if(!arrayOrderChanges) {
				this.arrayOrderChanges[arrayProperty]= arrayOrderChanges = [];
			}
			arrayOrderChanges.push({
					originalIndex: originalIndex,
					newIndex: newIndex
			});
		}
		return this;
	},
		
	_insertArray: function(arrayInsert, property, value, index) {
		var insertValues = arrayInsert[property];
		if (!insertValues) {
			insertValues = [];
			arrayInsert[property] = insertValues;
		}
		
		insertValues.push({
			value: value,
			beforeIndex: index
		});
		
		// support chaining
		return this;
	},
	
	insertArray: function(property, value, index) {
		return this._insertArray(this.arrayInsert, property, value, index);
	},
	
	insertArrayReference: function(property, value, index) {
		return this._insertArray(this.arrayInsertReference, property, value, index);
	},
	
	_removeArrayAt: function(arrayRemove, property, index, value) {
		var removeValues = arrayRemove[property];
		if (!removeValues) {
			removeValues = [];
			arrayRemove[property] = removeValues;
		}
		
		removeValues.push({
			beforeIndex: index,
			value: value
		});
		
		// support chaining
		return this;
	},
	
	_getReducedValue: function(value, reduceValue) {
		if (reduceValue && value.rdf_about) {
			// identify removed reference by its rdf_about - simplifies conversion to json
			return { rdf_about: value.rdf_about };
		} else {
			return value;
		}	
	},
	
	removeArrayAt: function(property, index, value, reduceValue) {	
		return this._removeArrayAt(this.arrayRemove, property, index, this._getReducedValue(value, reduceValue));
	},	
	
	removeArray: function(property, value, fromIndex, findLast, reduceValue) {
		var index = dojo.indexOf(this.target[property], value, fromIndex, findLast);
		if(index == -1) {
			throw "Item not found in array";
		}
		return this.removeArrayAt(property, index, value, reduceValue);
	},
	
	removeReferenceArrayAt: function(property, index, value, reduceValue) {
		return this._removeArrayAt(this.arrayRemoveReference, property, index, this._getReducedValue(value, reduceValue));
	},	
	
	removeReferenceArray: function(property, value, fromIndex, findLast, reduceValue) {
		if (!this._revert(this.arrayPushReference, property, value)) {
			var index = dojo.indexOf(this.target[property], value, fromIndex, findLast);
			if(index == -1) {
				throw "Item not found in array";
			}
			return this.removeReferenceArrayAt(property, index, value, reduceValue);
		}
		return this;
	},
	
	// if the same value is already being added/removed from this same Change, just cancel the add/remove
	_revert: function(array, property, value) {
		var values = array[property];
		var cancel = false;
		if (values) {
			for (var i = 0; i<values.length; i++) {
				var currentValue = values[i].value; 				
				if (currentValue && dojo.isString(currentValue)) {
					// use == instead of === as value may not be a string
					if (currentValue == value) {
						cancel = true;
					}
				} else if ((currentValue && currentValue.rdf_about == value.rdf_about) ||
						(values[i].rdf_about == value.rdf_about)) {
					cancel = true;
				}
				if (cancel) {
					values.splice(i,1);
					return true;
				}
			}
		}
		return false;
	},
	
	cleanNewReferences: function(property) {
		// clear any new references from the change as well
		this._cleanNewReferences(this.arrayPush[property]);
		this._cleanNewReferences(this.arrayPushReference[property]);
	},
	
	_cleanNewReferences: function(refArray) {
		if (refArray) {
			for (var i=0, lg=refArray.length; i<lg; i++) {
				var target = refArray[i];
				if (rmpw.common.util.IdentifierUtil.getIdentifier(target.rdf_about) === target.rdf_about) {
					refArray.splice(i,1);
					lg--;
					i--;
				}
			}
			
			if (refArray.length == 0) {
				delete refArray;
			}	
		}
	},

	removeReferenceArrayAndMapping: function(property, value, deletedRenderMapping, fromIndex, findLast) {
		if (deletedRenderMapping) {
			this.arrayRemoveReferenceSourcePathNames[property] = deletedRenderMapping.getSourcePathNames();
		}		
		return this.removeReferenceArray(property, value, fromIndex, findLast);
	},
	
	isPropertyChanged: function(property) {
		if (this.updated[property] || this.added[property] || this.deleted[property] ||
		    this.arrayPush[property] || this.arrayPushReference[property] || this.arrayInsert[property] ||
		    this.arrayInsertReference[property] || this.arrayRemove[property] || this.arrayRemoveReference[property]) {
			return true;
		}
		
		return false;
	},
	
	commit: function() {
		if(!this.original) {
			// capture inverse
			this.original = {};
		}
		for(var property in this.updated) {
			this.original[property] = this.target[property];
		}
		for(var property in this.deleted) {
			this.original[property] = this.target[property];
		}
		this.commitToTarget(this.target);
		
		// fire change event
		if(this.target.onChange) {
			this.target.onChange({
				added: this.added,
				updated: this.updated,
				deleted: this.deleted, 
				pushed: this.arrayPush, 
				pushedReference: this.arrayPushReference,
				pushedRefresh: this.arrayPushRefresh,
				before: this.original,
				insertArray: this.arrayInsert,
				insertArrayReference: this.arrayInsertReference,
				removeArray: this.arrayRemove,
				removeReferenceArray: this.arrayRemoveReference,
				removeReferenceSourcePathNames: this.arrayRemoveReferenceSourcePathNames
			});
		}
	},
	
	commitToTarget: function(target) {		
		// apply to target
		dojo.mixin(target, this.added);
		dojo.mixin(target, this.updated);
		for(var property in this.deleted) {
			delete target[property];
		}
		
		for(var property in this.arrayPush) {
			var pushValues = this.arrayPush[property];
			for(var i = 0, len = pushValues.length; i < len; i++) {
				if(!(property in target)) {
					target[property] = [];
				}
				target[property].push(pushValues[i]);
			}
		}
		
		for(var property in this.arrayPushReference) {
			var pushValues = this.arrayPushReference[property];
			for(var i = 0, len = pushValues.length; i < len; i++) {
				if(!(property in target)) {
					target[property] = [];
				}
				target[property].push(pushValues[i]);
			}
		}
		
		for(var property in this.arrayInsert) {
			var insertValues = this.arrayInsert[property];
			insertValues.sort(function(a, b) { return a.beforeIndex - b.beforeIndex; });
			for(var i = 0; insertValue = insertValues[i]; i++) {
				var afterIndex = insertValue.afterIndex = insertValue.beforeIndex + i;
				target[property].splice(afterIndex, 0, insertValue.value);
			}
		}
		
		for(var property in this.arrayInsertReference) {
			var insertValues = this.arrayInsertReference[property];
			insertValues.sort(function(a, b) { return a.beforeIndex - b.beforeIndex; });
			for(var i = 0; insertValue = insertValues[i]; i++) {
				var afterIndex = insertValue.afterIndex = insertValue.beforeIndex + i;
				target[property].splice(afterIndex, 0, insertValue.value);
			}
		}
		
		for(var property in this.arrayRemove) {
			var removeValues = this.arrayRemove[property];
			removeValues.sort(function(a, b) { return a.beforeIndex - b.beforeIndex; });
			for(var i = 0; removeValue = removeValues[i]; i++) {
				var index = dojo.indexOf(target[property], removeValue.value);
				target[property].splice(index, 1);
			}
		}
		
		for(var property in this.arrayRemoveReference) {
			var removeValues = this.arrayRemoveReference[property];
			removeValues.sort(function(a, b) { return a.beforeIndex - b.beforeIndex; });
			for(var i = 0; removeValue = removeValues[i]; i++) {
				// reference can be identified just by its rdf_about
				var removeValueAbout = removeValue.value.rdf_about;
				var index = -1;
				for (var i, value; value = target[property][i]; i++ ) {
					if (value.rdf_about === removeValueAbout) {
						index = i;
						break;
					}
				}
				target[property].splice(index, 1);
			}
		}
	},
		
	/**
	 * Applies the inverse change to the real target object, or a provided substitute
	 */
	rollback: function() {
		for(var property in this.original) {
			this.target[property] = this.original[property];
		}
		for(var property in this.added) {
			delete this.target[property];
		}
		
		for(var property in this.arrayPush) {
			this.target[property].length -= this.arrayPush[property].length;
		}
		
		for(var property in this.arrayPushReference) {
			this.target[property].length -= this.arrayPushReference[property].length;
		}
		
		for(var property in this.arrayRemove) {
			var insertValues = this.arrayRemove[property];
			for(var i = 0, len = insertValues.length; i < len; i++) {
				this.target[property].splice(insertValues[i].beforeIndex, 0, insertValues[i].value);
			}
		}
		
		for(var property in this.arrayRemoveReference) {
			var insertValues = this.arrayRemoveReference[property];
			for(var i = 0, len = insertValues.length; i < len; i++) {
				this.target[property].splice(insertValues[i].beforeIndex, 0, insertValues[i].value);
			}
		}
		
		for(var property in this.arrayInsert) {
			var removeValues = this.arrayInsert[property];
			for(var i = 0, len = removeValues.length; i < len; i++) {
				this.target[property].splice(removeValues[i].beforeIndex, 1);
			}
		}
		
		for(var property in this.arrayInsertReference) {
			var removeValues = this.arrayInsertReference[property];
			for(var i = 0, len = removeValues.length; i < len; i++) {
				this.target[property].splice(removeValues[i].beforeIndex, 1);
			}
		}
		
		// fire change event
		if(this.target.onChange) {
			// TODO: build a "before", fire an onBeforeChange() event, or just don't send it at all
			this.target.onChange({
				added: this.removed,
				updated: this.original,
				deleted: this.added,
				popped: this.arrayPush,
				poppedReference: this.arrayPushReference,
				pushedRefresh: this.arrayPushRefresh,
				before: this.updated,
				insertArray: this.arrayRemove,
				insertReferenceArray: this.arrayRemoveReference,
				removeArray: this.arrayInsert,
				removeReferenceArray: this.arrayInsertReference
			});
		}
	},
	
	/**
	 * Convenience method to delete a property, equivalent to set(property, undefined)
	 */
	remove: function(/* String */ property) {
		return this.set(property, undefined);
	},
	
	/**
	 * Convenience method to perform the equivalent of the ++ operator
	 */
	increment: function(/* String */ property) {
		return this.add(property, 1);
	},
	
	/**
	 * Convenience method to perform the equivalent of the -- operator
	 */
	decrement: function(/* String */ property) {
		return this.subtract(property, 1);
	},	
	
	/**
	 * Convenience method to perform the equivalent of the += operator
	 */
	add: function(/* String */ property, /* Number */ value) {
		return this.set(property, this.target[property] + value);
	},
	
	/**
	 * Convenience method to perform the equivalent of the -= operator
	 */
	subtract: function(/* String */ property, /* Number */ value) {
		return this.set(property, this.target[property] - value);
	},

	/**
	 * Convenience method to perform the equivalent of the *= operator
	 */
	multiply: function(/* String */ property, /* Number */ value) {
		return this.set(property, this.target[property] * value);
	},
	
	/**
	 * Convenience method to perform the equivalent of the /= operator
	 */
	divide: function(/* String */ property, /* Number */ value) {
		return this.set(property, this.target[property] / value);
	}	
});