/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2010.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.common.vcs.Repository");

dojo.require("dojo.DeferredList");
dojo.require("rmpw.common.vcs.ChangeSet");

/**
 * Object change tracking service.
 * 
 * Note:
 * The entries in the stack represent revisions; however, since each revision only has one change set associated with it,
 * change sets are used to represent revisions, even though they truly represent the transition from the previous revision
 * to the revision they represent.
 * 
 * This is why change sets are returned from getCurrentRevision(), and taken as arguments to rollbackTo() and recommitTo().
 * At some point, there may be some reason to model the revisions themselves, at which point those would change, so clients
 * should use change sets as references only when dealing with those functions.
 * 
 * @author Marc Gobeil
 */
dojo.declare("rmpw.common.vcs.Repository", null, {	
	constructor: function(args) {
		args = args || {};
		this.resetStack();
	},
	
	resetStack: function() {
		this.stack = [new rmpw.common.vcs.ChangeSet()];
		this.stackPosition = 0;
		this._fireRevisionChanged();
	},
	
	getStack: function(purge) {
		// if purge, pop off any potential redo's on the stack
		if (purge) {
			while (this.stack.length > this.stackPosition+1) {
				this.stack.pop();
			}
		}
		return this.stack;
	},
	
	getCurrentRevision: function() {
		return this.stack[this.stackPosition];
	},

	/**
	 * Fired before a commit.  Provides an opportunity to attach preconditions
	 * to the changeset.  changeSet.preconditions contains a list of promises
	 * that must be fulfilled before a commit will execute, and if any are
	 * rejected, the commit will be blocked.
	 */
	beforeCommit: function(changeSet) {
		// for listeners
	},
	
	/**
	 * Fired when a commit precondition has failed, and will not be applied to the model. 
	 */
	commitBlocked: function(error) {
		// for listeners
	},
	
	/**
	 * Fired when a fresh commit is made.  Note: will not fire when a
	 * commit is re-applied via redo.  
	 * 
	 * @param changeSet
	 */
	afterCommit: function(changeSet) {
		// for listeners
	},
	
	/**
	 * Note: Does not support progress notification because DeferredList doesn't support it.
	 * 
	 * @param {rmpw.common.vcs.ChangeSet} changeSet
	 */
	commit: function(changeSet) {
		changeSet.preconditions = [];
		this.beforeCommit(changeSet);
		
		if(changeSet.preconditions.length > 0) {
			this.wasDelayedCommit = true;
			var allPreconditions = new dojo.DeferredList(changeSet.preconditions, false, true, this.commitBlocked);
			allPreconditions.then(dojo.hitch(this, this._commit, changeSet), this.commitBlocked);
		} else {
			this._commit(changeSet);
		}
	},
	
	_commit: function(changeSet) {
		delete this.wasDelayedCommit;
		delete changeSet.preconditions;
		
		changeSet.commit();
		
		this.stack.length = ++this.stackPosition;
			
		// push change onto stack
		this.stack.push(changeSet);
		
		this.afterCommit(changeSet);
		this._fireRevisionChanged();
	},
	
	getLastChangeSet: function() {
		return this.stack[this.stackPosition];
	},
	
	rollback: function() {
		var wasRolledBack = false;
		if(this.canRollback()) {
			this.stack[this.stackPosition--].rollback();
			this._fireRevisionChanged();
			wasRolledBack = true;
		}
		return wasRolledBack;
	},
	
	canRollback: function() {
		// Note: revision 0 is the original model, which cannot be rolled back 
		return this.stackPosition > 0;
	},
	
	rollbackTo: function(changeSet) {
		var wasRolledBack = false;
		// find the state
		for(var i = this.stackPosition - 1; i >= 0; i--) {
			if(this.stack[i] === changeSet) {
				// rollback to it
				while(this.stackPosition > i) {
					this.stack[this.stackPosition--].rollback();	// TODO: undo, redo, undo can cause this to go undefined.  make test case, fix
				}
				wasRolledBack = true;
				break;
			}
		}
		if(wasRolledBack) {
			this._fireRevisionChanged();
		}
		return wasRolledBack;
	},
	
	recommit: function() {
		var wasRecommitted = false;
		if(this.canRecommit()) {
			this.stack[++this.stackPosition].commit();
			this._fireRevisionChanged();
			wasRecommitted = true;
		}
		return wasRecommitted;
	},
		
	canRecommit: function() {
		return this.stackPosition < this.stack.length;
	},
	
	recommitTo: function(changeSet) {
		var wasRecomitted = false;
		// find the state
		for(var i = this.stackPosition + 1, len = this.stack.length; i < len ; i++) {
			if(this.stack[i] === changeSet) {
				// recommit to it
				while(this.stackPosition < i) {
					this.stack[++this.stackPosition].commit();
				}
				wasRecomitted = true;
				break;
			}
		}
		if(wasRecomitted) {
			this._fireRevisionChanged();
		}
		return wasRecomitted;
	},
	
	_fireRevisionChanged: function() {
		this.onChange({
			newStackPosition: this.stackPosition
		});
	},
	
	onChange: function(e) {
		// for listeners
	},
	
	/**
	 *  Checks if there are any changes to the target identified by rdf_about that are committed in the repository, 
	 *	and if so returns the updated target (the last update, if there were more)
	 */
	getUpdatedTarget: function(rdf_about) {
		var target;
		for (var i = this.stackPosition; i >= 0 && !target; i--) {
			target = this.stack[i].getUpdatedTarget(rdf_about);
		}
		return target;
	},
	
	/**
	 * If there are any changes in this repository for given target.rdf_about, commits them to the target.
	 */	
	tryCommitToTarget: function(target) {
		var updatedTarget;
		for (var i = this.stackPosition; i >= 0 && !updatedTarget; i--) {
			updatedTarget = this.stack[i].tryCommitToTarget(target);
		}
		return updatedTarget ? updatedTarget : target;
	}
});