/*******************************************************************************
 * Licensed Materials - Property of IBM
 * Copyright IBM Corp. 2011.  All Rights Reserved.
 *
 * US Government Users Restricted Rights - Use, duplication or disclosure
 * restricted by GSA ADP Schedule Contract with IBM Corp.
 *******************************************************************************/

dojo.provide("rmpw.common.vcs.ServerChangeSet");
dojo.require("rmpw.common.vcs.ChangeSetWrapper");
dojo.require("rmpw.common.vcs.ChangeSet");
dojo.require("rmpw.common.vcs.Change");
dojo.require("rmpw.common.vcs.Patch");
dojo.require("rmps.core.error");
dojo.require("rmpw.common.util.IdentifierUtil");

/**
 * List of "rmpw.common.vcs.Change" without 'target', which reduces payload for json objects created from "change" object
 * which gets sent to the server.
 * Patch should do this job, but Patch does not maintain the order of operations which is needed
 * to avoid failure during server save operation. Until we fix we send the complete changeset stack to
 * the server in same the order the user executed.
 * @author Susanta Datta
 */
dojo.declare("rmpw.common.vcs.ServerChangeSet", null, {
	
	_visitedChanges: null,
	
	constructor: function(options) {
		this.changes = [];
		this._visitedChanges=[];
		
		if (options && options.changeSet) {
			this._add(options.changeSet);
		}
	},
	_add: function(incomingchange) {
		if (dojo.isArray(incomingchange)){
			for (var i = 0; i < incomingchange.length; i++) {
				this._add(incomingchange[i]);
			}
		}else if(incomingchange instanceof rmpw.common.vcs.ChangeSetWrapper){
			var currentChange= incomingchange.getWrapperObject();
			if(dojo.indexOf(this._visitedChanges, currentChange)<0){
				this._addChange(currentChange);
			}
		} else if(incomingchange instanceof rmpw.common.vcs.ChangeSet){
			for(var j = 0; j < incomingchange.changes.length; j++) {
				this._addChange(incomingchange.changes[j]);
			}
		} else if (incomingchange instanceof rmpw.common.vcs.Change) {
			this._addChange(incomingchange);
		} else if (incomingchange instanceof rmpw.common.vcs.Patch) {
			//don't add empty object
			if(!this._isEmptyObject(rmpw.common.vcs.Patch.changes)) {
				//FIXME: This condition will pass undefined object all the time. It seems a typo
				this._addChange(rmpw.common.vcs.Patch.changes);
			}
		}else {
			console.error ('ServerChangeSet.add param only takes ChangeSet or Change object or array of ChangeSet or Change');
		}
		
	},
	
	_addChange: function(incomingchange) {
		// If incomingchange has a target which hasn't been serialized yet, then the change is moot.
		// It will already be part of the target element which will be pushed to the server.
		// We can check this by seeing if the target has a fully qualified rdf_about
		var targetTypes = incomingchange.target.rdf_type;
		var isTargetDocument = false;
		if (targetTypes) {
			for (var i=0, lg=targetTypes.length; i<lg; i++) {
				if (targetTypes[i].rdf_about == "http://jazz.net/ns/dm/core#Document") {
					isTargetDocument = true;
					break;
				}
			}
		}
		if (!isTargetDocument && !rmpw.common.util.IdentifierUtil.isIdentifierFixed(incomingchange.target.rdf_about)) {
			// don't add the change 
			return;
		}
		
		var change = {};// should we create new object or use the change object directly ? 
		// the client may use incomingchange so we should not modify
		dojo.mixin(change, incomingchange); 
		if (change.target['rdf_about']!= undefined) { // in some cases we have rdf_about in target
			change['rdf_about'] = change.target['rdf_about'];
		}
		delete change.target;
		delete change.declaredClass;
		delete change.original;
		for (var property in change){
			if ((dojo.isObject(change[property]) && this._isEmptyObject(change[property])) || property.substring(0, 1) == "_"){
				delete change[property]; // delete empty object to send clean json data to the server
			}
		}
				
		if (change.updated|| change.added || change.deleted || change.arrayPush || change.arrayPushReference || change.arrayInsert ||
			change.arrayInsertReference || change.arrayRemove || change.arrayRemoveReference || 
			change.arrayRemoveReferenceSourcePathNames || change.arrayOrderChanges) {
			this.changes.push(change);
		}
		this._visitedChanges.push(incomingchange);	
	},
	
	_isEmptyObject: function(object) {
		for(var property in object)
			return false;
		return true;
	},
	length : function() {
		return this.changes.length;
	},
	getChanges: function() {
		return this.changes;
	},
	getChangeCommand: function(changeSet, commandArray) {
		if(changeSet instanceof rmpw.common.vcs.ChangeSet){
			for(var j = 0; j < changeSet.changes.length; j++) {
				var changes =changeSet.changes[j];
				for(var i=0; i<commandArray.length; i++){
					if (changes[commandArray[i]] !== undefined && 
							!this._isEmptyObject(changes[commandArray[i]]) ){
						return changes[commandArray[i]];
					}
				}
			}
		}
		return null;
	}
});